/* Header file for all code compiled by the compiler to C (not yet used).
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* Amendment to the GPL for this file, and this file only (codesupp.h):
   You may distribute this file along with code compiled with
   the HimML compiler to C, either in unmodified source form,
   or as part of binaries. You may use in private or commercial
   software without needing my prior consent, provided this
   file, and this file only, is left unchanged. The rest of the
   GNU General Public License remains in effect. In particular,
   this file comes with ABSOLUTELY NO WARRANTY, expressed or
   implied.
   The goal of this amendment is so that you may not only use
   HimML in any application you wish, but also use compiled
   HimML code without unreasonable restrictions.
*/

#include <stdio.h>
#include "eval.h"
#include "expr.h"
#include "evlproto.h"
#if OS==MAC_OS
#ifndef THINK_C
#include <types.h>
#include <memory.h>
#include <resources.h>
#endif
#endif

#undef FLOAT_DUMP_CONSTANTS
#define INLINE_UNSHARED_CONS
#define INLINE_UNSHARED_BOX

/* Declaration macros for execution of scofl as C instructions */

typedef CHAR *UBString;

#undef _N_EXPR
#undef _N_MAPGET
#undef _N__X

#define SYNC_PROTECTION() theStackFrames = (gcStackView *)&_x

#ifdef AGGR_INIT_OK
#define DECL_PROTECTION() \
  struct _x { COUNTER n; gcStackView *next; VOLATILE ExprPtr p[_N__X]; } _x = \
    { _N__X, theStackFrames, }; \
  SYNC_PROTECTION();
/* note that the rest of the structure will then be initialized to all zeroes,
   as the initializer is incomplete. GC counts on NOEXPR being 0 to handle
   this array correctly. */
#else
/* but local aggregate structure initialization is not legal under all compilers,
   so we define an alternative (which is slower in general). */
#define DECL_PROTECTION() \
  struct _x { COUNTER n; gcStackView *next; VOLATILE ExprPtr p[_N__X]; } _x = \
    { _N__X, theStackFrames, }; \
  { REGISTER ExprPtr *_pp,**_ppend; \
    for (_pp = (ExprPtr *)_x.p, _ppend = _pp+_N__X; _pp<_ppend; *_pp++ = NOEXPR); } \
  SYNC_PROTECTION();
#endif

#define STOP_PROTECTION() \
  theStackFrames = _x.next
#define ZAP(n) VAR(n) = NOEXPR
#define ZAPM(n) MAPGETY(n) = NOEXPR

  /* variable number n: */
#define VAR(n) _x.p[n-1]
#define LABVAR(n) MERGE(merge,n)
#define LAB(n) MERGE(merge_ms_,n)
#define qVAR(n) MERGE(qlab,n)

#define UNBOX_LISTSWEEP(l) l

#define START_LISTSWEEP(list,l) l = list

#define INC_LISTSWEEP(l,exitl) \
  if ((l = CDR(l))==NIL) \
    goto exitl;

  /* MapSweep is done by reserving MAP_CHAIN_LENGTH+1 consecutive entries
     inside the _x.p local array of protected cells. We must however declare
     an additional pointer to some entry in this array (this encodes a stack).
     This is done by declaring:
     REGISTER ExprPtr *_MPTR(n),
     where n is the number of the variable.
     */
#define _MPTR(n) MERGE(_m_ptr,n)
#define _MTOP(n) &_x.p[_N_EXPR+_N_MAPGET+(n-1)*(MAP_CHAIN_LENGTH+1)]

#define UNBOX_MAPSWEEP(m) _x.p[_N_EXPR+_N_MAPGET+(m-1)*(MAP_CHAIN_LENGTH+1)+MAP_CHAIN_LENGTH]

#define NEXT_MAPSWEEP(m) \
  while (UNBOX_MAPSWEEP(m)->kind!=(CHAR)E_MAPLET) \
    { \
      REGISTER ExprPtr _m_aux; \
 \
      _m_aux = MAPLEFT(UNBOX_MAPSWEEP(m)); \
      if (_m_aux==empty_map) \
	UNBOX_MAPSWEEP(m) = MAPRIGHT(UNBOX_MAPSWEEP(m)); \
      else \
	{ \
	  *_MPTR(m)++ = MAPRIGHT(UNBOX_MAPSWEEP(m)); /* push onto stack */ \
	  UNBOX_MAPSWEEP(m) = _m_aux; \
	} \
    }

#define START_MAPSWEEP(map,m) /* get to first maplet in map, and return it in map_m */ \
  _MPTR(m) = _MTOP(m); \
  UNBOX_MAPSWEEP(m) = map; \
  NEXT_MAPSWEEP(m)

#define INC_MAPSWEEP(m,exitm) /* get to next maplet, otherwise go to label exitm */ \
  do { \
    if (_MPTR(m)==_MTOP(m)) \
      goto exitm; \
    *_MPTR(m)-- = NOEXPR; /* allow this to be reclaimed by GC */ \
    UNBOX_MAPSWEEP(m) = *_MPTR(m); /* pop from stack */ \
  } while (UNBOX_MAPSWEEP(m)==empty_map); \
  NEXT_MAPSWEEP(m)

/* submap iterators: (for the strange-looking declarations of structs,
   see START_PROTECTION) */

IMPORT ExprPtr InitialSubMap1 P((IN CONST MapPtr map));

#define START_SUBMAPSWEEP(map,m) /* initialize sweep structure for powerset */ \
  { \
    REGISTER ExprPtr _i_submap; \
 \
    _i_submap = InitialSubMap1(map); \
    MKPAIR(E_CONS,_i_submap,empty_map,m); \
  }

#define UNBOX_SUBMAPSWEEP(m) CDR(m)

#define INC_SUBMAPSWEEP(m,exitm) /* get to next submap, otherwise go to label exitm */ \
  { \
    REGISTER ExprPtr _m_l,_m_e,_m_val; \
 \
    _m_l = CAR(m); \
    _m_val = CDR(m); \
    while (true) \
      { \
        if (_m_l==NIL) \
	  { \
	    SETCAR(m,NIL); /* free that at least */ \
	    goto exitm; \
	  } \
	_m_e = CAR(_m_l); \
	if (_m_e->kind==(CHAR)E_OR) \
	  { \
	    _m_e->kind = (CHAR)E_AND; \
	    _m_val = MapPutTrie(CAR(_m_e),CDR(_m_e),_m_val,(CODE)(1L<<NLOWBITS)); \
	    SETCDR(m,_m_val); \
	    break; \
	  } \
	_m_e->kind = (CHAR)E_OR; \
	_m_val = MapRemoveTrie(CAR(_m_e),_m_val,(CODE)(1L<<NLOWBITS)); \
	SETCDR(m,_m_val); \
	_m_l = CDR(_m_l); \
      } \
  }

/* exception handling: */

#define RAISE(exc) { _QUIT (void) Raise(exc); }

#if PROFILING & PROFILING_PROPER_TIMES
EXTERN ExprPtr currentProf;

#define START_HANDLE(abexit) /* push a handler, and go to label abexit if longjmp() */ \
   { \
     excHandler *_xh = (excHandler *)alloctmp(sizeof(excHandler)); \
     ExprPtr _cProf = currentProf; \
 \
     if (SETJUMP(_xh->jb)!=0) /* if abnormal exit */ \
       { \
	 setHandlerStack(excStack,_xh->next); \
	 syncProfProper(_cProf); \
	 goto abexit; \
       } \
       /* normal exit, evaluate body of handle clause */ \
     _xh->next = excStack; \
     _xh->tag = X_HANDLER; \
     _xh->code = _xh->data = NIL; \
     excStack = _xh; \
   }
#else
#define START_HANDLE(abexit) /* push a handler, and go to label abexit if longjmp() */ \
   { \
     excHandler *_xh = (excHandler *)alloctmp(sizeof(excHandler)); \
 \
     if (SETJUMP(_xh->jb)!=0) /* if abnormal exit */ \
       { \
	 setHandlerStack(excStack,_xh->next); \
	 goto abexit; \
       } \
       /* normal exit, evaluate body of handle clause */ \
     _xh->next = excStack; \
     _xh->tag = X_HANDLER; \
     _xh->code = _xh->data = NIL; \
     excStack = _xh; \
   }
#endif
#define END_HANDLE() /* pop a handler */ \
   excStack = NEXTEXCHANDLER(excStack)

  /* MapGet returns a virtual couple, i.e several values at once. */

#define MAPGETY(n) _x.p[_N_EXPR+n-1]
#define MAPGETF(n) MERGE(_m_f_,n)

#define MAPGET(x,map,n) MAPGETY(n) = MapGet(x,map,&MAPGETF(n))

  /* handling of floating-point values: */

#ifdef FLOAT_DUMP_CONSTANTS
#define DEF_EXTENDED(x,val,def) def

#define SET_XL(x,i,j) ((unsigned long *)&(x))[i] = j
#define SET_XS(x,i,j) ((unsigned short *)&(x))[i] = j
#define SET_XC(x,i,j) ((unsigned char *)&(x))[i] = j
#else
#define DEF_EXTENDED(x,val) x = val
#endif

  /* allocations that do not share: */
#ifdef INLINE_UNSHARED_CONS
#define MKPAIR(xkind,xcar,xcdr,e) \
  { REGISTER ExprPtr _K_aux; MAKEPAIR(xkind,xcar,xcdr,_K_aux); e = _K_aux; }
#else
#define MKPAIR(xkind,xcar,xcdr,e) e = MakePair(xkind,xcar,xcdr)
#endif

#ifdef INLINE_UNSHARED_BOX
#define MKSING(xkind,xref,e) \
  { REGISTER ExprPtr _K_aux; MAKESING(xkind,xref,_K_aux); e = _K_aux; }
#else
#define MKSING(xkind,xref,e) e = MakeSing(xkind,xref,0L)
#endif

#define BEGIN_VEC(xn) \
  { \
    ExprPtr *_K_ep; \
 \
    _K_ep = (ExprPtr *)alloctmp(xn*sizeof(ExprPtr))
#define PUSH_VEC(xval) *_K_ep++ = xval
#define END_VEC(e,xkind,xshape,xn) \
    e = MakeRecord(xkind,xshape,xn,_K_ep); \
  }
#define END_VEC_SHARE(e,xkind,xshape,xn) \
    e = Karray(xkind,xshape,xn,_K_ep); \
  }

#define Kbool(b) (b?trueValue:falseValue)

#define BVAL(b) (b==trueValue)

  /* environments: */
#define MKENV(enext,xn,e) \
  { \
    VecPtr _K_vp; \
    REGISTER ExprPtr _K_aux,*_K_ep; \
    REGISTER COUNTER _K_i; \
 \
    CHECKHETEROFREE(); \
    scAlloc(*(char **)&_K_vp,VECSIZE(xn)); \
    _K_vp->n = xn; \
    _K_ep = _K_vp->expr; \
    for (i=0; i<xn; i++) *_K_ep++ = NOEXPR; \
    MAKEENV1_NOGC(_K_vp,enext,_K_aux); \
    e = _K_aux; \
  }

#define FAKEENV(enext,e) \
  { \
    REGISTER ExprPtr _K_aux; \
 \
    MAKEENV1(EVARS(GLOBENV),enext,_K_aux); \
    e = _K_aux; \
  }
    
#define ELET(xenv,xval,xrank) \
  EVARS(xenv)->expr[xrank] = xval; \
  HITBARRIER(xenv)

  /* queues, as used in list comprehensions: */
#define MKQ(xarg,e) \
  { \
    REGISTER ExprPtr _K_q; \
 \
    EALLOC_HOMO(_K_q); \
    _K_q->kind = (CHAR)E_CONS; \
    CAR(_K_q) = _K_q; \
    CDR(_K_q) = xarg; \
    e = _K_q; \
  }
#define MK1Q(xarg,e) \
  { \
    REGISTER ExprPtr _K_l; \
 \
    BEGINQUICKKEEP(); \
    MKPAIR(E_CONS,xarg,NIL,_K_l); \
    QUICKKEEP(_K_l); \
    MKQ(_K_l,e); \
    ENDQUICKKEEP(); \
  }
#define Q2L(xq,xrest,e) e = Reverse1(CDR(xq),xrest)

#define CONCATQ(xq1,xq2,xe) \
  { \
    REGISTER ExprPtr _G_car2; \
 \
    _G_car2 = CAR(xq2); \
    if (_G_car2!=xq2) \
      { \
	REGISTER ExprPtr _G_car1: \
 \
	_G_car1 = CAR(q1); \
	SETCDR(_G_car1,CDR(q2)); \
	SETCAR(q1,_G_car2); \
     } \
     xe = xq1; \
  }

#define RMU1(xrec,xlab,xmod) mu1(xrec,xlab,xmod,true)
#define UNSHARED_RMU1(xrec,xlab,xmod) mu1(xrec,xlab,xmod,false)
#define RMUX(xrec,xmod) muGnrl(xrec,xmod,true)
#define UNSHARED_RMUX(xrec,xmod) muGnrl(xrec,xmod,false)

  /* Beamer functions are used when there is not enough stack space left
     to reliably execute the current function. */
#define DECL_BEAMER(fun) PRIVATE ExprPtr MERGE(_Beam_,fun) P((ExprPtr arg,ExprPtr env));

#define DEF_BEAMER(fun) \
PRIVATE ExprPtr MERGE(_Beam_,fun)(arg,env) \
IN CONST ExprPtr arg,env; \
{
#define START_BEAMER(fun) \
  REGISTER struct MERGE(_Beam_,fun) *b = (struct MERGE(_Beam_,fun) *)(env);

#define END_BEAMER() return NIL; }

#ifdef ASCENDING_STACK
#define TEST_STACK() \
{ \
  char _test_; \
 \
  if (&_test_>currentThread->lowMark) \
    goto _beam_stack_; \
}
#else
#define TEST_STACK() \
{ \
  char _test_; \
 \
  if (&_test_<currentThread->lowMark) \
    goto _beam_stack_; \
}
#endif

#define PREP_BEAM(fun) \
  struct MERGE(_Beam_,fun) *b = (struct MERGE(_Beam_,fun))alloctmp(sizeof(struct MERGE(_Beam_,fun))); \
 \
  _QUIT

#define BEAM(fun,data) CallCC(MERGE(_Beam_,fun),(ExprPtr)data,true);
