/* The HimML interpreter.
   Copyright (C) 1992-1997 Jean Goubault-Larrecq and Bull S.A.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "portable.h"
#if !TRANSLATOR
#include "mem.h"
#include "eval.h"
#include "expr.h"
#include "timing.h"
#include "debug.h"
#include "evlproto.h"
#include "typproto.h"
#include "errproto.h"
#include "exeproto.h"
#include "cmpproto.h"
#include "stkproto.h"
#include "dbgproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "hakproto.h"
#include "iter.h"
#include "reval.h"
#ifdef HAS_PROTOTYPES
#include "seval.h"
#endif

#define __SEG__ EVAL

	/* !!! WARNING:
	All functions here trash theEnvironment;
	this is done for efficiency, and to implement
	terminal recursion.
	This may impede garbage collection
	by KEEPing too large an environment sometimes
	(cases of EvalOf, EvalSimpleOf, EvalMap, EvalWhile),
	but I feel this is not serious.

	*In particular, Eval() almost never returns
	a meaningful theEnvironment*

	However, EvalBindingsRC() preserves theEnvironment.
	*/

PUBLIC ExprPtr CGlobals[N_C_GLOBALS];

EXTERN ExprPtr toplevelCont;

EXTERN ExprPtr stackException;
EXTERN SymPtr ione,itwo;
EXTERN MapPtr mapgetCache_e,mapgetCache_m,mapgetCache_r;
EXTERN Boolean mapgetCache_f;
EXTERN SymPtr itSym;
EXTERN Boolean itBarrier;

EXTERN ExprPtr theEnvironment;
EXTERN stackMark *currentThread;

EXTERN ExprPtr specialMarker;
EXTERN ExprPtr none,some;

#if DEBUGGING!=0
PUBLIC ExprPtr currentDebugZone;
#endif

#undef GETLOOKUPSTATS
#undef LOOKUP_BY_CASE

#define INITENV(vp,from,to) \
{ \
  REGISTER ExprPtr *_ep,*_epend; \
 \
  for (_ep = &vp->expr[from], _epend = &vp->expr[to]; _ep<_epend;) \
    *_ep++ = NOEXPR; \
}

PUBLIC ExprPtr InitialSubMap1 P((IN CONST MapPtr map));

PRIVATE ExprPtr Eval0 P((IN CONST ExprPtr e));
     /* Eval0() is as Eval(), except that it does not KEEP() its result. */

#if PROFILING
EXTERN long gc_utime,gc_stime,gc_rtime,gc_count;
#if PROFILING & PROFILING_PROPER_TIMES
PUBLIC ExprPtr currentProf = NOEXPR;
#endif

#if PROFILING & PROFILING_PROPER_TIMES
PRIVATE void profEnterProper(data,tms)
INOUT REGISTER Profile *data;
#if OS==AMIGA_OS
IN CONST unsigned int tms[];
#else
IN CONST struct tms *tms;
#endif
{
#if OS==AMIGA_OS
  data->last.proper.time.clock[0] = tms[0];
  data->last.proper.time.clock[1] = tms[1];
#else
  data->last.proper.time.tms_utime = tms->tms_utime;
  data->last.proper.time.tms_stime = tms->tms_stime;
#endif
  data->last.proper.gc_time.tms_utime = gc_utime;
  data->last.proper.gc_time.tms_stime = gc_stime;
  data->last.proper.gc_count = gc_count;
}
#else
#define profEnterProper(e,tms)
#endif

#if PROFILING & PROFILING_PROPER_TIMES
#if OS==AMIGA_OS
PRIVATE void subtime(c3,c1,c2)
IN CONST unsigned int c1[];
IN CONST unsigned int c2[];
OUT unsigned int c3[];
{
  unsigned int musec;

  musec = c1[1] - c2[1];
  if (musec>c1[1]) /* borrow */
    {
      c3[0] = c1[0]-c2[0]-1;
      c3[1] = 1000000+musec; /* this overflows, but the Amiga does not care */
    }
  else
    {
      c3[0] = c1[0]-c2[0];
      c3[1] = musec;
    }
}
#endif

PRIVATE void profExitProper(data,tms)
INOUT REGISTER Profile *data;
#if OS==AMIGA_OS
IN CONST unsigned int tms[];
#else
IN CONST struct tms *tms;
#endif
{
#if OS==AMIGA_OS
  subtime(data->stat.proper.time.clock,tms,data->last.proper.time.clock);
#else
  data->stat.proper.time.tms_utime += tms->tms_utime - data->last.proper.time.tms_utime;
  data->stat.proper.time.tms_stime += tms->tms_stime - data->last.proper.time.tms_stime;
#endif
  data->stat.proper.gc_time.tms_utime += gc_utime - data->last.proper.gc_time.tms_utime;
  data->stat.proper.gc_time.tms_stime += gc_stime - data->last.proper.gc_time.tms_stime;
  data->stat.proper.gc_count += gc_count - data->last.proper.gc_count;
}
#else
#define profExitProper(e,tms)
#endif

#if PROFILING & PROFILING_PROPER_TIMES
PUBLIC void syncProfProper(save)
IN CONST ExprPtr save;
{
#if OS==AMIGA_OS
  unsigned int clock[2];

  if (save!=NULL)
    {
      (void) timer(clock);
      profEnterProper(PROFDATA(save),clock);
    }
#else
  struct tms tms;

  if (save!=NULL)
    {
      (void) times(&tms);
      profEnterProper(PROFDATA(save),&tms);
    }
#endif
  currentProf = save;
}
#endif

#if PROFILING & PROFILING_TOTAL_TIMES
PRIVATE void profEnter(e,tms)
INOUT ExprPtr e;
#if OS==AMIGA_OS
IN CONST unsigned int tms[];
#else
IN CONST struct tms *tms;
#endif
{
  REGISTER struct Profile *data;

  data = PROFDATA(e);
  if (data->nEntries++==0)
    {
#if OS==AMIGA_OS
      data->last.total.time.clock[0] = tms[0];
      data->last.total.time.clock[1] = tms[1];
#else
      data->last.total.time.tms_utime = tms->tms_utime;
      data->last.total.time.tms_stime = tms->tms_stime;
#endif
      data->last.total.gc_time.tms_utime = gc_utime;
      data->last.total.gc_time.tms_stime = gc_stime;
      data->last.total.gc_count = gc_count;
    }
}
#else
#define profEnter(e,tms)
#endif

#if PROFILING & PROFILING_TOTAL_TIMES
PRIVATE void profExit(e,tms)
INOUT ExprPtr e;
#if OS==AMIGA_OS
IN CONST unsigned int tms[];
#else
IN CONST struct tms *tms;
#endif
{
  REGISTER struct Profile *data;

  data = PROFDATA(e);
  if (--data->nEntries==0)
    {
#if OS==AMIGA_OS
      subtime(data->stat.total.time.clock,tms,data->last.total.time.clock);
#else
      data->stat.total.time.tms_utime += tms->tms_utime - data->last.total.time.tms_utime;
      data->stat.total.time.tms_stime += tms->tms_stime - data->last.total.time.tms_stime;
#endif
      data->stat.total.gc_time.tms_utime += gc_utime - data->last.total.gc_time.tms_utime;
      data->stat.total.gc_time.tms_stime += gc_stime - data->last.total.gc_time.tms_stime;
      data->stat.total.gc_count += gc_count - data->last.total.gc_count;
    }
}
#else
#define profExit(e,tms)
#endif

PRIVATE ExprPtr profEval(e)
IN CONST ExprPtr e;
{
  ExprPtr val;
  excHandler *xh = (excHandler *)alloctmp(sizeof(excHandler));
#if PROFILING & PROFILING_PROPER_TIMES
  ExprPtr old = currentProf;
#endif
#if OS==AMIGA_OS
  unsigned int clock[2];
#define TMS clock
#else
  struct tms tms;
#define TMS &tms
#endif

  xh->next = excStack;
  xh->tag = X_PROF;
  xh->data = e;
  excStack = xh;

#if PROFILING & PROFILING_CALL_COUNTS
  PROFDATA(e)->nCalls++;
#endif
#if OS==AMIGA_OS
  (void) timer(clock);
#else
  times(&tms);
#endif
#if PROFILING & PROFILING_PROPER_TIMES
  if (old!=NULL)
    profExitProper(PROFDATA(old),TMS);
  profEnterProper(PROFDATA(e),TMS);
  currentProf = e;
#endif
  profEnter(e,TMS);
  val = Eval0(PROFCODE(e));
#if OS==AMIGA_OS
  (void) timer(clock);
#else
  times(&tms);
#endif
  profExit(e,TMS);
#if PROFILING & PROFILING_PROPER_TIMES
  profExitProper(PROFDATA(e),TMS);
  if (old!=NULL)
    profEnterProper(PROFDATA(old),TMS);
  currentProf = old;
#endif
  excStack = xh->next;
  return val;
#undef TMS
}
#endif

PUBLIC void resetHandlerStack()
{
#if PROFILING
  REGISTER excHandler *h;
  Boolean noTime = true;
#if OS==AMIGA_OS
  unsigned int clock[2];
#define TMS clock
#else
  struct tms tms;
#define TMS &tms
#endif

  if (excStack!=NULL)
    {
      for (h = excStack; h!=NULL; h = NEXTEXCHANDLER(h))
	{
	  if (h->tag==X_PROF)
	    {
	      if (noTime)
		{
#if OS==AMIGA_OS
		  (void) timer(clock);
#else
		  times(&tms);
#endif
		  noTime = false;
		}
	      profExit(h->data,TMS);
	    }
	}
    }
  if (currentProf!=NULL)
    {
      profExitProper(PROFDATA(currentProf),TMS);
      currentProf = NULL;
    }
#undef TMS
#endif
  excStack = NULL;
}

PUBLIC void setHandlerStack(stack)
IN CONST excHandler *stack;
{
#if PROFILING
  REGISTER excHandler *h;
  Boolean noTime = true;
#if OS==AMIGA_OS
  unsigned int clock[2];
#define TMS clock
#else
  struct tms tms;
#define TMS &tms
#endif

  if (excStack!=NULL && excStack!=stack)
    {
      for (h = excStack; h!=NULL && h!=stack; h = NEXTEXCHANDLER(h))
	{
	  if (h->tag==X_PROF)
	    {
	      if (noTime)
		{
#if OS==AMIGA_OS
		  (void) timer(clock);
#else
		  times(&tms);
#endif
		  noTime = false;
		}
	      profExit(h->data,TMS);
	    }
	}
    }
  if (h==NULL && stack!=NULL)
    {
      for (h = stack; h!=NULL; h = NEXTEXCHANDLER(h))
	{ /* should actually traverse excStack in reverse, but this
	     wouldn't change anything */
	  if (h->tag==X_PROF)
	    {
	      if (noTime)
		{
#if OS==AMIGA_OS
		  (void) timer(clock);
#else
		  times(&tms);
#endif
		  noTime = false;
		}
	      profEnter(h->data,TMS);
	    }
	}
    }
  if (currentProf!=NULL)
    {
      if (noTime)
	{
#if OS==AMIGA_OS
	  (void) timer(clock);
#else
	  times(&tms);
#endif
	}
      profExitProper(PROFDATA(currentProf),TMS);
      currentProf = NULL;
    }
#undef TMS
#endif
  excStack = stack;
}

#ifdef GETLOOKUPSTATS
PUBLIC INT lookupStats[500] = { 0, };
#endif

PRIVATE ExprPtr lookup(i,level)
IN REGISTER COUNTER i,level;
{
  REGISTER ExprPtr env;

#ifdef GETLOOKUPSTATS
  lookupStats[level]++;
#endif
#ifdef LOOKUP_BY_CASE
  env = theEnvironment;
  switch (level)
    {
    case 7: env = ENEXT(env); /*fallthrough*/
    case 6: env = ENEXT(env); /*fallthrough*/
    case 5: env = ENEXT(env); /*fallthrough*/
    case 4: env = ENEXT(env); /*fallthrough*/
    case 3: env = ENEXT(env); /*fallthrough*/
    case 2: env = ENEXT(env); /*fallthrough*/
    case 1: env = ENEXT(env); /*fallthrough*/
    case 0: return EVARS(env)->expr[i];
    default:
      for (; level!=0; level--, env=ENEXT(env));
      return EVARS(env)->expr[i];
    }
#else
  for (env = theEnvironment; level!=0; level--, env=ENEXT(env));
  return EVARS(env)->expr[i];
#endif
}

PRIVATE Boolean MatchValue P((INOUT ExprPtr env,
	IN CONST ExprPtr pat,IN CONST ExprPtr val));

PRIVATE ExprPtr patChoiceBody;

MAPSWEEPFUNC(do_mapadd_match)
{
  struct mapadd_match *data;

  data = MAPDATA(struct mapadd_match *);
  patChoiceBody = data->savePatChoiceBody;
  switch (data->maplet->kind)
    {
    case E_CONS:
      /*case E_UNSHARED_CONS:*/
      if (MatchValue(data->env,CAR(data->maplet),x))
	return false;
      if (MatchValue(data->env,CDR(data->maplet),y))
	return false;
      break;
    case E_PAT_REVCONS:
      if (MatchValue(data->env,CDR(data->maplet),y))
	return false;
      if (MatchValue(data->env,CAR(data->maplet),x))
	return false;
      break;
    }
  {
    struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;
    Boolean nomatch;

    _x.n = 1;
    _x.next = theStackFrames;
    _x.p[0] = MapKill(x,data->val);
    theStackFrames = (gcStackView *)&_x;

    nomatch = MatchValue(data->env,data->rest,_x.p[0]);
    theStackFrames = _x.next;
    return !nomatch;
  }
}

PRIVATE Boolean recordMapAddMatch(env,pat,val)
INOUT ExprPtr env;
IN CONST ExprPtr pat,val;
#define MATCH_ARG env
#define VP EVARS(env)
{
  /* pat is a E_PAT_RECORD_MAPADD list, terminated by an ellipsis
     (coded as E_PAT_ANY), with possible E_PATCHOICEs inside.
     It cannot be terminated by an end-of-record, since in that
     case the compiler would be able to determine the type and
     assign indexes to all the fields, so that we would have an
     E_PAT_RECORD_GNRL_CONS or E_PAT_REVRECORD_GNRL_CONS pattern.
     This list contains couples (label,pattern) to be matched.
  */
  Boolean found;
  ExprPtr idx,shape,maplet;

  /* val may be a record_gnrl, or a record_1; in the last case,
     CDR(pat) must be an ellipsis. */
  /*if (val==NIL)
    return false;*/
  if (val->kind==(CHAR)E_RECORD_1)
    return CDR(pat)!=patany ||
#if DEBUGGING!=0
	   CAR(CAR(pat))!=CAR(val) || /* compare labels */
#endif
	   MatchValue(MATCH_ARG,CDR(CAR(pat)),CDR(val));
#if DEBUGGING!=0
  if (val->kind!=(CHAR)E_RECORD_GNRL)
    return true;
#endif
  shape = ASHAPE(val);
again:
  switch (pat->kind)
    {
      case E_PAT_RECORD_MAPADD:
	maplet = CAR(pat);
	idx = CAR(maplet);
	found = (idx->kind==(CHAR)E_INT); /* pre-compiled index */
	if (!found)
	  idx = MapGet(CAR(maplet),shape,&found);
#if DEBUGGING!=0
	if (!found)
	  return true;
#endif
	if (MatchValue(MATCH_ARG,CDR(maplet),AVAL(val)->expr[IVAL(idx)-1]))
	  return true;
	pat = CDR(pat);
	goto again;
      case E_PATCHOICE:
	if (recordMapAddMatch(MATCH_ARG,CAR(pat),val))
	  { /* a choice is (primary pattern,(secondary pattern,secondary body)) */
	    pat = CDR(pat);
	    patChoiceBody = CDR(pat);
	    pat = CAR(pat);
	    goto again;
	  }
	return false;
      case E_DEBUGNAME:
#if DEBUGGING!=0
	VP->expr[COUNTEDN(DNINFO(pat))] = COUNTEDVAL(DNINFO(pat));
	HITBARRIER(env);
#endif
	pat = DNPAT(pat);
	goto again;
      default: /* should be an ellipsis */
	return false;
    }
#undef MATCH_ARG
#undef VP
}

PRIVATE Boolean vectorMatch P((INOUT ExprPtr env,
			       IN CONST ExprPtr pat,IN CONST ExprPtr val,
			       IN COUNTER idx));

DEF_PRIVATE_ACTION(vectormatch)
{
  struct vectormatch *data;

  data = (struct vectormatch *)env;
  data->res = vectorMatch(data->arg,data->pat,data->val,data->idx);
  return NIL;
}

PRIVATE Boolean vectorMatch1(arg,pat,val,idx)
INOUT ExprPtr arg;
IN CONST ExprPtr pat,val;
IN COUNTER idx;
{
  struct vectormatch *data = (struct vectormatch *)alloctmp(sizeof(struct vectormatch));
  ENTER;

  data->arg = arg;
  data->pat = pat;
  data->val = val;
  data->idx = idx;
  (void) CallCC(ACTION_NAME(vectormatch),(ExprPtr)data,true);
  QUIT;
  return data->res;
}

PRIVATE Boolean vectorMatch(env,pat,val,idx)
INOUT ExprPtr env;
IN CONST ExprPtr pat,val;
IN COUNTER idx;
#define MATCH_ARG env
#define VP EVARS(env)
{
  if (STKFULL(MATCH_ARG))
    return vectorMatch1(MATCH_ARG,pat,val,idx);
again:
  switch (pat->kind)
    {
      case E_PAT_TUPLE_GNRL_CONS:
      case E_PAT_RECORD_GNRL_CONS:
	if (MatchValue(MATCH_ARG,CAR(pat),AVAL(val)->expr[idx]))
	  return true;
	pat = CDR(pat);
	idx++;
	goto again;
      case E_PAT_REVTUPLE_GNRL_CONS:
      case E_PAT_REVRECORD_GNRL_CONS:
	if (vectorMatch(MATCH_ARG,CDR(pat),val,idx+1))
	  return true;
	return MatchValue(MATCH_ARG,CAR(pat),AVAL(val)->expr[idx]);
      case E_PATCHOICE:
	if (vectorMatch(MATCH_ARG,CAR(pat),val,idx))
	  { /* a choice is (primary pattern,(secondary pattern,secondary body)) */
	    pat = CDR(pat);
	    patChoiceBody = CDR(pat);
	    pat = CAR(pat);
	    goto again;
	  }
	return false;
      case E_DEBUGNAME:
#if DEBUGGING!=0
	VP->expr[COUNTEDN(DNINFO(pat))] = COUNTEDVAL(DNINFO(pat));
	HITBARRIER(env);
#endif
	pat = DNPAT(pat);
	goto again;
      case E_PAT_ANY: /* ellipsis */
	return false;
      default: /* should be an end-of-tuple (or end-of-vector) */
	return AVAL(val)->n!=idx; /* check that we are at the end */
    }
#undef MATCH_ARG
#undef VP
}

PRIVATE Boolean tupleMatch(env,pat,val)
INOUT ExprPtr env;
IN CONST ExprPtr pat,val;
#define MATCH_ARG env
{
  /* pat is a list of E_PAT_TUPLE_GNRL_CONS or E_PAT_REVTUPLE_GNRL_CONS.
     It is terminated by E_TUPLE_0 (fixed size tuples), or by
     E_PAT_ANY (ellipsis).
  */
  ExprPtr aux;
  /* val may be a tuple_gnrl, or a tuple_2; in the last case,
     CDR(pat) must be an ellipsis or a one-element list terminated
     by an ellipsis or tuple0. */

  /*if (val==NIL)
    return false;*/
  if (val->kind==(CHAR)E_TUPLE_2)
    {
      aux = CDR(pat);
      if (aux==patany)
	return MatchValue(MATCH_ARG,CAR(pat),CAR(val));
#if DEBUGGING!=0
      if ((aux->kind!=(CHAR)E_PAT_TUPLE_GNRL_CONS &&
	   aux->kind!=(CHAR)E_PAT_REVTUPLE_GNRL_CONS) ||
	  (CDR(aux)!=patany && CDR(aux)!=tuple0))
	return true;
#endif
      if (pat->kind==(CHAR)E_PAT_TUPLE_GNRL_CONS)
	return MatchValue(MATCH_ARG,CAR(pat),CAR(val)) ||
	       MatchValue(MATCH_ARG,CAR(aux),CDR(val));
      return MatchValue(MATCH_ARG,CAR(aux),CDR(val)) ||
	     MatchValue(MATCH_ARG,CAR(pat),CAR(val));
    }
#if DEBUGGING!=0
  if (val->kind!=(CHAR)E_TUPLE_GNRL)
    return true;
  /* there are no 1-tuples in HimML. */
#endif
  return vectorMatch(MATCH_ARG,pat,val,0);
#undef MATCH_ARG
}

PRIVATE Boolean recordMatch(env,pat,val)
INOUT ExprPtr env;
IN CONST ExprPtr pat,val;
{
  /* pat is a list of E_PAT_RECORD_GNRL_CONS or E_PAT_REVRECORD_GNRL_CONS.
     It is terminated by E_RECORD_0 (fixed size records), or by
     E_PAT_ANY (ellipsis).
  */
  /* val may be a record_gnrl, or a record_1; in the last case,
     CDR(pat) must be an ellipsis. */
  /*if (val==NIL)
    return false;*/
  if (val->kind==(CHAR)E_RECORD_1)
    return CDR(pat)!=patany ||
	   MatchValue(env,CAR(pat),CDR(val));
#if DEBUGGING
  if (val->kind!=(CHAR)E_RECORD_GNRL)
    return true;
#endif
  return vectorMatch(env,pat,val,0);
}

PRIVATE Boolean mapAddMatch(env,pat,val)
INOUT ExprPtr env;
IN CONST ExprPtr pat,val;
{
#if DEBUGGING
  if (ISNEMAP(val))
#else
    /* because of type checking, tags always match. */
    if (val!=NewMap())
#endif
      {
	ExprPtr x,y,xx;

	x = CAR(PATMAPADDXY(pat));
	/* particular case (frequent): x is a patmatch
	   or a constant value */
	/*if (x==NIL)
	  {
	  xx = NIL;
	  goto patmatch;
	  }*/
	switch (x->kind)
	  {
	  case E_QUOTE:
	    xx = REFVAL(x);
	    goto patmatch;
	  case E_VARREF:
	    xx = lookup(VARRANK(x),VARLEVEL(x));
	    goto patmatch;
	  case E_NIL: case E_EMPTY: case E_INT: case E_REAL:
	  case E_COMPLEX: case E_STRING: case E_EXCEPTION:
	  case E_TUPLE_0: case E_ARRAY_0: case E_RECORD_0:
	  case E_COMPILED_CLOSURE: case E_INTERPRETED_CLOSURE: case E_MEMO_CLOSURE:
	    xx = x;
patmatch:
	    {
	      ExprPtr yy;
	      struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;
	      Boolean found;
		
	      y = CDR(PATMAPADDXY(pat));
	      yy = MapGet(xx,val,&found);
	      if (!found || MatchValue(env,y,yy))
		return true;
	      _x.n = 1;
	      _x.next = theStackFrames;
	      _x.p[0] = MapKill(xx,val);
	      theStackFrames = (gcStackView *)&_x;
	      found = MatchValue(env,PATMAPADDREST(pat),_x.p[0]);
	      theStackFrames = _x.next;
	      return found;
	    }
	  default:
	    {
	      struct mapadd_match *data = (struct mapadd_match *)
		alloctmp(sizeof(struct mapadd_match));

	      data->maplet = PATMAPADDXY(pat);
	      data->rest = PATMAPADDREST(pat);
	      data->val = val;
	      data->savePatChoiceBody = patChoiceBody;
	      data->env = env;
	      return !MapSweep(val,do_mapadd_match,data);
	    }
	  }
      }
  return true;
}

DEF_PRIVATE_ACTION(matchvalue)
{
  struct matchvalue *data;

  data = (struct matchvalue *)env;
  data->res = MatchValue(data->arg,data->pat,data->val);
  return NIL;
}

PRIVATE Boolean MatchValue1(arg,pat,val)
INOUT ExprPtr arg;
IN CONST ExprPtr pat,val;
{
  struct matchvalue *data = (struct matchvalue *)alloctmp(sizeof(struct matchvalue));
  ENTER;

  data->arg = arg;
  data->pat = pat;
  data->val = val;
  (void) CallCC(ACTION_NAME(matchvalue),(ExprPtr)data,true);
  QUIT;
  return data->res;
}

PRIVATE Boolean MatchValue(env,pat,val)
INOUT ExprPtr env;
IN CONST ExprPtr pat,val;
	/* the policy for tag checking is the following.
		When the debugger is not present, the type
	system and the addition of E_PATCHECK tag checking
	patterns is enough to guarantee that the tags will
	be correct (the E_PATCHECK patterns are needed in
	the case of anonymous datatype constructors, where
	matching relies on tags, not on constructor names),
	so we don't check the tags, except for E_PATCHECK.
		When the debugger is present, the type
	system may become unsound, so we keep the type checks.
	*/
{
  if (STKFULL(env))
    return MatchValue1(env,pat,val);
again:
  /*if (pat==NIL)
    return val!=NIL;*/
  switch (pat->kind)
    {
    case E_VARINTRO:
      EVARS(env)->expr[VARRANK(pat)] = val;
      HITBARRIER(env);
      return false;
    case E_DEBUGNAME:
#if DEBUGGING
      EVARS(env)->expr[COUNTEDN(DNINFO(pat))] = pat; /*COUNTEDVAL(DNINFO(pat));*/
      HITBARRIER(env);
#endif
      pat = DNPAT(pat);
      goto again;
    case E_CONS:
    case E_UNSHARED_CONS:
#if DEBUGGING
      if (ISCONS(val))
#else
	/* we actually still need tag checking to distinguish nil
	   from cons cells, but we can directly compare to NIL. */
      if (val!=NIL)
#endif
	{
	  if (MatchValue(env,CAR(pat),CAR(val)))
	    return true;
	  pat = CDR(pat);
	  val = CDR(val);
	  goto again;
	}
      return true;
    case E_MAPLET:
#if DEBUGGING
      if (val->kind!=(CHAR)E_MAPLET)
	return true;
#endif
      if (MatchValue(env,MAPLETX(pat),MAPLETX(val)))
	return true;
      pat = CDR(pat);
      val = CDR(val);
      goto again;
    case E_PAT_REVCONS:
#if DEBUGGING
      if (ISCONS(val))
#else
	/* we actually still need tag checking to distinguish nil
	   from cons cells, but we can directly compare to NIL. */
      if (val!=NIL)
#endif
	{
	  if (MatchValue(env,CDR(pat),CDR(val)))
	    return true;
	  pat = CAR(pat);
	  val = CAR(val);
	  goto again;
	}
      return true;
    case E_PAT_CONS:
#if DEBUGGING
      if (ISCONS(val))
#else
	/* we actually still need tag checking to distinguish nil
	   from cons cells, but we can directly compare to NIL. */
      if (val!=NIL)
#endif
	{
	  val = Ktuple2(CAR(val),CDR(val));
	  (void) KEEP(val);
	  pat = REFVAL(pat);
	  goto again;
	}
      return true;
    case E_NIL:
      return val!=NIL;
    case E_TUPLE_0:
#if DEBUGGING
#if PLDI93_HACK
      return /*val==NIL || */val->kind!=(CHAR)E_TUPLE_0;
#else
      return val!=tuple0;
#endif
#else
      return false; /* because of type checking, this always matches. */
#endif
    case E_TUPLE_1:
    case E_UNSHARED_TUPLE_1:
#if DEBUGGING
      if (/*val==NIL || */val->kind!=(CHAR)E_TUPLE_1)
	return true;
#else
      /* because of type checking, tags always match. */
#endif
      pat = REFVAL(pat);
      val = REFVAL(val);
      goto again;
    case E_TUPLE_2:
    case E_UNSHARED_TUPLE_2:
#if DEBUGGING
      if (/*val==NIL || */val->kind!=(CHAR)E_TUPLE_2 ||
	  MatchValue(env,CAR(pat),CAR(val)))
	return true;
#else
      /* because of type checking, tags always match. */
      if (MatchValue(env,CAR(pat),CAR(val)))
	return true;
#endif
      pat = CDR(pat);
      val = CDR(val);
      goto again;
    case E_PAT_TUPLE_GNRL_CONS:
    case E_PAT_REVTUPLE_GNRL_CONS:
      return tupleMatch(env,pat,val);
    case E_PAT_REVTUPLE_2:
#if DEBUGGING
      if (/*val==NIL || */val->kind!=(CHAR)E_TUPLE_2 ||
	  MatchValue(env,CDR(pat),CDR(val)))
	return true;
#else
      /* because of type checking, tags always match. */
      if (MatchValue(env,CDR(pat),CDR(val)))
	return true;
#endif
      pat = CAR(pat);
      val = CAR(val);
      goto again;
    case E_RECORD_0:
#if DEBUGGING
#if PLDI93_HACK
      return /*val==NIL || */val->kind!=(CHAR)E_RECORD_0;
#else
      return val!=record0;
#endif
#else
      return false; /* because of type checking, this always matches. */
#endif
    case E_RECORD_1:
    case E_UNSHARED_RECORD_1:
#if DEBUGGING
      if (/*val==NIL || */val->kind!=(CHAR)E_RECORD_1
	  || CAR(pat)!=CAR(val)) /* check tag and field name */
	return true;
#else
      /* because of type checking, tags always match. */
#endif
      pat = CDR(pat);
      val = CDR(val);
      goto again;
    case E_PAT_RECORD_GNRL_CONS:
    case E_PAT_REVRECORD_GNRL_CONS: /* records with known indexes */
      return recordMatch(env,pat,val);
    case E_EMPTY: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EXCEPTION:
    case E_COMPILED_CLOSURE: case E_INTERPRETED_CLOSURE: case E_MEMO_CLOSURE:
#if PLDI93_HACK
      return !equal(pat,val);
#else
      return pat!=val;
#endif
    case E_REF:
#if DEBUGGING
      if (/*val==NIL || */val->kind!=(CHAR)E_REF)
	return true;
#else
      /* because of type checking, tags always match. */
#endif
      pat = REFVAL(pat);
      val = REFVAL(val);
      goto again;
    case E_QUOTE:
#if PLDI93_HACK
      return !equal(REFVAL(pat),val);
#else
      return REFVAL(pat)!=val;
#endif
    case E_VARREF: /* this is only used in patterns for exceptions */
#if PLDI93_HACK
	return !equal(lookup(VARRANK(pat),VARLEVEL(pat)),val);
#else
	return lookup(VARRANK(pat),VARLEVEL(pat))!=val;
#endif
    case E_APPL: /* this is only used for exceptions applied to arguments */
      if (val->kind!=(CHAR)E_CONS)
	return true;
      {
	ExprPtr exc;

	exc = CAR(pat);
	/* now the name of the exception as recorded in the pattern
	   is given by a E_VARREF that points to an entry in the environment
	   where we find a compiled closure with the exception in the
	   QENV part. */
#if PLDI93_HACK
	if (!equal(QENV(lookup(VARRANK(exc),VARLEVEL(exc))),CAR(val)))
	  return true;
#else
	if (QENV(lookup(VARRANK(exc),VARLEVEL(exc)))!=CAR(val))
	  return true;
#endif
	pat = CDR(pat);
	val = CDR(val);
	goto again;
      }
    case E_PAT_ANY:
      return false;
    case E_PAT_AS:
      if (MatchValue(env,PATAS2(pat),val))
	return true;
      pat = PATAS1(pat);
      goto again;
    case E_PAT_CHECK:
      /* if (val==NIL && IVAL(PATCHECKTAG(pat))!=E_NIL)
	 return true; */
      if (PATCHECKBITS(pat)[val->kind>>3] & (1 << (val->kind & 7)))
	{
	  pat = PATCHECKVAL(pat);
	  goto again;
	}
      return true;
    case E_PAT_PACK:
#if DEBUGGING
      if (/*val==NIL || */val->kind!=(CHAR)E_PACK)
	return true;
#else
      /* because of type checking, tags always match. */
#endif
      if (TypeMatch(PACKTYPE(val),PACKTYPE(pat)))
	return true;
      pat = PACKVAL(pat);
      val = PACKVAL(val);
      goto again;
    case E_PAT_CARD_EQ:
      if (
#if DEBUGGING
	  ISMAP(val) && 
#else
      /* because of type checking, tags always match. */
#endif
	  MapCardEq(val,COUNTEDN(pat)))
	{
	  pat = COUNTEDVAL(pat);
	  goto again;
	}
      return true;
    case E_PAT_CARD_GEQ:
      if (
#if DEBUGGING
	  ISMAP(val) && 
#else
      /* because of type checking, tags always match. */
#endif
	  MapCardGeq(val,COUNTEDN(pat)))
	{
	  pat = COUNTEDVAL(pat);
	  goto again;
	}
      return true;
    case E_PATCHOICE:
      if (MatchValue(env,CAR(pat),val))
	{ /* a choice is (primary pattern,(secondary pattern,secondary body)) */
	  pat = CDR(pat);
	  patChoiceBody = CDR(pat);
	  pat = CAR(pat);
	  goto again;
	}
      return false;
    case E_PAT_RECORD_MAPADD:
      return recordMapAddMatch(env,pat,val);
    case E_PAT_MAPADD:
      return mapAddMatch(env,pat,val);
    default:
      kinderror(message(400),pat->kind);
      break;
    }
}

PRIVATE Boolean Let P((IN CONST ExprPtr bindings));
PRIVATE Boolean EvalBindings P((IN CONST ExprPtr bindings,OUT ExprPtr env));

DEF_PRIVATE_ACTION(evalbinding)
{
  struct evalbinding1 *data;

  data = (struct evalbinding1 *)env;
  data->res = Let(LETBIND(data->binding)) || EvalBindings(LETBODY(data->binding),data->env);
  return NIL;
}

PRIVATE Boolean EvalBinding11(binding,env)
IN CONST ExprPtr binding;
INOUT ExprPtr env;
{
  struct evalbinding1 *data = (struct evalbinding1 *)alloctmp(sizeof(struct evalbinding1));
  ENTER;

  data->binding = binding;
  data->env = env;
  (void) CallCC(ACTION_NAME(evalbinding),(ExprPtr)data,true);
  QUIT;
  return data->res;
}

PRIVATE Boolean EvalBinding1(binding,env)
IN CONST ExprPtr binding;
INOUT ExprPtr env;
{
  if (binding->kind==(CHAR)E_LET) /* then this is a local list of bindings */
    {
      if (STKFULL(binding))
	return EvalBinding11(binding,env);
      return Let(LETBIND(binding)) || EvalBindings(LETBODY(binding),env);
    }
  {
    struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;
    Boolean nomatch;

    _x.n = 1;
    _x.next = theStackFrames;
    _x.p[0] = Eval0(CDR(binding));
    theStackFrames = (gcStackView *)&_x;
    nomatch = MatchValue(env,CAR(binding),_x.p[0]);
    theStackFrames = _x.next;
    return nomatch;
  }
  /* There was a comment here before that said:
     "! must KEEP the Evaluated value (see Let(),
     and yyparse()) !"
     I don't know why it was so, but I don't KEEP the value any longer
     anyway. */
}

PRIVATE Boolean MakePatVal(pat,val,nlocals)
IN CONST ExprPtr pat,val;
IN REGISTER COUNTER nlocals;
{
  VecPtr vp;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;
  long size;

  size = VECSIZE(nlocals);
  CHECKHETEROFREE();
  scAlloc(*(char **)&vp,size);
  vp->n = nlocals;
  INITENV(vp,0,nlocals);

  _x.n = 1;
  _x.next = theStackFrames;
  MAKEENV_NOGC(vp,theEnvironment,_x.p[0]);
  theStackFrames = (gcStackView *)&_x;

  if (MatchValue(_x.p[0],pat,val))
    {
      theStackFrames = _x.next;
      return false; /* error */
    }
  SETQ(theEnvironment,_x.p[0]);
  theStackFrames = _x.next;
  return true;
}

PRIVATE ExprPtr MakeSubMap(list)
IN CONST ExprPtr list;
{
  MapPtr maplet;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;
  
  _x.n = 1;
  _x.next = theStackFrames;
  _x.p[0] = NewMap();
  theStackFrames = (gcStackView *)&_x;
  while (list!=NIL)
    {
      maplet = CAR(list);
      list = CDR(list);
      if (maplet->kind==(CHAR)E_AND)
	_x.p[0] = MapAdd(PROGDO(maplet),PROGTHEN(maplet),_x.p[0]);
    }
  theStackFrames = _x.next;
  return _x.p[0];
}

PRIVATE Boolean SetPatVals(env,pats,reflist,bindings) /* used for iterators only */
INOUT ExprPtr env;
IN CONST ExprPtr pats,reflist,bindings;
{
  ExprPtr next,pat;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;
  Boolean nomatch;

  _x.n = 1;
  _x.next = theStackFrames;
  _x.p[0] = NOEXPR;
  theStackFrames = (gcStackView *)&_x;

  while (pats!=NIL)
    {
      pat = CAR(pats);
      switch (reflist->kind)
	{
	case E_PAT_RECORD_GNRL_CONS:
	  _x.p[0] = CAR(CAR(reflist));
	  next = CDR(reflist);
	  break;
	case E_PAT_RECORD_MAPADD:
	  _x.p[0] = MakeSubMap(MAPLETX(reflist));
	  next = MAPLETY(reflist);
	  break;
	case E_PAT_REVRECORD_GNRL_CONS:
	  _x.p[0] = CAR(CAR(reflist));
	  if (pat->kind==(CHAR)E_CONS)
	    {
	      if (MatchValue(env,CAR(pat),CAR(_x.p[0])))
		goto fail;
	      pat = CDR(pat);
	      _x.p[0] = CDR(_x.p[0]);
	    }
	  else
	    {
	      if (MatchValue(env,CDR(pat),CDR(_x.p[0])))
		goto fail;
	      pat = CAR(pat);
	      _x.p[0] = CAR(_x.p[0]);
	    }
	  next = CDR(reflist);
	  break;
	}
      if (MatchValue(env,pat,_x.p[0]))
	{
fail:
	  theStackFrames = _x.next;
	  return true;
	}
      pats = CDR(pats);
      reflist = next;
    }
  nomatch = (bindings==NIL)?false:EvalBindings(bindings,env);
  theStackFrames = _x.next;
  return nomatch;
}

PRIVATE Boolean EvalBindings(bindings,env)
IN CONST ExprPtr bindings;
OUT ExprPtr env;
{
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;

  _x.n = 1;
  _x.next = theStackFrames;
  _x.p[0] = theEnvironment;
  theStackFrames = (gcStackView *)&_x;

  while (bindings!=NIL)
    {
      SETQ(theEnvironment,_x.p[0]);
      if (EvalBinding1(CAR(bindings),env))
	{
	  SETQ(theEnvironment,_x.p[0]);
	  theStackFrames = _x.next;
	  return true; /* error */
	}
      bindings = CDR(bindings);
    }
  SETQ(theEnvironment,_x.p[0]);
  theStackFrames = _x.next;
  return false;
}

PUBLIC Boolean EvalBindingsRC(bindings,env,from,to)
IN CONST ExprPtr bindings;
OUT ExprPtr env;
IN COUNTER from,to;
/* evaluates bindings that are to be put into vp,
   from index 'from' to 'to'. There should be no
   values stored there. */
{
  Boolean nomatch;
  VecPtr vp;

  vp = EVARS(env);
  INITENV(vp,from,to);
  vp->n = to;
  nomatch = EvalBindings(bindings,env);
  return nomatch;
}

PUBLIC Boolean EvalBindingsRCIt(bindings,env,from,to)
IN CONST ExprPtr bindings;
OUT ExprPtr env;
IN COUNTER from,to;
/* same as EvalBindingsRC() of eval.c, but updates
   refcount of the special variable 'it'	*/
{
  ExprPtr saveit,it;
  VecPtr vp;
  Boolean nomatch;

  if (env==getTopEnv())
    {
      vp = EVARS(env);
      saveit = vp->expr[itIndex()];
      nomatch = EvalBindingsRC(bindings,env,from,to);
#if STOP_AND_COPY_VECTORS
      vp = EVARS(env);
#endif
      it = vp->expr[itIndex()];
      if (it!=saveit)
	{
	  itBarrier = true;
	}
      return nomatch;
    }
  return EvalBindingsRC(bindings,env,from,to);
}

PRIVATE ExprPtr MapToList(map)
IN CONST REGISTER MapPtr map;
{
  REGISTER ExprPtr *stkp,*top;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[1+MAP_CHAIN_LENGTH]; } _x;

  if (map==NewMap())
    return NIL;
  _x.n = 1;
  _x.next = theStackFrames;
  _x.p[0] = NIL;
  theStackFrames = (gcStackView *)&_x;

  stkp = top = _x.p+1;
  while (true)
    {
      while (map->kind!=(CHAR)E_MAPLET) /* find rightmost maplet */
	{
	  REGISTER ExprPtr aux;
	  
	  aux = MAPRIGHT(map);
	  if (aux==empty_map)
	    map = MAPLEFT(map);
	  else
	    {
	      *stkp++ = MAPLEFT(map);
	      _x.n++;
	      map = aux;
	    }
	}
      _x.p[0] = MakePair(E_CONS,map,_x.p[0]);
      do {
	if (stkp==top)
	  {
	    theStackFrames = _x.next;
	    return _x.p[0];
	  }
	map = *--stkp; _x.n--;
      } while (map==empty_map);
    }
  /*NOTREACHED*/
  return NIL;
}

MAPSWEEPFUNC(do_initialsubmap)
{
  ExprPtr *listp;
  ExprPtr aux,aux2;

  listp = MAPDATA(ExprPtr *VOLATILE);
  MAKEPROG(E_OR,x,y,aux);
  QUICKKEEP(aux);
  MAKECONS(aux,*listp,aux2);
  QUICKKEEP(aux2);
  *listp = aux2;
  return false;
}

PUBLIC ExprPtr InitialSubMap1(map)
IN CONST MapPtr map;
{
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;

  _x.n = 1;
  _x.next = theStackFrames;
  _x.p[0] = NIL;
  theStackFrames = (gcStackView *)&_x;

  (void) MapSweep(map,do_initialsubmap,(char *)&_x.p[0]);
  theStackFrames = _x.next;
  return _x.p[0];
}

PRIVATE ExprPtr EvalDomains(domains,pats,locals)
IN CONST ExprPtr domains;
OUT ExprPtr *pats,*locals; /* returns a reversed list of values;
			      stores in pats a reversed list of patterns (pats
			      should be initialized to NIL);
			      stores in locals the set of additional
			      local bindings to perform. */
{
  ExprPtr lists;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[3]; } _x;

  _x.n = 3;
  _x.next = theStackFrames;
  _x.p[0] = theEnvironment;
  _x.p[1] = _x.p[2] = NOEXPR;
  theStackFrames = (gcStackView *)&_x;

  lists = NIL;
again:
  switch (domains->kind)
    {
    case E_NIL:
      *locals = NIL;
      break;
    case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
    case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
    case E_PAT_RECORD_MAPADD: /* sub map domain descriptor */
      {
	ExprPtr pat,car,cdr,res;

	car = CAR(domains);
	cdr = CDR(domains);
	pat = CAR(car);
	SETQ(theEnvironment,_x.p[0]);

	_x.p[2] = lists;
	_x.p[1] = Eval0(CDR(car));
	switch (domains->kind)
	  {
	  case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
	    break;
	  case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
	    switch (_x.p[1]->kind)
	      {
	      case E_EMPTY:
		_x.p[1] = NIL;
		break;
	      case E_MAPLET:
		MAKECONS(_x.p[1],NIL,res);
		_x.p[1] = res;
		break;
	      case E_MAP:
		_x.p[1] = MapToList(_x.p[1]);
		break;
	      default:
		kinderror(message(402),_x.p[1]->kind);
		break;
	      }
	    break;
	  case E_PAT_RECORD_MAPADD: /* sub map domain descriptor */
	    _x.p[1] = InitialSubMap1(_x.p[1]);
	    break;
	  }
	MAKEPAIR(domains->kind,_x.p[1],_x.p[2],lists);
	_x.p[1] = lists;

	MAKECONS(pat,*pats,res);
	*pats = res; /* we assume that pats points to some variable in a gcStackView,
			and therefore won't be garbageable. */

	domains = cdr;
	goto again;
      }
    case E_TUPLE_1: /* list of local bindings */
      *locals = REFVAL(domains);
      break;
    default:
      kinderror(message(445),domains->kind);
      break;
    }
  theStackFrames = _x.next;
  return lists;
}

PRIVATE ExprPtr EvalDomainsSimple1 P((IN CONST ExprPtr domains,
				      IN CONST ExprPtr saveenv,
				      OUT ExprPtr *locals));

DEF_PRIVATE_ACTION(evaldomsim)
{
  ExprPtr e;
  struct evaldomsim *data;

  data = (struct evaldomsim *)env;
  e = EvalDomainsSimple1(data->domains,data->saveenv,&data->locals);
  return e;
}

PRIVATE ExprPtr EvalDomainsSimple1k(domains,saveenv,locals)
IN CONST ExprPtr domains;
IN CONST ExprPtr saveenv;
OUT ExprPtr *locals;
{
  struct evaldomsim *data = (struct evaldomsim *)alloctmp(sizeof(struct evaldomsim));
  ExprPtr res;
  ENTER;

  data->domains = domains;
  data->saveenv = saveenv;
  data->locals = *locals;
  res = CallCC(ACTION_NAME(evaldomsim),(ExprPtr)data,true);
  *locals = data->locals;
  QUIT;
  return res;
}

PRIVATE ExprPtr EvalDomainsSimple1(domains,saveenv,locals)
IN CONST ExprPtr domains;
IN CONST ExprPtr saveenv;
OUT ExprPtr *locals;
{
  if (STKFULL(locals))
    return EvalDomainsSimple1k(domains,saveenv,locals);
  if (domains==NIL)
    {
      *locals = NIL;
      return NIL;
    }
  SETQ(theEnvironment,saveenv);
  switch (domains->kind)
    {
    case E_PAT_RECORD_GNRL_CONS:
    case E_PAT_REVRECORD_GNRL_CONS:
    case E_PAT_RECORD_MAPADD:
      { /* in map, in list, sub map domain descriptors */
	struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;
	ExprPtr res;

	_x.n = 1;
	_x.next = theStackFrames;
	_x.p[0] = Eval0(CDR(CAR(domains)));
	theStackFrames = (gcStackView *)&_x;
	_x.p[1] = EvalDomainsSimple1(CDR(domains),saveenv,locals);
	_x.n = 2;
	MAKEPAIR(domains->kind,_x.p[0],_x.p[1],res);
	theStackFrames = _x.next;
	return res;
      }
    case E_TUPLE_1: /* local declarations */
      *locals = REFVAL(domains);
      return NIL;
    default:
      kinderror(message(481),domains->kind);
      /*NOTREACHED*/
      return NIL;
    }
}

PRIVATE ExprPtr EvalDomainsSimple(domains,locals)
IN CONST ExprPtr domains;
OUT ExprPtr *locals;
{
  ExprPtr doms;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;

  _x.n = 1;
  _x.next = theStackFrames;
  _x.p[0] = theEnvironment;
  theStackFrames = (gcStackView *)&_x;
  doms = EvalDomainsSimple1(domains,_x.p[0],locals);
  theStackFrames = _x.next;
  return doms;
}

PRIVATE Boolean IncSubMap(ll)
IN CONST ExprPtr ll;
{
  ExprPtr e;

  while (ll!=NIL)
    {
      e = CAR(ll);
      if (e->kind==(CHAR)E_OR)
	{
	  e->kind = (CHAR)E_AND;
	  return true;
	}
      e->kind = (CHAR)E_OR;
      ll = CDR(ll);
    }
  return false;
}

PRIVATE ExprPtr QueueToList(queue)
IN ExprPtr queue; /* a non-empty queue */
{
  if (queue==NIL)
    return NIL;
  /* If we didn't need to share the resulting list, it would be enough to do:
  e = CDR(queue);
  SETCDR(queue,NIL);
  */
  {
    REGISTER ExprPtr e,last,next,fin;

    /* Instead, we first reverse the circular list in place,
       while also breaking the circularity: */
    for (last = NIL, e = fin = CDR(queue);;)
      {
	next = CDR(e);
	SETCDR(e,last);
	if (next==fin)
	  break;
	last = e;
	e = next;
      }
    /* Note that queue is still protected against the GC, and is
       the reversed list itself now; so we don't need to KEEP anything. */
  }
  /* And then we reverse the resulting list, which involves sharing it: */
  return Reverse1(queue,NIL);
}

PRIVATE ExprPtr AppendQueue(q,list)
INOUT REGISTER ExprPtr q;
IN CONST REGISTER ExprPtr list;
{
  REGISTER ExprPtr e,end;

  if (list==NIL)
    return q;
  BEGINQUICKKEEP();
  if (q==NIL)
    {
      EALLOC_HOMO(e);
      e->kind = (CHAR)E_CONS;
      CAR(e) = CAR(list);
      CDR(e) = e;
      q = e;
      QUICKKEEP(e);
      list = CDR(list);
    }
  end = CDR(q);
  for (; list!=NIL; list = CDR(list))
    {
      EALLOC_HOMO(e);
      e->kind = (CHAR)E_CONS;
      CAR(e) = CAR(list);
      CDR(e) = end;
      SETCDR(q,e);
      q = e;
    }
  ENDQUICKKEEP();
  return q;
}

#ifdef OBSOLETE
PRIVATE ExprPtr RevQueueToList(queue)
IN ExprPtr queue; /* a non-empty queue */
{
  if (queue==NIL)
    return NIL;
  {
    struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;

    _x.n = 1;
    _x.next = theStackFrames;
    _x.p[0] = CDR(queue);
    SETCDR(queue,NIL);
    theStackFrames = (gcStackView *)&_x;

    _x.p[0] = Reverse1(_x.p[0],NIL);
    theStackFrames = _x.next;
    return _x.p[0];
  }
}
#endif

PRIVATE Boolean IncOf(reflist,demolist)
INOUT ExprPtr reflist;
IN CONST ExprPtr demolist;
{
  ExprPtr list;

  while (/*reflist!=NIL*/true)
    switch (reflist->kind)
      {
      case E_NIL:
	return false;
      case E_PAT_RECORD_GNRL_CONS:
      case E_PAT_REVRECORD_GNRL_CONS:
	list = CDR(CAR(reflist));
	if (list!=NIL)
	  {
	    SETCAR(reflist,list);
	    return true;
	  }
	SETCAR(reflist,CAR(demolist));
	reflist = CDR(reflist);
	demolist = CDR(reflist);
	break;
      case E_PAT_RECORD_MAPADD:
	if (IncSubMap(CAR(reflist)))
	  return true;
	/*ReinitSubMap(CAR(reflist));*/
	reflist = CDR(reflist);
	demolist = CDR(demolist);
	break;
      }
  /*return false;*/
}

#define INIT_ITER(gasp,e) \
switch (ik) \
{ \
 case C_PROG: \
  e = tuple0; \
  break; \
 case C_LIST: case C_UNSHARED_LIST: case C_APPEND: case C_UNSHARED_APPEND: \
  e = NIL; \
  break; \
 case C_SET: case C_UNION: case C_UNION_UNDER: case C_MAP: case C_MAP_UNDER: \
  e = NewMap(); \
  break; \
 case C_INTER: \
  e = NIL; \
  gasp = true; \
  break; \
 case C_SOME: case C_UNSHARED_SOME: \
  e = none; \
  gasp = true; \
  break; \
 case C_ALL: \
  e = trueValue; \
  break; \
 case C_EXISTS: \
  e = falseValue; \
  break; \
}

#define END_ITER(gasp,e) \
switch (ik) \
{ \
 case C_LIST: case C_APPEND: \
  e = Reverse1(e,NIL); \
  break; \
 case C_UNSHARED_LIST: case C_UNSHARED_APPEND: \
    if (e!=NIL) { aux = CDR(e); SETCDR(e,NIL); e = aux; } \
    break; \
 case C_INTER: \
  if (gasp) \
    e = Raise(EMPTYEXC); \
  break; \
 case C_SOME: \
  if (!gasp) \
    e = Kons(some,e); \
  break; \
 case C_UNSHARED_SOME: \
  if (!gasp) \
    e = MakePair(E_CONS,some,e); \
  break; \
}

#define COMBINE_ITER(gasp,res,e) \
switch (ik) \
{ \
 case C_PROG: \
  break; \
 case C_LIST: \
  MAKECONS(res,e,aux); \
  e = aux; \
  break; \
 case C_UNSHARED_LIST: \
    EALLOC_HOMO(aux); aux->kind = (CHAR)E_CONS; CAR(aux) = res; \
    if (e==NIL) \
      { e = CDR(aux) = aux; } \
    else { CDR(aux) = CDR(e); SETCDR(e,aux); e = aux; } \
    break; \
 case C_SET: \
  e = MapAdd(res,UNIT,e); \
  break; \
 case C_APPEND: \
  e = Nreverse1(res,e); \
  break; \
 case C_UNSHARED_APPEND: \
    if (res!=NIL) e = AppendQueue(e,res); \
    break; \
 case C_UNION: \
  e = MapOverwrite(e,res); \
  break; \
 case C_UNION_UNDER: \
  e = MapOverwrite(res,e); \
  break; \
 case C_INTER: \
  if (gasp) \
    { \
      e = res; \
      gasp = false; \
    } \
  else \
    e = MapDRT(e,res); \
  break; \
 case C_MAP: \
  e = MapPutMapletTrie(res,e,(CODE)(1L<<NLOWBITS)); \
  break; \
 case C_MAP_UNDER: \
  e = MapPutMapletBehindTrie(res,e,(CODE)(1L<<NLOWBITS)); \
  break; \
 case C_ALL: \
  if (res==falseValue) \
    { \
      e = res; \
      goto end; \
    } \
  break; \
 case C_EXISTS: \
  if (res!=falseValue) \
    { \
      e = res; \
      goto end; \
    } \
  break; \
 case C_SOME: case C_UNSHARED_SOME: \
  gasp = false; \
  e = res; \
  goto end; \
}

PRIVATE ExprPtr EvalOf(f,p,domains,ik)
IN CONST ExprPtr f,p,domains;
IN IterKind ik;
{
#if DEBUGGING
#define NVARS 7
#else
#define NVARS 6
#endif
  ExprPtr locals,ll,lastll,list;
  ExprPtr aux;
  COUNTER nlocals;
  Boolean gasp,cont;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[NVARS]; } _x;

  nlocals = IVAL(CAR(domains));
  domains = CDR(domains);

  _x.n = NVARS;
  _x.next = theStackFrames;
  _x.p[1] = theEnvironment;
#if DEBUGGING!=0
  _x.p[NVARS-1] = currentDebugZone;
#endif
  INIT_ITER(gasp,_x.p[0]);
  _x.p[2] = _x.p[5] = NOEXPR;
  _x.p[4] = _x.p[3] = NIL;
  theStackFrames = (gcStackView *)&_x;

  _x.p[2] = EvalDomains(domains,&_x.p[3],&locals);
  for (ll = _x.p[2], lastll = NIL; ll!=NIL; ll=CDR(ll))
    {
      switch (ll->kind)
	{
	case E_PAT_RECORD_GNRL_CONS:
	case E_PAT_REVRECORD_GNRL_CONS:
	  list = CAR(ll);
	  if (list==NIL)
	    goto end;
makepair:
	  MAKEPAIR(ll->kind,list,NIL,aux);
	  break;
	case E_PAT_RECORD_MAPADD:
	  list = CAR(ll);
	  goto makepair;
	}
      if (lastll==NIL)
	_x.p[4] = lastll = aux;
      else
	{
	  SETCDR(lastll,aux);
	  lastll = aux;
	}
    }
  {
    INT size;
    VecPtr vp;

    size = VECSIZE(nlocals);
    CHECKHETEROFREE();
    scAlloc(*(char **)&vp,size);
    vp->n = nlocals;
    INITENV(vp,0,nlocals);
    MAKEENV_NOGC(vp,_x.p[1],aux);
    _x.p[1] = aux;
  }
  do {
    SETQ(theEnvironment,_x.p[1]);
#if DEBUGGING!=0
    SETQ(currentDebugZone,_x.p[NVARS-1]);
#endif
    if (!SetPatVals(_x.p[1],_x.p[3],_x.p[4],locals))
      {
	if (Eval0(p)!=falseValue)
	  {
	    SETQ(theEnvironment,_x.p[1]);
	    _x.p[5] = Eval0(f);
	    COMBINE_ITER(gasp,_x.p[5],_x.p[0]);
	  }
      }
    cont = IncOf(_x.p[4],_x.p[2]);
  } while (cont);
end:
#if DEBUGGING!=0
  SETQ(currentDebugZone,_x.p[NVARS-1]);
#endif
  END_ITER(gasp,_x.p[0]);
  theStackFrames = _x.next;
  return _x.p[0];
#undef NVARS
}

PRIVATE ExprPtr identity P((ExprPtr));
PRIVATE ExprPtr identity(e)
ExprPtr e;
{
  return e;
}

#define neutralProg tuple0
PRIVATE ExprPtr combineProg P((ExprPtr,ExprPtr));
PRIVATE ExprPtr combineProg(e1,e2)
IN CONST ExprPtr e1,e2;
{
  return tuple0;
}
PRIVATE ExprPtr promoteProg P((ExprPtr));
PRIVATE ExprPtr promoteProg(e)
IN CONST ExprPtr e;
{
  return e;
}

#define neutralList NIL
PRIVATE ExprPtr combineList P((ExprPtr,ExprPtr));
PRIVATE ExprPtr combineList(e1,e2)
IN CONST ExprPtr e1,e2;
{
  if (e1==NIL)
    ;
  else if (e2==NIL)
    e2 = e1;
  else
    {
      REGISTER ExprPtr aux1,aux2;

      aux1 = CDR(e1);
      aux2 = CDR(e2);
      SETCDR(e1,aux2);
      SETCDR(e2,aux1);
    }
  return e2;
}
PRIVATE ExprPtr promoteList P((ExprPtr));
PRIVATE ExprPtr promoteList(e)
IN CONST ExprPtr e;
{
  REGISTER ExprPtr aux;

  EALLOC_HOMO(aux);
  aux->kind = (CHAR)E_CONS;
  CAR(aux) = e;
  CDR(aux) = aux;
  return aux;
}

#define neutralSet NewMap()
PRIVATE ExprPtr combineSet P((ExprPtr,ExprPtr));
PRIVATE ExprPtr combineSet(e1,e2)
IN CONST ExprPtr e1,e2;
{
  return MapOverwrite(e1,e2);
}
PRIVATE ExprPtr promoteSet P((ExprPtr));
PRIVATE ExprPtr promoteSet(e)
IN CONST ExprPtr e;
{
  return MapletKons(e,UNIT);
}

#ifdef NO_APPEND_EVALSIMPLEOF
#define neutralAppend NIL
#define combineAppend combineList
PRIVATE ExprPtr promoteAppend(list)
IN CONST ExprPtr list;
{
  if (list==NIL)
    return NIL;
  {
    REGISTER ExprPtr e,end,last;

    BEGINQUICKKEEP();
    EALLOC_HOMO(e);
    e->kind = (CHAR)E_CONS;
    CAR(e) = CAR(list);
    CDR(e) = e;
    QUICKKEEP(e);
    end = e;
    for (; last = e, list = CDR(list), list!=NIL; )
      {
	EALLOC_HOMO(e);
	e->kind = (CHAR)E_CONS;
	CAR(e) = CAR(list);
	CDR(e) = end;
	SETCDR(last,e);
      }
    ENDQUICKKEEP();
    return e;
  }
}
#endif

#define neutralUnsharedAppend NIL
#define combineUnsharedAppend combineList
PRIVATE ExprPtr promoteUnsharedAppend(list)
IN CONST ExprPtr list;
{
  if (list==NIL)
    return NIL;
  {
    REGISTER ExprPtr e,end,last;

    BEGINQUICKKEEP();
    EALLOC_HOMO(e);
    e->kind = (CHAR)E_CONS;
    CAR(e) = CAR(list);
    CDR(e) = e;
    QUICKKEEP(e);
    end = e;
    for (; last = e, list = CDR(list), list!=NIL; )
      {
	EALLOC_HOMO(e);
	e->kind = (CHAR)E_CONS;
	CAR(e) = CAR(list);
	CDR(e) = end;
	SETCDR(last,e);
      }
    ENDQUICKKEEP();
    return e;
  }
}

#define neutralUnion NewMap()
#define combineUnion combineSet
#define promoteUnion identity

#define neutralUnionUnder NewMap()
PRIVATE ExprPtr combineUnionUnder P((ExprPtr,ExprPtr));
PRIVATE ExprPtr combineUnionUnder(e1,e2)
IN CONST ExprPtr e1,e2;
{
  return MapOverwrite(e2,e1);
}
#define promoteUnionUnder promoteUnion

#define neutralInter NIL
PRIVATE ExprPtr combineInter P((ExprPtr,ExprPtr));
PRIVATE ExprPtr combineInter(e1,e2)
IN CONST ExprPtr e1,e2;
{
  if (e1==NIL)
    return e2;
  if (e2==NIL)
    return e1;
  return MapDRT(e1,e2);
}
#define promoteInter identity

#define neutralMap NewMap()
#define combineMap combineSet
#define promoteMap identity

#define neutralMapUnder NewMap()
#define combineMapUnder combineUnionUnder
#define promoteMapUnder identity

#ifndef HAS_PROTOTYPES
PRIVATE ExprPtr EvalSimpleOf1 P((INOUT struct evalsimpleof *data));
#endif

MAPFOLDFUNC(do_evalsimpleof)
{
  struct evalsimpleof *data;

  data = MAPDATA(struct evalsimpleof *);
  SETQ(theEnvironment,data->saveenv);
  if (data->mapsweep)
    {
      if (MatchValue(theEnvironment,CAR(data->pat),x))
	return data->neutral;
      if (MatchValue(theEnvironment,CDR(data->pat),y))
	return data->neutral;
    }
  else
    {
      if (MatchValue(theEnvironment,data->pat,x))
	return data->neutral;
    }
  if (data->domains==NIL)
    {
      if ((data->locals==NIL || !EvalBindings(data->locals,theEnvironment)) &&
	  Eval0(data->p)!=falseValue)
	{
	  struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;

	  SETQ(theEnvironment,data->saveenv);
	  _x.n = 1;
	  _x.next = theStackFrames;
	  _x.p[0] = Eval0(data->f);
	  theStackFrames = (gcStackView *)&_x;
	  _x.p[0] = (*data->promote)(_x.p[0]);
	  theStackFrames = _x.next;
	  return _x.p[0];
	}
      return data->neutral;
    }
  return EvalSimpleOf1(data);
}

DEF_PRIVATE_ACTION(evalsimof)
{
  return EvalSimpleOf1((struct evalsimpleof *)env);
}

PRIVATE ExprPtr EvalSimpleOf1k(data)
INOUT struct evalsimpleof *data;
{
  ExprPtr e;
  ENTER;

  e = CallCC(ACTION_NAME(evalsimof),(ExprPtr)data,true);
  QUIT;
  return e;
}

PRIVATE ExprPtr EvalSimpleOf1(data)
INOUT struct evalsimpleof *data;
{
  ExprPtr domains,dompats;
  Boolean savemapsweep;

  if (STKFULL(savemapsweep))
    return EvalSimpleOf1k(data);
  domains = data->domains;
  dompats = data->dompats;

  switch (dompats->kind)
    {
    case E_NIL:
      return data->neutral;
    case E_PAT_RECORD_GNRL_CONS: /* in list domain descriptor */
      {
	ExprPtr list,savepat;
	struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;

	_x.n = 2;
	_x.next = theStackFrames;
	_x.p[0] = _x.p[1] = NOEXPR;
	theStackFrames = (gcStackView *)&_x;

	list = CAR(domains);
	if (list==NIL)
	  _x.p[0] = data->neutral;
	else
	  {
	    data->domains = CDR(domains);
	    savepat = data->pat;
	    savemapsweep = data->mapsweep;
	    data->dompats = CDR(dompats);
	    data->pat = CAR(CAR(dompats));
	    data->mapsweep = false;
	    _x.p[0] = do_evalsimpleof(CAR(list),UNIT,(CHAR *)data);
	    while (list = CDR(list), list!=NIL)
	      {
		_x.p[1] = do_evalsimpleof(CAR(list),UNIT,(CHAR *)data);
		_x.p[0] = (*data->combine)(_x.p[0],_x.p[1]);
		_x.p[1] = NOEXPR;
	      }
	    data->mapsweep = savemapsweep;
	    data->pat = savepat;
	    data->domains = domains;
	  data->dompats = dompats;
	  }
	theStackFrames = _x.next;
	return _x.p[0];
      }
    case E_PAT_REVRECORD_GNRL_CONS: /* in map domain descriptor */
      {
	ExprPtr map,savepat,e;

	/* no need to build a _x structure, since the result of MapFold() below won't be GC'ed */
	map = CAR(domains);
	if (map==NewMap())
	  e = data->neutral;
	else
	  {
	    data->domains = CDR(domains);
	    savepat = data->pat;
	    savemapsweep = data->mapsweep;
	    data->dompats = CDR(dompats);
	    data->pat = CAR(CAR(dompats));
	    data->mapsweep = true;
	    e = MapFold(map,do_evalsimpleof,data,data->neutral,data->combine);
	    data->mapsweep = savemapsweep;
	    data->pat = savepat;
	    data->domains = domains;
	    data->dompats = dompats;
	  }
	return e;
      }
    case E_PAT_RECORD_MAPADD: /* sub map domain descriptor */
      {
	ExprPtr savepat;
	struct _x { COUNTER n; gcStackView *next; ExprPtr p[4]; } _x;


	_x.n = 4;
	_x.next = theStackFrames;
	_x.p[0] = _x.p[1] = _x.p[2] = _x.p[3] = NOEXPR;
	_x.p[1] = InitialSubMap1(CAR(domains));
	theStackFrames = (gcStackView *)&_x;

	data->domains = CDR(domains);
	savepat = data->pat;
	data->pat = CAR(CAR(dompats));
	data->dompats = CDR(dompats);
	do {
	  SETQ(theEnvironment,data->saveenv);
	  _x.p[2] = MakeSubMap(_x.p[1]);
	  if (!MatchValue(theEnvironment,data->pat,_x.p[2]))
	    {
	      if (data->domains==NIL)
		{
		  if ((data->locals==NIL || !EvalBindings(data->locals,theEnvironment)) &&
		      Eval0(data->p)!=falseValue)
		    {
		      SETQ(theEnvironment,data->saveenv);
		      _x.p[3] = Eval0(data->f);
		      _x.p[3] = (*data->promote)(_x.p[3]);
		      _x.p[0] = (*data->combine)(_x.p[0],_x.p[3]);
		    }
		}
	      else
		{
		  _x.p[3] = EvalSimpleOf1(data);
		  _x.p[0] = (*data->combine)(_x.p[0],_x.p[3]);
		}
	    }
	} while (IncSubMap(_x.p[1]));
	data->pat = savepat;
	data->domains = domains;
	data->dompats = dompats;
	theStackFrames = _x.next;
	return _x.p[0];
      }
    default:
      kinderror(message(463),dompats->kind);
      /*NOTREACHED*/
      return NIL;
    }
}

PRIVATE ExprPtr EvalSimpleOf(f,p,domains,ik)
IN CONST ExprPtr f,p,domains;
IN IterKind ik;
{
#if DEBUGGING
#define NVARS 3
#else
#define NVARS 2
#endif
  COUNTER nlocals;
  VecPtr vp;
  long size;
  ExprPtr e,aux;
  struct evalsimpleof *data = (struct evalsimpleof *)alloctmp(sizeof(struct evalsimpleof));
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[NVARS]; } _x;

  _x.n = NVARS;
  _x.next = theStackFrames;
  _x.p[0] = theEnvironment;
  _x.p[1] = NOEXPR;
#if DEBUGGING
  _x.p[NVARS-1] = currentDebugZone;
#endif
  theStackFrames = (gcStackView *)&_x;

  data->saveenv = _x.p[0];
  data->dompats = CDR(domains);
  data->domains = _x.p[1] = EvalDomainsSimple(data->dompats,&data->locals);
  nlocals = IVAL(CAR(domains));
  size = VECSIZE(nlocals);
  CHECKHETEROFREE();
  scAlloc(*(char **)&vp,size);
  vp->n = nlocals;
  INITENV(vp,0,nlocals);
  MAKEENV_NOGC(vp,data->saveenv,aux);
  data->saveenv = _x.p[0] = aux;
  SETQ(theEnvironment,aux);
  data->f = f;
  data->p = p;
  switch (ik)
    {
    case C_PROG:
      data->neutral = neutralProg;
      data->combine = combineProg;
      data->promote = promoteProg;
      break;
    case C_LIST:
    case C_UNSHARED_LIST:
      data->neutral = neutralList;
      data->combine = combineList;
      data->promote = promoteList;
      break;
    case C_SET:
      data->neutral = neutralSet;
      data->combine = combineSet;
      data->promote = promoteSet;
      break;
#ifdef NO_APPEND_EVALSIMPLEOF
    case C_APPEND:
      data->neutral = neutralAppend;
      data->combine = combineAppend;
      data->promote = promoteAppend;
      break;
#endif
    case C_UNSHARED_APPEND:
      data->neutral = neutralUnsharedAppend;
      data->combine = combineUnsharedAppend;
      data->promote = promoteUnsharedAppend;
      break;
    case C_UNION:
      data->neutral = neutralUnion;
      data->combine = combineUnion;
      data->promote = promoteUnion;
      break;
    case C_UNION_UNDER:
      data->neutral = neutralUnionUnder;
      data->combine = combineUnionUnder;
      data->promote = promoteUnionUnder;
      break;
    case C_INTER:
      data->neutral = neutralInter;
      data->combine = combineInter;
      data->promote = promoteInter;
      break;
    case C_MAP:
      data->neutral = neutralMap;
      data->combine = combineMap;
      data->promote = promoteMap;
      break;
    case C_MAP_UNDER:
      data->neutral = neutralMapUnder;
      data->combine = combineMapUnder;
      data->promote = promoteMapUnder;
      break;
    }
  e = EvalSimpleOf1(data);
#if DEBUGGING
  SETQ(currentDebugZone,_x.p[NVARS-1]);
#endif
  theStackFrames = _x.next;
  return e;
#undef NVARS
}

PRIVATE ExprPtr EvalMap(f,p,domains,ik)
IN CONST ExprPtr f,p,domains;
IN IterKind ik;
{
  ExprPtr locals,ll,aux;
#if DEBUGGING
#define NVARS 6
#else
#define NVARS 5
#endif
  Boolean cont;
  COUNTER nlocals;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[NVARS]; } _x;
  Boolean gasp;

  nlocals = IVAL(CAR(domains));
  domains = CDR(domains);

  _x.n = NVARS;
  _x.next = theStackFrames;
  _x.p[1] = theEnvironment;
  _x.p[2] = _x.p[4] = NOEXPR;
  _x.p[3] = NIL;
  INIT_ITER(gasp,_x.p[0]);
#if DEBUGGING
  _x.p[NVARS-1] = currentDebugZone;
#endif
  theStackFrames = (gcStackView *)&_x;

  _x.p[2] = EvalDomains(domains,&_x.p[3],&locals);
  if (_x.p[2]!=NIL)
    {
      INT len,len2;
      Boolean pow2;

      len = ListLength(CAR(_x.p[2]));
      ll = CDR(_x.p[2]);
      switch (_x.p[2]->kind)
	{
	case E_PAT_RECORD_GNRL_CONS: /* in list */
	case E_PAT_REVRECORD_GNRL_CONS: /* in map */
	  pow2 = false;
	  break;
	case E_PAT_RECORD_MAPADD: /* sub map */
	  pow2 = true;
	  break;
	}
      for (; ll!=NIL; ll=CDR(ll))
	{
	  switch (ll->kind)
	    {
	    case E_PAT_RECORD_GNRL_CONS: /* in list */
	    case E_PAT_REVRECORD_GNRL_CONS: /* in map */
	      len2 = ListLength(CAR(ll));
	      if (pow2)
		{
		  if (len2==0 || (1L<<len)!=len2)
		    {
#if DEBUGGING
		      SETQ(currentDebugZone,_x.p[NVARS-1]);
#endif
		      _x.p[0] = Raise(MAPEXC);
		      theStackFrames = _x.next;
		      return _x.p[0];
		    }
		}
	      else if (len!=len2)
		{
#if DEBUGGING
		  SETQ(currentDebugZone,_x.p[NVARS-1]);
#endif
		  _x.p[0] = Raise(MAPEXC);
		  theStackFrames = _x.next;
		  return _x.p[0];
		}
	      break;
	    case E_PAT_RECORD_MAPADD:
	      len2 = ListLength(CAR(ll));
	      if (pow2)
		{
		  if (len!=len2)
		    {
#if DEBUGGING
		      SETQ(currentDebugZone,_x.p[NVARS-1]);
#endif
		      _x.p[0] = Raise(MAPEXC);
		      theStackFrames = _x.next;
		      return _x.p[0];
		    }
		}
	      else if (len==0 || (1L<<len2)!=len)
		{
#if DEBUGGING
		  SETQ(currentDebugZone,_x.p[NVARS-1]);
#endif
		  _x.p[0] = Raise(MAPEXC);
		  theStackFrames = _x.next;
		  return _x.p[0];
		}
	      break;
	    }
	}
      if (len==0)
	goto end;
    }
  {
    INT size;
    VecPtr vp;

    size = VECSIZE(nlocals);
    CHECKHETEROFREE();
    scAlloc(*(char **)&vp,size);
    vp->n = nlocals;
    INITENV(vp,0,nlocals);
    MAKEENV_NOGC(vp,_x.p[1],aux);
    _x.p[1] = aux;
  }
  do {
#if DEBUGGING
    SETQ(currentDebugZone,_x.p[NVARS-1]);
#endif
    SETQ(theEnvironment,_x.p[1]);
    if (!SetPatVals(_x.p[1],_x.p[3],_x.p[2],locals))
      {
	SETQ(theEnvironment,_x.p[1]);
	if (Eval0(p)!=falseValue)
	  {
	    SETQ(theEnvironment,_x.p[1]);
	    _x.p[4] = Eval0(f);
	    COMBINE_ITER(gasp,_x.p[4],_x.p[0]);
	  }
      }
    if (_x.p[2]==NIL)
      cont = false;
    else
      {
	switch (_x.p[2]->kind)
	  {
	  case E_PAT_RECORD_GNRL_CONS: /* in list */
	  case E_PAT_REVRECORD_GNRL_CONS: /* in map */
	    aux = CDR(CAR(_x.p[2]));
	    SETCAR(_x.p[2],aux);
	    cont = (CAR(_x.p[2])!=NIL);
	    break;
	  case E_PAT_RECORD_MAPADD:
	    cont = IncSubMap(CAR(_x.p[2]));
	    break;
	  }
	ll = CDR(_x.p[2]);
	for (; ll!=NIL; ll = CDR(ll))
	  switch (ll->kind)
	    {
	    case E_PAT_RECORD_GNRL_CONS: /* in list */
	    case E_PAT_REVRECORD_GNRL_CONS: /* in map */
	      aux = CDR(CAR(ll));
	      SETCAR(ll,aux);
	      break;
	    case E_PAT_RECORD_MAPADD:
	      (void) IncSubMap(CAR(ll));
	      break;
	    }
      }
  } while (cont);
end:
#if DEBUGGING
  SETQ(currentDebugZone,_x.p[NVARS-1]);
#endif
  END_ITER(gasp,_x.p[0]);
  theStackFrames = _x.next;
  return _x.p[0];
#undef NVARS
}

PRIVATE ExprPtr EvalWhile(cond,suchthat,body,ik)
IN CONST ExprPtr cond,suchthat,body;
IN IterKind ik;
{
#if DEBUGGING
#define NVARS 4
#else
#define NVARS 3
#endif
  ExprPtr aux;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[NVARS]; } _x;
  Boolean gasp;

  _x.n = NVARS;
  _x.next = theStackFrames;
  INIT_ITER(gasp,_x.p[0]);
  _x.p[1] = theEnvironment;
  _x.p[2] = NOEXPR;
#if DEBUGGING
  _x.p[NVARS-1] = currentDebugZone;
#endif
  theStackFrames = (gcStackView *)&_x;

  while (true)
    {
      SETQ(theEnvironment,_x.p[1]);
      if (Eval0(cond)==falseValue)
	break;
      if (Eval0(suchthat)!=falseValue)
	{
	  SETQ(theEnvironment,_x.p[1]);
	  _x.p[2] = Eval0(body);
	  COMBINE_ITER(gasp,_x.p[2],_x.p[0]);
	}
    }
end:
#if DEBUGGING
  SETQ(currentDebugZone,_x.p[NVARS-1]);
#endif
  END_ITER(gasp,_x.p[0]);
  theStackFrames = _x.next;
  return _x.p[0];
#undef NVARS
}

/*** Exception handling ***/

PRIVATE ExprPtr HandleEval(e,handlers)
IN CONST ExprPtr e,handlers;
{
#if DEBUGGING
#define NVARS 2
#else
#define NVARS 1
#endif
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[NVARS]; } _x;
  excHandler *xh;
#if PROFILING & PROFILING_PROPER_TIMES
  ExprPtr saveCurrentProf = currentProf;
#endif
  ENTER;

  xh = (excHandler *)alloctmp(sizeof(excHandler));
  xh->next = excStack;
  xh->tag = X_HANDLER;
  xh->data = theEnvironment;
  xh->code = handlers;

  _x.n = NVARS;
  _x.next = theStackFrames;
  _x.p[0] = NOEXPR;
#if DEBUGGING
  _x.p[NVARS-1] = getApplStk();
#endif
  theStackFrames = (gcStackView *)&_x;
  switch (SETJUMP(xh->jb))
    {
    case 0: /* normal return, push handler and evaluate body */
      excStack = xh;
      _x.p[0] = NOEXPR;
      _x.p[0] = Eval0(e);
      excStack = xh->next; /* no need to setHandlerStack(), if Eval0() returns normally,
			      then it has not changed excStack. */
      break;
    default: /* an exception has been raised,
		and the handler body has not
		been evaluated, but theEnvironment
		has been set to the correct value. */
      setHandlerStack(xh->next);
#if DEBUGGING
      setApplStk(_x.p[NVARS-1]);
#endif
#if PROFILING & PROFILING_PROPER_TIMES
      syncProfProper(saveCurrentProf);
#endif
      _x.p[0] = NOEXPR;
      QUIT;
      break;
    }
  theStackFrames = _x.next;
  return _x.p[0];
#undef NVARS
}

PUBLIC ExprPtr Raise(exception)
IN CONST ExprPtr exception;
{
  excHandler *hstack;
  ExprPtr handlers,h;
  stackDesc *jb;
  ExprPtr oldCont;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;
#if DEBUGGING
  Boolean step = false;
#endif

  _x.n = 1;
  _x.next = theStackFrames;
  _x.p[0] = theEnvironment;
  theStackFrames = (gcStackView *)&_x;

  oldCont = NIL;
  for (hstack = excStack; hstack!=NULL; hstack = NEXTEXCHANDLER(hstack))
    {
      if (hstack->tag==X_CONT)
	{
	  oldCont = hstack->data; /* we shall have to reestablish an
				     old stack before raising the exception */
	  continue;
	}
#if PROFILING
      if (hstack->tag==X_PROF)
	continue;
#endif
#if DEBUGGING
      if (hstack->tag==X_STEP)
	{
	  step = true;
	  continue;
	}
#endif
      jb = &hstack->jb;
      handlers = hstack->code;
      if (handlers==NIL) /* if code is NIL, then we should directly longjmp() to the handler,
		       which is compiled inside a function and is responsible for the
		       pattern matching. */
	{
	  THEEXPR = exception; /* to be tested by the handler itself,
				  which is responsible for reraising the exception
				  if it does not match */
#if DEBUGGING
	  if (step)
	    stepMode = resumeStepMode; /* see debugger() */
#endif
	  if (oldCont==NIL)
	    LONGJUMP(*jb,1);
	  /* otherwise, reestablish old stack and raise exception
	     in this stack */
	  Throw1(oldCont,exception,jb);
	  /*NOTREACHED*/
	}
      SETQ(theEnvironment,hstack->data); /* get back saved theEnvironment */
      /* otherwise, handlers are E_POLYFUNs (interpreted handlers). */
      /*if (handlers->kind==(CHAR)E_POLYFUN)*/
      {
	patChoiceBody = CDR(COUNTEDVAL(handlers));
	if (MakePatVal(CAR(COUNTEDVAL(handlers)),exception,(COUNTER)COUNTEDN(handlers)))
	  {
	    h = patChoiceBody;
	    THEEXPR = h;
	    /* we can leave the evaluation of the handler body
	       to HandleEval(), this will spare some of the C stack. */
#if DEBUGGING
	    if (step)
	      stepMode = resumeStepMode; /* see debugger() */
#endif
	    if (oldCont==NIL)
	      LONGJUMP(*jb,1); /* -1 means "not evaluated" */
	    /* otherwise, reestablish old stack and raise exception
	       in this stack */
	    Throw1(oldCont,h,jb);
	    /*NOTREACHED*/
	  }
      }
    }
  SETQ(theEnvironment,_x.p[0]);
  theStackFrames = _x.next;
  return Error(exception);
}

PUBLIC NORETURN void Toplevel(errcond)
IN CONST ExprPtr errcond;
{
  excHandler *handler;
#if PROFILING
#if OS==AMIGA_OS
  unsigned int clock[2];
#define TMS clock

  (void) timer(clock);
#else
  struct tms tms;
#define TMS &tms

  times(&tms);
#endif

#if PROFILING & PROFILING_PROPER_TIMES
  if (currentProf!=NOEXPR)
    {
      profExitProper(PROFDATA(currentProf),TMS);
      currentProf = NOEXPR;
    }
#endif
  for (handler=excStack; handler!=NULL; handler=NEXTEXCHANDLER(handler))
    {
      if (handler->tag==X_PROF)
	profExit(handler->data,TMS);
    }
#undef TMS
#endif
  THEEXPR = NOEXPR;
#if DEBUGGING
  stepMode = resumeStepMode = STEP_NO;
  initApplStk();
#endif
  if (toplevelCont!=NOEXPR && toplevelCont!=NIL)
    Throw(toplevelCont,errcond);
  berror(message(397));
  /*NOTREACHED*/
}

PRIVATE Boolean Let(bindings)
IN CONST ExprPtr bindings;
{
  REGISTER COUNTER n;
  long size;
  VecPtr vp;
  ExprPtr aux;
#if DEBUGGING
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[1]; } _x;
#endif
  Boolean nomatch;

#if DEBUGGING
  _x.n = 1;
  _x.next = theStackFrames;
  _x.p[0] = currentDebugZone;
  theStackFrames = (gcStackView *)&_x;
#endif
  n = IVAL(CAR(bindings));
  if (n<0) /* then this is an imported module, and we allocate a fake copy of GLOBENV.
	      It is fake because its EVARS is not really copied; this is ok, because
	      this is a special case of which the GC is aware. */
    {
      CONSENV(EVARS(GLOBENV),aux);
      n = -1-n;
    }
  size = VECSIZE(n);
  CHECKHETEROFREE();
  scAlloc(*(char **)&vp,size);
  vp->n = n;
  /*INITENV(vp,0,n) done in EvalBindingsRC */
  CONSENV_NOGC(vp,aux);
  nomatch = EvalBindingsRC(CDR(bindings),theEnvironment,(COUNTER)0,n);
#if DEBUGGING
  SETQ(currentDebugZone,_x.p[0]);
  theStackFrames = _x.next;
#endif
  return nomatch;
}

PRIVATE ExprPtr EvalTuple(e,shared)
IN CONST ExprPtr e;
IN Boolean shared;
{
#if !STOP_AND_COPY_VECTORS
  ExprPtr *ep;
#endif
  REGISTER COUNTER i,n;
  REGISTER gcStackView *_x;
  REGISTER ExprPtr *VOLATILE epout;

  n = AVAL(e)->n;
#if !STOP_AND_COPY_VECTORS
  ep = AVAL(e)->expr;
#endif
  _x = (gcStackView *)alloctmp(VARSIZE(gcStackView,p,1+n));
  epout = &_x->p[1];

  _x->n = 1;
  _x->next = theHeapFrames;
  theHeapFrames = _x;
  _x->p[0] = theEnvironment;

  *epout++ = Eval0(
#if STOP_AND_COPY_VECTORS
		   AVAL(e)->expr[0]
#else
		   *ep++
#endif
		   );
  _x->n++;
  for (i=1; i<n; i++)
    {
      SETQ(theEnvironment,_x->p[0]);
      *epout++ = Eval0(
#if STOP_AND_COPY_VECTORS
		       AVAL(e)->expr[i]
#else
		       *ep++
#endif
		       );
      _x->n++;
    }
  _x->p[0] = shared?KtupleGnrl(n,&_x->p[1]):
    MakeRecord(E_TUPLE_GNRL,NIL,n,&_x->p[1]);

  theHeapFrames = _x->next;
  return _x->p[0];
}

PRIVATE ExprPtr EvalRecord(e,shared)
IN CONST ExprPtr e; /* e is not actually a record:
		       in the array.shape field, we find a
		       record_gnrl(shape,indices), where indices is an array
		       of E_INTs to indicate the place where successive values
		       have to be put. */
IN Boolean shared;
{
#if !STOP_AND_COPY_VECTORS
  ExprPtr *ep,*index;
#endif
  REGISTER COUNTER i,n;
  REGISTER gcStackView *_x;
  ExprPtr shape;
  REGISTER ExprPtr *epout;

  n = AVAL(e)->n;
  _x = (gcStackView *)alloctmp(VARSIZE(gcStackView,p,1+n));

  _x->n = 1;
  _x->next = theHeapFrames;
  theHeapFrames = (gcStackView *)_x;
  _x->p[0] = theEnvironment;
  for (i=0, epout = &_x->p[1]; i<n; i++)
    *epout++ = NOEXPR;
  epout = &_x->p[0];

  shape = ASHAPE(e);
#if STOP_AND_COPY_VECTORS
  epout[IVAL(AVAL(shape)->expr[0])] = Eval0(AVAL(e)->expr[0]);
#else
  index = AVAL(shape)->expr;
  ep = AVAL(e)->expr;
  epout[IVAL(*index++)] = Eval0(*ep++);
#endif
  _x->n++;
  for (i=1; i<n; i++)
    {
      SETQ(theEnvironment,_x->p[0]);
#if STOP_AND_COPY_VECTORS
      epout[IVAL(AVAL(shape)->expr[i])] = Eval0(AVAL(e)->expr[i]);
#else
      epout[IVAL(*index++)] = Eval0(*ep++);
#endif
      _x->n++;
    }
  _x->p[0] = shared?KrecordGnrl(ASHAPE(shape),n,&_x->p[1]):
    MakeRecord(E_RECORD_GNRL,ASHAPE(shape),n,&_x->p[1]);

  theHeapFrames = _x->next;
  return _x->p[0];
}

#if DEBUGGING!=0
EXTERN Boolean stopwanted;
EXTERN ExprPtr stepFile;
EXTERN COUNTER stepLine,stepPos;

PRIVATE ExprPtr BreakOnEntry(dz)
IN CONST ExprPtr dz;
{
  return debugger(dz,specialMarker,1);
}

PRIVATE ExprPtr BreakOnExit(dz,defaultval)
IN CONST ExprPtr dz;
IN CONST ExprPtr defaultval;
{
  return debugger(dz,defaultval,2);
}

PUBLIC ExprPtr Break(defaultval)
IN CONST ExprPtr defaultval;
{
  return debugger(currentDebugZone,defaultval,3);
}

#ifdef BUGGY
PRIVATE Boolean checkGuard(info,val,flag)
IN CONST ExprPtr info,val;
IN COUNTER flag;
{
  ExprPtr grdfun;
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[2]; } _x;
  Boolean t;

  if (!(TYPE(info) & flag))
    return false;
  grdfun = CAR(ARGS(info));
  grdfun = (flag & DZ_ACTIVE_ON_ENTRY)?CAR(grdfun):CDR(grdfun);

  _x.n = 1;
  _x.next = theStackFrames;
  _x.p[0] = MakeAppl(grdfun,val);
  theStackFrames = (gcStackView *)&_x;

  MAKECONS(izero,NIL,_x.p[1]);
  _x.n = 2;
  _x.p[0] = MakeLet(_x.p[1],_x.p[0]);
  _x.p[0] = MakeLet(_x.p[1],_x.p[0]);
  /* add two empty lets on top of non-closed
     expression, so as to simulate the extra
     level we got when building the grdfun
     function, under the debugger (one comes from the level of
     the fake _ variable, used for printing types and results
     under the debugger; the other comes from the fact that
     the compiler adds one level on its own) */
  t = (Eval0(_x.p[0])==trueValue);
  theStackFrames = _x.next;
  return t;
}
#endif
#endif

DEF_ACTION(datacon_unshared)
{
  ExprPtr res;

  MAKECONS(env,arg,res);
  return res;
}

DEF_ACTION(datacon_shared)
{
  ExprPtr res;

  res = Kons(env,arg);
  return res;
}

PRIVATE ExprPtr currentFunction = NOEXPR;
#if DEBUGGING
PUBLIC ExprPtr currentCallingEnv = NOEXPR;
/* just used as communication channel between apply_polyfun_closure or
   the E_APPL/E_GOTO case in Eval0() and the E_DEBUGFUNC case, which
   must push some info on the application stack. */
#endif

DEF_ACTION(apply_polyfun_closure)
{ /* a polyfun is a counted reference (refval+counter);
     the refval is a couple (pattern,primary body); the counter is
     the max number of slots in a local environment for all
     possible bodies. */
  ExprPtr polyfun;

#if DEBUGGING
  currentCallingEnv = theEnvironment;
#endif
  SETQ(theEnvironment,CDR(env));
  env = CAR(env);
  polyfun = COUNTEDVAL(env);
  patChoiceBody = CDR(polyfun);
  if (!MakePatVal(CAR(polyfun),arg,(COUNTER)COUNTEDN(env)))
    {
#if DEBUGGING
      if (!debugMode)
	PUSHAPPLSTK(env,theEnvironment,currentCallingEnv,currentDebugZone);
#endif
      arg = Raise(MATCHEXC);
    }
  else
    {
#if DEBUGGING
      currentFunction = env;
#endif
      arg = Eval0(patChoiceBody);
    }
  return arg;
}

DEF_PRIVATE_ACTION(apply_memo_fn)
{
  ExprPtr res,f;
  Boolean found;

  res = MapGet(arg,CAR(env),&found);
  if (!found)
    {
      ENTER;

      f = CDR(env);
      res = (*ACTION(f))(arg,QENV(f));
      (void) KEEP(res);
      setcarMapPut(arg,res,env);
      QUIT;
    }
  return res;
}

PUBLIC void mergeFieldLists(fields1,fields2,newfields,m,n,np)
     /* merge lists fields1 and fields2, putting result into newfields,
	returning length of newfields in *np, and replacing entries
	in fields1 and fields2 by the indexes to the new positions
	of fields inside newfields. */
INOUT REGISTER ExprPtr *fields1,*fields2;
OUT REGISTER ExprPtr *newfields;
IN REGISTER INT m,n;
OUT INT *np;
{
  INT pos;

  pos = 0;
  if (m==0)
    {
copyrest:
      while (n--!=0)
	{
	  *newfields++ = *fields2;
	  *fields2++ = (ExprPtr)pos++;
	}
      *np = pos;
      return;
    }
  if (n==0)
    {
      n = m;
      fields2 = fields1;
      goto copyrest;
    }
again:
  if (m==1)
    { /* insert *fields1 at a position we find by dichotomy inside fields2 */
      ExprPtr s1,s2;
      REGISTER INT i,j,k;

      s1 = *fields1;
      i = n>>1;
      j = 0;
      k = n;
      while (j!=k)
	{
	  s2 = fields2[i];
	  switch (CompareStrings(s1,s2))
	    {
	    case -1: k = i; break;
	    case 1: j = i+1; break;
	    case 0:
	      *fields1 = (ExprPtr)-1;
	      goto copyrest;
	    }
	  i = (j+k)>>1;
	}
      for (i=0; i<j; i++)
	{
	  *newfields++ = *fields2;
	  *fields2++ = (ExprPtr)pos++;
	}
      *newfields++ = *fields1; /* insert element */
      *fields1 = (ExprPtr)pos++;
      for (; i<n; i++)
	{
	  *newfields++ = *fields2;
	  *fields2++ = (ExprPtr)pos++;
	}
      *np = pos;
      return;
    }
  if (n==1)
    {
      ExprPtr s1,s2;
      REGISTER INT i,j,k;

      s2 = *fields2;
      i = m>>1;
      j = 0;
      k = m;
      while (j!=k)
	{
	  s1 = fields1[i];
	  switch (CompareStrings(s2,s1))
	    {
	    case -1: k = i; break;
	    case 1: j = i+1; break;
	    case 0:
	      for (j=0; j<i; j++)
		{
		  *newfields++ = *fields1;
		  *fields1++ = (ExprPtr)pos++;
		}
	      *newfields++ = s1; /* insert element */
	      *fields1++ = (ExprPtr)-1;
	      *fields2 = (ExprPtr)pos++;
	      for (; ++j<m;)
		{
		  *newfields++ = *fields1;
		  *fields1++ = (ExprPtr)pos++;
		}
	      *np = pos;
	      return;
	    }
	  i = (j+k)>>1;
	}
      for (i=0; i<j; i++)
	{
	  *newfields++ = *fields1;
	  *fields1++ = (ExprPtr)pos++;
	}
      *newfields++ = *fields2; /* insert element */
      *fields2 = (ExprPtr)pos++;
      for (; i<m; i++)
	{
	  *newfields++ = *fields1;
	  *fields1++ = (ExprPtr)pos++;
	}
      *np = pos;
      return;
    }
  { /* otherwise, advance in parallel in boths lists */
    ExprPtr s1,s2;

    s1 = *fields1;
    s2 = *fields2;
    switch (CompareStrings(s1,s2))
      {
      case -1:
	*newfields++ = s1;
	*fields1++ = (ExprPtr)pos++; m--;
	break;
      case 0:
	*newfields++ = s1;
	*fields1++ = (ExprPtr)-1; m--;
	*fields2++ = (ExprPtr)pos++; n--;
	break;
      case 1:
	*newfields++ = s2;
	*fields2++ = (ExprPtr)pos++; n--;
	break;
      }
  }
  goto again;
}

MAPSWEEPFUNC(do_dumpshape)
{
  ExprPtr *fields;

  fields = MAPDATA(ExprPtr *);
  fields[IVAL(y)-1] = x;
  return false;
}

PUBLIC void dumpShape(shape,fields)
IN CONST ExprPtr shape;
OUT ExprPtr *fields;
{
  (void) MapSweep(shape,do_dumpshape,fields);
}

PRIVATE MapPtr undumpShape1(fields,m,n)
IN REGISTER CONST ExprPtr *fields;
IN REGISTER INT m,n;
{
  if (n==0)
    return NewMap();
  if (n==1)
    {
      ExprPtr e;

      BEGINQUICKKEEP();
      e = Kint(m);
      QUICKKEEP(e);
      e = MapletKons(fields[0],e);
      ENDQUICKKEEP();
      return e;
    }
  {
    struct _x { COUNTER n; gcStackView *next; ExprPtr p[2];} _x;
    INT nn = n>>1;

    _x.p[0] = undumpShape1(fields,m,nn);
    _x.n = 1;
    _x.next = theStackFrames;
    theStackFrames = (gcStackView *)&_x;
    _x.p[1] = undumpShape1(fields+nn,m+nn,n-nn);
    _x.n = 2;
    _x.p[0] = MapOverwrite(_x.p[0],_x.p[1]);
    theStackFrames = _x.next;
    return _x.p[0];
  }
}

PUBLIC MapPtr undumpShape(fields,n)
IN CONST ExprPtr *fields;
IN INT n;
{
  MapPtr shape;

  shape = undumpShape1(fields,(INT)1,n);
  (void) KEEP(shape);
  return shape;
}

PUBLIC ExprPtr mu1(rec,modlab,modval,share)
IN CONST ExprPtr rec;
IN CONST ExprPtr modlab,modval;
IN Boolean share;
/* note that mu1() does not keep its result!
   This is done to respect the run-time scheme for compiled code,
   which keeps as few things as possible. */
{
  ENTER;

  switch (rec->kind)
    {
    case E_RECORD_0:
buildmod:
      if (share)
	rec = Krecord1(modlab,modval);
      else
	{
	  MAKERECORD1(modlab,modval,rec);
	}
      break;
    case E_RECORD_1:
      if (CAR(rec)==modlab) /* if same field, return mod */
	goto buildmod;
      { /* otherwise, build two-field record */
	ExprPtr shape;
	ExprPtr exprs[2];
	  
	if (CompareStrings(CAR(rec),modlab)<0)
	  {
	    shape = MapPair(CAR(rec),ione,modlab,itwo);
	    exprs[0] = CDR(rec);
	    exprs[1] = modval;
	  }
	else
	  {
	    shape = MapPair(modlab,ione,CAR(rec),itwo);
	    exprs[0] = modval;
	    exprs[1] = CDR(rec);
	  }
	rec = share?KrecordGnrl(shape,2,exprs):MakeRecord(E_RECORD_GNRL,shape,2,exprs);
	break;
      }
    case E_RECORD_GNRL:
      {
	ExprPtr shape;
	INT i,n,nn;
	ExprPtr *ep,*epp;

	shape = ASHAPE(rec);
	n = AVAL(rec)->n;
	{
	  ExprPtr *fields,*outfields,*exprs;

	  fields = (ExprPtr *)alloctmp((3*n+2)*sizeof(ExprPtr));
	  outfields = fields + n; /* fields has n elements */
	  exprs = outfields+1+n; /* outfields may have up to n+1 elements */
	  /* and exprs will have up to n+1 elements */
	  dumpShape(shape,fields);
	  mergeFieldLists(fields,&modlab,outfields,n,(INT)1,&nn);
	  shape = undumpShape1(outfields,(INT)1,nn);
	  (void) KEEP(shape);

	  for (i=0, ep=AVAL(rec)->expr, epp=exprs; i<(INT)modlab; i++)
	    *epp++ = *ep++;
	  *epp++ = modval; i++;
	  if (nn==n)
	    ep++;
	  for (; i<nn; i++)
	    *epp++ = *ep++;
	  rec = share?KrecordGnrl(shape,nn,exprs):MakeRecord(E_RECORD_GNRL,shape,nn,exprs);
	}
	break;
      }
    }
  QUIT;
  return rec;
}

PUBLIC ExprPtr muGnrl(rec,mod,share)
IN CONST ExprPtr rec;
IN CONST ExprPtr mod; /* mod is a E_GNRL_RECORD */
IN Boolean share;
/* note that muGnrl() does not keep its result!
   This is done to respect the run-time scheme for compiled code,
   which keeps as few things as possible. */
{
  ENTER;

  switch (rec->kind)
    {
    case E_RECORD_0:
      rec = mod;
      break;
    case E_RECORD_1:
      {
	ExprPtr shape;
	Boolean found;

	shape = ASHAPE(mod);
	if ((void) MapGet(CAR(rec),shape,&found), found)
	  rec = mod;
	else
	  {
	    INT i,n,nn;
	    ExprPtr lab;
	    ExprPtr *ep,*epp;

	    n = AVAL(mod)->n;
	    {
	      ExprPtr *fields,*outfields,*exprs;

	      fields = (ExprPtr *)alloctmp((3*n+2)*sizeof(ExprPtr));
	      outfields = fields + n; /* fields has n elements */
	      exprs = outfields+1+n; /* outfields may have up to n+1 elements */
	      /* and exprs will have up to n+1 elements */
	      dumpShape(shape,fields);
	      lab = CAR(rec);
	      mergeFieldLists(&lab,fields,outfields,(INT)1,n,&nn); /* here nn=n+1 */
	      shape = undumpShape1(outfields,(INT)1,nn);
	      (void) KEEP(shape);
	      for (i=0, ep=AVAL(mod)->expr, epp=exprs; i<(INT)lab; i++)
		*epp++ = *ep++;
	      *epp++ = CDR(rec); i++;
	      for (; i<nn; i++)
		*epp++ = *ep++;
	      rec = share?KrecordGnrl(shape,nn,exprs):MakeRecord(E_RECORD_GNRL,shape,nn,exprs);
	    }
	  }
      }
      break;
    case E_RECORD_GNRL:
      {
	INT i,m,n,nn;
	ExprPtr shape;
	ExprPtr *ep,*epp,*fp;

	m = AVAL(rec)->n;
	n = AVAL(mod)->n;
	{
	  ExprPtr *fields1,*fields2,*outfields,*exprs;

	  nn = m+n;
	  fields1 = (ExprPtr *)alloctmp(nn*(3*sizeof(ExprPtr)));
	  fields2 = fields1+m; /* fields1 has m elements */
	  outfields = fields2+n; /* fields2 has n elements */
	  exprs = outfields+nn; /* outfields may have up to m+n elements */
	  /* and exprs will have up to m+n elements */
	  dumpShape(ASHAPE(rec),fields1);
	  dumpShape(ASHAPE(mod),fields2);
	  mergeFieldLists(fields1,fields2,outfields,m,n,&nn);
	  shape = undumpShape1(outfields,(INT)1,nn);
	  (void) KEEP(shape);
	  epp = exprs;
	  for (i=0, ep=AVAL(rec)->expr, fp=fields1; i<m; i++, ep++)
	    {
	      COUNTER ii;

	      ii = (COUNTER)*fp++;
	      if (ii!=-1)
		SETQ(epp[ii],*ep);
	    }
	  for (i=0, ep=AVAL(mod)->expr, fp=fields2; i<n; i++, ep++)
	    {
	      SETQ(epp[(COUNTER)*fp++],*ep);
	    }
	  rec = share?KrecordGnrl(shape,nn,exprs):MakeRecord(E_RECORD_GNRL,shape,nn,exprs);
	}
	break;
      }
    }
  QUIT;
  return rec;
}

DEF_PRIVATE_ACTION(eval)
{
  return Eval0(env);
}

PRIVATE ExprPtr Eval1(e)
IN CONST ExprPtr e;
{
  ENTER;

  e = CallCC(ACTION_NAME(eval),e,true);
  QUIT;
  return e;
}

PUBLIC ExprPtr Eval(e)
IN CONST ExprPtr e;
{
  e = Eval0(e);
  return KEEP(e);
}

#ifdef PENDINGARGS
/* This scheme of pending args was supposed to do the following:
   When evaluating f x y, we first evaluate f x, which yields a
   closure, then we apply this closure to y.  Building the closure
   is actually useless, so the idea was to first push the expression
   y on a stack (pendingArgs), then evaluate f x; when Eval0()
   would evaluate a value, it would then check whether there are
   pending arguments to it in the stack, and if so, it would apply
   them; with this scheme, it is easy to optimize the case of
   lambda-expressions, which we can directly apply to their pending
   arguments instead of building the intermediate closure.
   This is however mostly useless, as lots of functions f
   are defined as: fun f x = let fun g y = ... in g end, which
   defeats the whole scheme. */
#define N_PENDING 16

typedef struct Pending {
  COUNTER n;
  gcStackView *next;
  ExprPtr env[N_PENDING];
  ExprPtr expr[N_PENDING];
#if DEBUGGING
  ExprPtr dz[N_PENDING];
#endif
} Pending;
#endif

#if DEBUGGING
PRIVATE ExprPtr EvalNext(e)
IN CONST ExprPtr e;
{
  excHandler *xh;

  xh = (excHandler *)alloctmp(sizeof(excHandler));
  xh->next = excStack;
  xh->tag = X_STEP;
  excStack = xh;

  stepMode = STEP_NO;
  resumeStepMode = STEP_NEXT;
  SETQ(currentDebugZone,e);
  e = Eval0(DZCODE(e));
  stepMode = resumeStepMode; /* see debugger() */
  excStack = xh->next;
  return e;
}
#endif

PRIVATE ExprPtr Eval0(e)
IN CONST ExprPtr e;
{
  struct _x { COUNTER n; gcStackView *next; ExprPtr p[4]; } _x;
#define f (_x.p[0])
#define saveenv (_x.p[1])
#define value (_x.p[2])
#ifdef PENDINGARGS
  Pending *pendingArgs;
#endif

#if DEBUGGING
  COUNTER applStkPushed;
#define PUSHAPPL(code,env,calling,zone) \
  {if (!debugMode) { applStkPushed++; PUSHAPPLSTK(code,env,calling,zone); }}
#endif

  if (e->kind==(CHAR)E_VARREF) /* special case to speed things up a bit */
    return lookup(VARRANK(e),VARLEVEL(e));

  if (STKFULL(e))
    /* (void) Raise(stackException); */
    return Eval1(e);
 /* instead of raising a "stack full" exception, save current stack, empty it,
    evaluate in the new empty stack, then reinstall old full stack and return.
    This is done by simply changing the current thread before evaluating e,
    the thread management in stack.c takes care of flushing old threads. */
#if DEBUGGING!=0
  applStkPushed = 0;
#endif
#ifdef PENDINGARGS
  pendingArgs = NULL;
#endif

  _x.n = 4;
  _x.next = theStackFrames;
  _x.p[0] = _x.p[1] = _x.p[2] = _x.p[3] = NOEXPR;
  theStackFrames = (gcStackView *)&_x;

again:
#if DEBUGGING!=0
  if (stopwanted)
    {
      value = Break(specialMarker);
      if (value!=specialMarker)
	goto final;
    }
#endif
  /*if (e==NIL)
    { value = e; goto final; }*/
  switch (e->kind)
    {
    case E_NIL: case E_INT: case E_REAL: case E_COMPLEX: case E_STRING:
    case E_EMPTY: case E_TUPLE_0: case E_RECORD_0:
    case E_ENV: case E_STACK:
    case E_COMPILED_CLOSURE: case E_INTERPRETED_CLOSURE: case E_MEMO_CLOSURE:
      value = e;
      goto final;
    case E_DATACON:
      f = COUNTEDVAL(e);
      if (COUNTEDN(e) & 2) /* generative: evaluate constructor */
	f = Eval0(f);
#ifdef PENDINGARGS
      if (pendingArgs!=NULL && pendingArgs->n!=0) /* if pending argument, apply it directly;
						   this is an optimization of what happens
						   at label final: */
	{
	  value = pendingArgs->expr[--pendingArgs->n];
	  SETQ(theEnvironment,pendingArgs->env[pendingArgs->n]);
#if DEBUGGING
	  SETQ(currentDebugZone,pendingArgs->dz[pendingArgs->n]);
#endif
	  value = Eval0(value); /* evaluate argument */
	  if (COUNTEDN(e) & 1)
	    { /* unshared */
	      MAKECONS(f,value,e);
	      value = e;
	    }
	  else
	    { /* shared */
	      value = Kons(f,value);
	    }
	}
      else
#endif
	{
	  value = MakeCompiled(E_COMPILED_CLOSURE,
			       (COUNTEDN(e) & 1)?ACTION_NAME(datacon_unshared)
			       :ACTION_NAME(datacon_shared),f);
	}
      goto final;
    case E_REF:
      f = Eval0(REFVAL(e));
      MAKEREF(f,value);
      goto final;
    case E_TUPLE_1:
      f = Eval0(REFVAL(e));
      value = Ktuple1(f);
      goto final;
    case E_UNSHARED_TUPLE_1:
      f = Eval0(REFVAL(e));
      value = MakeTuple1(f);
      goto final;
    case E_RECORD_1:
      f = Eval0(CDR(e));
      value = Krecord1(CAR(e),f);
      goto final;
    case E_UNSHARED_RECORD_1:
      f = Eval0(CDR(e));
      e = CAR(e);
      MAKERECORD1(e,f,value);
      goto final;
    case E_TUPLE_2:
      saveenv = theEnvironment;
      value = Eval0(CAR(e));
      SETQ(theEnvironment,saveenv);
      f = Eval0(CDR(e));
      value = Ktuple2(value,f);
      goto final;
    case E_UNSHARED_TUPLE_2:
      saveenv = theEnvironment;
      f = Eval0(CAR(e));
      SETQ(theEnvironment,saveenv);
      _x.p[3] = Eval0(CDR(e));
      MAKETUPLE2(f,_x.p[3],value);
      goto final;
    case E_CONS:
      saveenv = theEnvironment;
      value = Eval0(CAR(e));
      SETQ(theEnvironment,saveenv);
      f = Eval0(CDR(e));
      value = Kons(value,f);
      goto final;
    case E_UNSHARED_CONS:
      saveenv = theEnvironment;
      f = Eval0(CAR(e));
      SETQ(theEnvironment,saveenv);
      _x.p[3] = Eval0(CDR(e));
      MAKECONS(f,_x.p[3],value);
      goto final;
    case E_TUPLE_GNRL:
      value = EvalTuple(e,true);
      goto final;
    case E_UNSHARED_TUPLE_GNRL:
      value = EvalTuple(e,false);
      goto final;
    case E_RECORD_GNRL:
      value = EvalRecord(e,true);
      goto final;
    case E_UNSHARED_RECORD_GNRL:
      value = EvalRecord(e,false);
      goto final;
    case E_MAPLET:
      saveenv = theEnvironment;
      value = Eval0(MAPLETX(e));
      SETQ(theEnvironment,saveenv);
      f = Eval0(MAPLETY(e));
      value = MapletKons(value,f);
      goto final;
    case E_OVERWRITE:
      saveenv = theEnvironment;
      value = Eval0(CAR(e));
      SETQ(theEnvironment,saveenv);
      f = Eval0(CDR(e));
      value = MapOverwrite(value,f);
      goto final;
    case E_UNDERWRITE:
      saveenv = theEnvironment;
      value = Eval0(CAR(e));
      SETQ(theEnvironment,saveenv);
      f = Eval0(CDR(e));
      value = MapOverwrite(f,value);
      goto final;
    case E_QUOTE:
      value = REFVAL(e);
      goto final;
    case E_VARREF:
      value = lookup(VARRANK(e),VARLEVEL(e));
      goto final;
    case E_LET:
      if (Let(LETBIND(e)))
	{
	  value = Raise(BINDEXC);
	  goto final;
	}
      e = LETBODY(e);
      /*saveenv = theEnvironment;*/
      goto again;
    case E_PAT_AS:
      e = CAR(e);
      goto again;
    case E_POLYFUN:
#ifdef PENDINGARGS
      if (pendingArgs!=NULL && pendingArgs->n!=0) /* when some argument is pending,
						     avoid creating a closure;
						     this optimizes what happens here
						     and at label final: */
	{
	  saveenv = theEnvironment;
	  f = pendingArgs->expr[--pendingArgs->n];
	  SETQ(theEnvironment,pendingArgs->env[pendingArgs->n]);
#if DEBUGGING
	  SETQ(currentDebugZone,pendingArgs->dz[pendingArgs->n]);
#endif
	  f = Eval0(f); /* evaluate argument */
	  SETQ(theEnvironment,saveenv);
	  currentFunction = e;
#if DEBUGGING
	  currentCallingEnv = saveenv;
#endif
	  e = COUNTEDVAL(currentFunction);
	  patChoiceBody = CDR(e);
	  if (!MakePatVal(CAR(e),f,(COUNTER)COUNTEDN(currentFunction)))
	    {
#if DEBUGGING
	      PUSHAPPL(currentFunction,theEnvironment,currentCallingEnv,currentDebugZone);
#endif
	      value = Raise(MATCHEXC);
	      goto final;
	    }
	  e = patChoiceBody;
	  /*saveenv = theEnvironment;*/
	  goto again;
	}
      else
#endif
	{
	  MAKECONS(e,theEnvironment,f);
	  value = MakeCompiled(E_INTERPRETED_CLOSURE,ACTION_NAME(apply_polyfun_closure),f);
	}
      goto final;
    case E_MEMO_FN:
#ifdef PENDINGARGS
      if (pendingArgs!=NULL && pendingArgs->n!=0) /* when some argument is pending,
						     avoid creating a closure;
						     this optimizes what happens here
						     and at label final: */
	{ /* in fact, if a MEMO_FN is applied without being declared, memoization
	     is lost, so we do as through it was a POLYFUN. */
	  e = MEMOFUN(e);
	  goto again;
	}
      else
#endif
	{
	  f = Eval0(MEMOFUN(e));
	  MAKECONS(MEMOREM(e),f,value);
	  value = MakeCompiled(E_MEMO_CLOSURE,ACTION_NAME(apply_memo_fn),value);
	}
      goto final;
    case E_APPL: case E_GOTO:
      /* because Eval() does not reinstall
	 theEnvironment, all E_APPL code is tail
	 recursive; thus there is no difference
	 between E_APPL and E_GOTO! */
#ifdef PENDINGARGS
#if DEBUGGING
      if (stepMode==STEP_NEXT)
	stepMode = STEP_NO;
#endif
      if (pendingArgs==NULL)
	{
	  pendingArgs = (Pending *)alloctmp(sizeof(Pending));
	  pendingArgs->n = 1;
	  pendingArgs->next = theHeapFrames;
	  pendingArgs->env[0] = theEnvironment;
	  theHeapFrames = (gcStackView *)pendingArgs;

	  pendingArgs->expr[0] = APPLARG(e);
#if DEBUGGING
	  pendingArgs->dz[0] = currentDebugZone;
#endif
	  e = APPLFUN(e);
	  goto again;
	}
      if (pendingArgs->n==N_PENDING)
	{ /* if argument stack is full, then call ourselves recursively. */
	  value = Eval0(e);
	  goto final;
	}
      /* otherwise, just push argument on the argument stack. */
#if DEBUGGING
      pendingArgs->dz[pendingArgs->n] = currentDebugZone;
#endif
      pendingArgs->expr[pendingArgs->n] = APPLARG(e);
      pendingArgs->env[pendingArgs->n++] = theEnvironment;
      e = APPLFUN(e);
      goto again;
#else
      saveenv = theEnvironment;
#if DEBUGGING
      _x.p[3] = currentDebugZone;
#endif
      f = APPLFUN(e);
reapply:
      switch (f->kind)
	{
	case E_POLYFUN: /* case <expr> of ... */
	  /* a polyfun is a counted reference (refval+counter);
	     the refval is a couple (pattern,primary body); the counter is
	     the max number of slots in a local environment for all
	     possible bodies. */
	  value = Eval0(APPLARG(e));
	  SETQ(theEnvironment,saveenv);
#if DEBUGGING
	  SETQ(currentDebugZone,_x.p[3]);
	  /* do not push call on debug stack, as this is a case expression. */
#endif
	  patChoiceBody = CDR(COUNTEDVAL(f));
	  if (!MakePatVal(CAR(COUNTEDVAL(f)),value,(COUNTER)COUNTEDN(f)))
	    {
	      value = Raise(MATCHEXC);
	      goto final;
	    }
	  /*saveenv = theEnvironment;*/
	  e = patChoiceBody;
	  goto again;
	case E_MEMO_FN: /* a memo function as immediate value in the code does not
			   memoize (or it would memoize into a closure that would
			   be immediately garbageable) */
	  f = MEMOFUN(f);
	  goto reapply;
	case E_DELAY: /* same as for memo functions */
	  f = DCODE(f);
	  goto reapply;
	case E_DATACON:
	  value = COUNTEDVAL(f);
	  if (COUNTEDN(f) & 2) /* generative: evaluate constructor */
	    value = Eval0(value);
	  if (COUNTEDN(f) & 1) /* unshared */
	    {
	      MAKECONS(value,saveenv,f);
	      value = f;
	    }
	  else
	    value = Kons(value,saveenv);
	  goto final;
	case E_VARREF:
	  /* the most frequent case of function applications */
	  f = lookup(VARRANK(f),VARLEVEL(f));
	  /*fallthrough*/
	case E_COMPILED_CLOSURE: /* should not happen unless by fallthrough */
	case E_INTERPRETED_CLOSURE:
	case E_MEMO_CLOSURE:
apply:
	  value = Eval0(APPLARG(e));

#if DEBUGGING
	  SETQ(currentDebugZone,_x.p[3]);
#endif
	  if (f->kind==(CHAR)E_INTERPRETED_CLOSURE) /* do tail call */
	    { /* this is an optimization we can do in the interpreter only. */
	      f = QENV(f);
	      SETQ(theEnvironment,CDR(f));
	      currentFunction = CAR(f);
#if DEBUGGING
	      currentCallingEnv = saveenv;
#endif
	      e = COUNTEDVAL(currentFunction);
	      patChoiceBody = CDR(e);
	      if (!MakePatVal(CAR(e),value,(COUNTER)COUNTEDN(currentFunction)))
		{
#if DEBUGGING
		  PUSHAPPL(currentFunction,theEnvironment,currentCallingEnv,currentDebugZone);
#endif
		  value = Raise(MATCHEXC);
		  goto final;
		}
	      /*saveenv = theEnvironment;*/
	      e = patChoiceBody;
	      goto again;
	    }
	  value = (*ACTION(f))(value,QENV(f));
	  goto final;
	default:
	  f = Eval0(f);
	  SETQ(theEnvironment,saveenv);
#if DEBUGGING
	  SETQ(currentDebugZone,_x.p[3]);
#endif
	  goto apply;
	}
#endif
    case E_GET:
      value = Eval0(GETREF(e));
      value = REFVAL(value);
      goto final;
    case E_SET:
      saveenv = theEnvironment;
      f = Eval0(SETREF(e));
      SETQ(theEnvironment,saveenv);
      value = Eval0(SETVAL(e));
      SETREFVAL(f,value);
      value = tuple0;
      goto final;
    case E_MAPGET:
      saveenv = theEnvironment;
      f = Eval0(CAR(e));
      SETQ(theEnvironment,saveenv);
      value = Eval0(CDR(e));
      value = ACTION_NAME(q_mapapp)(value,f);
      goto final;
    case E_TSELECT:
      value = Eval0(CDR(e));
      f = CAR(e);
      switch (value->kind)
	{
	case E_TUPLE_1:
	  value = REFVAL(value);
	  break;
	case E_TUPLE_2:
	  value = (f==ione)?CAR(value):CDR(value);
	  break;
	case E_TUPLE_GNRL:
	  value = AVAL(value)->expr[IVAL(f)-1];
	  break;
	}
      goto final;
    case E_TSEL1:
      value = Eval0(CDR(e));
      value = REFVAL(value);
      goto final;
    case E_TSEL2:
      value = Eval0(CDR(e));
      f = CAR(e);
      value = (f==ione)?CAR(value):CDR(value);
      goto final;
    case E_TSELGNRL:
      value = Eval0(CDR(e));
      f = CAR(e);
      value = AVAL(value)->expr[IVAL(f)-1];
      goto final;
    case E_RSELECT:
      value = Eval0(CDR(e));
      f = CAR(e);
      switch (value->kind)
	{
	case E_RECORD_1:
	  value = CDR(value);
	  break;
	case E_RECORD_GNRL:
	  if (f->kind!=(CHAR)E_INT)
	    f = MapGetSure(f,ASHAPE(value));
	  value = AVAL(value)->expr[IVAL(f)-1];
	  break;
	}
      goto final;
    case E_RSEL1:
      value = Eval0(CDR(e));
      value = CDR(value);
      goto final;
    case E_RSELGNRL:
      value = Eval0(CDR(e));
      f = CAR(e);
      /*if (f->kind!=(CHAR)E_INT)
	f = MapGetSure(f,ASHAPE(value));*/
      value = AVAL(value)->expr[IVAL(f)-1];
      goto final;
    case E_RMU1:
      saveenv = theEnvironment;
      f = Eval0(CAR(e));
      SETQ(theEnvironment,saveenv);
      value = Eval0(CDR(CDR(e)));
      value = mu1(f,CAR(CDR(e)),value,true);
      goto final;
    case E_UNSHARED_RMU1:
      saveenv = theEnvironment;
      f = Eval0(CAR(e));
      SETQ(theEnvironment,saveenv);
      value = Eval0(CDR(CDR(e)));
      value = mu1(f,CAR(CDR(e)),value,false);
      goto final;
    case E_RMUX:
      saveenv = theEnvironment;
      f = Eval0(CAR(e));
      SETQ(theEnvironment,saveenv);
      value = Eval0(CDR(e));
      value = muGnrl(f,value,true);
      goto final;
    case E_UNSHARED_RMUX:
      saveenv = theEnvironment;
      f = Eval0(CAR(e));
      SETQ(theEnvironment,saveenv);
      value = Eval0(CDR(e));
      value = muGnrl(f,value,false);
      goto final;
    case E_IF:
      saveenv = theEnvironment;
      value = Eval0(TESTIF(e));
      SETQ(theEnvironment,saveenv);
      e = (value!=falseValue)?CAR(TESTFORK(e)):CDR(TESTFORK(e));
      goto again;
    case E_PROG:
      saveenv = theEnvironment;
      (void) Eval0(PROGDO(e));
      SETQ(theEnvironment,saveenv);
      e = PROGTHEN(e);
      goto again;
    case E_AND:
      saveenv = theEnvironment;
      if (Eval0(PROGDO(e))!=falseValue)
	{
	  SETQ(theEnvironment,saveenv);
	  e = PROGTHEN(e);
	  goto again;
	}
      value = falseValue;
      goto final;
    case E_OR:
      saveenv = theEnvironment;
      if (Eval0(PROGDO(e))!=falseValue)
	{
	  value = trueValue;
	  goto final;
	}
      SETQ(theEnvironment,saveenv);
      e = PROGTHEN(e);
      goto again;
    case E_EQUAL:
      saveenv = theEnvironment;
      f = Eval0(CAR(e));
      SETQ(theEnvironment,saveenv);
#if PLDI93_HACK
      value = (equal(f,Eval0(CDR(e))))?trueValue:falseValue;
#else
      value = (f==Eval0(CDR(e)))?trueValue:falseValue;
#endif
      goto final;
    case E_NEQUAL:
      saveenv = theEnvironment;
      f = Eval0(CAR(e));
      SETQ(theEnvironment,saveenv);
#if PLDI93_HACK
      value = (equal(f,Eval0(CDR(e))))?falseValue:trueValue;
#else
      value = (f!=Eval0(CDR(e)))?trueValue:falseValue;
#endif
      goto final;
    case E_NOT:
      value = (Eval0(REFVAL(e))==falseValue)?trueValue:falseValue;
      goto final;
    case E_PROG_OF:
      value = EvalSimpleOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_PROG);
      goto final;
    case E_LIST_OF:
      value = EvalSimpleOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_LIST);
      value = QueueToList(value);
      goto final;
    case E_UNSHARED_LIST_OF:
      value = EvalSimpleOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_UNSHARED_LIST);
      if (value!=NIL)
	{
	  e = CDR(value);
	  SETCDR(value,NIL);
	  value = e;
	}
      goto final;
    case E_SET_OF:
      value = EvalSimpleOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_SET);
      goto final;
    case E_APPEND_OF:
      value = EvalOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_APPEND);
      goto final;
    case E_UNSHARED_APPEND_OF:
      value = EvalSimpleOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_UNSHARED_APPEND);
      if (value!=NIL)
	{
	  e = CDR(value);
	  SETCDR(value,NIL);
	  value = e;
	}
      goto final;
    case E_UNION_OF:
      value = EvalSimpleOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_UNION);
      goto final;
    case E_UNION_UNDER_OF:
      value = EvalSimpleOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_UNION_UNDER);
      goto final;
    case E_INTER_OF:
      value = EvalSimpleOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_INTER);
      if (value==NIL)
	value = Raise(EMPTYEXC);
      goto final;
    case E_MAP_OF:
      value = EvalSimpleOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_MAP);
      goto final;
    case E_MAP_UNDER_OF:
      value = EvalSimpleOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_MAP_UNDER);
      goto final;
    case E_ALL_OF:
      value = EvalOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_ALL);
      goto final;
    case E_EXISTS_OF:
      value = EvalOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_EXISTS);
      goto final;
    case E_SOME_OF:
      value = EvalOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_SOME);
      goto final;
    case E_UNSHARED_SOME_OF:
      value = EvalOf(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_UNSHARED_SOME);
      goto final;
    case E_PROG_MAP:
      value = EvalMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_PROG);
      goto final;
    case E_LIST_MAP:
      value = EvalMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_LIST);
      goto final;
    case E_UNSHARED_LIST_MAP:
      value = EvalMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_UNSHARED_LIST);
      goto final;
    case E_SET_MAP:
      value = EvalMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_SET);
      goto final;
    case E_APPEND_MAP:
      value = EvalMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_APPEND);
      goto final;
    case E_UNSHARED_APPEND_MAP:
      value = EvalMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_UNSHARED_APPEND);
      goto final;
    case E_UNION_MAP:
      value = EvalMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_UNION);
      goto final;
    case E_UNION_UNDER_MAP:
      value = EvalMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_UNION_UNDER);
      goto final;
    case E_INTER_MAP:
      value = EvalMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_INTER);
      goto final;
    case E_MAP_MAP:
      value = EvalMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_MAP);
      goto final;
    case E_MAP_UNDER_MAP:
      value = EvalMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_MAP_UNDER);
      goto final;
    case E_ALL_MAP:
      value = EvalMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_ALL);
      goto final;
    case E_EXISTS_MAP:
      value = EvalMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_EXISTS);
      goto final;
    case E_SOME_MAP:
      value = EvalMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_SOME);
      goto final;
    case E_UNSHARED_SOME_MAP:
      value = EvalMap(CAR(OFFP(e)),CDR(OFFP(e)),OFDOM(e),C_UNSHARED_SOME);
      goto final;
    case E_PROG_WHILE:
      value = EvalWhile(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_PROG);
      goto final;
    case E_LIST_WHILE:
      value = EvalWhile(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_LIST);
      goto final;
    case E_UNSHARED_LIST_WHILE:
      value = EvalWhile(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_UNSHARED_LIST);
      goto final;
    case E_SET_WHILE:
      value = EvalWhile(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_SET);
      goto final;
    case E_APPEND_WHILE:
      value = EvalWhile(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_APPEND);
      goto final;
    case E_UNSHARED_APPEND_WHILE:
      value = EvalWhile(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_UNSHARED_APPEND);
      goto final;
    case E_UNION_WHILE:
      value = EvalWhile(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_UNION);
      goto final;
    case E_UNION_UNDER_WHILE:
      value = EvalWhile(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_UNION_UNDER);
      goto final;
    case E_INTER_WHILE:
      value = EvalWhile(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_INTER);
      goto final;
    case E_MAP_WHILE:
      value = EvalWhile(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_MAP);
      goto final;
    case E_MAP_UNDER_WHILE:
      value = EvalWhile(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_MAP_UNDER);
      goto final;
    case E_ALL_WHILE:
      value = EvalWhile(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_ALL);
      goto final;
    case E_EXISTS_WHILE:
      value = EvalWhile(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_EXISTS);
      goto final;
    case E_SOME_WHILE:
      value = EvalWhile(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_SOME);
      goto final;
    case E_UNSHARED_SOME_WHILE:
      value = EvalWhile(CAR(WCOND(e)),CDR(WCOND(e)),WBODY(e),C_UNSHARED_SOME);
      goto final;
    case E_HANDLE:
      value = HandleEval(HBODY(e),HHANDLERS(e));
      if (value==NOEXPR) /* then an exception was raised, and we have to evaluate THEEXPR */
	{
	  e = THEEXPR; /* no need to KEEP e, since it is a piece of code that won't be GC'ed */
	  goto again;
	}
      goto final;
    case E_RAISE:
      value = Eval0(RBODY(e));
      value = Raise(value);
      goto final;
    case E_EXCEPTION:
      if (e->mark & MARK_SHARED)
	value = e; /* extern exceptions are not generative */
      else
	{ /* exceptions are generative at run-time */
	  value = MakeException(EXCVAL(e));
	}
      goto final;
    case E_DELAY:
      value = Eval0(DCODE(e)); /* get closure for the promise */
      value = MakePromise(value);
      goto final;
    case E_PACK:
      f = Eval0(PACKVAL(e));
      value = MakePack(E_PACK,f,PACKTYPE(e));
      goto final;
    case E_DEBUGFUNC: /* like E_DEBUGZONE, but at entry in a function */
#if !DEBUGGING
      e = DZCODE(e);
      goto again;
#else
      if (stepMode==STEP_NEXT)
	{
stepnext:
	  saveenv = theEnvironment;
	  value = EvalNext(e);
	  switch (e->kind) /* see whether there is a breakpoint on exit (useless, see below) */
	    {
	    case E_DF_BRK_EXIT:
	    case E_DZ_BRK_EXIT:
	    case E_DF_BRK_BOTH:
	    case E_DZ_BRK_BOTH:
	      SETQ(theEnvironment,saveenv);
	      value = BreakOnExit(e,value); /* useless, as we can't set any breakpoints on exit */
	      break;
	    }
	  goto final;
	}
      PUSHAPPL(currentFunction,theEnvironment,currentCallingEnv,currentDebugZone);
      /*fallthrough*/
#endif
    case E_DEBUGZONE: /* placeholder for breakpoint which was not set (unless stepping) */
#if DEBUGGING
      if (stepMode!=STEP_NO &&
	  (f = DZINFO(e),
	   (ZONE(f)->line1!=stepLine ||
	    ZONE(f)->pos1!=stepPos ||
	    CAR(CDR(CDR(ARGS(f))))!=stepFile)))
	goto brkentry;
      SETQ(currentDebugZone,e);
#endif
      e = DZCODE(e);
      goto again;
    case E_DF_BRK_ENTRY:
#if !DEBUGGING
      e = DZCODE(e);
      goto again;
#else
      PUSHAPPL(currentFunction,theEnvironment,currentCallingEnv,currentDebugZone);
      /*fallthrough*/
#endif
    case E_DZ_BRK_ENTRY: /* breakpoint on entry */
#if !DEBUGGING
      e = DZCODE(e);
      goto again;
#else
brkentry:
      SETQ(currentDebugZone,e);
      saveenv = theEnvironment;
      value = BreakOnEntry(e);
      if (value!=specialMarker)
	goto final;
      switch (e->kind) /* check that the breakpoint on exit has not been positioned meanwhile */
	{
	case E_DF_BRK_EXIT:
	case E_DZ_BRK_EXIT:
	case E_DF_BRK_BOTH:
	case E_DZ_BRK_BOTH:
	  goto brkexit;
	}
      SETQ(theEnvironment,saveenv);
      SETQ(currentDebugZone,e);
      e = DZCODE(e);
      goto again;
#endif
    case E_DF_BRK_EXIT:
#if !DEBUGGING
      e = DZCODE(e);
      goto again;
#else
      PUSHAPPL(currentFunction,theEnvironment,currentCallingEnv,currentDebugZone);
      /*fallthrough*/
#endif
    case E_DZ_BRK_EXIT: /* breakpoint on exit */
#if !DEBUGGING
      e = DZCODE(e);
      goto again;
#else
      if (stepMode==STEP_NEXT)
	goto stepnext;
      if (stepMode!=STEP_NO &&
	  (f = DZINFO(e),
	   (ZONE(f)->line1!=stepLine ||
	    ZONE(f)->pos1!=stepPos ||
	    CAR(CDR(CDR(ARGS(f))))!=stepFile)))
	goto brkboth;
brkexit:
      SETQ(currentDebugZone,e);
      saveenv = theEnvironment;
      value = Eval0(DZCODE(e));
      switch (e->kind) /* check that the breakpoint on exit is still there */
	{
	case E_DF_BRK_EXIT:
	case E_DZ_BRK_EXIT:
	case E_DF_BRK_BOTH:
	case E_DZ_BRK_BOTH:
	  SETQ(theEnvironment,saveenv);
	  value = BreakOnExit(e,value); /* useless, as we can't set any breakpoints on exit */
	  break;
	}
      goto final;
#endif
    case E_DF_BRK_BOTH:
#if !DEBUGGING
      e = DZCODE(e);
      goto again;
#else
      PUSHAPPL(currentFunction,theEnvironment,currentCallingEnv,currentDebugZone);
      /*fallthrough*/
#endif
    case E_DZ_BRK_BOTH: /* breakpoint on exit */
#if !DEBUGGING
      e = DZCODE(e);
      goto again;
#else
brkboth:
      SETQ(currentDebugZone,e);
      saveenv = theEnvironment;
      value = BreakOnEntry(e);
      if (value!=specialMarker)
	goto final;
      goto brkexit;
#endif
#ifdef OBSOLETE
      f = DZINFO(e);
      saveenv = theEnvironment;
      SETQ(currentDebugZone,e);
      if ((stepMode!=STEP_NO &&
	   (ZONE(f)->line1!=stepLine ||
	    ZONE(f)->pos1!=stepPos ||
	    CAR(CDR(CDR(ARGS(f))))!=stepFile)) ||
	  checkGuard(f,tuple0,DZ_ACTIVE_ON_ENTRY))
	{
	  SETQ(theEnvironment,saveenv);
	  value = BreakOnEntry(e);
	  if (value!=specialMarker)
	    goto final;
	}
      SETQ(theEnvironment,saveenv);
      SETQ(currentDebugZone,e);
      if (stepMode!=STEP_NO || (TYPE(f) & DZ_ACTIVE_ON_EXIT))
	{
	  if (stepMode==STEP_NEXT)
	    {
	      stepMode = STEP_NO;
	      resumeStepMode = STEP_NEXT;
	      value = Eval0(DZCODE(e));
	      stepMode = resumeStepMode;
	      /* so that, if we break on entry, then do
		 a next(), and if we break again before the end of evaluation
		 of the expression, and if we do a resume then it does not
		 reinstall the next mode, but correctly resumes computation:
		 see debugger() */
	    }
	  else
	    {
	      value = Eval0(DZCODE(e));
	    }
	  SETQ(theEnvironment,saveenv);
	  SETQ(currentDebugZone,e);
	  if (stepMode!=STEP_NO ||
	      (f = MakeQuote(value),
	       checkGuard(DZINFO(e),f,DZ_ACTIVE_ON_EXIT)))
	    {
	      SETQ(theEnvironment,saveenv);
	      value = BreakOnExit(e,value);
	    }
	  goto final;
	}
      e = DZCODE(e);
      goto again;
#endif
    case E_PROFILE:
#if PROFILING==0
      e = CAR(e);
      goto again;
#else
      f = CDR(e);
      if (TYPE(f)==EA_PROFILE)
	value = profEval(e);
      else
	{
	  e = CAR(e);
	  goto again;
	}
      goto final;
#endif
    default:
error:
      kinderror(message(404),e->kind);
      break;
    }
final:
#ifdef PENDINGARGS
  if (pendingArgs!=NULL)
    {
      while (pendingArgs->n!=0)
	{
	  e = pendingArgs->expr[--pendingArgs->n];
	  SETQ(theEnvironment,pendingArgs->env[pendingArgs->n]);
#if DEBUGGING
	  SETQ(currentDebugZone,pendingArgs->dz[pendingArgs->n]);
#endif
	  f = Eval0(e); /* evaluate argument */
	  switch (value->kind) /* and apply function (value) */
	    {
	    case E_COMPILED_CLOSURE:
	    case E_MEMO_CLOSURE:
	      value = (*ACTION(value))(f,QENV(value)); /* call primitive */
	      break;
	    case E_INTERPRETED_CLOSURE: /* do tail call */
	      value = QENV(value);
	      SETQ(theEnvironment,CDR(value));
	      currentFunction = CAR(value);
#if DEBUGGING
	      currentCallingEnv = saveenv;
#endif
	      e = COUNTEDVAL(currentFunction);
	      patChoiceBody = CDR(e);
	      if (!MakePatVal(CAR(e),f,(COUNTER)COUNTEDN(currentFunction)))
		{
#if DEBUGGING
		  PUSHAPPL(currentFunction,theEnvironment,currentCallingEnv,currentDebugZone);
#endif
		  value = Raise(MATCHEXC);
		  goto final;
		}
	      e = patChoiceBody;
	      /*saveenv = theEnvironment;*/
	      goto again;
	    default:
	      kinderror(message(443),value->kind);
	      /*NOTREACHED*/
	      break;
	    }
	}
      theHeapFrames = pendingArgs->next;
    }
#endif
#if DEBUGGING
  if (!debugMode)
    while (applStkPushed!=0)
      {
	POPAPPLSTK();
	applStkPushed--;
      }
#endif
  theStackFrames = _x.next;
  return value;
}
#endif
