/* Definitions for bytecode intepreter (unused).
   Copyright (C) 1992-1997 Jean Goubault-Larrecq and Bull S.A.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* The byte code is organized as a sequence of shorts.
   (So it's a "short code").
   The idea is that it should be as close as possible
   to the representation we use in translate.c, so as
   to be translated to C in the easiest way possible.
   (Translation to C should be done by an independent
   program taking a module (a .mlx), and should be
   as simple as possible.)
   But it should be also rather efficient.
*/

#define OPCODE SHORT

typedef struct byteCodeChunk {
  COUNTER refcount; /* number of bFunction's pointing to here */
  COUNTER nConsts; /* number of constant ExprPtrs that we need */
  COUNTER rawDataOff; /* number of bytes from start to beginning of rawdata section */
  COUNTER instrOff; /* number of bytes from start to beginning of instructions */
  COUNTER size; /* number of bytes in structure */
  ExprPtr consts[1]; /* we have nConsts ExprPtrs here */
} byteCodeChunk;

typedef struct byteCode {
  byteCodeChunk *chunk;
  COUNTER start;
} byteCode;

typedef struct bFunction {
  ExprPtr (*ufun) P((struct bFunction *myself,...));
  /* a universal function, for use by the interpreter or by indirect calls;
     this universal function is then responsible for calling the real function
     (if compiled) or for interpreting the following bytecode (then the
     universal function is Interpret()). */
  char *Cname; /* for the compiler to C; name of the real function */
  INT fxtypes; /* types of side-effects; see optimize.h */
  INT flags;
  INT storage; /* storage class of the result */
  COUNTER arity;
  COUNTER gcflags; /* 1: alloced on the heap; 2: interpreted */
  struct byteCode bc; /* for interpreted code */
} bFunction, *bFunPtr;

typedef ExprPtr (*univFun) P((bFunPtr myself,...));

IMPORT ExprPtr Interpret P((bFunPtr myself,...));

#define K_RETURN_EXPR 0
#define K_RETURN_BOOL 1
#define K_RETURN_INT 2
#define K_RETURN_STRING 3
#define K_RETURN_CODE 4
#define K_RETURN_REAL 5
#define K_RETURN_VOID 6

#define K_PARAM_EXPR 7
#define K_PARAM_BOOL 8
#define K_PARAM_INT 9
#define K_PARAM_STRING 10
#define K_PARAM_CODE 11
#define K_PARAM_REAL 12

#define K_END_PARAM 13

#define K_CONST_EXPR 14
#define K_CONST_BOOL 15
#define K_CONST_INT 16
#define K_CONST_STRING 17
#define K_CONST_CODE 18
#define K_CONST_REAL 19

#define K_ASSIGN_EXPR 20
#define K_ASSIGN_BOOL 21
#define K_ASSIGN_INT 22
#define K_ASSIGN_STRING 23
#define K_ASSIGN_CODE 24
#define K_ASSIGN_REAL 25

#define K_GOTO 30
#define K_IF 31
#define K_UNLESS 32
#define K_INC_LISTSWEEP 33
#define K_INC_MAPSWEEP 34
#define K_INC_SUBMAPSWEEP 35
#define K_START_HANDLE 36

#define K_TEST_STACK 37
#define K_ALLOC_EXPRS 38
#define K_ALLOC_NOEXPR 39
#define K_STOP_PROTECTION 40
