/* Optimizer, both for the bytecode compiler and the compiler to C (unused).
   Copyright (C) 1992-1997 Jean Goubault-Larrecq and Bull S.A.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#if TRANSLATOR
#include "expr.h"
#include "refcount.h"
#include "mem.h"
#include "translate.h"
#include "optimize.h"
#include "timing.h"
#include "debug.h"
#include "transfer.h"
#include "evlproto.h"
#include "errproto.h"
#include "typproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "ropt.h"
#ifdef HAS_PROTOTYPES
#include "sopt.h"
#endif

#define __SEG__ OPTIMIZE

PUBLIC ExprPtr maxInlineRef;
EXTERN int errno;

PRIVATE void ppInfoSweep(extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  ppInfo *pp;

  pp = (ppInfo *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      outInt(pp->storage);
      if (errno) return;
      outChars(pp->tags,NTAGS);
      if (errno) return;
      /*fallthrough*/
    case X_MARK:
    case X_MARKDATA:
      (*doext->action)(pp->uses);
      (*doext->action)(pp->def);
      (*doext->action)(pp->eqns);
      (*doext->action)(pp->fragile);
      (*doext->action)(pp->valset);
      break;
    case X_FREE:
      MARKFREE(extdata); /* free the C structure */
      break;
    }
}

PRIVATE ExprPtr ppInfoLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  ppInfo *pp;
  ExprPtr e;
  ENTER;

  pp = (ppInfo *)MALLOC(sizeof(ppInfo));
  pp->storage = inInt();
  if (errno)
    {
err:
      MARKFREE(pp);
      return NIL;
    }
  inChars(pp->tags,NTAGS);
  if (errno) goto err;
  pp->uses = (*load)(sharedref);
  if (errno) goto err;
  pp->def = (*load)(sharedref);
  if (errno) goto err;
  pp->eqns = (*load)(sharedref);
  if (errno) goto err;
  pp->fragile = (*load)(sharedref);
  if (errno) goto err;
  pp->valset = (*load)(sharedref);
  if (errno) goto err;
  e = (shared?Kexternal:MakeExternal)(desc,(char *)pp);
  KEEPONLY(e);
  return e;
}

PUBLIC extDesc ppInfoDesc = {
  (CHAR *)"program-point-info",
  'p','p','?',' ', NULL,
  NULL, /* not shared */
  NULL, /* not shared */
  NULL, /* print as default */
  ppInfoSweep,
  ppInfoLoad
  };

PRIVATE ExprPtr ppInfoNew(ppmodel)
IN CONST ppInfo *ppmodel;
{
  ppInfo *pp;
  ExprPtr e;

  pp = (ppInfo *)MALLOC(sizeof(ppInfo));
  *pp = *ppmodel;
  e = MakeExternal(&ppInfoDesc,(char *)pp);
  return KEEP(e);
}

PRIVATE void clearTags(pp)
INOUT ppInfo *pp;
{
  COUNTER i;
  CHAR *t;

  for (i=0, t=pp->tags; i<NTAGS; i++)
    *t++ = 0;
}

PRIVATE void setTags(pp)
INOUT ppInfo *pp;
{
  COUNTER i;
  CHAR *t;

  for (i=0, t=pp->tags; i<NTAGS; i++)
    *t++ = 255;
}

PRIVATE void ppDecoSweep(extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  ppDecoration *pd;

  pd = (ppDecoration *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      outInt(pd->fxtypes);
      if (errno) return;
      /*fallthrough*/
    case X_MARK:
    case X_MARKDATA:
      (*doext->action)(pd->info);
      (*doext->action)(pd->lambda);
      break;
    case X_FREE:
      MARKFREE(extdata); /* free the C structure */
      break;
    }
}

PRIVATE ExprPtr ppDecoLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  ppDecoration *pd;
  ExprPtr e;
  ENTER;

  pd = (ppDecoration *)MALLOC(sizeof(ppDecoration));
  pd->fxtypes = inInt();
  if (errno)
    {
err:
      MARKFREE(pd);
      return NIL;
    }
  pd->info = (*load)(sharedref);
  if (errno) goto err;
  pd->lambda = (*load)(sharedref);
  if (errno) goto err;
  e = (shared?Kexternal:MakeExternal)(desc,(char *)pd);
  KEEPONLY(e);
  return e;
}

PUBLIC extDesc ppDecoDesc = {
  (CHAR *)"program-decoration",
  'p','d','e','c', NULL,
  NULL, /* not shared */
  NULL, /* not shared */
  NULL, /* print as default */
  ppDecoSweep,
  ppDecoLoad
  };

PUBLIC ExprPtr ppDecoInit()
{
  ppDecoration *pd;
  ExprPtr e;

  pd = (ppDecoration *)MALLOC(sizeof(ppDecoration));
  pd->info = NewMap();
  pd->lambda = NewMap();
  pd->fxtypes = 0;
  e = MakeExternal(&ppDecoDesc,(char *)pd);
  return KEEP(e);
}

MAPSWEEPFUNC(do_fundata)
{
  struct fundata *data;
  Code *code;
  INT i;

  data = MAPDATA(struct fundata *);
  switch (x->kind)
    {
    case E_CONS: /* constant value is a closure, coded as a couple (f,env) */
      x = CAR(x);
      /*fallthrough*/
    default: /* now, constant value should be a Code */
      code = (Code *)EXTDATA(x);
      data->storage |= code->storage;
      data->fxtypes |= code->fxtypes;
      for (i=0; i<NTAGS; i++)
	data->tags[i] |= code->tags[i];
      break;
    }
  return false;
}

PRIVATE void getApplyData(e,deco,data)
IN CONST ExprPtr e;
IN CONST ExprPtr deco;
OUT struct fundata *data;
{
  ExprPtr arg;
  ExprPtr fvalset,f;
  COUNTER i;

  arg = COUNTEDVAL(e);
  switch (arg->kind)
    {
    case E_TUPLE_2: f = CAR(arg); break;
    case E_TUPLE_GNRL: f = AVAL(arg)->expr[0]; break;
    default: f = arg; break;
    }
  fvalset = getPpInfo(f,deco)->valset;
  if (fvalset==NIL)
    {
      data->storage = SCF_ALL_TYPES | SCF_BOXED | SCF_NOINIT;
      /* may have any type; cannot be computed at init time */
      data->fxtypes = FXF_ALL; /* may destroy all possible fragile equations */
      for (i=0; i<NTAGS; i++)
	data->tags[i] = (CHAR)-1;
    }
  else
    {
      data->storage = SCF_TOP;
      data->fxtypes = 0L;
      for (i=0; i<NTAGS; i++)
	data->tags[i] = 0;
      (void) MapSweep(fvalset,do_fundata,data);
    }
}

PRIVATE Boolean hasVisibleFx(e,deco)
IN CONST ExprPtr e;
IN CONST ExprPtr deco;
{ /* returns true if the expression e may have side-effects that
     would be visible from outside the current procedure. This includes
     Z_REF_SET for example since this may be used to change the value
     of a reference passed in argument, or Z_RAISE since it simulates
     the return of a value.
     This does not include local side-effects like Z_LOCAL_REF_SET
     or Z_INC_* which are used only internally and do not change the
     final result and output state of the procedure.
     This does not include either side-effects that are not visible
     because of strict lexical rules. */
  switch (COUNTEDN(e))
    {
    case Z_APPLY: /* may have side effects if impure and its touched regions are visible
		   from the outside. */
    case Z_CALL: /* same */
      {
	struct fundata data;

	getApplyData(e,deco,&data);
	return data.fxtypes!=0;
      }
    case Z_START_HANDLE:
    case Z_END_HANDLE: /* visible to called functions that may raise */
    case Z_RAISE: /* always visible */
    case Z_REF_SET: /* may be visible if the region of this ref is visible from the outside */
    case Z_LET: /* this is subtler; elimination of useless Z_LETs is done by
		   patchLet() */
    case Z_SET_C_GLOBAL:
    case Z_SET_C_CONST:
      return true;
    default:
      return false;
    }
}

PRIVATE ExprPtr newPpInfo()
{
  ppInfo *ppmodel = (ppInfo *)alloctmp(sizeof(ppInfo));
  ExprPtr info;

  ppmodel->uses = NewMap();
  ppmodel->def = NIL; /* = top */
  ppmodel->eqns = NewMap(); /* nothing known */
  ppmodel->fragile = NewMap(); /* subset of the above */
  ppmodel->valset = NIL; /* = top */
  ppmodel->storage = SCF_TOP;
  clearTags(ppmodel);
  info = ppInfoNew(ppmodel);
  return info;
}

PUBLIC ExprPtr getPpInfo1(stmt,deco)
IN CONST ExprPtr stmt;
INOUT ExprPtr deco;
{
  ExprPtr info;
  ppDecoration *pd;
  Boolean found;

  pd = PD(deco);
  info = MapGet(stmt,pd->info,&found);
  if (!found)
    {
      info = newPpInfo();
      setMapPut(stmt,info,&pd->info);
      HITBARRIER(deco);
    }
  return info;
}

PUBLIC ppInfo *getPpInfo(stmt,deco)
IN CONST ExprPtr stmt;
INOUT ExprPtr deco;
{
  return PP(getPpInfo1(stmt,deco));
}

PUBLIC Boolean argSweep(arg,fun,data)
IN CONST ExprPtr arg; /* an argument list of an expression (not a statement) */
IN CONST argSweepFunc fun;
INOUT VOIDPTR data;
{
again:
  switch (arg->kind)
    {
    case E_NIL: /* no argument */
      break;
    case E_TUPLE_2: /* non-empty list of variables, terminated by last value */
      if ((*fun)(CAR(arg),data))
	return true;
      arg = CDR(arg);
      goto again;
    case E_TUPLE_GNRL: /* vector of variables */
      {
	COUNTER i,n;

	n = AVAL(arg)->n;
	for (i=0; i<n; i++)
	  if ((*fun)(AVAL(arg)->expr[i],data))
	    return true;
	break;
      }
    default: /* function code (struct Code, in an E_EXTERNAL), or
		other constant argument, like strings or ints or reals or ... */
      break;
    }
  return false;
}

PRIVATE ExprPtr argMapVec(arg,fun,data)
IN CONST ExprPtr arg; /* an argument list of an expression (not a statement) */
IN CONST argMapFunc fun;
INOUT VOIDPTR data;
{
  COUNTER i,n;

  n = AVAL(arg)->n;
  {
    ExprPtr *ep,e;
    ExprPtr *vals;

    vals = (ExprPtr *)alloctmp(n*sizeof(ExprPtr));
    for (i=0, ep=vals; i<n; i++)
      {
	*ep++ = e = (*fun)(AVAL(arg)->expr[i],data);
      }
    e = KtupleGnrl(n,vals);
    (void) KEEP(e);
    return e;
  }
}

PUBLIC ExprPtr argMap(arg,fun,data)
IN CONST ExprPtr arg; /* an argument list of an expression (not a statement) */
IN CONST argMapFunc fun;
INOUT VOIDPTR data;
{
  switch (arg->kind)
    {
    case E_NIL: /* no argument */
      break;
    case E_TUPLE_2: /* non-empty list of variables, terminated by last value */
      arg = Ktuple2((*fun)(CAR(arg),data),argMap(CDR(arg),fun,data));
      (void) KEEP(arg);
      break;
    case E_TUPLE_GNRL: /* vector of variables */
      arg = argMapVec(arg,fun,data);
      break;
    default: /* constant argument, like strings or ints or reals or ... */
      break;
    }
  return arg;
}

ARGSWEEPFUNC(do_newuse)
{
  struct newuse *data;

  data = MAPDATA(struct newuse *);
  (void) newUse(x,data->stmt,data->deco);
  return false;
}

MAPSWEEPFUNC(do_newuse_merge)
{
  struct newuse *data;

  data = MAPDATA(struct newuse *);
  (void) newUse(y,data->stmt,data->deco);
  return false;
}

DEF_PRIVATE_ACTION(newusek)
{
  struct newusek *data;

  data = (struct newusek *)env;
  data->res = newUse(data->stmt,data->use,data->deco);
  return NIL;
}

PRIVATE Boolean newUsek(stmt,use,deco)
IN CONST ExprPtr stmt,use;
INOUT ExprPtr deco;
{
  struct newusek *data = (struct newusek *)alloctmp(sizeof(struct newusek));

  data->stmt = stmt;
  data->use = use;
  data->deco = deco;
  (void) CallCC(ACTION_NAME(newusek),(ExprPtr)data,true);
  return data->res;
}

PUBLIC Boolean newUse(stmt,use,deco)
IN CONST ExprPtr stmt,use;
INOUT ExprPtr deco;
{
  MapPtr olduses;
  ExprPtr info;
  ppInfo *pp;
  Boolean found;

  if (STKFULL(found))
    return newUsek(stmt,use,deco);
  info = MapGet(stmt,PD(deco)->info,&found);
  if (!found)
    return false; /* if stmt not reached, then we don't record the DEFUSE chain */
  pp = PP(info);
  olduses = pp->uses;
  setSetAdd(use,&pp->uses); HITBARRIER(info);
  if (pp->uses==olduses) /* DEFUSE chain already recorded */
    return false; /* unchanged */
  HITBARRIER(deco);
  {
    struct newuse *data = (struct newuse *)alloctmp(sizeof(struct newuse));
    ExprPtr e;

    data->stmt = stmt;
    data->deco = deco;
    switch (stmt->kind)
      {
      case E_CONS: /* assignment */
	e = CAR(stmt);
	if (e->kind==(CHAR)E_COUNTED)
	  {
	    if (COUNTEDN(e)==Z_ASSERT_EQUAL)
	      (void) newUse(CAR(COUNTEDVAL(e)),stmt,deco);
	    /* in the case Z_ASSERT_EQUAL, the second argument is meant as a hint,
	       never as useful code */
	    else (void) argSweep(COUNTEDVAL(e),do_newuse,data);
	  }
	else (void) newUse(e,stmt,deco);
	break;
      case E_PROG: /* function header: should not happen */
	break;
      case E_IF: /* tests, loops; should only be used as dummies (do nothing) */
	break;
      case E_GOTO: /* merge points */
	(void) MapSweep(COUNTEDVAL(CAR(stmt)),do_newuse_merge,data);
	break;
      default:
	kinderror(message(486),stmt->kind);
	/*NOTREACHED*/
	break;
      }
  }
  return true;
}

ARGMAPFUNC(do_rewrite_arg)
{
  ExprPtr deco,val;
  ppInfo *xinfo;

  deco = MAPDATA(ExprPtr);
  xinfo = getPpInfo(x,deco);
  val = xinfo->def;
  if (val->kind!=(CHAR)E_COUNTED) /* if value is a variable, take it */
    x = val;
  return x;
}

ARGSWEEPFUNC(do_noninit)
{
  ExprPtr deco;

  deco = MAPDATA(ExprPtr);
  return (getPpInfo(x,deco)->storage & SCF_NOINIT)!=0;
}

PRIVATE void ppCopy(pp,info)
INOUT ExprPtr pp;
IN CONST ppInfo *info;
{
  ppInfo *info2;

  info2 = PP(pp);
  SETQ(info2->def,info->def);
  SETQ(info2->eqns,info->eqns);
  SETQ(info2->fragile,info->fragile);
  SETQ(info2->valset,info->valset);
  info2->storage = info->storage;
  memcpy(info2->tags,info->tags,NTAGS*sizeof(CHAR));
  HITBARRIER(pp);
}

PRIVATE ExprPtr deConstruct(x,deco,pp)
IN CONST ExprPtr x;
INOUT ExprPtr deco;
OUT ExprPtr pp;
{ /* e = f~1(x), we want to see whether x is f(y) for some y, and then to simplify e to y */
  /* updates all fields of pp except uses. */
  ExprPtr e,arg,y;
  Boolean found;
  ppInfo *info;

  e = CAR(x);
  info = PP(pp);
  arg = argMap(COUNTEDVAL(e),do_rewrite_arg,deco); /* get simplified argument list */
  y = MapGet(zApply(COUNTEDN(e),arg),info->eqns,&found);
  /* get simplified whole expression */
  if (found)
    ppCopy(pp,getPpInfo(y,deco));
  else
    {
      info->def = x;
      info->eqns = NewMap();
      info->fragile = NewMap();
      info->valset = NIL;
      info->storage = SCF_TOP; /* do not know yet, and we must take ors, so we start at the top */
      HITBARRIER(pp);
      if (argSweep(COUNTEDVAL(e),do_noninit,deco))
	info->storage |= SCF_NOINIT;
      clearTags(info);
     }
  return arg;
}

PRIVATE INT maxCrossCard = 16;

PRIVATE MapPtr crossProduct1(x1,map2,fun,data)
IN CONST ExprPtr x1;
IN CONST MapPtr map2;
IN crossFunc fun;
INOUT char *data;
{
  switch (map2->kind)
    {
    case E_EMPTY:
      return map2;
    case E_MAPLET:
      return Singleton((*fun)(x1,MAPLETX(map2),data));
    default:
      return SetUnion(crossProduct1(x1,MAPLEFT(map2),fun,data),
		      crossProduct1(x1,MAPRIGHT(map2),fun,data));
    }
}

/* do a cross-product, applying fun in couples first, and
   limiting the number of elements to maxCrossCard. */
PRIVATE MapPtr crossProduct(map1,map2,fun,data)
IN CONST MapPtr map1,map2;
IN crossFunc fun;
INOUT char *data;
{
  MapPtr aux;

  if (map1==NewMap() || map2==NewMap())
    return NewMap();
  if (map1==NIL || map2==NIL)
    return NIL;
  if (map1->kind==(CHAR)E_MAPLET)
    return crossProduct1(MAPLETX(map1),map2,fun,data);
  aux = crossProduct(MAPLEFT(map1),map2,fun,data);
  if (aux==NIL)
    return NIL;
  map1 = crossProduct(MAPRIGHT(map1),map2,fun,data);
  if (map1==NIL)
    return NIL;
  map1 = SetUnion(aux,map1);
  return (MapCardLeq(map1,maxCrossCard))?map1:NIL;
}

PRIVATE MapPtr xProd(arg,deco,fun,data)
IN CONST ExprPtr arg;
IN CONST ExprPtr deco;
IN crossFunc fun;
INOUT char *data;
{
  return crossProduct(getPpInfo(CAR(arg),deco)->valset,
		      getPpInfo(CDR(arg),deco)->valset,
		      fun,data);
}

PRIVATE ExprPtr cross_cons(x1,x2,data)
IN CONST ExprPtr x1,x2;
IN CONST char *data;
{
  x1 = Kpair((INT)data,x1,x2);
  return KEEP(x1);
}

PRIVATE MapPtr xCouple(arg,deco)
IN CONST ExprPtr arg;
IN CONST ExprPtr deco;
{
  return xProd(arg,deco,cross_cons,(char *)E_CONS);
}

PRIVATE MapPtr xProd3(arg,deco,fun,data)
IN CONST ExprPtr arg;
IN CONST ExprPtr deco;
IN crossFunc fun;
INOUT char *data;
{
  return crossProduct(getPpInfo(CAR(arg),deco)->valset,
		      xCouple(CDR(arg),deco),
		      fun,data);
}

PRIVATE ExprPtr cross_ncross_1(x1,x2,_data)
IN CONST ExprPtr x1,x2;
INOUT char *_data;
{
  struct nxprod *data;

  data = (struct nxprod *)_data;
  data->args[data->pos[0]] = x2;
  return (*data->fun)(data->n,data->args,data->data);
}

PRIVATE ExprPtr cross_ncross_2(x1,x2,_data)
IN CONST ExprPtr x1,x2;
INOUT char *_data;
{
  struct nxprod *data;

  data = (struct nxprod *)_data;
  data->args[data->pos[0]] = x1;
  data->args[data->pos[1]] = x2;
  return (*data->fun)(data->n,data->args,data->data);
}

PRIVATE MapPtr nxProd(arg,deco,fun,data)
IN CONST ExprPtr arg;
IN CONST ExprPtr deco;
IN ncrossFunc fun;
INOUT char *data;
{
  INT i;
  ExprPtr e;
  struct nxprod *ndata = (struct nxprod *)alloctmp(sizeof(struct nxprod));

  ndata->fun = fun;
  ndata->data = data;
  ndata->n = AVAL(arg)->n;
  ndata->npos = 0;
  {
    ExprPtr *args;

    args = (ExprPtr *)alloctmp(ndata->n*sizeof(ExprPtr));
    ndata->args = args;
    for (i=0; i<ndata->n; i++)
      {
	e = AVAL(arg)->expr[i];
	switch (e->kind)
	  {
	  case E_NIL:
	  case E_EMPTY:
	    return e;
	  case E_MAPLET:
	    e = MAPLETX(e);
	    break;
	  case E_MAP:
	    if (ndata->npos>1) /* to avoid a possible combinatorial explosion,
				 we go to bottom as soon as more than two arguments
				 vary */
	      return NIL;
	    ndata->pos[ndata->npos++] = i;
	    break;
	  }
	args[i] = e;
      }
    switch (ndata->npos)
      {
      case 0:
	return (*fun)(ndata->n,args,data);
      case 1:
	return crossProduct1(NIL,args[ndata->pos[0]],cross_ncross_1,(char *)ndata);
      case 2:
	return crossProduct(args[ndata->pos[0]],args[ndata->pos[1]],
			    cross_ncross_2,(char *)ndata);
      }
    /*NOTREACHED*/
  }
  return NIL;
}

PRIVATE ExprPtr cross_complex(x1,x2,data)
IN CONST ExprPtr x1,x2;
IN CONST char *data;
{
  if (x2==zzero)
    return x1;
  x1 = Kcomplex(x1,x2);
  return KEEP(x1);
}

MAPCOLLECTFUNC(do_strlen)
{
  return Singleton(KINT(SLEN(x)));
}

MAPCOLLECTFUNC(do_ktuple1)
{
  x = Ktuple1(x);
  return Singleton(KEEP(x));
}

MAPCOLLECTFUNC(do_not)
{
  x = (x==falseValue)?trueValue:falseValue;
  return Singleton(x);
}

MAPCOLLECTFUNC(do_knumtype)
{
  x = KnumType(x);
  return Singleton(KEEP(x));
}

PRIVATE ExprPtr ncross_tuple(n,args,data)
IN INT n;
IN CONST ExprPtr *args;
INOUT char *data;
{
  return buildGnrlArg(n,args);
}

MAPCOLLECTFUNC(do_tuple_nth)
{
  return Singleton(AVAL(x)->expr[IVAL(MAPDATA(ExprPtr))-1]);
}

PRIVATE ExprPtr cross_array_nth(x1,x2,data)
IN CONST ExprPtr x1,x2;
INOUT char *data;
{
  return Singleton(AVAL(x1)->expr[IVAL(x2)-1]);
}

MAPCOLLECTFUNC(do_tuple_len)
{
  return Singleton(KINT(AVAL(x)->n));
}

MAPSWEEPFUNC(do_settag)
{
  INT tag;

  tag = (INT)x->kind;
  SETTAG(MAPDATA(ppInfo *),tag);
  return false;
}

PRIVATE void SetTags(pp,valset)
INOUT ppInfo *pp;
IN CONST ExprPtr valset;
{
  if (valset==NIL)
    setTags(pp); /* do not know the tag */
  else
    (void) MapSweep(valset,do_settag,(char *)pp);
}

PRIVATE Boolean HasOnlyTag(pp,tag)
IN CONST ppInfo *pp;
IN COUNTER tag;
{
  COUNTER offset,i;
  CHAR mask;

  offset = tag >> 3;
  mask = 1 << (tag & 7);
  for (i=0; i<NTAGS; i++)
    {
      if (i==offset)
	{
	  if (pp->tags[i] != mask)
	    return false;
	}
      else
	{
	  if (pp->tags[i] != 0)
	    return false;
	}
    }
  return true;
}

PRIVATE ExprPtr ncross_record(n,args,data)
IN INT n;
IN CONST ExprPtr *args;
INOUT char *data;
{
  ExprPtr e;

  e = KrecordGnrl(args[0],n-1,args+1);
  (void) KEEP(e);
  return e;
}

PRIVATE ExprPtr cross_overwrite(x1,x2,data)
IN CONST ExprPtr x1,x2;
INOUT char *data;
{
  return MapMerge(x1,x2);
}

PRIVATE ExprPtr cross_equal(x1,x2,data)
IN CONST ExprPtr x1,x2;
INOUT char *data;
{
  return (x1==x2)?trueValue:falseValue;
}

PRIVATE Boolean noCommonTags(x1,x2,deco)
IN CONST ExprPtr x1,x2;
IN CONST ExprPtr deco;
{
  ppInfo *pp1,*pp2;
  CHAR *t1,*t2;
  COUNTER i;

  pp1 = getPpInfo(x1,deco);
  pp2 = getPpInfo(x2,deco);
  for (i=0, t1=pp1->tags, t2=pp2->tags; i<NTAGS; i++)
    if ((*t1++) & (*t2++))
      return false;
  return true;
}

MAPCOLLECTFUNC(do_car)
{
  return Singleton(CAR(x));
}

MAPCOLLECTFUNC(do_cdr)
{
  return Singleton(CDR(x));
}

PRIVATE ExprPtr cross_type_match(x1,x2,data)
IN CONST ExprPtr x1,x2;
INOUT char *data;
{
  return KINT(TypeMatch(x1,x2));
}

MAPCOLLECTFUNC(do_map_card_dec)
{
  return KINT(MapCardDec(x,IVAL(MAPDATA(ExprPtr))));
}

PRIVATE ExprPtr cross_int_leq(x1,x2,data)
IN CONST ExprPtr x1,x2;
INOUT char *data;
{
  return (IVAL(x1)<=IVAL(x2))?trueValue:falseValue;
}

MAPCOLLECTFUNC(do_get_shape)
{
  return Singleton(ASHAPE(x));
}

PRIVATE ExprPtr cross_mapget(x1,x2,data)
IN CONST ExprPtr x1,x2;
INOUT char *data;
{
  ExprPtr val;
  Boolean found;

  val = MapGet(x1,x2,&found);
  return Singleton(KONS(val,found?trueValue:falseValue));
}

PRIVATE ExprPtr cross_mapremove(x1,x2,data)
IN CONST ExprPtr x1,x2;
INOUT char *data;
{
  ExprPtr val;

  val = MapRemove(x1,x2);
  return Singleton(val);
}

PRIVATE ExprPtr cross_mapput(x1,x2,data)
IN CONST ExprPtr x1,x2;
INOUT char *data;
{
  ExprPtr val;

  val = MapPut(x1,CAR(x2),CDR(x2));
  return Singleton(val);
}

PRIVATE ExprPtr cross_mapput_behind(x1,x2,data)
IN CONST ExprPtr x1,x2;
INOUT char *data;
{
  ExprPtr val;

  val = MapPutBehind(x1,CAR(x2),CDR(x2));
  return Singleton(val);
}

MAPCOLLECTFUNC(do_get_destroyed_known)
{
  struct getdestroyed_known *data;
  ExprPtr ref;

  data = MAPDATA(struct getdestroyed_known *);
  switch (COUNTEDN(x))
    {
    case Z_UNBOX_REF:
      ref = getPpInfo(COUNTEDVAL(x),data->deco)->def;
      if (ref!=data->def && ref->kind==(CHAR)E_CONS && CAR(ref)->kind==(CHAR)E_COUNTED
	  && COUNTEDN(CAR(ref))==Z_BOX_REF)
	return NewMap();
#ifdef ARRAY_CODES_NOT_YET_DEFINED
    case Z_ARRAY_NTH:
      ref = getPpInfo(COUNTEDVAL(CAR(x)),data->deco)->def;
      if (ref!=data->def && ref->kind==(CHAR)E_CONS && CAR(ref)->kind==(CHAR)E_COUNTED
	  && (COUNTEDN(CAR(ref))==Z_BOX_ARRAY_FROM_LIST ||
	      COUNTEDN(CAR(ref))==Z_BOX_ARRAY))
	return NewMap();
#endif
    default:
      return NewMap();
    }
}

MAPCOLLECTFUNC(do_get_destroyed)
{
  struct getdestroyed *data;

  data = MAPDATA(struct getdestroyed *);
  switch (COUNTEDN(x))
    {
    case Z_UNBOX_REF:
      if (data->fxtypes & FXF_REF)
	break;
      return NewMap();
#ifdef ARRAY_CODES_NOT_YET_DEFINED
    case Z_ARRAY_NTH:
      if (data->fxtypes & FXF_ARRAY)
	break;
      return NewMap();
#endif
    default:
      return NewMap();
    }
  return MapSingleton(x,y);
}

PRIVATE ExprPtr cross_append(x1,x2,data)
IN CONST ExprPtr x1,x2;
INOUT char *data;
{
  ExprPtr val;

  val = Append(x1,x2);
  return Singleton(val);
}

PRIVATE ExprPtr cross_reverse1(x1,x2,data)
IN CONST ExprPtr x1,x2;
INOUT char *data;
{
  ExprPtr val;

  val = Reverse1(x1,x2);
  (void) KEEP(val); /* Reverse1() does not KEEP its result */
  return Singleton(val);
}

PRIVATE ExprPtr cross_domrestrto(x1,x2,data)
IN CONST ExprPtr x1,x2;
INOUT char *data;
{
  ExprPtr val;

  val = MapDomRestrTo(x1,x2);
  return Singleton(val);
}

PRIVATE ExprPtr cross_inter(x1,x2,data)
IN CONST ExprPtr x1,x2;
INOUT char *data;
{
  ExprPtr val;

  val = MapInter(x1,x2);
  return Singleton(val);
}

PRIVATE ExprPtr cross_domrestrby(x1,x2,data)
IN CONST ExprPtr x1,x2;
INOUT char *data;
{
  ExprPtr val;

  val = MapDomRestrBy(x1,x2);
  return Singleton(val);
}

PRIVATE ExprPtr cross_rmu1(x1,x2,data)
IN CONST ExprPtr x1,x2;
INOUT char *data;
{
  ExprPtr val;

  val = mu1(x1,CAR(x2),CDR(x2),true);
  (void) KEEP(val);
  return Singleton(val);
}

PRIVATE ExprPtr cross_rmux(x1,x2,data)
IN CONST ExprPtr x1,x2;
INOUT char *data;
{
  ExprPtr val;

  val = muGnrl(x1,x2,true);
  (void) KEEP(val);
  return Singleton(val);
}

PRIVATE ExprPtr cross_cmpstr(x1,x2,data)
IN CONST ExprPtr x1,x2;
INOUT char *data;
{
  int val;

  val = CompareStrings(x1,x2);
  return Singleton(KINT(val));
}

MAPSWEEPFUNC(do_readmapget1)
{
  struct readmapget *data;
  Boolean found;

  data = MAPDATA(struct readmapget *);
  (void) MapGet(data->x,x,&found);
  data->found |= (found?Q_YES:Q_NO);
  return false;
}

MAPSWEEPFUNC(do_readmapget)
{
  struct readmapget *data;

  data = MAPDATA(struct readmapget *);
  data->x = x;
  (void) MapSweep(data->mapvalset,do_readmapget1,data);
  return false;
}

PRIVATE ExprPtr readMapGet(x,map,deco,eqns,foundp)
IN CONST ExprPtr x,map;
IN CONST MapPtr eqns;
INOUT ExprPtr deco;
OUT COUNTER *foundp;
{
  struct readmapget *data = (struct readmapget *)alloctmp(sizeof(struct readmapget));
  ExprPtr xvalset,get;
  MapPtr mm;
  Boolean found;

  mm = MapGet(zApply(Z_FAKE_MAPGET,map),eqns,&found);
  /* FAKE_MAPGET is a fake opcode, used so that we know what is mapped to what;
     instead of asking whether Z_FAKE_MAPGET(x,m) returns some value y,
     we store this as a map from m to the corresponding map from x to y. */
  if (found)
    {
      get = MapGet(x,mm,&found);
      if (found)
	{
	  *foundp = Q_YES;
	  return get;
	}
    }
  (void) MapGet(zApplyBin(Z_FAKE_NOT_MAPGET,x,map),eqns,&found);
  /* NOT_MAPGET is a fake opcode, used so that we know what is mapped to nothing */
  if (found)
    {
      *foundp = Q_NO;
      return UNIT;
    }
  xvalset = getPpInfo(x,deco)->valset;
  data->mapvalset = getPpInfo(map,deco)->valset;
  if (xvalset==NIL || data->mapvalset==NIL)
    {
      *foundp = Q_MAYBE;
      return UNIT;
    }
  data->found = 0;
  data->deco = deco;
  data->eqns = eqns;
  (void) MapSweep(xvalset,do_readmapget,data);
  *foundp = data->found;
  return UNIT; /* means we don't know which variable in the program holds the result. */
}

PRIVATE void computeOverwrite(x,arg,deco,eqns,pp)
IN CONST ExprPtr x,arg;
INOUT ExprPtr deco;
IN CONST MapPtr eqns;
INOUT ExprPtr pp;
{
  MapPtr mm,mm2;
  Boolean found;

  mm2 = MapGet(zApply(Z_FAKE_MAPGET,CDR(arg)),eqns,&found);
  if (!found)
    mm2 = NewMap();
  mm = MapGet(zApply(Z_FAKE_MAPGET,CAR(arg)),eqns,&found);
  if (!found)
    mm = NewMap();
  mm = Dom(mm);
  /* add all mappings (from x to UNIT) that we are sure are present in the first map;
   we don't keep what is mapped to these x, since we don't know whether they
   will be overwritten by elements in the second map (we cannot decide equality,
   this is why) */
  mm = MapMerge(mm,mm2); /* now, this is the set of bindings we're sure of inside
			    the overwritten map */
  setMapPut(zApply(Z_FAKE_MAPGET,x),mm,&PP(pp)->eqns);
  HITBARRIER(pp);
  mm = MapGet(zApply(Z_FAKE_NOT_MAPGET,CAR(arg)),eqns,&found);
  if (!found)
    mm = NewMap();
  mm2 = MapGet(zApply(Z_FAKE_NOT_MAPGET,CDR(arg)),eqns,&found);
  if (!found)
    mm2 = NewMap();
  mm = SetInter(mm,mm2); /* now, this is the set of items we're sure are not in the domain
			    of the overwritten map */
  setMapPut(zApply(Z_FAKE_NOT_MAPGET,x),mm,&PP(pp)->eqns);
  HITBARRIER(pp);
}

PRIVATE void computeMapRemove(x,arg,deco,eqns,pp)
IN CONST ExprPtr x,arg;
INOUT ExprPtr deco;
IN CONST MapPtr eqns;
INOUT ExprPtr pp;
{
  MapPtr mm;
  Boolean found;

  /* we cannot say much about Z_FAKE_MAPGET, because of possible aliases */
  mm = MapGet(zApply(Z_FAKE_NOT_MAPGET,CDR(arg)),eqns,&found);
  if (!found)
    mm = NewMap();
  mm = SetAdd(CAR(arg),mm);
  setMapPut(zApply(Z_FAKE_NOT_MAPGET,x),mm,&PP(pp)->eqns);
  HITBARRIER(pp);
}

PRIVATE void computeMapPut(x,arg,deco,eqns,pp) /* modeled after computeOverwrite() */
IN CONST ExprPtr x,arg;
INOUT ExprPtr deco;
IN CONST MapPtr eqns;
INOUT ExprPtr pp;
{
  MapPtr mm,mm2;
  Boolean found;

  mm2 = MapSingleton(CAR(x),CAR(CDR(x)));
  mm = MapGet(zApply(Z_FAKE_MAPGET,CDR(CDR(arg))),eqns,&found);
  if (!found)
    mm = NewMap();
  mm = Dom(mm);
  /* add all mappings (from x to UNIT) that we are sure are present in the first map;
   we don't keep what is mapped to these x, since we don't know whether they
   will be overwritten by elements in the second map (we cannot decide equality,
   this is why) */
  mm = MapMerge(mm,mm2); /* now, this is the set of bindings we're sure of inside
			    the overwritten map */
  setMapPut(zApply(Z_FAKE_MAPGET,x),mm,&PP(pp)->eqns);
  HITBARRIER(pp);
  /* cannot say much about Z_FAKE_NOT_MAPGET */
}

PRIVATE void computeMapPutBehind(x,arg,deco,eqns,pp) /* modeled after computeOverwrite() */
IN CONST ExprPtr x,arg;
INOUT ExprPtr deco;
IN CONST MapPtr eqns;
INOUT ExprPtr pp;
{
  MapPtr mm,mm2;
  Boolean found;

  mm = MapSingleton(CAR(x),UNIT);
  mm2 = MapGet(zApply(Z_FAKE_MAPGET,CDR(CDR(arg))),eqns,&found);
  if (!found)
    mm2 = NewMap();
  mm = MapMerge(mm,mm2); /* now, this is the set of bindings we're sure of inside
			    the overwritten map */
  setMapPut(zApply(Z_FAKE_MAPGET,x),mm,&PP(pp)->eqns);
  HITBARRIER(pp);
  /* cannot say much about Z_FAKE_NOT_MAPGET */
}

PRIVATE void computeMapDomRestrTo(x,arg,deco,eqns,pp)
IN CONST ExprPtr x,arg;
INOUT ExprPtr deco;
IN CONST MapPtr eqns;
INOUT ExprPtr pp;
{
  MapPtr mm,mm2;
  Boolean found;

  mm = MapGet(zApply(Z_FAKE_MAPGET,CAR(arg)),eqns,&found);
  if (!found)
    mm = NewMap();
  mm2 = MapGet(zApply(Z_FAKE_MAPGET,CDR(arg)),eqns,&found);
  if (!found)
    mm2 = NewMap();
  mm = MapDomRestrTo(mm,mm2);
  setMapPut(zApply(Z_FAKE_MAPGET,x),mm,&PP(pp)->eqns);
  HITBARRIER(pp);
  /* aliases do not pose any problem here */
  mm = MapGet(zApply(Z_FAKE_NOT_MAPGET,CAR(arg)),eqns,&found);
  if (!found)
    mm = NewMap();
  mm2 = MapGet(zApply(Z_FAKE_NOT_MAPGET,CDR(arg)),eqns,&found);
  if (!found)
    mm2 = NewMap();
  mm = SetUnion(mm,mm2);
  setMapPut(zApply(Z_FAKE_NOT_MAPGET,x),mm,&PP(pp)->eqns);
  HITBARRIER(pp);
  /* and the entries sure not to be in the result are the ones sure not
     to be in either one of the arguments. */
}

PRIVATE void computeMapInter(x,arg,deco,eqns,pp)
IN CONST ExprPtr x,arg;
INOUT ExprPtr deco;
IN CONST MapPtr eqns;
INOUT ExprPtr pp;
{
  MapPtr mm,mm2;
  Boolean found;

  mm = MapGet(zApply(Z_FAKE_MAPGET,CAR(arg)),eqns,&found);
  if (!found)
    mm = NewMap();
  mm2 = MapGet(zApply(Z_FAKE_MAPGET,CDR(arg)),eqns,&found);
  if (!found)
    mm2 = NewMap();
  mm = MapInter(mm,mm2);
  setMapPut(zApply(Z_FAKE_MAPGET,x),mm,&PP(pp)->eqns);
  HITBARRIER(pp);
  /* aliases do not pose any problem here */
  mm = MapGet(zApply(Z_FAKE_NOT_MAPGET,CAR(arg)),eqns,&found);
  if (!found)
    mm = NewMap();
  mm2 = MapGet(zApply(Z_FAKE_NOT_MAPGET,CDR(arg)),eqns,&found);
  if (!found)
    mm2 = NewMap();
  mm = SetUnion(mm,mm2);
  setMapPut(zApply(Z_FAKE_NOT_MAPGET,x),mm,&PP(pp)->eqns);
  HITBARRIER(pp);
  /* and the entries sure not to be in the result are the ones sure not
     to be in either one of the arguments. */
}

PRIVATE void computeMapDomRestrBy(x,arg,deco,eqns,pp)
IN CONST ExprPtr x,arg;
INOUT ExprPtr deco;
IN CONST MapPtr eqns;
INOUT ExprPtr pp;
{
  MapPtr mm,mm2;
  Boolean found;

  /* we cannot say much about Z_FAKE_MAPGET */
  mm = MapGet(zApply(Z_FAKE_NOT_MAPGET,CAR(arg)),eqns,&found);
  if (!found)
    mm = NewMap();
  mm2 = MapGet(zApply(Z_FAKE_MAPGET,CDR(arg)),eqns,&found);
  if (!found)
    mm2 = NewMap();
  mm = MapDomRestrBy(mm,mm2);
  setMapPut(zApply(Z_FAKE_NOT_MAPGET,x),mm,&PP(pp)->eqns);
  HITBARRIER(pp);
}

PRIVATE void computeInfo(x,deco,fragile,eqns,pp)
IN CONST ExprPtr x; /* compute new info for variable x, and put it in pp */
INOUT ExprPtr deco;
IN CONST MapPtr fragile; /* equations that should be erased when any side-effect
			    might occur */
IN CONST MapPtr eqns; /* equations valid just before executing the statement */
OUT ExprPtr pp;
{
  ExprPtr e,arg,valset;
  ppInfo *info;
  INT tag;

  info = PP(pp);
  info->eqns = eqns;
  info->fragile = fragile;
  HITBARRIER(pp);
  e = CAR(x);
  if (e->kind!=(CHAR)E_COUNTED)
    { /* assignment of the form x:=y */
      ppCopy(pp,getPpInfo(e,deco));
      return;
    }
  arg = deConstruct(x,deco,pp);
  if (info->def!=x) /* already simplified */
    return;
  switch (COUNTEDN(e))
    {
    case Z_NIL:
      valset = Singleton(NIL);
      SETTAG(info,E_NIL);
      goto empty;
    case Z_EMPTY:
      valset = Singleton(NewMap());
      SETTAG(info,E_EMPTY);
      goto empty;
    case Z_TUPLE_0:
      valset = Singleton(tuple0);
      SETTAG(info,E_TUPLE_0);
      goto empty;
    case Z_RECORD_0:
      valset = Singleton(record0);
      SETTAG(info,E_RECORD_0);
empty:
      info->storage = SCF_COMPOSITE | SCF_BOXED;
nofx:
      SETQ(info->valset,valset);
      HITBARRIER(pp);
      if (info->def->kind!=(CHAR)E_COUNTED) /* if def is a variable, record it */
	{
	  setMapPut(zApply(COUNTEDN(e),arg),info->def,&info->eqns);
	  HITBARRIER(pp);
	}
fin:
      break;
    case Z_UNBOXED_INT:
      valset = Singleton(arg);
      info->storage = SCF_INT | SCF_UNBOXED;
      goto nofx;
    case Z_UNBOX_INT:
      valset = getPpInfo(arg,deco)->valset; /* the unboxed x has the same values as
					       the boxed one; moreover, it has no tag. */
      info->storage |= SCF_INT | SCF_UNBOXED;
      setMapPut(zApply(Z_BOX_INT,info->def),arg,&info->eqns);
      HITBARRIER(pp);
      /* BOX_INT(UNBOX_INT x) = x */
      goto nofx;
    case Z_BOX_INT:
      valset = getPpInfo(arg,deco)->valset; /* the boxed x has the same values as
					       the unboxed one. */
      SETTAG(info,E_INT);
      info->storage |= SCF_INT | SCF_BOXED;
      setMapPut(zApply(Z_UNBOX_INT,info->def),arg,&info->eqns);
      HITBARRIER(pp);
      /* UNBOX_INT(BOX_INT x) = x */
      goto nofx;
    case Z_UNBOXED_REAL:
      valset = Singleton(arg);
      info->storage = ((arg==zzero)?SCF_REAL0:SCF_REALN) | SCF_UNBOXED;
      goto nofx;
    case Z_UNBOX_REAL:
      {
	ppInfo *ppp;

	ppp = getPpInfo(arg,deco);
	valset = ppp->valset; /* the unboxed x has the same values as
				 the boxed one; moreover, it has no tag. */
	info->storage |= ((SCF_REAL & ppp->storage) | SCF_UNBOXED);
	setMapPut(zApply(Z_BOX_REAL,info->def),arg,&info->eqns);
	HITBARRIER(pp);
	/* BOX_REAL(UNBOX_REAL x) = x */
      }
      goto nofx;
    case Z_BOX_REAL:
      {
	ppInfo *ppp;

	ppp = getPpInfo(arg,deco);
	valset = ppp->valset; /* the unboxed x has the same values as
				 the boxed one. */
	SETTAG(info,E_REAL);
	info->storage |= ((SCF_REAL & ppp->storage) | SCF_BOXED);
	setMapPut(zApply(Z_UNBOX_REAL,info->def),arg,&info->eqns);
	HITBARRIER(pp);
	/* UNBOX_REAL(BOX_REAL x) = x */
      }
      goto nofx;
    case Z_BOX_COMPLEX:
      {
	ppInfo *ppp;

	ppp = getPpInfo(CDR(arg),deco);
	switch (ppp->storage & SCF_REAL)
	  {
	  case SCF_REAL0: /* then number is real, and equal to its real part */
	    ppCopy(pp,getPpInfo(CAR(arg),deco));
	    goto fin;
	  case SCF_REALN: /* then number is not real */
	    valset = info->valset;
	    if (valset==NIL)
	      valset = xProd(arg,deco,cross_complex,NULL);
	    SETTAG(info,E_COMPLEX);
	    goto couple1;
	  }
      }
      valset = info->valset;
      if (valset==NIL)
	valset = xProd(arg,deco,cross_complex,NULL);
      SETTAG(info,E_REAL);
      SETTAG(info,E_COMPLEX);
      info->storage |= SCF_COMPOSITE | SCF_BOXED; /* do not know more */
      goto nofx;
    case Z_BOX_TYPE_APPL:
      tag = E_TYPE_APPL;
      goto couple;
    case Z_BOX_RECORD_1:
    case Z_BOX_UNSHARED_RECORD_1:
      tag = E_RECORD_1;
      goto couple;
    case Z_BOX_CLOSURE:
      tag = E_COMPILED_CLOSURE;
      valset = info->valset;
      if (valset==NIL)
	valset = xProd(arg,deco,cross_cons,(char *)tag);
      SETTAG(info,tag);
      info->storage |= SCF_COMPOSITE | SCF_BOXED;
      setMapPut(zApply(Z_UNBOX_CLOSURE_FN,info->def),CAR(arg),&info->eqns);
      HITBARRIER(pp);
      /* UNBOX_CLOSURE_FN(BOX_CLOSURE(f,e)) = f */
      setMapPut(zApply(Z_UNBOX_CLOSURE_ENV,info->def),CDR(arg),&info->eqns);
      HITBARRIER(pp);
      /* UNBOX_CLOSURE_ENV(BOX_CLOSURE(f,e)) = e */
      goto nofx;
    case Z_UNBOX_CLOSURE_FN:
      valset = info->valset;
      if (valset==NIL)
	{
	  ExprPtr consvalset;

	  consvalset = getPpInfo(arg,deco)->valset;
	  if (consvalset!=NIL)
	    valset = MapCollect(consvalset,do_car,NULL);
	}
      SETTAG(info,E_COMPILED_CLOSURE);
      info->storage |= SCF_CODE | SCF_UNBOXED;
      goto nofx;
    case Z_UNBOX_CLOSURE_ENV:
      valset = info->valset;
      if (valset==NIL)
	{
	  ExprPtr consvalset;

	  consvalset = getPpInfo(arg,deco)->valset;
	  if (consvalset!=NIL)
	    valset = MapCollect(consvalset,do_cdr,NULL);
	}
      SETTAG(info,E_ENV);
      SETTAG(info,E_NIL);
      info->storage |= SCF_COMPOSITE | SCF_BOXED;
      goto nofx;
    case Z_BOX_CONS:
    case Z_BOX_UNSHARED_CONS:
      tag = E_CONS;
      goto couple;
    case Z_BOX_TUPLE_2:
    case Z_BOX_UNSHARED_TUPLE_2:
      tag = E_TUPLE_2;
      goto couple;
    case Z_BOX_SINGLETON:
      tag = E_MAPLET;
      setMapPut(zApply(Z_FAKE_MAPGET,x),MapSingleton(CAR(arg),CDR(arg)),&info->eqns);
      HITBARRIER(pp);
      /* so that we know the binding later on */
couple:
      valset = info->valset;
      if (valset==NIL)
	valset = xProd(arg,deco,cross_cons,(char *)tag);
      SETTAG(info,tag);
couple1:
      info->storage |= SCF_COMPOSITE | SCF_BOXED;
      setMapPut(zApply(Z_CAR,info->def),CAR(arg),&info->eqns); /* CAR(x,y) = x */
      HITBARRIER(pp);
      setMapPut(zApply(Z_CDR,info->def),CDR(arg),&info->eqns); /* CDR(x,y) = y */
      HITBARRIER(pp);
      goto nofx;
    case Z_BOX_PACK:
      tag = E_PACK;
      goto couple;
    case Z_UNBOXED_STRING: /* C-like string (not 0-terminated), without length */
      valset = Singleton(arg);
      info->storage = SCF_STRING | SCF_UNBOXED;
      goto nofx;
    case Z_UNBOX_STRING_LEN:
      valset = info->valset;
      if (valset==NIL)
	{
	  ExprPtr strvalset;

	  strvalset = getPpInfo(arg,deco)->valset;
	  if (strvalset!=NIL)
	    valset = MapCollect(strvalset,do_strlen,NULL);
	}
      info->storage |= SCF_INT | SCF_UNBOXED;
      goto nofx;
    case Z_UNBOX_STRING_TEXT:
      valset = getPpInfo(arg,deco)->valset; /* possible values for the text
					       are all possible values for
					       the string itself; there is no
					       possible tag, since the data
					       is unboxed. */
      info->storage |= SCF_STRING | SCF_UNBOXED;
      goto nofx;
    case Z_BOX_STRING:
      valset = getPpInfo(CDR(arg),deco)->valset; /* possible values for the boxed
						    string are all values for
						    the unboxed text. This means
						    that the length of the string
						    must equal the length of the
						    text as we know it. */
      SETTAG(info,E_STRING);
      info->storage |= SCF_COMPOSITE | SCF_BOXED;
      setMapPut(zApply(Z_UNBOX_STRING_LEN,info->def),CAR(arg),&info->eqns);
      HITBARRIER(pp);
      /* UNBOX_STRING_LEN(BOX_STRING (l,s)) = l */
      setMapPut(zApply(Z_UNBOX_STRING_TEXT,info->def),CDR(arg),&info->eqns);
      HITBARRIER(pp);
      /* UNBOX_STRING_TEXT(BOX_STRING (l,s)) = s */
      goto nofx;
    case Z_BOX_REF:
      SETTAG(info,E_REF);
      info->storage |= SCF_COMPOSITE | SCF_BOXED | SCF_NOINIT; /* cannot be created at init time */
      e = zApply(Z_UNBOX_REF,info->def);
      setMapPut(e,arg,&info->eqns);
      HITBARRIER(pp);
      /* UNBOX_REF(BOX_REF x) = x */
      /* this may evolve during the course of the program, because of SET_REFs */
      setMapPut(e,arg,&info->fragile);
      HITBARRIER(pp);
      /* put it somewhere so that we know how to kill it after
	 impure function calls */
      break;
    case Z_MAKE_PROMISE:
      SETTAG(info,E_COMPILED_CLOSURE);
      info->storage |= SCF_COMPOSITE | SCF_BOXED;
      /* no equations, we don't know anything about promises */
      break;
    case Z_BOX_TUPLE_1:
    case Z_BOX_UNSHARED_TUPLE_1:
      valset = info->valset;
      if (valset==NIL)
	{
	  ExprPtr argvalset;

	  argvalset = getPpInfo(arg,deco)->valset;
	  if (argvalset!=NIL)
	    valset = MapCollect(argvalset,do_ktuple1,NULL);
	}
      SETTAG(info,E_TUPLE_1);
tuple1:
      info->storage |= SCF_COMPOSITE | SCF_BOXED;
      e = zApply(Z_UNBOX_REF,info->def);
      setMapPut(e,arg,&info->eqns);
      HITBARRIER(pp);
      /* UNBOX_REF(BOX_TUPLE_1 x) = x */
      goto nofx;
    case Z_CODE:
      optimizeCode(COUNTEDVAL(e));
      /* Look inside the code, decorate it and patch it.
	 This must be done as though these codes were lambda-lifted, i.e. without
	 keeping any equations from the outside (most interesting equations, notably
	 known functions from the outer environments, should have been inserted
	 inside the code itself by assertions or just by inserting the value). */
      SETTAG(info,E_COMPILED_CLOSURE);
      info->storage = SCF_CODE | SCF_UNBOXED;
      /* and record the closure in the set of local lambdas of deco. */
      setSetAdd(COUNTEDVAL(e),&PD(e)->lambda);
      HITBARRIER(e);
      /*setSetUnion(&PD(e)->lambda,((Code *)EXTDATA(COUNTEDVAL(e)))->lambda);*/
      /*HITBARRIER(e);*/
      valset = Singleton(COUNTEDVAL(e));
      goto nofx;
    case Z_BOX_TUPLE_GNRL:
    case Z_BOX_UNSHARED_TUPLE_GNRL:
      valset = info->valset;
      if (valset==NIL)
	valset = nxProd(arg,deco,ncross_tuple,NULL);
      SETTAG(info,E_TUPLE_GNRL);
      info->storage |= SCF_COMPOSITE | SCF_BOXED;
      {
	COUNTER i,n;

	n = AVAL(arg)->n;
	for (i=0; i<n; i++)
	  {
	    setMapPut(zApplyBin(Z_UNBOX_TUPLE_NTH,info->def,KINT((INT)(i+1))),
		      AVAL(arg)->expr[i],&info->eqns);
	    HITBARRIER(pp);
	  }
	/* UNBOX_TUPLE_NTH(BOX_TUPLE_GNRL (x1,...,xn),k) = xk */
	setMapPut(zApply(Z_FAKE_UNBOX_TUPLE_LEN,info->def),KINT(n),&info->eqns);
	HITBARRIER(pp);
	/* Z_FAKE_UNBOX_TUPLE_LEN is a fake opcode, put here to get back
	   the length in case Z_UNBOX_TUPLE_NTH. */
      }
      goto nofx;
    case Z_UNBOX_TUPLE_NTH:
      valset = info->valset;
      if (valset==NIL)
	{
	  ExprPtr argvalset;

	  argvalset = getPpInfo(CAR(arg),deco)->valset;
	  if (argvalset!=NIL)
	    valset = MapCollect(argvalset,do_tuple_nth,(char *)CDR(arg));
	}
      SetTags(info,valset);
      info->storage |= SCF_ALL_TYPES | SCF_BOXED;
      goto nofx;
    case Z_BOX_RECORD_GNRL:
    case Z_BOX_UNSHARED_RECORD_GNRL:
      valset = info->valset;
      if (valset==NIL)
	valset = nxProd(arg,deco,ncross_record,NULL);
      SETTAG(info,E_RECORD_GNRL);
      info->storage |= SCF_COMPOSITE | SCF_BOXED;
      {
	COUNTER i,n;

	n = AVAL(arg)->n;
	setMapPut(zApply(Z_GET_SHAPE,info->def),AVAL(arg)->expr[0],&info->eqns);
	HITBARRIER(pp);
	/* GET_SHAPE(BOX_RECORD_GNRL (shape,x1,...,xn)) = shape */
	for (i=1; i<n; i++)
	  {
	    setMapPut(zApplyBin(Z_UNBOX_TUPLE_NTH,info->def,KINT((INT)i)),
		      AVAL(arg)->expr[i],&info->eqns);
	    HITBARRIER(pp);
	  }
	/* UNBOX_TUPLE_NTH(BOX_RECORD_GNRL (shape,x1,...,xn),k) = xk */
	setMapPut(zApply(Z_FAKE_UNBOX_TUPLE_LEN,info->def),KINT(n-1),&info->eqns);
	HITBARRIER(pp);
	/* Z_FAKE_UNBOX_TUPLE_LEN is a fake opcode, put here to get back
	   the length in case Z_UNBOX_TUPLE_NTH. */
      }
      goto nofx;
    case Z_OVERWRITE:
      {
	ppInfo *pp1,*pp2;

	pp1 = getPpInfo(CAR(arg),deco);
	if (CAR(arg)==CDR(arg))
	  {
	    ppCopy(pp,pp1); /* e ++ e == e */
	    break;
	  }
	pp2 = getPpInfo(CDR(arg),deco);
	if (pp1->valset->kind==(CHAR)E_MAPLET && MAPLETX(pp1->valset)==NewMap())
	  {
	    ppCopy(pp,pp2); /* {} ++ e = e */
	    break;
	  }
	if (pp2->valset->kind==(CHAR)E_MAPLET && MAPLETX(pp2->valset)==NewMap())
	  {
	    ppCopy(pp,pp1); /* e ++ {} = e */
	    break;
	  }
	computeOverwrite(x,arg,deco,eqns,pp);
	valset = info->valset;
	if (valset==NIL)
	  valset = xProd(arg,deco,cross_overwrite,NULL);
map:
	if (valset!=NIL)
	  (void) MapSweep(valset,do_settag,info);
	else
	  {
	    SETTAG(info,E_EMPTY);
	    SETTAG(info,E_MAPLET);
	    SETTAG(info,E_MAP);
	  }
      }
      info->storage |= SCF_COMPOSITE | SCF_BOXED;
      goto nofx;
    case Z_ENEXT:
      SETTAG(info,E_ENV);
      info->storage |= SCF_COMPOSITE | SCF_BOXED;
      break;
    case Z_CONSENV:
      SETTAG(info,E_ENV);
      info->storage |= SCF_COMPOSITE | SCF_BOXED | SCF_NOINIT;
      /* Z_CONSENV produces new environments each time */
      setMapPut(zApply(Z_ENEXT,info->def),CDR(arg),&info->eqns);
      HITBARRIER(pp);
      /* ENEXT(CONSENV(x,y)) = y */
      break;
    case Z_FAKEENV:
      SETTAG(info,E_ENV);
      info->storage |= SCF_COMPOSITE | SCF_BOXED | SCF_NOINIT;
      /* Z_FAKEENV produces new environments each time, although its EVARS is that of GLOBENV */
      setMapPut(zApply(Z_ENEXT,info->def),arg,&info->eqns);
      HITBARRIER(pp);
      /* ENEXT(FAKEENV(x)) = x */
      /* We should add the fact that EVARS(FAKEENV(x)) = EVARS(GLOBENV), but
	 this would be hard to do (EVARS cannot be submitted to common subexpression
	 elimination, as GC may move them around, so we would have to add all
	 equations Z_VAR(FAKEENV(x),i) = Z_VAR(GLOBENV,i); on the other hand,
	 compileEnvIndex() in translate.c does all needed optimizations already... */
      break;
    case Z_VAR:
      SetTags(info,info->valset);
      info->storage |= SCF_ALL_TYPES | SCF_BOXED;
      break;
    case Z_LET:
      info->storage = SCF_TOP; /* not stored */
      setMapPut(zApplyBin(Z_VAR,CAR(arg),CDR(CDR(arg))),
		CAR(CDR(arg)),&info->eqns);
      HITBARRIER(pp);
      /* after Z_LET(env,val,rank), we know that Z_VAR(env,rank)=val */
      /* moreover, it is not fragile, since modifications of environments
	 occur inside a function, and are never visible outside. */
      break;
    case Z_RAISE:
      info->storage = SCF_TOP; /* not stored */
      break;
    case Z_NOT:
      valset = info->valset;
      if (valset==NIL)
	{
	  ExprPtr argvalset;

	  argvalset = getPpInfo(arg,deco)->valset;
	  if (argvalset!=NIL)
	    valset = MapCollect(argvalset,do_not,NULL);
	}
      /* no tag, since value is unboxed */
      info->storage |= SCF_BOOL | SCF_UNBOXED;
      goto boolnofx;
    case Z_EQUAL:
      valset = info->valset;
      if (valset==NIL)
	xProd(arg,deco,cross_equal,NULL);
      /* no tag, since value is unboxed */
      info->storage |= SCF_BOOL | SCF_UNBOXED;
      if (CAR(arg)==CDR(arg))
	valset = Singleton(trueValue);
      else if (noCommonTags(CAR(arg),CDR(arg),deco))
	valset = Singleton(falseValue);
boolnofx:
      if (valset==NIL)
	valset = Pair(trueValue,falseValue);
      goto nofx;
    case Z_CAR:
      valset = info->valset;
      if (valset==NIL)
	{
	  ExprPtr consvalset;

	  consvalset = getPpInfo(arg,deco)->valset;
	  if (consvalset!=NIL)
	    valset = MapCollect(consvalset,do_car,NULL);
	}
      SetTags(info,valset);
      info->storage |= SCF_ALL_TYPES | SCF_BOXED;
      goto nofx;
    case Z_CDR:
      valset = info->valset;
      if (valset==NIL)
	{
	  ExprPtr consvalset;

	  consvalset = getPpInfo(arg,deco)->valset;
	  if (consvalset!=NIL)
	    valset = MapCollect(consvalset,do_cdr,NULL);
	}
      SetTags(info,valset);
      info->storage |= SCF_ALL_TYPES | SCF_BOXED;
      goto nofx;
    case Z_UNBOX_REF:
deref:
      valset = info->valset;
      SetTags(info,valset);
      info->storage |= SCF_ALL_TYPES | SCF_BOXED;
      goto nofx;
    case Z_HAS_TAG:
      valset = info->valset;
      if (valset==NIL)
	{
	  ppInfo *ppp;

	  ppp = getPpInfo(CAR(arg),deco);
	  if (HASTAG(ppp,IVAL(CDR(arg))))
	    {
	      if (HasOnlyTag(ppp,IVAL(CDR(arg))))
		valset = Singleton(trueValue);
	      else
		valset = Pair(trueValue,falseValue);
	    }
	  else
	    valset = Singleton(falseValue);
	}
      info->storage |= SCF_BOOL | SCF_UNBOXED;
      goto boolnofx;
    case Z_GET_VECTOR_LEN:
      {
	Boolean found;

	e = MapGet(zApply(Z_FAKE_UNBOX_TUPLE_LEN,arg),eqns,&found);
	if (found)
	  valset = Singleton(e);
	else
	  {
	    ExprPtr argvalset;

	    valset = info->valset;
	    argvalset = getPpInfo(info->def,deco)->valset;
	    if (argvalset!=NIL)
	      valset = MapCollect(argvalset,do_tuple_len,NULL);
	  }
	info->storage |= SCF_INT | SCF_UNBOXED;
	goto nofx;
      }
    case Z_TYPE_MATCH:
      valset = info->valset;
      if (valset==NIL)
	valset = xProd(arg,deco,cross_type_match,NULL);
      info->storage |= SCF_BOOL | SCF_UNBOXED;
      goto boolnofx;
      break;
    case Z_MAP_CARD_DEC: /* second argument is always an integer constant */
      {
	ppInfo *ppp;

	ppp = getPpInfo(CAR(arg),deco);
	valset = info->valset;
	if (valset==NIL && ppp->valset!=NIL)
	    valset = MapCollect(ppp->valset,do_map_card_dec,CDR(arg));
	if (HASTAG(ppp,E_EMPTY))
	  {
	    if (!HASTAG(ppp,E_MAPLET) && !HASTAG(ppp,E_MAP))
	      valset = Singleton(CDR(arg)); /* MapCardDec({},n) = n */
	  }
	else if (HASTAG(ppp,E_MAPLET))
	  {
	    if (!HASTAG(ppp,E_MAP))
	      valset = Singleton(KINT(IVAL(arg)-1)); /* MapCardDec({x},n) = n-1 */
	  }
	}
      info->storage |= SCF_INT | SCF_UNBOXED;
      goto nofx;
    case Z_INT_LEQ:
      valset = info->valset;
      if (valset==NIL)
	valset = xProd(arg,deco,cross_int_leq,NULL);
      info->storage |= SCF_BOOL | SCF_UNBOXED;
      goto boolnofx;
    case Z_GET_SHAPE:
      valset = info->valset;
      if (valset==NIL)
	{
	  ExprPtr argvalset;

	  argvalset = getPpInfo(arg,deco)->valset;
	  if (argvalset!=NIL)
	    valset = MapCollect(argvalset,do_get_shape,NULL);
	}
      if (valset!=NIL)
	(void) MapSweep(valset,do_settag,info);
      else
	{
	  SETTAG(info,E_EMPTY);
	  SETTAG(info,E_MAPLET);
	  SETTAG(info,E_MAP);
	}
      info->storage |= SCF_COMPOSITE | SCF_BOXED;
      goto nofx;
    case Z_MAPGET:
      valset = info->valset;
      if (valset==NIL)
	valset = xProd(arg,deco,cross_mapget,NULL);
      {
	ExprPtr get;
	COUNTER found;

	get = readMapGet(CAR(arg),CDR(arg),deco,eqns,&found);
	if (found==Q_YES)
	  {
	    if (get!=UNIT) /* and if we know what it is mapped to */
	      {
		setMapPut(zApply(Z_GET_MAPGET_Y,x),get,&info->eqns);
		HITBARRIER(pp);
	      }
	    setMapPut(zApply(Z_FAKE_MAPGET_FOUND,x),trueValue,&info->eqns);
	    HITBARRIER(pp);
	    /* Z_FAKE_MAPGET_FOUND is a fake opcode to establish communication with
	       Z_GET_MAPGET_FOUND */
	  }
	else if (found==Q_NO)
	  {
	    setMapPut(zApply(Z_FAKE_MAPGET_FOUND,x),falseValue,&info->eqns);
	    HITBARRIER(pp);
	    /* Z_FAKE_MAPGET_FOUND is a fake opcode to establish communication with
	       Z_GET_MAPGET_FOUND */
	  }
      }
      /* no tags, Z_MAPGET returns a virtual couple */
      info->storage |= SCF_COMPOSITE | SCF_LOCAL;
      goto nofx;
    case Z_GET_MAPGET_Y:
      valset = info->valset;
      if (valset==NIL)
	{
	  ExprPtr compvalset;

	  compvalset = getPpInfo(arg,deco)->valset;
	  if (compvalset!=NIL)
	    valset = MapCollect(compvalset,do_car,NULL);
	}
      SetTags(info,valset);
      info->storage |= SCF_ALL_TYPES | SCF_BOXED;
      goto nofx;
    case Z_GET_MAPGET_FOUND:
      valset = info->valset;
      if (valset==NIL)
	{
	  ExprPtr compvalset;
	  Boolean found;

	  compvalset = MapGet(zApply(Z_FAKE_MAPGET_FOUND,arg),eqns,&found);
	  if (found)
	    valset = Singleton(compvalset);
	  else
	    {
	      compvalset = getPpInfo(arg,deco)->valset;
	      if (compvalset!=NIL)
		valset = MapCollect(compvalset,do_cdr,NULL);
	    }
	}
      info->storage |= SCF_BOOL | SCF_UNBOXED;
      goto boolnofx;
    case Z_TUPLE_SUB_NTH:
      valset = info->valset;
      if (valset==NIL)
	{
	  ExprPtr argvalset;

	  argvalset = getPpInfo(CDR(arg),deco)->valset;
	  if (argvalset->kind==(CHAR)E_MAPLET) /* if we know the index */
	    { /* then try to compute it from Z_UNBOX_TUPLE_NTH */
	      Boolean found;

	      e = MapGet(zApplyBin(Z_UNBOX_TUPLE_NTH,CAR(arg),
				   MAPLETX(argvalset)),eqns,&found);
	      if (found)
		{
		  ppCopy(pp,getPpInfo(e,deco));
		  break;
		}
	    }
	  valset = xProd(arg,deco,cross_array_nth,NULL);
	}
      SetTags(info,valset);
      info->storage |= SCF_ALL_TYPES | SCF_BOXED;
      goto nofx;
    case Z_MAPREMOVE:
      {
	COUNTER found;
	ppInfo *ppp;

	ppp = getPpInfo(CDR(arg),deco);
	if ((ppp->valset->kind==(CHAR)E_MAPLET && MAPLETX(ppp->valset)==NewMap()) ||
	    ((void) readMapGet(CAR(arg),CDR(arg),deco,eqns,&found), found==Q_NO))
	  { /* removing x from m such that m={} or x not in m yields m */
	    ppCopy(pp,ppp);
	    break;
	  }
	valset = info->valset;
	if (valset==NIL)
	  valset = xProd(arg,deco,cross_mapremove,NULL);
	if (HASTAG(ppp,E_MAP))
	  { SETTAG(info,E_MAP); SETTAG(info,E_MAPLET); }
	if (HASTAG(ppp,E_MAPLET))
	  { SETTAG(info,E_MAPLET); SETTAG(info,E_EMPTY); }
	if (HASTAG(ppp,E_EMPTY))
	  { SETTAG(info,E_EMPTY); }
	info->storage |= SCF_COMPOSITE | SCF_BOXED;
	computeMapRemove(x,arg,deco,eqns,pp);
	goto nofx;
      }
    case Z_MAPPUT:
      valset = info->valset;
      if (valset==NIL)
	valset = xProd3(arg,deco,cross_mapput,NULL);
      computeMapPut(x,arg,deco,eqns,pp);
mapput:
      {
	ppInfo *ppp;

	ppp = getPpInfo(CDR(CDR(arg)),deco);
	if (HASTAG(ppp,E_MAP))
	  { SETTAG(info,E_MAP); }
	if (HASTAG(ppp,E_MAPLET))
	  { SETTAG(info,E_MAPLET); SETTAG(info,E_MAP); }
	if (HASTAG(ppp,E_EMPTY))
	  { SETTAG(info,E_MAPLET); }
	info->storage |= SCF_COMPOSITE | SCF_BOXED;
	goto nofx;
      }
    case Z_MAPPUTBEHIND:
      {
	COUNTER found;

	(void) readMapGet(CAR(arg),CDR(CDR(arg)),deco,eqns,&found);
	if (found==Q_YES)
	  { /* if a in c, putbehind(a,b,c) -> c */
	    ppCopy(pp,getPpInfo(CDR(CDR(arg)),deco));
	    break;
	  }
      }
      computeMapPutBehind(x,arg,deco,eqns,pp);
      valset = info->valset;
      if (valset==NIL)
	valset = xProd3(arg,deco,cross_mapput_behind,NULL);
      goto mapput;
    case Z_BOX_QUEUE:
      SETTAG(info,E_CONS); /* useless, since queues are not first-class citizens */
      info->storage |= SCF_COMPOSITE | SCF_BOXED | SCF_NOINIT; /* cannot be created at init time */
      /*e = zApply(Z_CDR,info->def);
	setMapPut(e,arg,&info->eqns);
	HITBARRIER(pp);
	// CDR(BOX_QUEUE x) = x; x==NIL, normally.
	// this may evolve during the course of the program, because of CONCAT_QUEUEs
	// don't put it in fragile, since it is kept purely locally
	*/
      /* no equations; what's inside a queue is done purely for side-effects */
      break;
    case Z_BOX_1_QUEUE:
      SETTAG(info,E_CONS); /* useless, since queues are not first-class citizens */
      info->storage |= SCF_COMPOSITE | SCF_BOXED | SCF_NOINIT; /* cannot be created at init time */
      /* no equations: what's inside a queue is done purely for side-effects. */
      break;
    case Z_UNBOXED_FALSE:
      valset = Singleton(falseValue);
      info->storage = SCF_BOOL | SCF_UNBOXED;
      goto nofx;
    case Z_UNBOXED_TRUE:
      valset = Singleton(trueValue);
      info->storage = SCF_BOOL | SCF_UNBOXED;
      goto nofx;
    case Z_REF_SET:
      {
	Boolean found;

	e = zApply(Z_UNBOX_REF,CAR(arg));
	(void) MapGet(e,eqns,&found);
	{
	  MapPtr destroyed,aux;

	  if (found)
	    {
	      struct getdestroyed_known *data = (struct getdestroyed_known *)
		alloctmp(sizeof(struct getdestroyed_known));

	      data->def = info->def;
	      data->deco = deco;
	      destroyed = MapCollect(fragile,do_get_destroyed_known,data);
	    }
	  else
	    /* if we don't know precisely what ref is modified,
	       find all possible candidates; they are in fragile. */
	    {
	      struct getdestroyed *data = (struct getdestroyed *)
		alloctmp(sizeof(struct getdestroyed));

	      data->fxtypes = FXF_REF;
	      /*data->obj = CAR(arg);*/
	      data->deco = deco;
	      destroyed = MapCollect(fragile,do_get_destroyed,data);
	    }
	  aux = MapDomRestrBy(fragile,destroyed);
	  SETQ(info->fragile,aux);
	  HITBARRIER(pp);
	  aux = MapDomRestrBy(eqns,destroyed);
	  SETQ(info->eqns,aux);
	  HITBARRIER(pp);
	}
	setMapPut(e,arg,&info->fragile);
	HITBARRIER(pp);
	setMapPut(e,arg,&info->eqns);
	HITBARRIER(pp);
      }
      PD(deco)->fxtypes |= FXF_REF;
unitfx:
      info->storage = SCF_TOP;
      valset = NIL;
      SETQ(info->valset,valset);
      HITBARRIER(pp);
      break;
    case Z_BOX_NUMTYPE:
      if (valset==NIL)
	{
	  ExprPtr argvalset;

	  argvalset = getPpInfo(arg,deco)->valset;
	  if (argvalset!=NIL)
	    valset = MapCollect(argvalset,do_knumtype,NULL);
	}
      SETTAG(info,E_NUMTYPE);
      goto tuple1;
    case Z_BOX_TYPEFX:
      tag = E_TYPEFX;
      goto couple;
    case Z_APPLY:
    case Z_CALL: /* call of a known C function (with arbitrary prototype) */
      /* format: Z_CALL ([z_code,other arguments]), where z_code is a variable
	 referencing a Code (encapsulated in an external), and [...] means a
	 n-tuple. */
      valset = NIL; /* cannot compute application (must inline to do that) */
      {
	struct fundata data;
	struct getdestroyed *ddata = (struct getdestroyed *)alloctmp(sizeof(struct getdestroyed));
	MapPtr destroyed,aux;

	getApplyData(e,deco,&data);
	if ((data.fxtypes & FXF_ALL)==FXF_ALL) /* all fragile eqns are out */
	  destroyed = fragile;
	else
	  {
	    ddata->fxtypes = data.fxtypes;
	    ddata->deco = deco;
	    destroyed = MapCollect(fragile,do_get_destroyed,ddata);
	  }
	aux = MapDomRestrBy(fragile,destroyed);
	SETQ(info->fragile,aux);
	HITBARRIER(pp);
	aux = MapDomRestrBy(eqns,destroyed);
	SETQ(info->eqns,aux);
	HITBARRIER(pp);
	info->storage |= data.storage;
	memcpy(info->tags,data.tags,NTAGS);
	PD(deco)->fxtypes |= data.fxtypes;
      }
      valset = NIL; /* do not know the value */
      SETQ(info->valset,valset);
      HITBARRIER(pp);
      break;
    case Z_APPEND:
      {
	ppInfo *pp1,*pp2;

	pp1 = getPpInfo(CAR(arg),deco);
	pp2 = getPpInfo(CDR(arg),deco);
	if (pp1->valset->kind==(CHAR)E_MAPLET && MAPLETX(pp1->valset)==NIL)
	  {
	    ppCopy(pp,pp2); /* nil @ e = e */
	    break;
	  }
	if (pp2->valset->kind==(CHAR)E_MAPLET && MAPLETX(pp2->valset)==NIL)
	  {
	    ppCopy(pp,pp1); /* e @ nil = e */
	    break;
	  }
      }
      valset = info->valset;
      if (valset==NIL)
	valset = xProd(arg,deco,cross_append,NULL);
list:
      info->storage |= SCF_COMPOSITE | SCF_BOXED;
      if (valset!=NIL)
	(void) MapSweep(valset,do_settag,info);
      else
	{
	  SETTAG(info,E_NIL);
	  SETTAG(info,E_CONS);
	}
      goto nofx;
    case Z_MAPDOMRESTRTO:
      computeMapDomRestrTo(x,arg,deco,eqns,pp);
      valset = info->valset;
      if (valset==NIL)
	valset = xProd(arg,deco,cross_domrestrto,NULL);
inter:
      {
	ppInfo *pp1,*pp2;

	pp1 = getPpInfo(CAR(arg),deco);
	if (CAR(arg)==CDR(arg))
	  {
	    ppCopy(pp,pp1); /* e <| e = e */
	    break;
	  }
	pp2 = getPpInfo(CDR(arg),deco);
	if (pp1->valset->kind==(CHAR)E_MAPLET && MAPLETX(pp1->valset)==NewMap())
	  {
	    ppCopy(pp,pp1); /* {} <| e = {} */
	    break;
	  }
	if (pp2->valset->kind==(CHAR)E_MAPLET && MAPLETX(pp2->valset)==NewMap())
	  {
	    ppCopy(pp,pp2); /* e <| {} = {} */
	    break;
	  }
      }
      goto map;
    case Z_MAPINTER:
      computeMapInter(x,arg,deco,eqns,pp);
      valset = info->valset;
      if (valset==NIL)
	valset = xProd(arg,deco,cross_inter,NULL);
      goto inter;
    case Z_MAPDOMRESTRBY:
      {
	ppInfo *pp1,*pp2;

	pp1 = getPpInfo(CAR(arg),deco);
	pp2 = getPpInfo(CDR(arg),deco);
	if ((pp1->valset->kind==(CHAR)E_MAPLET && MAPLETX(pp1->valset)==NewMap()) ||
	    (pp2->valset->kind==(CHAR)E_MAPLET && MAPLETX(pp2->valset)==NewMap()))
	  {
	    ppCopy(pp,pp2); /* {} <-| e = e, e <-| {} = {} */
	    break;
	  }
	if (pp1->def==pp2->def)
	  { /* e <-| e = {} */
	    valset = Singleton(NewMap());
	    goto map;
	  }
      }
      computeMapDomRestrBy(x,arg,deco,eqns,pp);
      valset = info->valset;
      if (valset==NIL)
	valset = xProd(arg,deco,cross_domrestrby,NULL);
      goto map;
    case Z_START_LISTSWEEP: /* initialize a local ref to sublists */
    case Z_START_MAPSWEEP: /* or to reference points in a map */
    case Z_START_SUBMAPSWEEP: /* or to submaps of a map */
      /* as this cannot be constant and is not a 1st-class citizen,
	 we just say that we do not know anything on it. */
      info->storage |= SCF_COMPOSITE | SCF_LOCAL | SCF_NOINIT; /* cannot be created at init time */
      break;
    case Z_INC_LISTSWEEP: /* done purely for side-effects */
    case Z_INC_MAPSWEEP:
    case Z_INC_SUBMAPSWEEP:
    case Z_START_HANDLE:
      info->storage |= SCF_BOOL | SCF_UNBOXED;
      valset = Pair(trueValue,falseValue);
      SETQ(info->valset,valset);
      HITBARRIER(pp);
      SETTAG(info,E_STRING);
      break;
    case Z_UNBOX_LISTSWEEP: /* get next element of the sequence in the local ref */
    case Z_UNBOX_MAPSWEEP:
      /* we don't keep any equations on these UNBOXes, as they are created only
	 by translate.c and are already optimized there. */
      goto deref;
    case Z_UNBOX_SUBMAPSWEEP: /* get next submap */
      SETTAG(info,E_EMPTY);
      SETTAG(info,E_MAPLET);
      SETTAG(info,E_MAP);
      info->storage |= SCF_ALL_TYPES | SCF_BOXED;
      goto nofx;
    case Z_BOX_LOCAL_REF: /* a fake, local ref; treated as an unboxed ref */
      {
	ppInfo *ppp;

	ppp = getPpInfo(arg,deco);
	memcpy(info->tags,ppp->tags,NTAGS);
	info->storage = ppp->storage;
	info->storage |= SCF_LOCAL | SCF_NOINIT; /* cannot be created at init time */
      }
      e = zApply(Z_UNBOX_LOCAL_REF,info->def);
      setMapPut(e,arg,&info->eqns);
      HITBARRIER(pp);
      /* UNBOX_LOCAL_REF(BOX_LOCAL_REF x) = x */
      /* this may evolve, because of SET_LOCAL_REFs, but it is not fragile
       as only a LOCAL_REF_SET may change it. */
      break;
    case Z_UNBOX_LOCAL_REF:
      {
	ppInfo *ppp;

	ppp = getPpInfo(arg,deco);
	memcpy(info->tags,ppp->tags,NTAGS);
	valset = info->valset;
	info->storage = ppp->storage & ~SCF_LOCAL;
      }
      goto nofx;
    case Z_LOCAL_REF_SET:
      e = zApply(Z_UNBOX_LOCAL_REF,CAR(arg));
      /* as local refs are purely local and not 1st class, such an
	 equation must be present in eqns. We overwrite it now. */
      setMapPut(e,arg,&info->eqns);
      HITBARRIER(pp);
      goto unitfx;
    case Z_REVERSE1:
      {
	ppInfo *pp1;

	pp1 = getPpInfo(CAR(arg),deco);
	if (pp1->valset->kind==(CHAR)E_MAPLET && MAPLETX(pp1->valset)==NIL)
	  {
	    ppInfo *pp2;

	    pp2 = getPpInfo(CDR(arg),deco);
	    ppCopy(pp,pp2); /* revtail(nil,e) = e */
	    break;
	  }
      }
      valset = info->valset;
      if (valset==NIL)
	valset = xProd(arg,deco,cross_reverse1,NULL);
      goto list;
    case Z_QUEUE_TO_LIST:
      valset = info->valset;
      goto list;
    case Z_CONCAT_QUEUE:
      goto unitfx; /* done purely for side-effects */
    case Z_BOX_UNSHARED_EXCEPTION:
      SETTAG(info,E_EXCEPTION);
      info->storage |= SCF_COMPOSITE | SCF_BOXED | SCF_NOINIT; /* cannot be created at init time */
      /* no equations, since we never need to look inside an E_EXCEPTION. */
      break;
    case Z_BOX_SHARED_EXCEPTION:
      SETTAG(info,E_EXCEPTION);
      info->storage |= SCF_COMPOSITE | SCF_BOXED;
      /* no equations, since we never need to look inside an E_EXCEPTION. */
      break;
    case Z_BOX_BOOL:
      valset = getPpInfo(arg,deco)->valset; /* the boxed x has the same values as
					       the unboxed one. */
      SETTAG(info,E_STRING);
      info->storage |= SCF_BOOL | SCF_BOXED;
      setMapPut(zApply(Z_UNBOX_BOOL,info->def),arg,&info->eqns);
      HITBARRIER(pp);
      /* UNBOX_BOOL(BOX_BOOL x) = x */
      goto nofx;
    case Z_UNBOX_BOOL:
      valset = getPpInfo(arg,deco)->valset; /* the unboxed x has the same values as
					       the boxed one; moreover, it has no tag. */
      info->storage |= SCF_BOOL | SCF_UNBOXED;
      setMapPut(zApply(Z_BOX_BOOL,info->def),arg,&info->eqns);
      HITBARRIER(pp);
      /* BOX_BOOL(UNBOX_BOOL x) = x */
      goto nofx;
    case Z_END_HANDLE:
      goto unitfx;
    case Z_C_GLOBAL:
    case Z_C_CONST:
      valset = info->valset;
      info->storage = IVAL(CDR(arg));
      goto nofx;
    case Z_GETTHEEXPR:
      valset = NIL;
      info->storage = (SCF_COMPOSITE | SCF_BOXED | SCF_NOINIT);
      goto nofx;
    case Z_SET_C_GLOBAL:
    case Z_SET_C_CONST:
      goto unitfx;
    case Z_RMU1:
    case Z_UNSHARED_RMU1:
      /* we should add equations of the form #a (x ++|[a = e]|) = e
	 and #b (x ++|[a = e]|) = #b x, but this seems to be hard to do
	 other by assertions. */
      valset = info->valset;
      if (valset==NIL)
	valset = xProd3(arg,deco,cross_rmu1,NULL);
rec:
      if (valset!=NIL)
	(void) MapSweep(valset,do_settag,info);
      else
	{
	  SETTAG(info,E_RECORD_0);
	  SETTAG(info,E_RECORD_1);
	  SETTAG(info,E_RECORD_GNRL);
	}
      info->storage |= SCF_COMPOSITE | SCF_BOXED;
      goto nofx;
    case Z_RMUX:
    case Z_UNSHARED_RMUX:
      /* we should add equations of the form #a (x ++|[a = e]|) = e
	 and #b (x ++|[a = e]|) = #b x, but this seems to be hard to do
	 other by assertions. */
      valset = info->valset;
      if (valset==NIL)
	valset = xProd(arg,deco,cross_rmux,NULL);
      goto rec;
    case Z_CMPSTR:
      valset = info->valset;
      if (valset==NIL)
	valset = xProd(arg,deco,cross_cmpstr,NULL);
      SETTAG(info,E_INT);
      info->storage |= SCF_INT | SCF_UNBOXED;
      goto nofx;
    case Z_ASSERT_EQUAL: /* assert(a,b) means we must rewrite a into b */
      {
	ppInfo *pp1,*pp2;

	pp1 = getPpInfo(CAR(arg),deco);
	if (pp1->def==CAR(arg)) /* only if a is not already rewritten into
				   something else. */
	  {
	    pp2 = getPpInfo(CDR(arg),deco);
	    ppCopy(pp,pp2);
	  }
      }
      break;
    default:
      kinderror(message(485),COUNTEDN(e));
      break;
    }
}

PRIVATE MapPtr EqnInter1x(map1,map2,mask)
IN CONST REGISTER MapPtr map1,map2;
IN CONST REGISTER CODE mask;
{ /* map1 is a E_MAPLET */
  ExprPtr x,y;
  Boolean found;

  x = MAPLETX(map1);
  y = MapGet1(x,map2,mask,&found);
  if (!found)
    return NewMap();
  if (y==MAPLETY(map1))
    return map1;
  if (x->kind==(CHAR)E_COUNTED)
    switch (COUNTEDN(x))
      {
      case Z_FAKE_MAPGET:
      case Z_FAKE_NOT_MAPGET:
	return MapSingleton(x,MapInter(MAPLETY(map1),y));
      default:
	return NewMap();
      }
}

PRIVATE MapPtr EqnInter(map1,map2,mask)
IN CONST REGISTER MapPtr map1,map2;
IN CONST REGISTER CODE mask;
{
  ExprPtr left,right;

  switch (map1->kind)
    {
    case E_EMPTY:
      return NewMap();
    case E_MAPGET:
      return EqnInter1x(map1,map2,mask);
    case E_MAP:
      switch (map2->kind)
	{
	case E_EMPTY:
	  return NewMap();
	case E_MAPGET:
	  return EqnInter1x(map2,map1,mask);
	case E_MAP:
	  mask <<= 1;
	  left = EqnInter(MAPLEFT(map1),MAPLEFT(map2),mask);
	  right = EqnInter(MAPRIGHT(map1),MAPRIGHT(map2),mask);
	  if (left==NewMap() && right->kind!=(CHAR)E_MAP)
	    return right;
	  if (right==NewMap() && left->kind!=(CHAR)E_MAP)
	    return left;
	  right = MapKons(left,right);
	  return KEEP(right);
	default:
	  map1 = map2;
	  /*fallthrough*/
	}
    default:
      kinderror(message(523),map1->kind);
      /*NOTREACHED*/
    }
}

PRIVATE Boolean ppMerge(pp,newpp,stmt)
INOUT ExprPtr pp;
IN CONST ExprPtr newpp;
IN CONST ExprPtr stmt;
{
  ExprPtr eqns;
  ppInfo *info,*newinfo;
  Boolean changed;

  changed = false;
  info = PP(pp);
  newinfo = PP(newpp);
  if (newinfo->def!=info->def)
    {
      changed = true;
      SETQ(info->def,stmt);
      HITBARRIER(pp);
    }
  if (newinfo->eqns!=info->eqns)
    {
      changed = true;
      eqns = EqnInter(info->eqns,newinfo->eqns,(CODE)(1L<<NLOWBITS));
      SETQ(info->eqns,eqns);
      HITBARRIER(pp);
    }
  if (newinfo->fragile!=info->fragile)
    {
      changed = true;
      eqns = EqnInter(info->fragile,newinfo->fragile,(CODE)(1L<<NLOWBITS));
      SETQ(info->fragile,eqns);
      HITBARRIER(pp);
    }
  if (newinfo->valset!=info->valset)
    {
      changed = true;
      if (newinfo->valset==NIL)
	info->valset = NIL;
      else
	{
	  if (info->valset==NIL)
	    { SETQ(info->valset,newinfo->valset); }
	  else
	    setSetUnion(&info->valset,newinfo->valset);
	  HITBARRIER(pp);
	}
    }
  if (newinfo->storage!=info->storage)
    {
      changed = true;
      info->storage |= newinfo->storage;
    }
  if (memcmp(newinfo->tags,info->tags,NTAGS))
    {
      COUNTER i;

      changed = true;
      for (i=0; i<NTAGS; i++)
	info->tags[i] |= newinfo->tags[i];
    }
  return changed;
}

PRIVATE Boolean ppDecoAssign(x,deco,fragile,eqns)
IN CONST ExprPtr x;
INOUT ExprPtr deco;
IN CONST MapPtr fragile,eqns;
{
  ExprPtr e;
  ExprPtr newinfo;
  ExprPtr pp;
  Boolean changed;

  e = CAR(x); /* we analyze x: x:=e, where x is both a label and a variable */
  changed = false;
  if (hasVisibleFx(e,deco))
    changed |= newUse(x,NIL,deco); /* x is used, by NIL i.e. by the return statement */
  pp = getPpInfo1(x,deco);
  newinfo = newPpInfo();
  computeInfo(x,deco,fragile,eqns,newinfo);
  changed |= ppMerge(pp,newinfo,x);
  return changed;
}

MAPSWEEPFUNC(do_ppdecomerge)
{
  Boolean found;
  ExprPtr pp;
  struct ppdecomerge *data;

  data = MAPDATA(struct ppdecomerge *);
  pp = MapGet(y,PD(data->deco)->info,&found);
  if (found) /* if some branches do not extend to their merge point, ignore them */
    {
      data->changed |= ppMerge(data->info,pp,data->merge);
    }
  return false;
}

PRIVATE Boolean ppDecoMerge(stmt,deco,fragile,eqns)
IN CONST ExprPtr stmt;
INOUT ExprPtr deco;
IN CONST MapPtr fragile,eqns;
{
  struct ppdecomerge *data = (struct ppdecomerge *)alloctmp(sizeof(struct ppdecomerge));

  data->changed = false;
  data->info = getPpInfo1(stmt,deco);
  data->deco = deco;
  data->merge = stmt;
  (void) MapSweep(COUNTEDVAL(CAR(stmt)),do_ppdecomerge,data);
  return data->changed;
}

PRIVATE MapPtr ppDecorate1 P((IN CONST ExprPtr x,IN CONST ExprPtr last,
			      INOUT ExprPtr deco,IN CONST MapPtr fragile,
			      IN CONST MapPtr eqns,IN CONST MapPtr frontier));

DEF_PRIVATE_ACTION(ppdeco)
{
  struct ppdecok *data;

  data = (struct ppdecok *)env;
  return ppDecorate1(data->x,data->last,data->deco,data->fragile,data->eqns,data->frontier);
}

PRIVATE MapPtr ppDecorate1k(x,last,deco,fragile,eqns,frontier)
IN CONST ExprPtr x;
IN CONST ExprPtr last;
INOUT ExprPtr deco;
IN CONST MapPtr fragile;
IN CONST MapPtr eqns;
IN CONST MapPtr frontier;
{
  struct ppdecok *data = (struct ppdecok *)alloctmp(sizeof(struct ppdecok));

  data->x = x;
  data->last = last;
  data->deco = deco;
  data->fragile = fragile;
  data->eqns = eqns;
  data->frontier = frontier;
  return CallCC(ACTION_NAME(ppdeco),(ExprPtr)data,true);
}

PRIVATE MapPtr ppDecorate1(x,last,deco,fragile,eqns,frontier)
IN CONST ExprPtr x; /* scofl statement that we analyze */
IN CONST ExprPtr last; /* return statement */
INOUT ExprPtr deco; /* external containing a ppDecoration *, which we update */
IN CONST MapPtr fragile; /* equations that could be killed by side-effects in x */
IN CONST MapPtr eqns; /* equations valid just before x */
IN CONST MapPtr frontier; /* set of statements on which analysis was stopped
			     (they are analysed, but not their uses; these are
			     the merge points, in general) */
{
  if (STKFULL(x))
    return ppDecorate1k(x,last,deco,fragile,eqns,frontier);
again:
  frontier = MapRemove(x,frontier);
  if (x==last)
    (void) newUse(x,NIL,deco);
  /* last is always used by NIL, the fake return statement */
  switch (x->kind)
    {
    case E_NIL:
      break;
    case E_CONS: /* assignment */
      if (ppDecoAssign(x,deco,fragile,eqns))
	{
	  ppInfo *pp;

	  pp = getPpInfo(x,deco);
	  fragile = pp->fragile;
	  eqns = pp->eqns;
	  x = CDR(x);
	  goto again; /* loop only if something changed */
	}
      break;
    case E_PROG: /* function header */
      {
	ExprPtr var;
	ppInfo *pp;

	for (var=CAR(x); var!=NIL; var=CDR(var))
	  {
	    pp = getPpInfo(var,deco);
	    pp->storage = IVAL(CAR(var));
	    pp->def = var;
	    setTags(pp);
	  }
	/* initialize ppInfos for parameters; we don't know anything yet,
	   but everything we need to know should be given by assertions. */
      }
      x = CDR(x);
      goto again;
    case E_IF: /* tests, loops */
      {
	ExprPtr cond;

	cond = CAR(CAR(x));
	if (cond->kind==(CHAR)E_COUNTED)
	  cond = COUNTEDVAL(cond);
	(void) newUse(cond,x,deco);
      }
      /* !!! should assert equations temporarily on both branches, if the test
	 is a variable, and analyze the then and else part; this could be done
	 by adding new equations in eqns, which we could infer simply from
	 equations var==true or var==false. */
      {
	ExprPtr valset;
	Boolean found;

	valset = getPpInfo(CAR(CAR(x)),deco)->valset;
	if (valset==NIL || ((void) MapGet(trueValue,valset,&found), found))
	  /* analyze the then part */
	  frontier = ppDecorate1(CDR(CAR(x)),last,deco,fragile,eqns,frontier);
	if (valset==NIL || ((void) MapGet(falseValue,valset,&found), found))
	  /* analyze the else part */
	  frontier = ppDecorate1(CDR(x),last,deco,fragile,eqns,frontier);
      }
      break;
    case E_GOTO: /* merge points */
      if (ppDecoMerge(x,deco,fragile,eqns))
	{
	  frontier = SetAdd(x,frontier);
	  /* so that we can call ppDecorate1 again to propagate info forward;
	     we do this instead of calling ppDecorate1 directly since then
	     the code is analyzed in mostly breadth-first style, which is
	     more economical in terms of re-traversals of the code. */
	}
      break;
    default:
      kinderror(message(484),x->kind);
      /*NOTREACHED*/
      break;
    }
  return frontier;
}

MAPSWEEPFUNC(do_countparents)
{
  ppDecoration *pd;
  struct countfrontier *data;
  Boolean found;

  data = MAPDATA(struct countfrontier *);
  pd = PD(data->deco);
  (void) MapGet(y,pd->info,&found);
  if (!found)
    data->n++;
  return false;
}

/* get the node from the frontier that has the least number of parents not
   already analyzed */
MAPSWEEPFUNC(do_getfrontier)
{
  struct getfrontier *data;
  struct countfrontier aux;

  data = MAPDATA(struct getfrontier *);
  aux.n = 0;
  aux.deco = data->deco;
  /* assumes x is a E_GOTO node, ie that the frontier only contains merge nodes. */
  (void) MapSweep(COUNTEDVAL(CAR(x)),do_countparents,&aux);
  if (aux.n<data->n)
    {
      data->n = aux.n;
      data->nodes = Singleton(x);
    }
  else data->nodes = SetAdd(x,data->nodes);
  return false;
}

PUBLIC ExprPtr getHighestMergeNode(frontier,deco,beforep)
IN CONST MapPtr frontier;
IN CONST ExprPtr deco;
INOUT MapPtr *beforep;
{
  struct getfrontier data;
  ExprPtr cand,x;
  ENTER;

  data.deco = deco;
  data.n = MAX_SIGNED_INT;
  (void) MapSweep(frontier,do_getfrontier,&data);
  /* compute the set cand of all nodes in the frontier that have already been visited once,
     mapped to the nodes we have encountered before. */
  cand = MapDomRestrTo(*beforep,data.nodes);
  if (cand!=NewMap())
    { /* if there is one, choose some that has been visited as anciently as possible
	 (this means it is probably closer to the root of the program) */
      do {
	x = ChooseInDom(cand);
	cand = SetInter(data.nodes,ChooseInRng(cand));
      } while (cand!=NewMap());
    }
  else
    { /* otherwise, choose any node in the frontier that have the least number of
	 parents not already analyzed, and record it in before. */
      x = ChooseInDom(data.nodes);
      *beforep = MapPut(x,*beforep,*beforep);
    }
  KEEPONLY(*beforep);
  return x;
}

PUBLIC void ppDecorate(x,last,deco,fragile,eqns)
IN CONST ExprPtr x; /* scofl statement that we analyze */
IN CONST ExprPtr last; /* return statement */
INOUT ExprPtr deco; /* external containing a ppDecoration *, which we update */
IN CONST MapPtr fragile; /* equations that could be killed by side-effects in x */
IN CONST MapPtr eqns; /* equations valid just before x */
{
  MapPtr frontier,before;
  ENTER;

  /* the whole point of this complicated iteration around a single call to ppDecorate1()
     is to traverse the control flow graph as few times possible. This is done by
     always preferring to continue the analysis on merge nodes in the frontier such that:
     1) they have as few unanalyzed parents as possible;
     2) among those, we choose the one that has already been visited earliest (if there is
     one) or any one of them if there is none.
     The result is that we do something like on-the-fly computation of the domination graph
     (although this may not be exactly it). */
  frontier = NewMap();
  before = NewMap(); /* map from nodes that have already been encountered to the set
			of nodes we had encountered before. */
  while (true)
    {
      frontier = ppDecorate1(x,last,deco,fragile,eqns,frontier);
      if (frontier==NewMap())
	break;
      x = getHighestMergeNode(frontier,deco,&before);
      x = CDR(x); /* make the analysis progress: analyze the node just after the chosen merge node. */
    }
  QUIT;
}

/***** after decorating, we must patch to take into account optimizations
  permitted by the equations we discovered in the decoration phase:
*****/

MAPSWEEPFUNC(do_remove_merge)
{
  ppDecoration *pd;
  struct remove_merge *data;
  Boolean found;

  data = MAPDATA(struct remove_merge *);
  pd = PD(data->deco);
  (void) MapGet(y,pd->info,&found);
  if (!found)
    {
      setMapRemove(x,&COUNTEDVAL(data->merge));
      HITBARRIER(data->merge);
      --COUNTEDN(data->merge);
    }
  return false;

}

PRIVATE Boolean ppPatchAssign P((INOUT ExprPtr x,IN CONST ExprPtr deco,INOUT ExprPtr last));

PRIVATE Boolean patchReverse1(x,deco,last)
INOUT ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr last;
{
  ExprPtr e,arg,a1,ee,y;
  Boolean changed;

  changed = false;
again:
  e = CAR(x);
  arg = COUNTEDVAL(e);
  a1 = CDR(arg);
  if (a1->kind==(CHAR)E_CONS &&
      (ee = CAR(a1), ee->kind==(CHAR)E_COUNTED && COUNTEDN(ee)==Z_APPEND))
    { /* reverse1(a,append(b,c)) -> reverse1(reverse1(b,a),c) */
      y = tackApplyBin(Z_REVERSE1,CAR(COUNTEDVAL(ee)),CAR(arg),last);
      SETCDR(y,x); /* insert computation of reverse1(b,a) before x */
      ee = zApplyBin(Z_REVERSE1,y,CDR(COUNTEDVAL(ee)));
      SETCAR(x,ee); /* and patch x */
      (void) ppPatchAssign(y,deco,last); /* propagate patches on the newly created REVERSE1 node */
      changed = true;
      for (last = y; CDR(last)!=x; last=CDR(last));
      goto again; /* and on the old one, too */
    }
  a1 = CAR(arg);
  if (a1->kind==(CHAR)E_CONS &&
      (ee = CAR(a1), ee->kind==(CHAR)E_COUNTED &&
       (COUNTEDN(ee)==Z_APPEND || COUNTEDN(ee)==Z_BOX_CONS || COUNTEDN(ee)==Z_BOX_UNSHARED_CONS)))
    { /* reverse1(append(a,b),c) -> reverse1(b,reverse1(a,c)) */
      /* reverse1(cons(a,b),c) -> reverse1(b,cons(a,c)) [converting unshared cons to shared cons] */
      y = tackApplyBin((COUNTEDN(ee)!=Z_APPEND)?Z_BOX_CONS:Z_REVERSE1,
		       CAR(COUNTEDVAL(ee)),CDR(arg),last);
      SETCDR(y,x); /* insert computation of reverse1(a,c) or cons(a,c) before x */
      ee = zApplyBin(Z_REVERSE1,CDR(COUNTEDVAL(ee)),y);
      SETCAR(x,ee);
      (void) ppPatchAssign(y,deco,last); /* propagate patches on the newly created REVERSE1 node */
      changed = true;
      for (last = y; CDR(last)!=x; last=CDR(last));
      goto again; /* and on the old one, too */
    }
  return changed;
}

PRIVATE Boolean patchAppend(x,deco,last)
INOUT ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr last;
{
  ExprPtr e,arg,a1,ee,y,arg2,valset;
  Boolean changed;

  changed = false;
again:
  e = CAR(x);
  arg = COUNTEDVAL(e);
  a1 = CAR(arg);
  if (a1->kind==(CHAR)E_CONS &&
      (ee = CAR(a1), ee->kind==(CHAR)E_COUNTED &&
       (COUNTEDN(ee)==Z_APPEND || COUNTEDN(ee)==Z_QUEUE_TO_LIST || COUNTEDN(ee)==Z_BOX_CONS)) &&
      (arg2 = COUNTEDVAL(ee), valset = getPpInfo(CDR(arg2),deco)->valset,
       valset->kind==(CHAR)E_MAPLET && MAPLETX(valset)==NIL))
    { /* append(reverse1(a,nil),b) -> reverse1(a,b) */
      /* append(queue_to_list(a,nil),b) -> queue_to_list(a,b) */
      /* append(cons(a,nil),b) -> cons(a,b) [shared or unshared] */
      ee = zApplyBin(COUNTEDN(ee),CAR(arg2),CDR(arg));
      SETCAR(x,ee);
      (void) ppPatchAssign(x,deco,last);
      changed = true;
    }
  else if (a1->kind==(CHAR)E_CONS &&
	   (ee = CAR(a1), ee->kind==(CHAR)E_COUNTED &&
	    (COUNTEDN(ee)==Z_BOX_CONS || COUNTEDN(ee)==Z_BOX_UNSHARED_CONS
	     || COUNTEDN(ee)==Z_APPEND)))
    { /* append(cons(a,b),c) -> cons (a,append(b,c)) [shared or unshared] */
      /* append(append(a,b),c) -> append(a,append(b,c)) */
      y = tackApplyBin(Z_APPEND,CDR(COUNTEDVAL(ee)),CDR(arg),last);
      SETCDR(y,x); /* insert computation of append(b,c) before x */
      ee = zApplyBin(COUNTEDN(ee),CAR(COUNTEDVAL(ee)),y);
      SETCAR(x,ee);
      (void) ppPatchAssign(y,deco,last); /* propagate patches on the newly created APPEND node */
      for (last = y; CDR(last)!=x; last=CDR(last));
      (void) ppPatchAssign(x,deco,last);
      changed = true;
    }
  return changed;
}

PRIVATE Boolean patchDomRestrBy(x,deco,last)
INOUT ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr last;
{
  ExprPtr e,arg,a1,ee;
  Boolean changed;

  changed = false;
  e = CAR(x);
  arg = COUNTEDVAL(e);
  a1 = CDR(arg);
  if (a1->kind==(CHAR)E_CONS &&
      (ee = CAR(a1), ee->kind==(CHAR)E_COUNTED && COUNTEDN(ee)==Z_BOX_SINGLETON))
    { /* restrby(a,maplet(b,c)) -> remove(b,a) */
      ee = zApplyBin(Z_MAPREMOVE,CAR(arg),CAR(COUNTEDVAL(ee)));
      SETCAR(x,ee);
      (void) ppPatchAssign(x,deco,last);
      changed = true;
    }
  return changed;
}

PRIVATE Boolean patchOverwrite(x,deco,last)
INOUT ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr last;
{
  ExprPtr e,arg,a1,ee;
  Boolean changed;

  changed = false;
again:
  e = CAR(x);
  arg = COUNTEDVAL(e);
  a1 = CAR(arg);
  if (a1->kind==(CHAR)E_CONS &&
      (ee = CAR(a1), ee->kind==(CHAR)E_COUNTED && COUNTEDN(ee)==Z_BOX_SINGLETON))
    { /* overwrite(maplet(a,b),c) -> putbehind(a,b,c) */
      ee = zApplyTer(Z_MAPPUTBEHIND,CAR(COUNTEDVAL(ee)),CDR(COUNTEDVAL(ee)),CDR(arg));
      SETCAR(x,ee);
      (void) ppPatchAssign(x,deco,last);
      changed = true;
    }
  else
    {
      a1 = CDR(arg);
      if (a1->kind==(CHAR)E_CONS &&
	  (ee = CAR(a1), ee->kind==(CHAR)E_COUNTED && COUNTEDN(ee)==Z_BOX_SINGLETON))
	{ /* overwrite(a,maplet(b,c)) -> put(b,c,a) */
	  ee = zApplyTer(Z_MAPPUT,CAR(COUNTEDVAL(ee)),CDR(COUNTEDVAL(ee)),CAR(arg));
	  SETCAR(x,ee);
	  (void) ppPatchAssign(x,deco,last);
	  changed = true;
	}
    }
  return changed;
}

PRIVATE Boolean patchMapPut(x,deco,last)
INOUT ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr last;
{
  ExprPtr e,arg,a1,ee,valset;
  Boolean changed;

  changed = false;
again:
  e = CAR(x);
  arg = COUNTEDVAL(e);
  a1 = CDR(CDR(arg));
  if (a1->kind==(CHAR)E_CONS &&
      (valset = getPpInfo(a1,deco)->valset,
       valset->kind==(CHAR)E_MAPLET && MAPLETX(valset)==NewMap()))
    { /* put(a,b,{}) -> maplet(a,b) */
      ee = zApplyBin(Z_BOX_SINGLETON,CAR(arg),CAR(CDR(arg)));
      SETCAR(x,ee);
      (void) ppPatchAssign(x,deco,last);
      changed = true;
    }
  else if (a1->kind==(CHAR)E_CONS &&
	   (ee = CAR(a1), ee->kind==(CHAR)E_COUNTED))
    {
      if (COUNTEDN(ee)==Z_BOX_SINGLETON && CAR(arg)==CAR(COUNTEDVAL(ee)))
	{ /* put(a,b,maplet(a,c)) -> maplet(a,b) */
	  ee = zApplyBin(Z_BOX_SINGLETON,CAR(arg),CAR(CDR(arg)));
	  SETCAR(x,ee);
	  (void) ppPatchAssign(x,deco,last);
	  changed = true;
	}
      else if ((COUNTEDN(ee)==Z_MAPPUT || COUNTEDN(ee)==Z_MAPPUTBEHIND) &&
	       CAR(arg)==CAR(COUNTEDVAL(ee)))
	{ /* put(a,b,put(a,c,d)) -> put(a,b,d) */
	  /* put(a,b,putbehind(a,c,d)) -> put(a,b,d) */
	  ee = zApplyTer(Z_MAPPUT,CAR(arg),CAR(CDR(arg)),CDR(CDR(COUNTEDVAL(ee))));
	  SETCAR(x,ee);
	  (void) ppPatchAssign(x,deco,last);
	  changed = true;
	}
    }
  return changed;
}

PRIVATE Boolean patchMapPutBehind(x,deco,last)
INOUT ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr last;
{
  ExprPtr e,arg,a1,ee,valset;
  Boolean changed;

  changed = false;
again:
  e = CAR(x);
  arg = COUNTEDVAL(e);
  a1 = CDR(CDR(arg));
  if (a1->kind==(CHAR)E_CONS &&
      (valset = getPpInfo(a1,deco)->valset,
       valset->kind==(CHAR)E_MAPLET && MAPLETX(valset)==NewMap()))
    { /* putbehind(a,b,{}) -> maplet(a,b) */
      ee = zApplyBin(Z_BOX_SINGLETON,CAR(arg),CAR(CDR(arg)));
      SETCAR(x,ee);
      (void) ppPatchAssign(x,deco,last);
      changed = true;
    }
  return changed;
}

PRIVATE ExprPtr findShape(rec,deco,eqns,foundp)
IN CONST ExprPtr rec;
IN CONST ExprPtr deco;
IN CONST MapPtr eqns;
OUT Boolean *foundp;
{
  ExprPtr shape,shapeval;
  Boolean found;
  ENTER;

  shape = MapGet(zApply(Z_GET_SHAPE,rec),eqns,&found);
  if (found &&
      (shapeval = getPpInfo(shape,deco)->valset,
       shapeval->kind==(CHAR)E_MAPLET))
    {
      *foundp = true;
      QUIT;
      return MAPLETX(shapeval);
    }
  *foundp = false;
  QUIT;
  return NIL;
}

PRIVATE Boolean patchMu1(x,deco,last,sharedp)
INOUT ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr last;
IN Boolean sharedp;
{
  MapPtr eqns;
  ExprPtr ee,e,a,rec,elselast;
  INT len;
  ppInfo *pp;
  Boolean found,changed;

  changed = false;
  e = CAR(x);
  rec = CAR(e);
  a = CDR(e);
  eqns = getPpInfo(x,deco)->eqns;
  pp = getPpInfo(rec,deco);
  if (HasOnlyTag(pp,E_RECORD_0))
    {
      ee = zApplyBin((sharedp?Z_BOX_RECORD_1:Z_BOX_UNSHARED_RECORD_1),CAR(a),CDR(a));
patch:
      SETCAR(x,ee);
      (void) ppPatchAssign(x,deco,last);
      changed = true;
    }
  else if (HasOnlyTag(pp,E_RECORD_1))
    {
      ExprPtr outlast,lab,z;
      ExprPtr args[3];

      outlast = NIL;
      lab = tackApply(Z_CAR,rec,last);
      last = tackApplyBin(Z_EQUAL,last,CAR(a),lab);
      last = compileIf(last,last,&elselast);
      /* if shapes are the same, then build record1 value */
      last = tackApplyBin((sharedp?Z_BOX_RECORD_1:Z_BOX_UNSHARED_RECORD_1),
			  CAR(a),CDR(a),last);
      outlast = mergeContinuations(last,last,outlast);
      /* otherwise, build general record */
      ee = tackApplyBin(Z_CMPSTR,lab,CAR(a),elselast);
      z = tackApply(Z_UNBOXED_INT,izero,ee);
      last = tackApplyBin(Z_INT_LEQ,ee,z,z);
      last = compileIf(last,last,&elselast);
      /* if CAR(rec)<=modlab: */
      last = tackApply(Z_UNBOXED_INT,ione,last);
      ee = tackApplyBin(Z_BOX_SINGLETON,lab,last,last);
      last = tackApply(Z_UNBOXED_INT,itwo,ee);
      args[0] = tackApplyTer(Z_MAPPUT,CAR(a),last,ee,last);
      args[1] = tackApply(Z_CDR,rec,args[0]);
      args[2] = CDR(a);
      last = tackApply((sharedp?Z_BOX_RECORD_GNRL:Z_BOX_UNSHARED_RECORD_GNRL),
		       buildGnrlArg(3,args),args[1]);
      outlast = mergeContinuations(last,last,outlast);
      /* if CAR(rec)>modlab: */
      last = tackApply(Z_UNBOXED_INT,itwo,last);
      ee = tackApplyBin(Z_BOX_SINGLETON,lab,last,last);
      last = tackApply(Z_UNBOXED_INT,ione,ee);
      args[0] = tackApplyTer(Z_MAPPUT,CAR(a),last,ee,last);
      args[2] = tackApply(Z_CDR,rec,args[0]);
      args[1] = CDR(a);
      last = tackApply((sharedp?Z_BOX_RECORD_GNRL:Z_BOX_UNSHARED_RECORD_GNRL),
		       buildGnrlArg(3,args),args[2]);
      outlast = mergeContinuations(last,last,outlast);
      /* finally, merge all paths: */
      last = outlast;
      ee = last;
      SETCDR(last,x);
      goto patch;
    }
  else if (HasOnlyTag(pp,E_RECORD_GNRL))
    {
      /* and we do as in mu1() in eval.c */
      /* But we only patch the code if we know the shapes */
      ExprPtr shape,modlab;

      shape = findShape(rec,deco,eqns,&found);
      if (found &&
	  (modlab = getPpInfo(CAR(a),deco)->valset,
	   modlab->kind==(CHAR)E_MAPLET))
	{
	  ExprPtr *fields,*outfields,*args,*argp;
	  COUNTER i;
	  INT nn,j;

	  len = MapCard(shape);
	  modlab = MAPLETX(modlab);
	  fields = (ExprPtr *)alloctmp(len*sizeof(ExprPtr));
	  outfields = (ExprPtr *)alloctmp((len+1)*sizeof(ExprPtr));
	  dumpShape(shape,fields);
	  mergeFieldLists(fields,&modlab,outfields,len,(INT)1,&nn);
	  shape = undumpShape(outfields,nn);
	  args = (ExprPtr *)alloctmp((nn+1)*sizeof(ExprPtr));
	  last = args[0] = compileConstant(shape,last,true);
	  for (i=0, j=1, argp = args+1; i<(INT)modlab; i++)
	    {
	      last = tackApplyBin(Z_UNBOX_TUPLE_NTH,rec,KINT(j),last);
	      *argp++ = last;
	    }
	  *argp++ = CDR(a); i++;
	  if (nn==len)
	    j++;
	  for (; i<nn; i++, j++)
	    {
	      last = tackApplyBin(Z_UNBOX_TUPLE_NTH,rec,KINT(j),last);
	      *argp++ = last;
	    }
	  ee = zApply((sharedp?Z_BOX_RECORD_GNRL:Z_BOX_UNSHARED_RECORD_GNRL),
		      buildGnrlArg(nn,args));
	  SETCDR(last,x);
	  goto patch;
	}
    }
  return changed;
}

PRIVATE Boolean patchMuGnrl(x,deco,last,sharedp)
INOUT ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr last;
IN Boolean sharedp;
{
  MapPtr eqns;
  ExprPtr ee,e,rec,mod,shaperec,shapemod;
  ppInfo *pp;
  Boolean found,changed;

  changed = false;
  e = CAR(x);
  rec = CAR(e);
  mod = CDR(e);
  pp = getPpInfo(rec,deco);
  if (HasOnlyTag(pp,E_RECORD_0))
    {
patchmod:
      ee = mod;
patch:
      SETCAR(x,ee);
      (void) ppPatchAssign(x,deco,last);
      changed = true;
    }
  else
    {
      /* and we do as in mu1() in eval.c */
      /* But we only patch the code if we know the shapes */
      eqns = getPpInfo(x,deco)->eqns;
      shapemod = findShape(mod,deco,eqns,&found);
      if (found)
	{
	  INT len;

	  len = MapCard(shapemod);
	  if (HasOnlyTag(pp,E_RECORD_1))
	    {
	      ExprPtr lab;

	      lab = MapGet(zApply(Z_CAR,rec),eqns,&found);
	      if (found && (lab = getPpInfo(lab,deco)->valset,
			    lab->kind==(CHAR)E_MAPLET))
		{
		  ExprPtr *fields,*outfields,*args,*argp;
		  COUNTER i;
		  INT nn,j;
		      
		  lab = MAPLETX(lab);
		  if ((void) MapGet(lab,shapemod,&found), found)
		    goto patchmod;
		  fields = (ExprPtr *)alloctmp(len*sizeof(ExprPtr));
		  outfields = (ExprPtr *)alloctmp((len+1)*sizeof(ExprPtr));
		  dumpShape(shapemod,fields);
		  mergeFieldLists(&lab,fields,outfields,len,(INT)1,&nn);
		  shaperec = undumpShape(outfields,nn);
		  args = (ExprPtr *)alloctmp((nn+1)*sizeof(ExprPtr));
		  last = args[0] = compileConstant(shaperec,last,true);
		  for (i=0, j=1, argp = args+1; i<(INT)lab; i++, j++)
		    {
		      last = tackApplyBin(Z_UNBOX_TUPLE_NTH,mod,KINT(j),last);
		      *argp++ = last;
		    }
		  *argp++ = last = tackApply(Z_CDR,rec,last); i++;
		  for (; i<nn; i++, j++)
		    {
		      last = tackApplyBin(Z_UNBOX_TUPLE_NTH,mod,KINT(j),last);
		      *argp++ = last;
		    }
		  ee = zApply((sharedp?Z_BOX_RECORD_GNRL:Z_BOX_UNSHARED_RECORD_GNRL),
			      buildGnrlArg(nn,args));
		  SETCDR(last,x);
		  goto patch;
		}
	    }
	  else if (HasOnlyTag(pp,E_RECORD_GNRL))
	    {
	      shaperec = findShape(rec,deco,eqns,&found);
	      if (found)
		{
		  INT i,j,m,nn;
		  ExprPtr shape;
		  ExprPtr *fields1,*fields2,*fp,*outfields,*args;

		  m = MapCard(shaperec);
		  fields1 = (ExprPtr *)alloctmp(m*sizeof(ExprPtr));
		  fields2 = (ExprPtr *)alloctmp(len*sizeof(ExprPtr));
		  outfields = (ExprPtr *)alloctmp((m+len)*sizeof(ExprPtr));
		  dumpShape(shaperec,fields1);
		  dumpShape(shapemod,fields2);
		  mergeFieldLists(fields1,fields2,outfields,m,len,&nn);
		  shape = undumpShape(outfields,nn);
		  args = (ExprPtr *)alloctmp((nn+1)*sizeof(ExprPtr));
		  last = args[0] = compileConstant(shaperec,last,true);
		  for (i=0, j=1, fp=fields1; i<m; i++, j++)
		    {
		      last = tackApplyBin(Z_UNBOX_TUPLE_NTH,rec,KINT(j),last);
		      args[1+(COUNTER)*fp++] = last;
		    }
		  for (i=0, j=1, fp=fields2; i<len; i++, j++)
		    {
		      last = tackApplyBin(Z_UNBOX_TUPLE_NTH,mod,KINT(j),last);
		      args[1+(COUNTER)*fp++] = last;
		    }
		  ee = zApply((sharedp?Z_BOX_RECORD_GNRL:Z_BOX_UNSHARED_RECORD_GNRL),
			      buildGnrlArg(nn,args));
		  SETCDR(last,x);
		  goto patch;
		}
	    }
	}
    }
}

MAPSWEEPFUNC(do_escape_ref)
{
  /* a BOX_REF escapes if it is used for anything but REF_SET and UNBOX_REF;
     this includes Z_LET, which stores the reference in the current environment,
     so this condition is a bit crude. This also includes merge points, so it is
     really crude. */
  ExprPtr e;

  if (x->kind!=(CHAR)E_CONS)
    return true;
  e = CAR(x);
  if (e->kind!=(CHAR)E_COUNTED)
    return true;
  switch (COUNTEDN(e))
    {
    case Z_UNBOX_REF:
      return false;
    case Z_REF_SET:
      return CDR(COUNTEDVAL(e))!=MAPDATA(ExprPtr);
    }
  return true;
}

MAPSWEEPFUNC(do_patch_ref)
{
  ExprPtr e;

  e = CAR(x);
  switch (COUNTEDN(e))
    {
    case Z_UNBOX_REF:
      e = zApply(Z_UNBOX_LOCAL_REF,COUNTEDVAL(e));
      SETCAR(x,e);
      break;
    case Z_REF_SET:
      e = zApply(Z_LOCAL_REF_SET,COUNTEDVAL(e));
      SETCAR(x,e);
      break;
    }
  return false;
}

PRIVATE Boolean patchBoxRef(x,deco,last)
INOUT ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr last;
{
  ExprPtr e;
  ppInfo *pp;
  Boolean changed;

  changed = false;
  e = CAR(x);
  pp = getPpInfo(x,deco);
  if (!MapSweep(pp->uses,do_escape_ref,x))
    {
      (void) MapSweep(pp->uses,do_patch_ref,NULL);
      e = zApply(Z_BOX_LOCAL_REF,COUNTEDVAL(e));
      SETCAR(x,e);
      changed = true;
    }
  return changed;
}

MAPSWEEPFUNC(do_useenv)
{
  switch (x->kind)
    {
    case E_CONS:
      x = CAR(x);
      /*fallthrough*/
    default: /* x should be a Code */
      return (((Code *)EXTDATA(x))->flags & CODE_DOESNOTUSE_ENV)==0;
    }
}

PRIVATE ExprPtr copyArg(arg,ivars,vars)
IN CONST ExprPtr arg;
IN CONST MapPtr ivars,vars;
{
  switch (arg->kind)
    {
    case E_NIL: /* no argument */
      break;
    case E_CONS: /* variable */
      {
	ExprPtr x;
	Boolean found;

	x = MapGet(arg,ivars,&found);
	if (found)
	  arg = x;
	else
	  arg = MapGetSure(arg,vars);
      }
      break;
    case E_TUPLE_2: /* non-empty list of variables, terminated by last value */
      arg = Ktuple2(copyArg(CAR(arg),ivars,vars),copyArg(CDR(arg),ivars,vars));
      (void) KEEP(arg);
      break;
    case E_TUPLE_GNRL: /* vector of variables */
      {
	COUNTER i,n;
	ExprPtr *tmp,*tp;

	n = AVAL(arg)->n;
	tmp = (ExprPtr *)alloctmp(n*sizeof(ExprPtr));
	for (i=0, tp=tmp; i<n; i++, tp++)
	  *tp = copyArg(AVAL(arg)->expr[i],ivars,vars);
	arg = buildGnrlArg(n,tmp);
	break;
      }
    default: /* function code (struct Code, in an E_EXTERNAL), or
		other constant argument, like strings or ints or reals or ... */
      break;
    }
  return arg;
}

PRIVATE ExprPtr copyExpression(e,ivars,vars)
IN CONST ExprPtr e;
IN CONST MapPtr ivars,vars;
{
  if (e->kind==(CHAR)E_COUNTED)
    return zApply(COUNTEDN(e),copyArg(COUNTEDVAL(e),ivars,vars));
  return copyArg(e,ivars,vars);
}

PRIVATE MapPtr copyCode P((IN CONST ExprPtr x,IN CONST ExprPtr inlast,
			   INOUT ExprPtr outlast,IN CONST MapPtr initvars,
			   IN CONST MapPtr vars));

DEF_PRIVATE_ACTION(copycode)
{
  struct copycode *data;

  data = (struct copycode *)env;
  return copyCode(data->x,data->inlast,data->outlast,data->initvars,data->vars);
}

PRIVATE MapPtr copyCode1(x,inlast,outlast,initvars,vars)
IN CONST ExprPtr x;
IN CONST ExprPtr inlast;
INOUT ExprPtr outlast;
IN CONST MapPtr initvars;
IN CONST MapPtr vars;
{
  struct copycode *data = (struct copycode *)alloctmp(sizeof(struct copycode));

  data->x = x;
  data->inlast = inlast;
  data->outlast = outlast;
  data->initvars = initvars;
  data->vars = vars;
  return CallCC(ACTION_NAME(copycode),(ExprPtr)data,true);
}

PRIVATE MapPtr copyCode(x,inlast,outlast,initvars,vars)
IN CONST ExprPtr x;
IN CONST ExprPtr inlast;
INOUT ExprPtr outlast;
IN CONST MapPtr initvars;
IN CONST MapPtr vars;
{
  ExprPtr e;
  Boolean found;
  ENTER;

  if (STKFULL(found))
    return copyCode1(x,inlast,outlast,initvars,vars);
again:
  e = MapGet(x,vars,&found);
  if (found)
    {
      if (x->kind==(CHAR)E_CONS)
	goto merge;
    }
  else
    switch (x->kind)
      {
      case E_NIL:
	break;
      case E_CONS:
	e = copyExpression(CAR(x),initvars,vars);
	outlast = tackAssign(e,outlast);
copy:
	vars = MapPut(x,outlast,vars);
	KEEPONLY(vars); /* last is kept in a translation structure, so it won't be GC'ed */
	inlast = x;
	x = CDR(x);
	goto again;
	/*case E_PROG:*/
      case E_GOTO:
	e = NIL;
merge:
	outlast = mergeContinuations(copyExpression(MapGetSure(inlast,COUNTEDVAL(CAR(x))),
						    initvars,vars),
				     outlast,NIL);
	goto copy;
      case E_IF:
	outlast = compileIf(copyExpression(CAR(CAR(x)),initvars,vars),outlast,&e);
	vars = copyCode(CDR(x),x,e,initvars,vars); /* copy else part */
	/* and now copy then part: */
	inlast = CAR(x);
	x = CDR(inlast);
	goto again;
      default:
	kinderror(message(526),x->kind);
	/*NOTREACHED*/
	break;
      }
  return vars;
}

PRIVATE Boolean patchApply(x,deco,last)
INOUT ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr last;
{
  ExprPtr e,arg,f,env;
  ppInfo *pp;
  Boolean changed;

  changed = false;
  e = CAR(x);
  arg = COUNTEDVAL(e);
  f = CAR(arg);
  arg = CDR(arg);
  env = CDR(arg);
  arg = CAR(arg);
  pp = getPpInfo(f,deco);
  if (pp->valset->kind==(CHAR)E_MAPLET)
    {
      ExprPtr c;
      Code *code;

      c = MAPLETX(pp->valset);
      code = (Code *)EXTDATA(c);
      if (code->flags & CODE_INLINE)
	{
	  MapPtr ivars,vars;
	  ExprPtr parms;
	  Boolean found;

	  parms = CAR(code->prgm);
	  ivars = MapPair(parms,arg,CDR(parms),env);
	  vars = copyCode(CDR(code->prgm),code->prgm,last,ivars,NewMap());
	  last = MapGet(code->last,vars,&found);
	  if (!found)
	    last = NIL;
	  else SETCDR(last,CDR(x));
	  changed = true;
	  return changed;
	}
    }
  /* otherwise, if called function does not use its environment, substitute
   a dummy environment (first check that the dummy has not been installed already). */
  if ((env->kind!=(CHAR)E_CONS || CAR(env)->kind!=(CHAR)E_COUNTED ||
       COUNTEDN(CAR(env))!=Z_NIL) &&
      pp->valset!=NIL && !MapSweep(pp->valset,do_useenv,NULL))
    {
      last = tackApply(Z_NIL,NIL,last); /* the dummy */
      SETCDR(last,x);
      e = zApplyTer(Z_APPLY,f,arg,env);
      SETCAR(x,e);
      changed = true;
    }
  return changed;
}
 
MAPSWEEPFUNC(do_useful)
{ /* a stmt producing a constant value is deemed useful (i.e, worthwhile
     to replace by an explicit definition of the constant) if at least
     one of its uses is a non-constant or a side-effecting instruction.
     As all side-effecting instructions are defined so as to return nothing,
     they cannot be constant. */
  ExprPtr deco;
  ppInfo *pp;

  if (x==NIL)
    return true;
  deco = MAPDATA(ExprPtr);
  pp = getPpInfo(x,deco);
  return pp->valset->kind!=E_MAPLET;
}

PRIVATE Boolean ppPatchConstant(x,deco,last,val,pp)
INOUT ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr last;
IN CONST ExprPtr val;
IN CONST ppInfo *pp;
{
  ExprPtr y;
  Boolean changed;

  changed = false;
  if (MapSweep(pp->uses,do_useful,deco))
    {
      ENTER;

      (void) KEEP(x); /* x is assumed to be CDR(last) */
      y = compileConstant(val,last,(pp->storage & SCF_UNBOXED)==0);
      if (CAR(y)==CAR(x)) /* if the constant was generated with the same expression
			     as the one already used for x, then destroy this code. */
	{
	  SETCDR(last,x); /* reinstall old code */
	}
      else
	{
	  SETCAR(x,CAR(y)); /* patch x so as to represent result */
	  while (CDR(last)!=y)
	    last = CDR(last); /* assumes compileConstant() only compiles assignments */
	  SETCDR(last,x); /* and reconnect everything */
	  changed = true;
	}
    }
  return changed;
}

PRIVATE Boolean constantPatchable(x)
IN CONST ExprPtr x;
{
  ExprPtr e;

  if (x==NIL)
    return false;
  if (x->kind!=(CHAR)E_CONS)
    return false;
  e = CAR(x);
  if (e->kind!=(CHAR)E_COUNTED)
    return false;
  switch (COUNTEDN(e))
    {
    case Z_NIL:
    case Z_UNBOXED_INT:
    case Z_UNBOXED_REAL:
    case Z_UNBOXED_STRING:
    case Z_EMPTY:
    case Z_TUPLE_0:
    case Z_RECORD_0:
      /* Z_NIL and other tags that only produce constants are deemed not patchable,
	 since they would be patched by themselves; this would be pointless. */
    case Z_CODE:
    case Z_BOX_CLOSURE:
    case Z_UNBOX_CLOSURE_FN:
    case Z_UNBOX_CLOSURE_ENV:
      /* Z_CODE, Z_BOX_CLOSURE are considered not patchable by a constant,
	 as they are already notations for a constant pieces of code,
	 which cannot be represented easily as a boxed value. */
    case Z_MAPGET:
      /* we cannot represent results of Z_MAPGET by constants, as these results are
	 fake couples; but Z_GET_MAPGET_Y and Z_GET_MAPGET_FOUND can be patched. */
    case Z_APPLY:
    case Z_CALL:
      /* function applications may have all sorts of side-effects, so we don't consider
	 we can replace a function call by a constant (unless the function is inlined). */
    case Z_ENEXT:
    case Z_CONSENV:
    case Z_FAKEENV:
    case Z_RAISE:
    case Z_ASSERT_EQUAL:
    case Z_LET:
    case Z_BOX_QUEUE:
    case Z_BOX_1_QUEUE:
    case Z_BOX_REF:
    case Z_REF_SET:
    case Z_START_LISTSWEEP:
    case Z_INC_LISTSWEEP:
    case Z_START_MAPSWEEP:
    case Z_INC_MAPSWEEP:
    case Z_START_SUBMAPSWEEP:
    case Z_INC_SUBMAPSWEEP:
    case Z_BOX_LOCAL_REF:
    case Z_LOCAL_REF_SET:
    case Z_CONCAT_QUEUE:
    case Z_BOX_UNSHARED_EXCEPTION: /* but Z_BOX_SHARED_EXCEPTION are patchable by constants */
    case Z_START_HANDLE:
    case Z_END_HANDLE:
    case Z_MAKE_PROMISE:
      return false;
    }
  return true;
}

MAPSWEEPFUNC(do_patch_consenv_p)
{
  ExprPtr e;

  if (x->kind!=(CHAR)E_CONS)
    return true;
  e = CAR(x);
  if (e->kind==(CHAR)E_COUNTED)
    return COUNTEDN(e)!=Z_LET;
  return true;
}

PRIVATE Boolean patchConsEnv(x,deco,last)
INOUT ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr last;
{
  ExprPtr pp;
  ppInfo *info;
  Boolean changed;

  changed = false;
  pp = getPpInfo1(x,deco);
  info = PP(pp);
  if (!MapSweep(info->uses,do_patch_consenv_p,deco))
    { /* in this case, we consenv an environment that we patch but never read
	 and never use to build closures, construct other environments or anything. */
      SETQ(info->uses,NewMap()); /* signal by this trick that all Z_LETs on this env
				    are useless. */
      HITBARRIER(pp);
      SETCDR(last,CDR(x)); /* and eliminate the Z_CONSENV or Z_FAKEENV instruction altogether. */
      changed = true;
    }
  return changed;
}

PRIVATE Boolean patchLet(x,deco,last)
INOUT ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr last;
{
  ppInfo *pp;
  ExprPtr env;
  Boolean changed;

  changed = false;
  env = CAR(COUNTEDVAL(CAR(x)));
  pp = getPpInfo(env,deco);
  if (pp->uses==NewMap()) /* The environment is useless */
    { /* then eliminate the Z_LET instruction. */
      SETCDR(last,CDR(x));
      changed = true;
    }
  return changed;
}

DEF_PRIVATE_ACTION(pppatch)
{
  struct pppatch *data;

  data = (struct pppatch *)env;
  data->res = ppPatchAssign(data->x,data->deco,data->last);
  return NIL;
}

PRIVATE Boolean ppPatchAssign1(x,deco,last)
INOUT ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr last;
{
  struct pppatch *data = (struct pppatch *)alloctmp(sizeof(struct pppatch));

  data->x = x;
  data->deco = deco;
  data->last = last;
  (void) CallCC(ACTION_NAME(pppatch),(ExprPtr)data,true);
  return data->res;
}

PRIVATE Boolean ppPatchAssign(x,deco,last)
INOUT ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr last;
{
  ppInfo *pp;
  MapPtr valset;
  Boolean changed;

  if (STKFULL(changed))
    return ppPatchAssign1(x,deco,last);
  pp = getPpInfo(x,deco);
  valset = pp->valset;
  if (valset->kind==(CHAR)E_MAPLET && /* if expression is constant */
      constantPatchable(x))
    { /* then patch instruction so as to build the constant explicitly;
	 as this may change the storage field (the SCF_NOINIT flag),
	 changed must in general be set to true. */
      changed = ppPatchConstant(x,deco,last,MAPLETX(valset),pp);
    }
  else switch (CAR(x)->kind)
    {
    case Z_MAPDOMRESTRBY:
      changed = patchDomRestrBy(x,deco,last);
      break;
    case Z_APPEND:
      changed = patchAppend(x,deco,last);
      break;
    case Z_REVERSE1:
      changed = patchReverse1(x,deco,last);
      break;
    case Z_OVERWRITE:
      changed = patchOverwrite(x,deco,last);
      break;
    case Z_MAPPUT:
      changed = patchMapPut(x,deco,last);
      break;
    case Z_MAPPUTBEHIND:
      changed = patchMapPutBehind(x,deco,last);
      break;
    case Z_RMU1:
      changed = patchMu1(x,deco,last,true);
      break;
    case Z_UNSHARED_RMU1:
      changed = patchMu1(x,deco,last,false);
      break;
    case Z_RMUX:
      changed = patchMuGnrl(x,deco,last,true);
      break;
    case Z_UNSHARED_RMUX:
      changed = patchMuGnrl(x,deco,last,false);
      break;
    case Z_APPLY:
      changed = patchApply(x,deco,last);
      break;
    case Z_CONSENV:
    case Z_FAKEENV:
      changed = patchConsEnv(x,deco,last);
      break;
    case Z_LET:
      changed = patchLet(x,deco,last);
      break;
    default:
      changed = false;
    }
  return changed;
}

PRIVATE Boolean unaryBoxing(x)
IN CONST ExprPtr x;
{
  ExprPtr e;

  if (x->kind!=(CHAR)E_CONS)
    return false;
  e = CAR(x);
  if (e->kind==(CHAR)E_COUNTED)
    switch (COUNTEDN(e))
      {
	/* boxing functions: */
      case Z_BOX_INT:
      case Z_BOX_REAL:
      case Z_BOX_TUPLE_1:
      case Z_BOX_UNSHARED_TUPLE_1:
      case Z_BOX_NUMTYPE:
      case Z_BOX_BOOL:
      case Z_BOX_REF:
      case Z_BOX_UNSHARED_EXCEPTION:
      case Z_BOX_SHARED_EXCEPTION:
	/* and also unboxing functions: */
      case Z_UNBOX_INT:
      case Z_UNBOX_REAL:
      case Z_UNBOX_BOOL:
	return true;
      }
  return false;
}

MAPSWEEPFUNC(do_unarymerge_p)
{
  struct unarymerge_p *data;
  INT tag;
  ppInfo *pp;

  data = MAPDATA(struct unarymerge_p *);
  if (!unaryBoxing(y))
    return true;
  pp = getPpInfo(y,data->deco);
  if (pp->uses->kind!=(CHAR)E_MAPLET)
    return true; /* if the boxing operation produces a value that is only used by
		    the merge node, it is ok; otherwise, it may be useful to box it
		    prior to the merge node. */
  tag = COUNTEDN(CAR(y));
  if (data->tag==0)
    data->tag = tag;
  else if (data->tag==Z_BOX_UNSHARED_TUPLE_1 && tag==Z_BOX_TUPLE_1)
    data->tag = tag;
  else if (data->tag!=tag)
    return true;
  return false;
}

MAPSWEEPFUNC(do_unarymerge)
{
  struct unarymerge *data;

  data = MAPDATA(struct unarymerge *);
  data->outlast = mergeContinuations(COUNTEDVAL(CAR(y)),x,data->outlast);
  return false;
}

PRIVATE Boolean patchUnaryMerge(x,deco)
INOUT ExprPtr x;
IN CONST ExprPtr deco;
{
  struct unarymerge_p data;
  struct unarymerge ddata;
  ExprPtr e;
  Boolean changed;

  changed = false;
  data.tag = 0;
  data.deco = deco;
  if (!MapSweep(COUNTEDVAL(CAR(x)),do_unarymerge_p,&data))
    { /* if the merge node is a merge of unary boxing operations that are not
	 needed before coming to the merge node, then the boxing operation can
	 be transported to just after the merge node. This will be useful in
	 Deforest(), where functions like fn x => if x=0 then 1 else 2 will have
	 their boxing instruction (here Z_BOX_INT) at the end, so that a specialized
	 function returning unboxed ints instead of boxed ones will be produced. */
      ddata.outlast = NIL;
      (void) MapSweep(COUNTEDVAL(CAR(x)),do_unarymerge,&ddata);
      SETCDR(ddata.outlast,x); /* link in new merge node */
      e = zApply(data.tag,ddata.outlast);
      x->kind = (CHAR)data.tag;
      SETCAR(x,e); /* and patch old merge node to become the boxing node. */
      changed = true;
    }
  return changed;
}

PRIVATE ExprPtr augmentHandler(x,handlers)
IN CONST ExprPtr x;
IN CONST ExprPtr handlers;
{
  ExprPtr aux;

  /* Z_START_HANDLE nodes are always followed at some
     distance by a merge node, used for tacking goto's
     to it when patching Z_RAISE nodes. */
  while (x->kind!=(CHAR)E_GOTO)
    x = CDR(x);
  MAKECONS(x,handlers,aux);
  return KEEP(aux);
}

PRIVATE Boolean ppPatch1 P((INOUT ExprPtr x,IN CONST ExprPtr deco,
			    INOUT MapPtr *remp,INOUT ExprPtr handlers,
			    IN CONST MapPtr last));

DEF_PRIVATE_ACTION(pppatch1)
{
  struct pppatch1 *data;

  data = (struct pppatch1 *)env;
  data->res = ppPatch1(data->x,data->deco,&data->rem,data->handlers,data->last);
  return NIL;
}

PRIVATE Boolean ppPatch1k(x,deco,remp,handlers,last)
INOUT ExprPtr x;
IN CONST ExprPtr deco;
INOUT MapPtr *remp;
INOUT ExprPtr handlers;
IN CONST MapPtr last;
{
  struct pppatch1 *data = (struct pppatch1 *)alloctmp(sizeof(struct pppatch1));

  data->x = x;
  data->deco = deco;
  data->rem = *remp;
  data->handlers = handlers;
  data->last = last;
  (void) CallCC(ACTION_NAME(pppatch1),(ExprPtr)data,true);
  *remp = data->rem;
  return data->res;
}

PRIVATE Boolean ppPatch1(x,deco,remp,handlers,last)
INOUT ExprPtr x; /* first scofl statement that we patch */
IN CONST ExprPtr deco; /* external containing a ppDecoration *, which we update */
INOUT MapPtr *remp;
INOUT ExprPtr handlers;
IN CONST MapPtr last;
{
  ExprPtr e;
  ppInfo *info;
  Boolean found,changed;

  if (STKFULL(changed))
    return ppPatch1k(x,deco,remp,handlers,last);
  changed = false;
again:
  if (x==NIL)
    return changed;
  (void) MapGet(x,*remp,&found);
  if (found)
    return changed;
  *remp = SetAdd(x,*remp);
  switch (x->kind)
    {
    case E_NIL:
      break;
    case E_CONS: /* assignment */
      e = CAR(x);
      if (e->kind==(CHAR)E_COUNTED)
	switch (COUNTEDN(e))
	  {
	  case Z_START_HANDLE:
	    handlers = augmentHandler(x,handlers);
	    break;
	  case Z_END_HANDLE:
	    handlers = CDR(handlers);
	    break;
	  case Z_RAISE:
	    (void) mergeContinuations(COUNTEDVAL(e),last,CAR(handlers));
	    /* short-circuit Z_RAISE instruction and replace
	       it by a GOTO node leading directly to the
	       current exception handler. This allows us
	       to ensure that no Z_RAISE opcode can ever
	       be handled by a handler inside the same C code
	       for an optimized function. */
	    changed = true;
	    x = CAR(handlers);
	    goto again;
	  }
      info = getPpInfo(x,deco);
      if (info->def!=x && e!=info->def)
	{ /* then rewrite e to its simplification */
	  SETCAR(x,info->def);
	  /* this is not enough to consider the code changed, since this preserves
	     the semantics, so we leave changed as is. */
	}
      else changed |= ppPatchAssign(x,deco,last);
      /*fallthrough*/
    case E_PROG: /* function header */
      last = x;
      x = CDR(x);
      goto again;
    case E_IF: /* tests, loops */
      {
	ExprPtr valset;
	Boolean found;

	valset = getPpInfo(CAR(CAR(x)),deco)->valset;
	if (valset!=NIL && ((void) MapGet(trueValue,valset,&found), !found))
	  /* if the then part is never reached */
	  {
	    SETCDR(last,CDR(x)); /* proceed directly to else part */
	    x = CDR(last);
	    /* the code was not really changed, ie we would not get more information
	       by redecorating parts of the code, so we leave changed as is. */
	    goto again;
	  }
	if (valset!=NIL && ((void) MapGet(falseValue,valset,&found), !found))
	  /* if the else part is never reached */
	  {
	    SETCDR(last,CDR(CAR(x))); /* proceed directly to then part */
	    x = CDR(last);
	    /* the code was not really changed, ie we would not get more information
	       by redecorating parts of the code, so we leave changed as is. */
	    goto again;
	  }
	changed |= ppPatch1(CDR(x),deco,remp,handlers,x); /* patch the else part */
	last = CAR(x);
	x = CDR(x); /* patch the then part */
	goto again;
      }
      break;
    case E_GOTO: /* merge points */
      { /* remove unreachable parents of this merge node */
	struct remove_merge data;

	data.deco = deco;
	data.merge = CAR(x);
	(void) MapSweep(COUNTEDVAL(CAR(x)),do_remove_merge,&data);
      }
      changed |= patchUnaryMerge(x,deco);
      last = x;
      x = CDR(x);
      /* the code was not really changed, ie we would not get more information
	 by redecorating parts of the code, so we leave changed as is. */
      goto again;
    default:
      kinderror(message(522),x->kind);
      /*NOTREACHED*/
      break;
    }
  return changed;
}

PUBLIC Boolean ppPatch(x,deco)
INOUT ExprPtr x; /* first scofl statement that we patch */
IN CONST ExprPtr deco; /* external containing a ppDecoration *, which we update */
  /* returns true if some part of the code was really changed, ie if recomputing
     a new decoration from scratch would get us more information. */
{
  MapPtr rem;
  Boolean changed;
  ENTER;

  rem = NewMap();
  changed = ppPatch1(x,deco,&rem,NIL,NIL);
  QUIT;
  return changed;
}

PRIVATE Boolean deforesting(x,deco)
IN CONST ExprPtr x;
IN CONST ExprPtr deco;
{
  ppInfo *pp;
  ExprPtr e;

  switch (x->kind)
    {
    case E_NIL:
      return true;
    case E_CONS: /* assignment */
      pp = getPpInfo(x,deco);
      if ((pp->storage & SCF_NOINIT)!=0 && pp->uses!=NewMap())
	{ /* constant expressions are already deforested, since actually
	     they will be compiled in the separate initialization function.
	     Similarly, unused expressions won't be compiled, so they are
	     considered deforested. */
	  e = CAR(x);
	  if (e->kind==(CHAR)E_COUNTED)
	    switch (COUNTEDN(e))
	      {
	      case Z_UNBOX_INT:
	      case Z_UNBOX_REAL:
	      case Z_UNBOX_TUPLE_NTH:
	      case Z_UNBOX_STRING_LEN:
	      case Z_UNBOX_STRING_TEXT:
	      case Z_ENEXT:
	      case Z_VAR:
	      case Z_RAISE:
	      case Z_EQUAL:
	      case Z_CAR:
	      case Z_CDR:
	      case Z_UNBOX_REF:
	      case Z_HAS_TAG:
	      case Z_GET_VECTOR_LEN:
	      case Z_MAP_CARD_DEC:
	      case Z_INT_LEQ:
	      case Z_GET_SHAPE:
	      case Z_MAPGET:
	      case Z_GET_MAPGET_Y:
	      case Z_GET_MAPGET_FOUND:
	      case Z_TUPLE_SUB_NTH:
	      case Z_REF_SET:
	      case Z_UNBOX_CLOSURE_FN:
	      case Z_UNBOX_CLOSURE_ENV:
	      case Z_UNBOX_BOOL:
	      case Z_C_GLOBAL:
	      case Z_C_CONST:
	      case Z_CMPSTR:
	      case Z_NOT:
	      case Z_GETTHEEXPR:
		return true;
	      default:
		return false;
	  }
	}
    case E_PROG: /* function header */
    case E_GOTO: /* merge points */
    case E_IF: /* tests, loops */
      return true;
    }
}

MAPSWEEPFUNC(do_deforested)
{
  return !deforesting(x,MAPDATA(ExprPtr));
}

PRIVATE Boolean deforested(x,deco)
IN CONST ExprPtr x;
IN CONST ExprPtr deco;
/* x is deforested iff all its uses are deforesting uses */
{
  return !MapSweep(getPpInfo(x,deco)->uses,do_deforested,deco);
}

ARGSWEEPFUNC(arg_deforested)
{
  return !deforested(x,MAPDATA(ExprPtr));
}

PRIVATE Boolean from_deforested(x,deco)
IN CONST ExprPtr x;
IN CONST ExprPtr deco;
/* x is from_deforested iff all its arguments are deforested,
   i.e. iff it is built from objects that will serve to be deforested
   only. */
{
  ExprPtr e;

  e = CAR(x);
  if (e->kind==(CHAR)E_COUNTED)
    return !argSweep(COUNTEDVAL(e),arg_deforested,deco);
  return !arg_deforested(e,(VOIDPTR)deco);
}

ARGSWEEPFUNC(arg_deforest_vars)
{
  struct deforest *data;

  data = MAPDATA(struct deforest *);
  data->vars = SetRemove(x,data->vars); /* x is deforested, so should not be taken as input */
  data->vars = SetUnion(getPpInfo(x,data->deco)->uses,data->vars);
  /* but then we have to take as input all its used components. */
  return false;
}

PRIVATE MapPtr deforest_vars(x,deco,vars)
IN CONST ExprPtr x;
IN CONST ExprPtr deco;
IN CONST MapPtr vars;
{
  struct deforest data;
  ExprPtr e;
  ENTER;

  e = CAR(x);
  data.deco = deco;
  data.vars = vars;
  if (e->kind==(CHAR)E_COUNTED)
    (void) argSweep(COUNTEDVAL(e),arg_deforest_vars,&data);
  else (void) arg_deforest_vars(e,(VOIDPTR)&data);
  KEEPONLY(data.vars);
  return data.vars;
}

MAPCOLLECTFUNC(do_rel_union)
{
  return MapSingleton(x,SetUnion(y,MapGetSure(x,MAPDATA(MapPtr))));
}

PRIVATE MapPtr RelUnion(map1,map2)
IN CONST MapPtr map1,map2;
{
  return MapMerge(MapDelta(map1,map2),
		  MapCollect(MapDomRestrTo(map1,map2),do_rel_union,map2));
}

PRIVATE MapPtr RelAdd(x,y,map)
IN CONST ExprPtr x,y;
IN CONST MapPtr map;
{
  ExprPtr yy;
  Boolean found;

  yy = MapGet(x,map,&found);
  if (found)
    return MapPut(x,SetUnion(y,yy),map);
  return MapPut(x,y,map);
}

PRIVATE MapPtr toDeforest P((IN CONST ExprPtr x,IN CONST ExprPtr deco,
			     IN CONST ExprPtr last,IN CONST MapPtr vars));

DEF_PRIVATE_ACTION(todeforest)
{
  struct todeforest *data;

  data = (struct todeforest *)env;
  return toDeforest(data->x,data->deco,data->last,data->vars);
}

PRIVATE MapPtr toDeforest1(x,deco,last,vars)
IN CONST ExprPtr x;
IN CONST ExprPtr deco;
IN CONST ExprPtr last;
IN CONST MapPtr vars;
{
  struct todeforest *data = (struct todeforest *)alloctmp(sizeof(struct todeforest));

  data->x = x;
  data->deco = deco;
  data->last = last;
  data->vars = vars;
  return CallCC(ACTION_NAME(todeforest),(ExprPtr)data,true);
}

PRIVATE MapPtr toDeforest(x,deco,last,vars)
     /* return frontier (set of last nodes) before which allocations begin;
	actually, the frontier is a map from last nodes to sets of parameters
	for the new local function to create. */
IN CONST ExprPtr x;
IN CONST ExprPtr deco;
IN CONST ExprPtr last;
IN CONST MapPtr vars;
{
  MapPtr frontier;
  MapPtr newvars;
  ppInfo *pp;

  if (STKFULL(pp))
    return toDeforest1(x,deco,last,vars);
  frontier = NewMap();
again:
  switch (x->kind)
    {
    case E_NIL:
end:
      frontier = RelAdd(last,vars,frontier);
      break;
    case E_CONS: /* assignment */
      pp = getPpInfo(x,deco);
      if ((pp->storage & SCF_NOINIT)==0 || pp->uses==NewMap())
	; /* constant or unused assignments are considered to deforest (leave in
	     front of the call to the current function). */
      else if (!deforesting(x,deco) || !from_deforested(x,deco))
	goto end;
      /* We estimate that the assignments that we should deforest (leave in front
	 of the call to the current function) are the ones that:
	 - are deforesting, ie, they do not allocate memory;
	 - are from_deforested, ie, their arguments are objects that lie there
	 only to be deforested (structurally decomposed), and not to be stored away
	 or passed to functions.
	 */
      newvars = deforest_vars(x,deco,vars);
      if (MapCard(newvars)>=MAX_ARITY)
	goto end; /* never deforest past the maximum arity */
      vars = newvars;
      /*fallthrough*/
    case E_PROG: /* function header */
    case E_GOTO: /* merge points */
      last = x;
      x = CDR(x);
      goto again;
    case E_IF: /* tests, loops; should mostly identify conditional pattern matches
		  in front of functions. */
      frontier = RelUnion(frontier,toDeforest(CDR(x),deco,x,vars));
      last = CAR(x);
      x = CDR(x);
      goto again;
    }
  return frontier;
}

PRIVATE MapPtr prgmVars(list)
IN CONST ExprPtr list;
{
  MapPtr vars;
  ENTER;

  vars = NewMap();
  for (; list!=NIL; list=CDR(list))
    vars = SetAdd(list,vars);
  KEEPONLY(vars);
  return vars;
}

MAPSWEEPFUNC(do_dump_frontier)
{
  struct frontiercmp **fpp,*fp;

  fpp = MAPDATA(struct frontiercmp **);
  fp = (*fpp)++;
  fp->last = x;
  fp->vars = y;
  return false;
}

PRIVATE int frontiercmp(f1,f2)
IN CONST ANSI_CONST struct frontiercmp *f1,*f2;
{
  if (f1->vars==f2->vars) /* then we should be able to deforest them independently */
    return 0;
  if (SetIncludedp(f1->vars,f2->vars)) /* then we should prefer to deforest f2 first,
					  since then the code for f1 will just call that
					  for f2, if the latter is reachable from the former. */
    return 1;
  if (SetIncludedp(f2->vars,f1->vars))
    return -1;
  return 0; /* otherwise, we don't know which should be deforested first. */
}

MAPCOLLECTFUNC(do_build_ivars)
{
  ExprPtr res,storage;
  ppInfo *pp;
  struct build_ivars *data;

  data = MAPDATA(struct build_ivars *);
  res = MapSingleton(x,data->newvar);
  pp = getPpInfo(x,data->deco);
  storage = KINT(pp->storage);
  SETCAR(data->newvar,storage);
  data->newvar = CDR(data->newvar);
  return res;
}

PRIVATE void giveName(c)
INOUT ExprPtr c;
{
  Code *code;

  code = (Code *)EXTDATA(c);
  if (code->name==NULL) /* then we must invent a name for it */
    { /* and if we didn't have a name, then it was a function local to
	 some other function, so we declare the resulting piece of C code
	 to be PRIVATE (should avoid cluttering the C compiler's symbol table). */
      char s[32];

      code->flags |= CODE_PRIVATE;
      /* now, as the function is private, and all private functions have Z_CODE
	 chunks in memory, we can take any encoding of the address of the chunk
	 as a correct C name. */
      sprintf(s,"LoCaL__%x",(long)code);
      code->name = (CHAR *)MALLOC(strlen(s)+1);
      strcpy((char *)code->name,s);
    }
}

PUBLIC void fillCode(c,deco)
INOUT ExprPtr c;
IN CONST ExprPtr deco;
{
  Code *code;

  code = (Code *)EXTDATA(c);
  if (code->last==NIL)
    {
      code->storage = SCF_TOP;
      /*leave tags as they are, i.e all cleared. */
    }
  else
    {
      ppInfo *pp;

      pp = getPpInfo(code->last,deco);
      code->storage = pp->storage;
      memcpy(pp->tags,pp->tags,NTAGS);
    }
  code->lambda = PD(deco)->lambda;
  code->deco = deco;
  HITBARRIER(c);
  giveName(c);
  code->fxtypes = PD(deco)->fxtypes;
  code->size = MapCard(PD(deco)->info);
}

PRIVATE Boolean inlinable(c)
IN CONST ExprPtr c;
{
  Code *code;
  Boolean found;

  code = (Code *)EXTDATA(c);
  if (code->size<=IVAL(REFVAL(maxInlineRef)) &&
      ((void)MapGet(c,code->lambda,&found), !found))
    /* inline if: not too large, and: not recursive (but mutually recursive is ok
       if the function through which it is called is itself not inlined) */
    return true;
  return false;
}

PRIVATE ExprPtr getBoxingsOut(c,last)
INOUT ExprPtr c; /* decorated */
INOUT ExprPtr last;
{
  Code *code;
  ExprPtr pp,pp2;
  ppInfo *ppp;
  ExprPtr ll,tags,e;
  ENTER;

  code = (Code *)EXTDATA(c);
  ll = code->last;
  tags = NIL;
  while (unaryBoxing(ll) &&
	 (pp = getPpInfo1(ll,code->deco),
	  ppp = PP(pp),
	  ppp->uses->kind==(CHAR)E_MAPLET))
    {
      e = CAR(ll);
      tags = MakeCounted(E_COUNTED,tags,COUNTEDN(e));
      (void) KEEP(tags);
      ppp->uses = NewMap(); /* not used any more */
      HITBARRIER(pp);
      ll = COUNTEDVAL(e);
    }
  if (ll!=code->last)
    { /* add an assignment at the end */
      code->last = tackAssign(ll,code->last);
      HITBARRIER(c);
      pp = getPpInfo1(ll,code->deco);
      ppp = PP(pp);
      pp2 = getPpInfo1(code->last,code->deco);
      ppCopy(pp2,ppp);
      (void) newUse(code->last,NIL,code->deco);
    }
  for (; tags!=NIL; tags = COUNTEDVAL(tags))
    last = tackApply(COUNTEDN(tags),last,last);
  KEEPONLY(last);
  return last;
}

PRIVATE ExprPtr Deforest(c,deco)
INOUT ExprPtr c; /* an E_EXTERNAL node of type "code" */
IN CONST ExprPtr deco;
{
  Code *code;
  MapPtr frontier,ivars;
  ExprPtr last,outlast;
  struct frontiercmp *lasts,*lastsp;
  INT i,n;
  ENTER;

  code = (Code *)EXTDATA(c);
  ivars = prgmVars(CAR(code->prgm));
  outlast = NIL;
  frontier = toDeforest(CDR(code->prgm),deco,code->prgm,ivars);
  n = MapCard(frontier);
  if (n==0 || (n==1 && SetIncludedp(MAPLETY(frontier),ivars) &&
	       !unaryBoxing(code->last)))
    return deco; /* if there is nothing to deforest, just return. */
  lasts = (struct frontiercmp *)alloctmp(n*sizeof(struct frontiercmp));
  lastsp = lasts;
  (void) MapSweep(frontier,do_dump_frontier,&lastsp);
  qsort(lasts,(long)n,(long)sizeof(struct frontiercmp),frontiercmp);
  for (i=0, lastsp = lasts; i<n; i++, lastsp++)
    {
      if (CDR(lastsp->last)==NIL)
	last = lastsp->last;
      else
	{
	  struct translation *fun = (struct translation *)alloctmp(sizeof(struct translation));
	  COUNTER m;
	  struct build_ivars data;
	  MapPtr vars;
	  ExprPtr res,auxc,zcode,auxdeco;
	  Boolean found;
	  ENTER;

	  m = MapCard(lastsp->vars);
	  initTranslation(m,fun);
	  data.newvar = CAR(fun->prgm);
	  data.deco = deco;
	  ivars = MapCollect(lastsp->vars,do_build_ivars,&data);
	  vars = copyCode(CDR(lastsp->last),lastsp->last,fun->prgm,ivars,NewMap());
	  res = MapGet(code->last,vars,&found);
	  if (!found)
	    res = NIL;
	  last = zcode = transToCode(fun,res,lastsp->last);
	  auxc = COUNTEDVAL(CAR(zcode));
	  /* patch code to call newly generated function */
	  {
	    /* generate calling sequence */
	    ExprPtr *args,*argp,arg;
	    COUNTER j;
	    ExprPtr ll;

	    args = (ExprPtr *)alloctmp((m+1)*sizeof(ExprPtr));
	    argp = args;
	    *argp++ = last;
	    ivars = MapInverse0(ivars);
	    (void) KEEP(ivars);
	    for (j=0, ll = CAR(fun->prgm); j<m; j++, ll = CDR(ll))
	      *argp++ = MapGetSure(ll,ivars);
	    arg = buildGnrlArg(m+1,args);
	    last = tackApply(Z_CALL,arg,last);
	  }
	  /* finally, decorate */
	  auxdeco = ppDecoInit();
	  ppDecorate(fun->prgm,res,auxdeco,NewMap(),NewMap());
	  /* and get unary boxings out of the called function */
	  last = getBoxingsOut(auxc,last);
	  fillCode(auxc,auxdeco);
	  if (inlinable(auxc)) /* if the generated function is too small, reinline
				  it after all */
	    {
	      vars = copyCode(CDR(fun->prgm),fun->prgm,lastsp->last,ivars,NewMap());
	      last = MapGet(((Code *)EXTDATA(auxc))->last,vars,&found);
	      if (!found)
		last = NIL;
	      if (last!=NIL)
		{
		  SETCDR(last,CDR(zcode));
		  while (CDR(last)!=NIL)
		    last = CDR(last);
		}
	    }
	}
      outlast = mergeContinuations(last,last,outlast);
      KEEPONLY(outlast);
    }
  code->last = outlast;
  code->flags |= CODE_INLINE;
  HITBARRIER(c);
  deco = ppDecoInit();
  ppDecorate(code->prgm,code->last,deco,NewMap(),NewMap());
  KEEPONLY(deco);
  return deco;
}

PUBLIC void optimizeCode(c)
INOUT ExprPtr c; /* an E_EXTERNAL node of type "code" */
{
  Code *code;

  code = (Code *)EXTDATA(c);
  if ((code->flags & (CODE_INUSE | CODE_ANALYZED))==0 &&
      code->prgm!=NIL) /* otherwise, code is on disk, and presumably has been analyzed */
    {
      ExprPtr deco;
      Boolean changed;
      ENTER;

      changed = true;
      code->flags |= CODE_INUSE; /* prevents infinite loops */
      while (deco = ppDecoInit(),
	     ppDecorate(code->prgm,code->last,deco,NewMap(),NewMap()),
	     changed)
	{
	  changed = ppPatch(code->prgm,deco);
	  QUIT;
	}
      deco = Deforest(c,deco);
      fillCode(c,deco);
      {
	ExprPtr env;
	ppInfo *pp;

	env = CDR(CAR(code->prgm));
	pp = getPpInfo(env,deco);
	if (pp->uses==NewMap())
	  code->flags |= CODE_DOESNOTUSE_ENV; /* frequent case, with big returns
						 at application time */
      }
      /*!!! tail-recursion elimination */
      if (inlinable(c))
	code->flags |= CODE_INLINE;
      code->flags ^= (CODE_INUSE | CODE_ANALYZED); /* set flag to CODE_ANALYZED */
      QUIT;
    }
}

#endif
