/* Stack and continuation handling functions.
   This is mostly black magic.
   Copyright (C) 1992-1997 Jean Goubault-Larrecq and Bull S.A.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "stack.h"
#include "eval.h"
#include "evlproto.h"
#include "stkproto.h"
#ifdef MPW
#include <memory.h>
#endif
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#if OS==AMIGA_OS
#undef SHORT
/* SHORT is defined in types.h as short */
#include <exec/types.h>
#include <exec/tasks.h>
#include <proto/exec.h>
#endif
#ifdef HAS_PROTOTYPES
#include "sstack.h"
#endif

#define __SEG__ STACK

PUBLIC char *highMark,*lowMark;
EXTERN ExprPtr theEnvironment,basisEnv,handlerStack;
EXTERN gcStackView *theStackFrames;

PRIVATE jmp_buf topStackBuf;
PRIVATE jmp_buf botStackBuf;

PRIVATE StackChunk *theChunk;
PRIVATE stackDesc theStackDesc;
PRIVATE int botData;

/* the increment by which we progressively deepen the stack: */
#define STACK_INC 100
/* the number of bytes needed to make safe calls to
   establishBotStack() and longjmp(), counting the space
   needed to save registers: */
#define SECURITY	300

PRIVATE void establishBotStack()
{
  char reserve[STACK_INC];

  reserve[0] = 0; /* to prevent optimizers from freeing the space of the array */
#ifdef ASCENDING_STACK
  if (reserve+STACK_INC+SECURITY<=lowMark)
    {
      establishBotStack(); /* recurse until enough stack depth */
      return;
    }
  lowMark = reserve+STACK_INC-SAFETY_SIZE;
#else
  if (reserve-STACK_INC-SECURITY>=lowMark)
    {
      establishBotStack(); /* recurse until enough stack depth */
      return;
    }
  lowMark = reserve+SAFETY_SIZE;
#endif
  if ((botData = SETJMP(botStackBuf))==0)
    {
      return; /* ok */
    }
  else
    { /* don't use any locals, and don't simply return ! */
      /* the reason is that this portion of code is called in
	 an artificially lowered stack (thus no stack is really
	 present yet); see stackJmp()
	 */
#ifdef ASCENDING_STACK
      memcpy(theChunk->highMark,theChunk->contents,
	     theChunk->lowMark-theChunk->highMark);
#else
      memcpy(theChunk->lowMark,theChunk->contents,
	     theChunk->highMark-theChunk->lowMark);
#endif
      /* smash stack with contents of the chunk */
      LONGJUMP(theStackDesc,botData);
    }
}

#if OS==UNIX_SYS_V || OS==UNIX_BSD
PUBLIC void expandStack(mark)
IN CONST char *mark;
{
#ifdef ASCENDING_STACK
  if (mark>lowMark)
    {
      lowMark = mark;
      establishBotStack();
    }
#else
  if (mark<lowMark)
    {
      lowMark = mark;
      establishBotStack();
    }
#endif
}
#endif

PUBLIC NORETURN void stackJmp(chunk,jb,data)
IN CONST StackChunk *chunk;
IN CONST stackDesc *jb;
IN CONST int data;
{
  if (chunk==(StackChunk *)0L)
    LONGJUMP(*jb,data);
  theChunk = chunk;
  theStackDesc = *jb;
  LONGJMP(botStackBuf,data);
  /* 1) pass control to establishBotStack(), where stack is lowered
     to the bottom of the ML stack (lowMark-SAFETY_SIZE);
     2) there, the stack described by the chunk is copied inside
     the processor stack, and control is passed to the
     jmp_buf jb (which are above lowMark)
     */
}

PRIVATE jmp_buf theReturnBuf;
PRIVATE ExprPtr (*theFun) P((INOUT char *data));
PRIVATE char *theArg;
PRIVATE ExprPtr theResult;

PRIVATE void establishStack()
{
  char reserve[TOP_STACK_OFFSET]; /* to allocate space on the stack */

  reserve[0] = 0; /* to prevent optimizers from freeing the space of the array */
#ifdef ASCENDING_STACK
  highMark = reserve+TOP_STACK_OFFSET;
  lowMark = reserve+STACK_SIZE+SAFETY_SIZE;
#else
  highMark = reserve;
#if OS==MAC /* a Mac has a descending stack and a limit to the stack */
  lowMark = GetApplLimit();
#else
#if OS==AMIGA_OS /* an Amiga has a descending stack and a limit, too */
  {
    struct Task *myself;

    myself = FindTask(NULL);
    lowMark = myself->tc_SPLower;
  }
#else
  lowMark = reserve-STACK_SIZE-SAFETY_SIZE;
#endif
#endif
#endif
  if (SETJMP(topStackBuf)==0)
    {
      establishBotStack();
    }
  else
    { /* don't use any locals, and don't simply return ! */
      /* the reason is that this portion of code is called in
	 an artificially lowered stack (thus no stack is really
	 present yet); see doML()
	 */
      theResult = (*theFun)(theArg);
      LONGJMP(theReturnBuf,1);
    }
}

PUBLIC void flushstk(fun,data)
     IN mlAction fun;
     /* to use with precautions; mainly, (*fun) must never return,
	because otherwise it would get to the longjmp(theReturnBuf,1)
	at the end of establishStack(), but this would return to the
	end of doML(), terminating all computations */
     INOUT char *data;
{
  theFun = fun;
  theArg = data;
  LONGJMP(topStackBuf,1);
}

PUBLIC ExprPtr doML(fun,data) /* computes fun(data) in the ML stack */
     IN mlAction fun;
     /* !!! not reentrant, because it uses theReturnBuf, which is global */
     INOUT char *data;
{
  SETQ(theEnvironment,basisEnv);
  resetHandlerStack();
  theStackFrames = (gcStackView *)0L;
  theFun = fun;
  theArg = data;
  if (SETJMP(theReturnBuf)==0)
    LONGJMP(topStackBuf,1);
  /* 1) pass control to establishStack(), where stack is lowered
     to the top ML stack (highMark);
     2) there, the function is applied, and a longjmp(theReturnBuf,1)
     is executed, which returns control here, with the stack
     at its original (above highMark) level.
     */
  SETQ(theEnvironment,basisEnv);
  resetHandlerStack();
  return theResult;
}

#undef __SEG__
#define __SEG__ INIT

PUBLIC void InitStack()
{
  establishStack();
}
