
infix 0 $

datatype term =
    V of string (* named variable *)
  | $ of string * term list (* application *)

datatype form = A of term
  | FALSE
  | AND of form * form
  | OR of form * form
  | NOT of form
  | IMP of form * form
  | EQV of form * form
  | ALL of term * string * int * form
  | EX of term * string * int * form;

val ctors = { "0" => (nil, "nat"),
	      "S" => (["nat"], "nat"),
	      "nil" => (nil, "nat list"),
	      "cons" => (["nat", "nat list"], "nat list"),
	      "K" => (["key"], "base"),
	      "B" => (["base"], "msg"),
	      "tuple" => (["msg", "msg"], "msg"),
	      "crypt" => (["msg", "key"], "msg"),
	      "inftree" => (["inftree", "nat", "inftree"], "inftree")
	      };

val funs = { "+" => (["nat", "nat"], "nat") };
val defsort = "msg";

val functs = ref {} : (string -m> bool list set) ref;
val constrs = ref {} : (term -m> term) ref;

exception Munge;

fun tc env =
    let fun tc1 (x as V _) = (?env x handle MapGet => raise Munge)
	  | tc1 (f $ l) =
	    let val types = [tc1 t | t in list l]
	    in
		if f inset ctors
		    then let val (argtypes, restype) = ?ctors f
			 in
			     if argtypes=types
				 then restype
			     else raise Munge
			 end
		else if f inset funs
			 then let val (argtypes, restype) = ?funs f
			      in
				  if argtypes=types
				      then restype
				  else raise Munge
			      end
		     else if all t=defsort | t in list types end
			      then defsort
			  else raise Munge
	    (* unknown function symbols are assumed of signature
	     S x ... x S -> S, where S=defsort, and there are
	     as many arguments as specified. *)
	    end
    in
	tc1
    end;

val env = ref {} : (term -m> string) ref (* typing environment *)
val newenv = ref {} : (term -m> string) ref (* same, but only for fresh variables *)

val genvarcnt = ref 0;

fun genvar prefix = let val sf as |[convert, put, ...]| = outstring prefix
		    in
			inc genvarcnt;
			print sf (pack (!genvarcnt));
			V (convert ())
		    end;

fun trues n = let val r = ref nil
		  val cnt = ref n
	      in
		  while !cnt>0 do
		      (r := true :: !r;
		       dec cnt);
		      !r
	      end;

fun munge1 (x as V _) = x
  | munge1 (t as (f $ l)) =
    if f inset ctors (* if f is a constructor, then we leave it in *)
	then f $ munge_list l
    else if t inset !constrs
	     then
		 (case ?(!constrs) t of
		      _ $ [x, ...] => x)
	 else let val l' = munge_list l
		  val typ = tc (!env) (f $ l')
		  val x = genvar "$"
		  val Pf = "is-" ^ f
		  val c = Pf $ (x :: l')
	      in
		  env := !env ++ {x => typ};
		  newenv := !newenv ++ {x => typ};
		  functs := !functs ++ {Pf =>
					 (if Pf inset !functs
					      then ?(!functs) Pf
					  else {})
					      U {false::trues (len l)}};
		  constrs := !constrs ++ {t => c};
		  x
	      end
and munge_list l = [munge1 t | t in list l];

fun tc_atom (_ $ l) =
    let val tc1 = tc (!env)
    in
	iterate
	  tc1 t
	| t in list l
	end
    end;

fun flush_ex F = let val Fr = ref F
		 in
		     iterate
		       Fr := EX (x, ?(!newenv) x, 1, AND (A c, !Fr))
		     | _ => c as _ $ (x :: _) in map !constrs
		     end;
		     constrs := {};
		     env := {};
		     newenv := {};
		     !Fr
		 end;

fun fmunge (e, A (P $ l)) = (env := e;
			     let val t' = P $ munge_list l
			     in
				 tc_atom t';
				 A t'
			     end)
  | fmunge (e, ALL (x, d, m, F)) =
    ALL (x, d, m,
	 flush_ex (fmunge (e ++ {x => d}, F)))
    ;

val F = ALL (V "p", "nat", 3, A ("=" $ ["+" $ ["0" $ nil, V "p"], V "p"]));

fmunge ({}, F);

ok "dt1" true;
