(* An implementation of futures a la MultiLisp.
   Copyright (C) 1992-1995 Jean Goubault and Bull S.A.
   All rights reserved.

   Redistribution and use in source and binary forms are permitted provided
   that: (1) source distributions retain this entire copyright notice and
   comment, and (2) distributions including binaries display the following
   acknowledgement:  ``This product includes software developed by Jean
   Goubault as a contributor of Bull S.A.'' in the
   documentation or other materials provided with the distribution and in
   all advertising materials mentioning features or use of this software.
   Neither the name of Bull S.A. or of its corporate affiliates,
   nor the names of its contributors may be used to endorse or promote
   products derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*)

(* requires mutex.ml *)

abstype ('a,''env,'rs) future =
    FUTURE of (''env,'rs) condition * 'a cell ref
and 'a cell =
    BUSY (* the future is currently performing the computation *)
  | DONE of 'a (* the result is available *)
  | EXN of exn (* the future has raised an exception *)
with
  fun future_manager (tm as |[fork,...]| : (''env thread,'rest,'rs) thread_manager,
		      new_sched : unit -> (''env thread,'rs) scheduler) =
      let val |[mutex,condition,with_condition,broadcast,wait,...]| =
		mutex_manager (tm,new_sched)
	  fun future (func,arg) =
	      let val c = condition (mutex ())
		  val f = ref BUSY
		  fun wrapper () =
		      let val res = DONE (func arg)
					handle exn => EXN exn
		      in
			 with_condition (c,
					 (fn () => (f := res;
						    broadcast c)))
		      end
	      in
		 fork wrapper; FUTURE (c,f)
	      end
	  fun touch (FUTURE (c,f)) =
	      let fun touch' () =
		      case !f of
			   BUSY => (wait c; touch' ())
			 | DONE x => x
			 | EXN exn => raise exn
	      in
		 with_condition (c, touch')
	      end
      in
	 |[future = future,
	   touch = touch,
	   cobegin = fn funlist =>
			iterate touch (future (f,()))
			| f in list funlist
			end
	   ]|
      end
end
