(* An implementation of reader/writer locks (many readers, one writer).
   Copyright (C) 1992-1995 Jean Goubault and Bull S.A.
   All rights reserved.

   Redistribution and use in source and binary forms are permitted provided
   that: (1) source distributions retain this entire copyright notice and
   comment, and (2) distributions including binaries display the following
   acknowledgement:  ``This product includes software developed by Jean
   Goubault as a contributor of Bull S.A.'' in the
   documentation or other materials provided with the distribution and in
   all advertising materials mentioning features or use of this software.
   Neither the name of Bull S.A. or of its corporate affiliates,
   nor the names of its contributors may be used to endorse or promote
   products derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*)

(* requires mutex.ml *)

abstype (''env,'rs) rw =
    RW of |[free : (''env,'rs) condition,
	    readers : int ref,
	    write : bool ref
	    ]|
with
  fun rw_manager (tm : (''env thread,'rest,'rs) thread_manager,
		  new_sched : unit -> (''env thread,'rs) scheduler) =
      let val |[mutex,condition,await,broadcast,with_condition,...]| =
		mutex_manager (tm,new_sched)
	  fun rw () =
	      RW |[free = condition (mutex ()),
		   readers = ref 0,
		   write = ref false
		   ]|
	  fun lock_read (RW |[free,readers,write]|) =
	      with_condition (free,
			      fn () => (await (free, fn () => not (!write));
					readers := !readers+1))
	  fun unlock_read (RW |[free,readers,...]|) =
	      with_condition (free,
			      fn () => (readers := !readers-1;
					if !readers=0
					    then broadcast free
					else ()))
	  fun lock_write (RW |[free,readers,write]|) =
	      with_condition (free,
			      fn () => (await (free,
					       fn () => !readers=0
					       andalso not (!write));
					write := true))
	  fun unlock_write (RW |[free,readers,write]|) =
	      with_condition (free,
			      fn () => (write := false;
					if !readers=0
					    then broadcast free
					else ()))
	  fun with_lock (lockfun,unlockfun) (lock,f : unit -> 'a) =
	      (lockfun lock;
	       f () handle exn => (unlockfun lock; raise exn))
	      before unlockfun lock
	  val read = with_lock (lock_read,unlock_read)
	  val write = with_lock (lock_write,unlock_write)
      in
	 |[rw = rw,
	   read_with_lock = read,
	   write_with_lock = write
	   ]|
      end
end
