(* FIFO queues, useful in particular to implement breadth-first search.
   Copyright (C) 1992-1995 Jean Goubault and Bull S.A.
   All rights reserved.

   Redistribution and use in source and binary forms are permitted provided
   that: (1) source distributions retain this entire copyright notice and
   comment, and (2) distributions including binaries display the following
   acknowledgement:  ``This product includes software developed by Jean
   Goubault as a contributor of Bull S.A.'' in the
   documentation or other materials provided with the distribution and in
   all advertising materials mentioning features or use of this software.
   Neither the name of Bull S.A. or of its corporate affiliates,
   nor the names of its contributors may be used to endorse or promote
   products derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*)

(* new_fifo() creates a new first-in first-out queue, as
 and object with methods empty (testing whether the queue
 is empty), put (to put an element on one end), get (to
 get an element from the other end), and flush (to empty
 the queue).
 Exception FifoEmpty is raised on getting from an empty queue.
*)

datatype '_a mutlist = Mnil | Mcons of '_a * '_a mutlist ref

type ('a,'rest) fifo = |[empty : unit -> bool,
			 put : 'a -> unit,
			 get : unit -> 'a,
			 flush : unit -> unit,
			 ... : 'rest]|

exception FifoEmpty

fun new_fifo () =
    let val l = ref Mnil (* head of the list *)
	val lp = ref l (* tail of the list *)
    in
	|[flush = (fn () => (l := Mnil; (* to empty the fifo *)
			     lp := l)),
	  empty = (fn () => !lp=l), (* tests if it is empty,
				     could be !l=Mnil at the expense
				     of restricting use to equality types *)
	  put = (fn x => let val ender = ref Mnil (* puts x at end *)
			 in
			     !lp := Mcons (x,ender);
			     lp := ender
			 end),
	  get = (fn () => let val Mcons(x,r) = !l (* gets x from head *)
			  in
			      l := !r;
			      if !lp=r
				  then (lp := l;
					x)
			      else x
			  end handle Bind => raise FifoEmpty)
	  ]|
    end
