(* Fluid bindings on references.
   Copyright (C) 1992-1995 Jean Goubault and Bull S.A.
   All rights reserved.

   Redistribution and use in source and binary forms are permitted provided
   that: (1) source distributions retain this entire copyright notice and
   comment, and (2) distributions including binaries display the following
   acknowledgement:  ``This product includes software developed by Jean
   Goubault as a contributor of Bull S.A.'' in the
   documentation or other materials provided with the distribution and in
   all advertising materials mentioning features or use of this software.
   Neither the name of Bull S.A. or of its corporate affiliates,
   nor the names of its contributors may be used to endorse or promote
   products derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*)

(* A fluid binding of a reference to a value
 is a local binding with dynamic scope. *)

fun fluid (r as ref oldv,v,func) =
    (r := v;
     (func () handle exn => (r := oldv;
			     raise exn))
     before r := oldv)

(* usage:
 fluid (my_ref, my_new_value, function)
 sets my_ref to my_new_value during the execution of the body of function,
 and restores it afterwards.
*)

fun fluids (bindings,func) =
    let val oldvals = [(r := newv; oldv)
		      | (r as ref oldv,newv) in list bindings]
	fun restore () =
	    iterate r := oldv
	     || (r,_) in list bindings and
		oldv in list oldvals
	    end
    in
	(func () handle exn => (restore ();
				raise exn))
	    before restore ()
    end

(* usage:
 fluids (list of (my_ref, my_new_value), function)
 sets all the my_refs to the values my_new_value during
 the execution of the body of function, and restores them
 afterwards.
*)
