/* Bytecode interpreter support functions.
   Copyright (C) 2005,2011 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "expr.h"
#include "hakproto.h"
#include "bciaux.h"

#define __SEG__ EVAL

PUBLIC void mergeFieldLists(fields1,fields2,newfields,m,n,np)
     /* merge lists fields1 and fields2, putting result into newfields,
	returning length of newfields in *np, and replacing entries
	in fields1 and fields2 by the indexes to the new positions
	of fields inside newfields. */
INOUT REGISTER ExprPtr *fields1,*fields2;
OUT REGISTER ExprPtr *newfields;
IN REGISTER INT m,n;
OUT INT *np;
{
  INT pos;

  pos = 0;
  if (m==0)
    {
copyrest:
      while (n--!=0)
	{
	  *newfields++ = *fields2;
	  *fields2++ = (ExprPtr)pos++;
	}
      *np = pos;
      return;
    }
  if (n==0)
    {
      n = m;
      fields2 = fields1;
      goto copyrest;
    }
again:
  if (m==1)
    { /* insert *fields1 at a position we find by dichotomy inside fields2 */
      ExprPtr s1,s2;
      REGISTER INT i,j,k;

      s1 = *fields1;
      i = n>>1;
      j = 0;
      k = n;
      while (j!=k)
	{
	  s2 = fields2[i];
	  switch (CompareStrings(s1,s2))
	    {
	    case -1: k = i; break;
	    case 1: j = i+1; break;
	    case 0:
	      *fields1 = (ExprPtr)-1;
	      goto copyrest;
	    }
	  i = (j+k)>>1;
	}
      for (i=0; i<j; i++)
	{
	  *newfields++ = *fields2;
	  *fields2++ = (ExprPtr)pos++;
	}
      *newfields++ = *fields1; /* insert element */
      *fields1 = (ExprPtr)pos++;
      for (; i<n; i++)
	{
	  *newfields++ = *fields2;
	  *fields2++ = (ExprPtr)pos++;
	}
      *np = pos;
      return;
    }
  if (n==1)
    {
      ExprPtr s1,s2;
      REGISTER INT i,j,k;

      s2 = *fields2;
      i = m>>1;
      j = 0;
      k = m;
      while (j!=k)
	{
	  s1 = fields1[i];
	  switch (CompareStrings(s2,s1))
	    {
	    case -1: k = i; break;
	    case 1: j = i+1; break;
	    case 0:
	      for (j=0; j<i; j++)
		{
		  *newfields++ = *fields1;
		  *fields1++ = (ExprPtr)pos++;
		}
	      *newfields++ = s1; /* insert element */
	      *fields1++ = (ExprPtr)-1;
	      *fields2 = (ExprPtr)pos++;
	      for (; ++j<m;)
		{
		  *newfields++ = *fields1;
		  *fields1++ = (ExprPtr)pos++;
		}
	      *np = pos;
	      return;
	    }
	  i = (j+k)>>1;
	}
      for (i=0; i<j; i++)
	{
	  *newfields++ = *fields1;
	  *fields1++ = (ExprPtr)pos++;
	}
      *newfields++ = *fields2; /* insert element */
      *fields2 = (ExprPtr)pos++;
      for (; i<m; i++)
	{
	  *newfields++ = *fields1;
	  *fields1++ = (ExprPtr)pos++;
	}
      *np = pos;
      return;
    }
  { /* otherwise, advance in parallel in boths lists */
    ExprPtr s1,s2;

    s1 = *fields1;
    s2 = *fields2;
    switch (CompareStrings(s1,s2))
      {
      case -1:
	*newfields++ = s1;
	*fields1++ = (ExprPtr)pos++; m--;
	break;
      case 0:
	*newfields++ = s1;
	*fields1++ = (ExprPtr)-1; m--;
	*fields2++ = (ExprPtr)pos++; n--;
	break;
      case 1:
	*newfields++ = s2;
	*fields2++ = (ExprPtr)pos++; n--;
	break;
      }
  }
  goto again;
}

MAPSWEEPFUNC(do_dumpshape)
{
  ExprPtr *fields;

  fields = MAPDATA(ExprPtr *);
  fields[IVAL(y)-1] = x;
  return false;
}

PUBLIC void dumpShape(shape,fields)
IN CONST ExprPtr shape;
OUT ExprPtr *fields;
{
  (void) MapSweep(shape,do_dumpshape,fields);
}

PRIVATE MapPtr undumpShape1(fields,m,n)
IN REGISTER CONST ExprPtr *fields;
IN REGISTER INT m,n;
{
  if (n==0)
    return NewMap();
  if (n==1)
    {
      ExprPtr e;

      BEGINQUICKKEEP();
      e = Kint(m);
      QUICKKEEP(e);
      e = MapletKons(fields[0],e);
      ENDQUICKKEEP();
      return e;
    }
  {
    struct _x { COUNTER n; gcStackView *next; ExprPtr p[2];} _x;
    INT nn = n>>1;

    _x.p[0] = undumpShape1(fields,m,nn);
    _x.n = 1;
    _x.next = theStackFrames;
    theStackFrames = (gcStackView *)&_x;
    _x.p[1] = undumpShape1(fields+nn,m+nn,n-nn);
    _x.n = 2;
    _x.p[0] = MapOverwrite(_x.p[0],_x.p[1]);
    theStackFrames = _x.next;
    return _x.p[0];
  }
}

PUBLIC MapPtr undumpShape(fields,n)
IN CONST ExprPtr *fields;
IN INT n;
{
  MapPtr shape;

  shape = undumpShape1(fields,(INT)1,n);
  (void) KEEP(shape);
  return shape;
}

#define MU1(name,share) \
PUBLIC ExprPtr name(rec,modlab,modval) \
IN CONST ExprPtr rec; \
IN CONST ExprPtr modlab,modval; \
{ \
  ENTER; \
 \
  switch (rec->kind) \
    { \
    case E_RECORD_0: \
buildmod: \
      if (share) \
	rec = Krecord1(modlab,modval); \
      else \
	{ \
	  MAKERECORD1(modlab,modval,rec); \
	} \
      break; \
    case E_RECORD_1: \
      if (CAR(rec)==modlab) /* if same field, return mod */ \
	goto buildmod; \
      { /* otherwise, build two-field record */ \
	ExprPtr shape; \
	ExprPtr exprs[2]; \
	   \
	if (CompareStrings(CAR(rec),modlab)<0) \
	  { \
	    shape = MapPair(CAR(rec),ione,modlab,itwo); \
	    exprs[0] = CDR(rec); \
	    exprs[1] = modval; \
	  } \
	else \
	  { \
	    shape = MapPair(modlab,ione,CAR(rec),itwo); \
	    exprs[0] = modval; \
	    exprs[1] = CDR(rec); \
	  } \
	rec = share?KrecordGnrl(shape,2,exprs):MakeRecord(E_RECORD_GNRL,shape,2,exprs); \
	break; \
      } \
    case E_RECORD_GNRL: \
      { \
	ExprPtr shape; \
	INT i,n,nn; \
	ExprPtr *ep,*epp; \
 \
	shape = ASHAPE(rec); \
	n = AVAL(rec)->n; \
	{ \
	  ExprPtr *fields,*outfields,*exprs; \
 \
	  fields = (ExprPtr *)alloctmp((3*n+2)*sizeof(ExprPtr)); \
	  outfields = fields + n; /* fields has n elements */ \
	  exprs = outfields+1+n; /* outfields may have up to n+1 elements */ \
	  /* and exprs will have up to n+1 elements */ \
	  dumpShape(shape,fields); \
	  mergeFieldLists(fields,&modlab,outfields,n,(INT)1,&nn); \
	  shape = undumpShape1(outfields,(INT)1,nn); \
	  (void) KEEP(shape); \
 \
	  for (i=0, ep=AVAL(rec)->expr, epp=exprs; i<(INT)modlab; i++) \
	    *epp++ = *ep++; \
	  *epp++ = modval; i++; \
	  if (nn==n) \
	    ep++; \
	  for (; i<nn; i++) \
	    *epp++ = *ep++; \
	  rec = share?KrecordGnrl(shape,nn,exprs):MakeRecord(E_RECORD_GNRL,shape,nn,exprs); \
	} \
	break; \
      } \
    } \
  QUIT; \
  return rec; \
}

MU1(Kmu1,true)
MU1(Mu1,false)

#define MUGNRL(name,share) \
PUBLIC ExprPtr name(rec,mod) \
IN CONST ExprPtr rec; \
IN CONST ExprPtr mod; /* mod is a E_GNRL_RECORD */ \
{ \
  ENTER; \
 \
  switch (rec->kind) \
    { \
    case E_RECORD_0: \
      rec = mod; \
      break; \
    case E_RECORD_1: \
      { \
	ExprPtr shape; \
	Boolean found; \
 \
	shape = ASHAPE(mod); \
	if ((void) MapGet(CAR(rec),shape,&found), found) \
	  rec = mod; \
	else \
	  { \
	    INT i,n,nn; \
	    ExprPtr lab; \
	    ExprPtr *ep,*epp; \
 \
	    n = AVAL(mod)->n; \
	    { \
	      ExprPtr *fields,*outfields,*exprs; \
 \
	      fields = (ExprPtr *)alloctmp((3*n+2)*sizeof(ExprPtr)); \
	      outfields = fields + n; /* fields has n elements */ \
	      exprs = outfields+1+n; /* outfields may have up to n+1 elements */ \
	      /* and exprs will have up to n+1 elements */ \
	      dumpShape(shape,fields); \
	      lab = CAR(rec); \
	      mergeFieldLists(&lab,fields,outfields,(INT)1,n,&nn); /* here nn=n+1 */ \
	      shape = undumpShape1(outfields,(INT)1,nn); \
	      (void) KEEP(shape); \
	      for (i=0, ep=AVAL(mod)->expr, epp=exprs; i<(INT)lab; i++) \
		*epp++ = *ep++; \
	      *epp++ = CDR(rec); i++; \
	      for (; i<nn; i++) \
		*epp++ = *ep++; \
	      rec = share?KrecordGnrl(shape,nn,exprs):MakeRecord(E_RECORD_GNRL,shape,nn,exprs); \
	    } \
	  } \
      } \
      break; \
    case E_RECORD_GNRL: \
      { \
	INT i,m,n,nn; \
	ExprPtr shape; \
	ExprPtr *ep,*epp,*fp; \
 \
	m = AVAL(rec)->n; \
	n = AVAL(mod)->n; \
	{ \
	  ExprPtr *fields1,*fields2,*outfields,*exprs; \
 \
	  nn = m+n; \
	  fields1 = (ExprPtr *)alloctmp(nn*(3*sizeof(ExprPtr))); \
	  fields2 = fields1+m; /* fields1 has m elements */ \
	  outfields = fields2+n; /* fields2 has n elements */ \
	  exprs = outfields+nn; /* outfields may have up to m+n elements */ \
	  /* and exprs will have up to m+n elements */ \
	  dumpShape(ASHAPE(rec),fields1); \
	  dumpShape(ASHAPE(mod),fields2); \
	  mergeFieldLists(fields1,fields2,outfields,m,n,&nn); \
	  shape = undumpShape1(outfields,(INT)1,nn); \
	  (void) KEEP(shape); \
	  epp = exprs; \
	  for (i=0, ep=AVAL(rec)->expr, fp=fields1; i<m; i++, ep++) \
	    { \
	      INT ii; \
 \
	      ii = (INT)*fp++; \
	      if (ii!=-1) \
		SETQ(epp[ii],*ep); \
	    } \
	  for (i=0, ep=AVAL(mod)->expr, fp=fields2; i<n; i++, ep++) \
	    { \
	      SETQ(epp[(INT)*fp++],*ep); \
	    } \
	  rec = share?KrecordGnrl(shape,nn,exprs):MakeRecord(E_RECORD_GNRL,shape,nn,exprs); \
	} \
	break; \
      } \
    } \
  QUIT; \
  return rec; \
}

MUGNRL(KmuGnrl,true)
MUGNRL(MuGnrl,false)

PUBLIC ExprPtr envCons(n,env)
IN REGISTER COUNTER n;
IN CONST ExprPtr env;
{
  VecPtr vp;

  CHECKHETEROFREE();
  scAlloc(*(char **)&vp,VECSIZE(n));
  vp->n = n;
  {
    REGISTER ExprPtr *ep,*epend;

    for (ep = vp->expr, epend = ep+n; ep<epend; ) *ep++ = NOEXPR;
  }
  {
    REGISTER ExprPtr res;

    MAKEENV1_NOGC(vp,env,res);
    return res;
  }
}

PUBLIC ExprPtr QueueToList(queue)
IN ExprPtr queue; /* a non-empty queue */
{
  /* If we didn't need to share the resulting list, it would be enough to do:
  e = CDR(queue);
  SETCDR(queue,NIL);
  */
  {
    REGISTER ExprPtr e,last,next,fin;

    /* Instead, we first reverse the circular list in place,
       while also breaking the circularity: */
    for (last = NIL, e = fin = CDR(queue);;)
      {
	next = CDR(e);
	SETCDR(e,last);
	if (next==fin)
	  break;
	last = e;
	e = next;
      }
    /* Note that queue is still protected against the GC, and is
       the reversed list itself now; so we don't need to KEEP anything. */
  }
  /* And then we reverse the resulting list, which involves sharing it: */
  return Reverse1(queue,NIL);
}

PUBLIC ExprPtr ListToQueue(list)
IN CONST ExprPtr list;
{
  REGISTER ExprPtr e,end,last;

  BEGINQUICKKEEP();
  EALLOC_HOMO(e);
  e->kind = (CHAR)E_CONS;
  CAR(e) = CAR(list);
  CDR(e) = e;
  QUICKKEEP(e);
  end = e;
  for (; last = e, list = CDR(list), list!=NIL; )
    {
      EALLOC_HOMO(e);
      e->kind = (CHAR)E_CONS;
      CAR(e) = CAR(list);
      CDR(e) = end;
      SETCDR(last,e);
    }
  ENDQUICKKEEP();
  return e;
}

PUBLIC Boolean do_initialsubmap(x,y,_data)
IN CONST ExprPtr x,y;
INOUT CHAR *_data;
{
  ExprPtr *listp;
  ExprPtr aux,aux2;

  listp = MAPDATA(ExprPtr *VOLATILE);
  aux = MapletKons(x,y); /* no need to keep it, it is already shared by the whole map
			    we are traversing. */
  MAKEPROG(E_OR,aux,*listp,aux2);
  *listp = aux2;
  return false;
}

DEF_ACTION_COLLECT(datacon_shared)
{
  ExprPtr res;
  ExprPtr uenv;

  uenv = ENV_UNDERLYING(env);
  res = Kdata(uenv->magic, CAR(uenv), arg);
  return res;
}
ACTION_END

PUBLIC void resetHandlerStack()
{
  excStack = NULL;
}

PUBLIC void setHandlerStack(stack)
IN CONST excHandler *stack;
{
  excStack = stack;
}
