%{
/* himmldep dependency calculator.
   Copyright (C) 2001-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdlib.h>
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif

#define KW_OPEN 257
#define KW_USE 258
#define KW_STRING 259

%}

SPACE [\ \t\f]
DIG [0-9]
INT (\~?{DIG}+)
FMT [\ \t\n\f]
LETTER [A-Za-z]
ALPHA [A-Za-z0-9\'\_]
SYMB [\!\%\&\$\#\+\-\/\:\<\=\>\?\@\\\~\'\^\|\*\`]
ALPHA_ID ({LETTER}{ALPHA}*)
SYMB_ID ({SYMB}+)
ID ({SYMB_ID}|{ALPHA_ID})
XID ({ID}|\_{ID})
PREFIX ({XID}{FMT}*\.{FMT}*)

%x comment

%option noyywrap

%%

<INITIAL>open { return KW_OPEN; }
<INITIAL>use { return KW_USE; }
<INITIAL>\(\* { BEGIN(comment); }
<comment>\*\) { BEGIN(INITIAL); }
<INITIAL>\"([^\"\\]|(\\([\"\\ntvf]|(\^[\@A-Z\[\\\]\^\_])|({DIG}{DIG}{DIG})|({FMT}*\\))))*\" { return KW_STRING; }
<INITIAL>{SPACE}+ { ; }
<comment>\"([^\"\\]|(\\([\"\\ntvf]|(\^[\@A-Z\[\\\]\^\_])|({DIG}{DIG}{DIG})|({FMT}*\\))))*\" { ; }
<comment>. { ; }
<INITIAL>. { return yytext[0]; }
<INITIAL,comment><<EOF>> { return -1; }
<INITIAL,comment>\n { ; }

%%

int getString(text)
register char *text;
{
  char *begin;
  register int len;
  register char *dest,c;

  begin = text;
  len = 0;
  for (dest = text, text++; c = *text++, c!='\"'; )
    {
      switch (c)
	{
	case '\\':
	  c = *text++;
	  switch (c)
	    {
	    case '\"':   case '\\':
	      *dest++ = c;
	      len++;
	      break;
	    case 'n':
	      *dest++ = '\n';
	      len++;
	      break;
	    case 't':
	      *dest++ = '\t';
	      len++;
	      break;
	    case 'v':
	      *dest++ = '\v';
	      len++;
	      break;
	    case 'f':
	      *dest++ = '\f';
	      len++;
	      break;
	    case '^':
	      *dest++ = (*text++)-'@';
	      len++;
	      break;
	    case '0':	case '1':	case '2':	case '3':
	    case '4':	case '5':	case '6':	case '7':
	    case '8':	case '9':
	      *dest = (c-'0')*100;
	      *dest += ((*text++)-'0')*10;
	      *dest++ += ((*text++)-'0');
	      len++;
	      break;
	    default:
	      while (c = *text++, c!='\\');
	      break;
	    }
	  break;
	case '\n':
	  break;
	default:
	  *dest++ = c;
	  len++;
	  break;
	}
    }
  *dest = 0;
  return len;
}

struct filelist {
  struct filelist *next;
  YY_BUFFER_STATE state;
  char name[1]; /* variable-length string. */
};

struct filelist *theFileList = NULL;

int main (argc, argv)
int argc;
char *argv[];
{
  int i,tok;

  for (i=1; i<argc; i++)
    {
      yyin = fopen (argv[i], "r");
      if (yyin==NULL)
	{
	  fprintf (stderr, "%s: cannot open '%s'.\n", argv[0], argv[i]);
	  exit(10);
	}
      fprintf (stdout, "%sx : %s", argv[i], argv[i]);
      for (;;)
	switch (tok = yylex ())
	  {
	  case KW_OPEN:
	    switch (yylex ())
	      {
	      case -1: goto close;
	      case KW_STRING:
		(void) getString (yytext);
		fprintf (stdout, " %s.mlx", yytext);
		break;
	      }
	    break;
	  case KW_USE:
	    switch (yylex ())
	      {
	      case -1: goto close;
	      case KW_STRING:
		(void) getString (yytext);
		fprintf (stdout, " %s", yytext);
		{
		  int len = strlen ((char *)yytext);
		  int size = sizeof(struct filelist)+len;
		  struct filelist *fl;
		  FILE *usefile;

		  for (fl=theFileList; fl!=NULL; fl=fl->next)
		    if (strcmp (yytext, fl->name)==0)
		      {
			fprintf (stderr, "%s: recursive include '%s'.\n", argv[0], yytext);
			break;
		      }
		  usefile = fopen (yytext, "r");
		  if (usefile==NULL)
		    {
		      fprintf (stderr, "%s: cannot open '%s'.\n", argv[0], yytext);
		      break;
		    }
		  fl = malloc (size);
		  if (fl==NULL)
		    {
		      fprintf (stderr, "%s: cannot allocate %d bytes.\n", argv[0], size);
		      fclose (usefile);
		      break;
		    }
		  fl->next = theFileList;
		  fl->state = YY_CURRENT_BUFFER;
		  strcpy ((char *)fl->name, yytext);
		  theFileList = fl;
		  yyin = usefile;
		  yy_switch_to_buffer (yy_create_buffer (yyin, YY_BUF_SIZE));
		  BEGIN(INITIAL);
		}
		break;
	      }
	    break;
	  case -1:
	  close:
	  {
	    struct filelist *fl;

	    if (theFileList==NULL)
	      goto exit;
	    yy_delete_buffer (YY_CURRENT_BUFFER);
	    yy_switch_to_buffer (theFileList->state);
	    fl = theFileList->next;
	    free (theFileList);
	    theFileList = fl;
	    BEGIN(INITIAL);
	  }
	  }
    exit:
      fclose (yyin);
      fputs ("\n", stdout);
      fflush (stdout);
    }
  exit(0);
}
