/* Error signalling functions.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#include "evlproto.h"
#include "errproto.h"
#include "exeproto.h"
#include "prtproto.h"
#if OS==MAC
#ifndef THINK_C
#include <types.h>
#include <toolutils.h>
#endif
#endif
#ifdef __STDC__
#include <signal.h>
#endif
#ifdef HAS_STDARG
#include <stdarg.h>
#else
#ifdef HAS_VARARGS
#include <varargs.h>
#endif
#endif
#include "filproto.h"
#ifdef HAS_PROTOTYPES
#include "serrors.h"
#endif

#define __SEG__ ERROR

EXTERN ExprPtr toplevelCont;
EXTERN CHAR *yyfilename;
EXTERN char *yytext;
EXTERN Boolean syntax_analysis;
EXTERN int yynerrs;

EXTERN ExprPtr exnType;

#if CORE_TRACE
EXTERN hFILE *coreTraceFile;
#endif

#if DEBUGGING
EXTERN Boolean debugEnabled;
EXTERN Boolean evaluatingp,stopwanted;
EXTERN COUNTER debugMode;
#endif

EXTERN ExprPtr nomemException,stackException,specialMarker;
EXTERN Boolean inGC;

PUBLIC NORETURN ExprPtr Error(exception)
IN CONST ExprPtr exception; /* = cons(exc,arg) or exc, where exc:E_EXCEPTION */
{
#ifdef OBSOLETE
  ExprPtr res;
#endif

  inGC = false;
  hfprintf(hstderr,message(406));
  if (exception==nomemException)
    hfputs(message(200),hstderr); /* to avoid consuming memory we don't have */
  else
    valPrint(exception,exnType,do_printfile,(char *)hstderr,0);
  hfprintf(hstderr,message(407));
  hfflush(hstderr);
#if DEBUGGING
  debugMode = 0; /* = DBM_NONE */
#ifdef OBSOLETE
  if (evaluatingp && exception!=nomemException/* && exception!=stackException*/)
    { /* stackException allowed to enter the debugger, since it
    	is never actually raised for a stack overflow */
      res = Break(specialMarker);
      if (res!=specialMarker)
        return res;
    }
#endif
#endif
  if (toplevelCont!=NOEXPR && toplevelCont!=NIL)
    Toplevel(itwo); /* itwo is code for no memory */
  /*CRASH();*/
  ml_done(10);
  /*NOTREACHED*/
#ifndef HAS_NORETURN
  return NIL;
#endif
}

#ifdef HAS_STDARG
PUBLIC void bmsg(char *msg,...)
{
  va_list args;

  va_start(args,msg);
  (void) hvfprintf(hstderr,msg,args);
  va_end(args);
  hfflush(hstderr);
}
#else
#ifdef HAS_VARARGS
/*VARARGS*/
PUBLIC void bmsg(msg,va_alist)
char *msg;
va_dcl
{
  va_list args;
  
  va_start(args);
  (void) hvfprintf(hstderr,msg,args);
  va_end(args);
  hfflush(hstderr);
}
#else
PUBLIC void bmsg(msg,a,b,c,d,e,f,g,h)
IN char *msg;
IN long a,b,c,d,e,f,g,h;
{
  hfprintf(hstderr,msg,a,b,c,d,e,f,g,h);
  hfflush(hstderr);
}
#endif
#endif

#ifdef HAS_STDARG
PUBLIC NORETURN void berror(char *msg,...)
{
  va_list args;

  va_start(args,msg);
  (void) hvfprintf(hstderr,msg,args);
  va_end(args);
#else
#ifdef HAS_VARARGS
/*VARARGS*/
PUBLIC NORETURN void berror(msg,va_alist)
char *msg;
va_dcl
{
  va_list args;
  
  va_start(args);
  (void) hvfprintf(hstderr,msg,args);
  va_end(args);
#else
PUBLIC NORETURN void berror(msg,a,b,c,d,e,f,g,h)
IN CONST char *msg;
IN long a,b,c,d,e,f,g,h;
{
  bmsg(msg,a,b,c,d,e,f,g,h);
#endif
#endif
  hfputc('\n',hstderr);
  hfflush(hstderr);
#if CORE_TRACE
  if (coreTraceFile!=NULL)
    hfflush(coreTraceFile);
#endif
  CRASH();
}

#ifdef HAS_STDARG
PUBLIC NORETURN void bfatal(char *msg,...)
{
  va_list args;

  va_start(args,msg);
  (void) hvfprintf(hstderr,msg,args);
  va_end(args);
  hfflush(hstderr);
#else
#ifdef HAS_VARARGS
/*VARARGS*/
PUBLIC NORETURN void bfatal(msg,va_alist)
char *msg;
va_dcl
{
  va_list args;
  
  va_start(args);
  (void) hvfprintf(hstderr,msg,args);
  va_end(args);
  hfflush(hstderr);
#else
PUBLIC NORETURN void bfatal(msg,a,b,c,d,e,f,g,h)
IN CONST char *msg;
IN long a,b,c,d,e,f,g,h;
{
  bmsg(msg,a,b,c,d,e,f,g,h);
#endif
#endif
  hfputc('\n',hstderr);
  ml_done(10);
}

#ifdef HAS_PROTOTYPES
PUBLIC NORETURN void kinderror(IN CONST char *name,IN CHAR kind)
#else
PUBLIC NORETURN void kinderror(name,kind)
IN CONST char *name;
IN CHAR kind;
#endif
{
  berror("%s on bad kind: %d",name,kind);
}

PUBLIC void printZone(filename,zone,out,trailingp)
IN CONST CHAR *filename;
IN CONST TZone *zone;
INOUT hFILE *out;
IN Boolean trailingp;
{
  if (filename!=NULL && zone!=(TZone *)0L)
    {
#if OS==UNIX_SYS_V || OS==UNIX_BSD
      hfprintf(out,"%s, line%s %d",filename,((zone->line1!=zone->line2)?"s":""),
	       zone->line1+1);
      if (zone->pos1!=0)
	hfprintf(out,"(%d)",zone->pos1);
      if (zone->line2!=zone->line1 ||
	   zone->pos2!=zone->pos1)
	{
	  hfprintf(out,"-%d",zone->line2+1);
	  if (zone->pos2!=0)
	    hfprintf(out,"(%d)",zone->pos2);
	}
      if (trailingp)
	hfprintf(out,": ");
#else
#if OS==AMIGA_OS
      hfprintf(out," %s, line%s %d",filename,
	       ((zone->line1!=zone->line2)?"s":""),
	       zone->line1+1);
      if (zone->pos1!=0)
	hfprintf(out,"(%d)",zone->pos1);
      if (zone->line2!=zone->line1 ||
	   zone->pos2!=zone->pos1)
	{
	  hfprintf(out,"-%d",zone->line2+1);
	  if (zone->pos2!=0)
	    hfprintf(out,"(%d)",zone->pos2);
	}
      if (trailingp)
	hfprintf(out,": ");
#else
      /* error messages, with the Mac MPW conventions */
      hfprintf(out,"File \"%s\"; Line %d",filename,zone->line1);
      if (zone->pos1!=0)
	hfprintf(out,"!%d",zone->pos1);
      if (zone->line2!=zone->line1 ||
	   zone->pos2!=zone->pos1)
	{
	  hfprintf(out,":%d",zone->line2);
	  if (zone->pos2!=0)
	    hfprintf(out,"!%d",zone->pos2);
	}
      if (trailingp)
	hfprintf(out,"\t# ");
#endif
#endif
    }
}

PRIVATE void type_warn1(zone,msg,hdr)
IN CONST TZone *zone;
IN CONST CHAR *msg;
IN CONST char *hdr;
{
#if OS==UNIX_SYS_V || OS==UNIX_BSD || OS==AMIGA_OS
  hfputs(hdr,hstderr);
#endif
  printZone(yyfilename,zone,hstderr,true);
  hfputs((char *)msg,hstderr);
  if (syntax_analysis && yytext[0]!=0)
    {
      hfputs(" (at '",hstderr);
      hfputs(yytext,hstderr);
      hfputs("')",hstderr);
    }
  hputc('\n',hstderr);
  hfflush(hstderr);
}

PUBLIC void type_warn(zone,msg)
IN CONST TZone *zone;
IN CONST CHAR *msg;
{
  type_warn1(zone,msg,"Warning: ");
}

PUBLIC NORETURN void type_error(zone,msg)
IN CONST TZone *zone;
IN CONST CHAR *msg;
{
  if (msg!=NULL)
    type_warn1(zone,msg,"Error: ");
  if (toplevelCont!=NOEXPR && toplevelCont!=NIL)
    Toplevel(ione);
  berror(message(409));
}

PUBLIC void type_soft_error(zone,msg)
IN CONST TZone *zone;
IN CONST CHAR *msg;
{
  type_warn1(zone,msg,"Error: ");
  if (++yynerrs>=20)
    type_error(zone,(CHAR *)message(565));
}

PUBLIC void ml_break(n)
int n;
{
  signal(n,ml_break);
#if DEBUGGING
  if (debugEnabled && evaluatingp)
    stopwanted = true; /* so that Eval() will stop on the next debugzone */
  else
#endif
    if (toplevelCont!=NOEXPR && toplevelCont!=NIL)
      {
	hfputs(message(410),hstderr);
	hfflush(hstderr);
	Toplevel(ione);
      }
    else exit(2);
}

PUBLIC NORETURN void yy_error(l1,c1,l2,c2,msg)
IN int l1,c1,l2,c2;
IN CONST char *msg;
{
  TZone zone;

  zone.line1 = l1;
  zone.pos1 = c1;
  zone.line2 = l2;
  zone.pos2 = c2;
  type_error(&zone,(CHAR *)msg);
}

PUBLIC NORETURN void crash_error(zone,msg)
IN CONST TZone *zone;
IN CONST CHAR *msg;
{
  type_warn1(zone,msg,"Fatal error: ");
#if CORE_TRACE
  if (coreTraceFile!=NULL)
    hfflush(coreTraceFile);
#endif
  CRASH();
}

#if OS==MAC
PRIVATE char **messages=(char **)0L;

PUBLIC char *message(n)
int n;
{
  if (messages==(char **)0L)
    {
      short i,j,m;
      long len;
      Handle h;
      char *s,**msg,c;

      h = GetResource('STR#',128);
      DetachResource(h);
      HNoPurge(h);
      MoveHHi(h);
      HLock(h);
      m = *(short *)*h;
      messages = (char **)NewPtr(((long)m)*sizeof(char *));
      if (messages==(char **)0L)
	berror("cannot load message table");
      for (msg = messages, i = 1, s = (*h)+2; i<=m; i++, msg++)
	{
	  len = (long)*s;
	  *msg = s;
	  for (j=0; j<len; j++)
	    {
	      c = s[1];
	      if (c=='\r')
		c = '\n';
	      *s++ = c;
	    }
	  *s++ = 0;
	}
    }
  return messages[n-1];
}
#else
PRIVATE char *messages[] = {
#include "messages"
};

PUBLIC char *message(n)
int n;
{
  return messages[n-1];
}
#endif

#if OS==AMIGA_OS
PUBLIC NORETURN void crash()
{
  hfputs("Crash!\n",hstderr);
  hfflush(hstderr);
  ml_done(20);
}
#endif

PUBLIC void longjmperror()
{
  /* do nothing;
     this is the entry point on some BSD systems of the procedure
     that is called when it detects that we use longjmp() on a
     higher stack than the corresponding setjmp().
     We should probably use this to backpatch longjmp() in a
     way or another. */
}
