/* Lexer initialization.
   Copyright (C) 2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/types.h>
#endif
#include "expr.h"
#include "exeproto.h"
#include "ex2proto.h"
#include "grmproto.h"
#include "hlexproto.h"
#include "typproto.h"
#include "cmpproto.h"
#include "type.h"
#ifdef HAS_PROTOTYPES
#include "shlexinit.h"
#endif

IMPORT void InitHlex P((void));

EXTERN extDesc yytblExtDesc, yybufExtDesc, yydataExtDesc;
EXTERN ExprPtr flexErrExc;

EXTERN SymPtr betaeq,gama,gamaeq,numalpha,numbeta;
EXTERN ExprPtr num_alpha,num_beta;

EXTERN SymPtr arraySym, optSym;
EXTERN SymPtr readsym, readlinesym;
EXTERN ExprPtr intType, boolType, stringType, exnType, iarrayType;

#define __SEG__ INIT

/* there is no garbage collection during initialization,
	so KEEP() is a noop */
#undef KEEP
#define KEEP(x) x
#undef KONS
#define KONS Kons

PRIVATE ExprPtr KSYMBOL(n)
IN COUNTER n;
{
  ExprPtr sym;

  sym = Ksymbol(message(n));
  return sym;
}

void InitHlex ()
{
  SymPtr flexErrSym;
  SymPtr yydatasym, yydataSym, yytblsym, yytblSym, yybufsym, yybufSym;
  ExprPtr yydataType, yytbl_alpha, yybufType;
  ExprPtr alphaarg, alphahdr;
  ExprPtr aux, aux2, type;
  ExprPtr instream;
  ENTER;

  registerExtDesc (&yytblExtDesc);
  registerExtDesc (&yybufExtDesc);
  registerExtDesc (&yydataExtDesc);
  flexErrSym = KSYMBOL(249);
  setConstructor (flexErrSym);
  flexErrExc = MakeException(flexErrSym,EXC_BENIGN);
  type = funType (intType, exnType);
  addIdType (flexErrSym, type);
  aux = MakeCompiled (E_COMPILED_CLOSURE,
		      ACTION_NAME (make_io_exc),
		      flexErrExc);
  addVal (flexErrSym, aux, flexErrExc);
  QUIT;
  alphaarg = KONS(alpha,NIL);
  alphahdr = KONS(alphaarg,MapSingleton(alpha,falseValue));
  yydatasym = KSYMBOL(383);
  yydataSym = KSYMBOL(384);
  yydataType = KtypeAppl (yydataSym, NIL);
  yytblsym = KSYMBOL(419);
  yytblSym = KSYMBOL(420);
  yytbl_alpha = KtypeAppl (yytblSym, alphaarg);
  yybufsym = KSYMBOL(681);
  yybufSym = KSYMBOL(682);
  yybufType = KtypeAppl (yybufSym, NIL);
  aux2 = KONS(NIL,NewMap());
  aux = MapPut (yybufsym, KONS (aux2, yybufType),
		MapPair (yydatasym, KONS (aux2, yydataType),
			 yytblsym, KONS (alphahdr, yytbl_alpha)));
  setMapMerge (&A_TYFCNS(basisAssum), aux);
  type = funType (coupleType (yydataType, intType), NIL);
  ADDFUN(250, type, hbegin);
  ADDFUN(679, type, hunput);
  ADDFUN(685, type, hpushstate);
  ADDFUN(689, type, hless);
  QUIT;
  type = funType (yydataType, intType);
  ADDFUN(378, type, hstart);
  ADDFUN(680, type, hinput);
  ADDFUN(687, type, htopstate);
  ADDFUN(690, type, hlen);
  QUIT;
  type = funType (coupleType (yydataType, boolType), NIL);
  ADDFUN(380, type, hsetinteractive);
  ADDFUN(411, type, hsetbol);
  QUIT;
  type = funType (yydataType, boolType);
  ADDFUN(385, type, hisinteractive);
  ADDFUN(413, type, hatbol);
  QUIT;
  type = funType (yydataType, stringType);
  ADDFUN(415, type, htext);
  QUIT;
  aux = listType (intType);
  type = funType (KONS (listType (funType (yydataType,
					   KtypeAppl (optSym, alphaarg))),
			KONS (intType,
			      KONS (intType,
				    KONS (aux,
					  KONS (aux,
						KONS (aux,
						      KONS (aux,
							    KONS (aux,
								  KONS (aux,
									KONS (aux,
									      KONS (aux,
										    KONS (intType,
											  coupleType (intType,
												      intType))))))))))))),
		  yytbl_alpha);
  ADDFUN(421, type, hbuildyytbl);
  QUIT;
  type = funType(intType,stringType);
  ADDFUN(703, type, herrormsg);
  instream = MapPut (NIL, alpha,
		     MapPair(readsym,type,
			     readlinesym,funType(NIL,stringType))); /* 'a instream */
  type = funType (coupleType (instream, funType (NIL, boolType)), yydataType);
  ADDFUN(422, type, hnewdata);
  QUIT;
  type = funType (yytbl_alpha, funType (yydataType, alpha));
  ADDFUN(442, type, hlex);
  QUIT;
  type = funType (coupleType (yydataType, yybufType), NIL);
  ADDFUN(683, type, hswitchbuf);
  ADDFUN(688, type, hflushbuf);
  QUIT;
  type = funType (coupleType (yydataType, instream), yybufType);
  ADDFUN(684, type, hnewbuf);
  QUIT;
  type = funType (yydataType, NIL);
  ADDFUN(686, type, hpopstate);
  QUIT;
  type = funType (yydataType, iarrayType);
  ADDFUN(691, type, hyylloc);
  QUIT;
  type = funType (yydataType, yybufType);
  ADDFUN(733, type, hbuf);
  QUIT;
}
