/* Parser initialization.
   Copyright (C) 2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/types.h>
#endif
#include "expr.h"
#include "exeproto.h"
#include "ex2proto.h"
#include "grmproto.h"
#include "typproto.h"
#include "cmpproto.h"
#include "hyaccproto.h"
#include "type.h"
#ifdef HAS_PROTOTYPES
#include "shyaccinit.h"
#endif

IMPORT void InitHyacc P((void));

EXTERN extDesc hytblExtDesc, hydataExtDesc;
EXTERN ExprPtr hyaccErrExc, parseErrExc;

EXTERN SymPtr betaeq,gama,gamaeq,numalpha,numbeta;
EXTERN ExprPtr num_alpha,num_beta;

EXTERN SymPtr arraySym, optSym;
EXTERN SymPtr readsym, readlinesym;
EXTERN ExprPtr intType, boolType, stringType, exnType, iarrayType;

#define __SEG__ INIT

/* there is no garbage collection during initialization,
	so KEEP() is a noop */
#undef KEEP
#define KEEP(x) x
#undef KONS
#define KONS Kons

PRIVATE ExprPtr KSYMBOL(n)
IN COUNTER n;
{
  ExprPtr sym;

  sym = Ksymbol(message(n));
  return sym;
}

void InitHyacc ()
{
  SymPtr hyaccErrSym, parseErrSym;
  ExprPtr type, aux;
  /*ExprPtr alphaarg, alphahdr;*/
  ExprPtr betaarg, betahdr;
  ExprPtr abarg, abhdr;
  SymPtr hydatasym, hydataSym, hytblsym, hytblSym;
  ExprPtr hydataType, hytblType;
  ExprPtr valoptType;
  ENTER;

  registerExtDesc (&hytblExtDesc);
  registerExtDesc (&hydataExtDesc);
  hyaccErrSym = KSYMBOL(704);
  setConstructor (hyaccErrSym);
  hyaccErrExc = MakeException (hyaccErrSym,EXC_BENIGN);
  type = funType (intType, exnType);
  addIdType (hyaccErrSym, type);
  aux = MakeCompiled (E_COMPILED_CLOSURE,
		      ACTION_NAME (make_io_exc),
		      hyaccErrExc);
  addVal (hyaccErrSym, aux, hyaccErrExc);
  parseErrSym = KSYMBOL(705);
  setConstructor (parseErrSym);
  parseErrExc = MakeException (parseErrSym,EXC_BENIGN);
  type = funType (listType (stringType), exnType);
  addIdType (parseErrSym, type);
  aux = MakeCompiled (E_COMPILED_CLOSURE,
		      ACTION_NAME (make_io_exc),
		      parseErrExc);
  addVal (parseErrSym, aux, parseErrExc);
  QUIT;
  {
    MapPtr am, bm;

    /*alphaarg = KONS(alpha,NIL);*/
    am = MapSingleton(alpha,falseValue);
    /*alphahdr = KONS(alphaarg, am);*/
    betaarg = KONS(beta,NIL);
    betahdr = KONS(betaarg, bm = MapSingleton(beta,falseValue));
    abarg = KONS(alpha,betaarg);
    abhdr = KONS(abarg, MapMerge (am, bm));
  }
  hydatasym = KSYMBOL(706);
  hydataSym = KSYMBOL(707);
  hydataType = KtypeAppl (hydataSym, abarg);
  hytblsym = KSYMBOL(708);
  hytblSym = KSYMBOL(709);
  hytblType = KtypeAppl (hytblSym, betaarg);
  aux = MapPair (hydatasym, KONS (abhdr, hydataType),
		 hytblsym, KONS (betahdr, hytblType));
  setMapMerge (&A_TYFCNS(basisAssum), aux);
  type = funType (hydataType, intType);
  ADDFUN(710, type, hmaxdepth);
  ADDFUN(732, type, hychar);
  QUIT;
  aux = coupleType (hydataType, intType);
  type = funType (aux, NIL);
  ADDFUN(711, type, hsetmaxdepth);
  QUIT;
  type = funType (hydataType, funType (intType, beta));
  ADDFUN(712, type, hyval);
  QUIT;
  type = funType (hydataType, beta);
  ADDFUN(730, type, hydefval);
  QUIT;
  type = funType (hydataType, funType (intType, iarrayType));
  ADDFUN(713, type, hyloc);
  QUIT;
  type = funType (hydataType, NIL);
  ADDFUN(714, type, hyerrok);
  ADDFUN(715, type, hyclearin);
  QUIT;
  type = funType (hydataType, boolType);
  ADDFUN(716, type, hyrecovering);
  QUIT;
  type = funType (coupleType (hydataType, boolType), NIL);
  ADDFUN(731, type, hysetdebug);
  QUIT;
  valoptType = KtypeAppl (optSym, betaarg);
  type = funType (hytblType, funType (KtypeAppl (hydataSym, Kons (intType, betaarg)), valoptType));
  ADDFUN(717, type, hyparse);
  {
    ExprPtr aux2;
    COUNTER i;

    aux = NIL;
    aux2 = listType (intType);
    for (i=8; i<=19; i++)
      aux = Kons (aux2, aux);
    for (i=2; i<=7; i++)
      aux = Kons (intType, aux);
    aux = Kons (KtypeAppl (arraySym, Kons (stringType, NIL)), aux);
    aux = Kons (listType (funType (hydataType, valoptType)), aux);
  }
  type = funType (aux, hytblType);
  ADDFUN(718, type, hbuildhytbl);
  QUIT;
  type = funType (Kons (gama,
			Kons (funType (gama, alpha),
			      Kons (beta,
				    Kons (KtypeAppl (refSym, betaarg),
					  coupleType (iarrayType,
						      funType (listType (stringType), NIL) ))))),
		  hydataType);
  ADDFUN(719, type, hynewdata);
  QUIT;
  type = funType (intType, stringType);
  ADDFUN(720, type, hyerrmsg);
  QUIT;
}
