/* Set and map functions.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "eval.h"
#include "expr.h"
#include "errproto.h"

struct compose {
  MapPtr map1;
};

#ifdef HAS_PROTOTYPES
#include "smap.h"
#endif

#define __SEG__ EXPR

#undef DEBUG

PUBLIC MapPtr empty_map;

PUBLIC MapPtr MapSingleton(x,y)
IN CONST ExprPtr x,y;
{
  MapPtr map;

  map = MapletKons(x,y);
  return KEEP(map);
}

PUBLIC MapPtr MapPair(x1,y1,x2,y2)
IN CONST ExprPtr x1,y1,x2,y2;
{
  MapPtr map;

  map = MapSingleton(x2,y2);
  if (x1!=x2)
    map = MapPut(x1,y1,map);
  return map;
}

PUBLIC ExprPtr MapGet1(elem,map,mask,found)
IN REGISTER CONST ExprPtr elem;
IN REGISTER CONST MapPtr map;
IN REGISTER CODE mask;
OUT Boolean *found;
{
  while (map->kind==(CHAR)E_MAP)
    {
      map = (((CODE)elem) & mask)?MAPRIGHT(map):MAPLEFT(map);
      mask <<= 1;
      if (map->kind!=(CHAR)E_MAP) /* unroll loop. */
	break;
      map = (((CODE)elem) & mask)?MAPRIGHT(map):MAPLEFT(map);
      mask <<= 1;
      if (map->kind!=(CHAR)E_MAP) /* unroll loop. */
	break;
      map = (((CODE)elem) & mask)?MAPRIGHT(map):MAPLEFT(map);
      mask <<= 1;
      if (map->kind!=(CHAR)E_MAP) /* unroll loop. */
	break;
      map = (((CODE)elem) & mask)?MAPRIGHT(map):MAPLEFT(map);
      mask <<= 1;
      if (map->kind!=(CHAR)E_MAP) /* unroll loop. */
	break;
      map = (((CODE)elem) & mask)?MAPRIGHT(map):MAPLEFT(map);
      mask <<= 1;
      if (map->kind!=(CHAR)E_MAP) /* unroll loop. */
	break;
      map = (((CODE)elem) & mask)?MAPRIGHT(map):MAPLEFT(map);
      mask <<= 1;
      if (map->kind!=(CHAR)E_MAP) /* unroll loop. */
	break;
      map = (((CODE)elem) & mask)?MAPRIGHT(map):MAPLEFT(map);
      mask <<= 1;
      if (map->kind!=(CHAR)E_MAP) /* unroll loop. */
	break;
      map = (((CODE)elem) & mask)?MAPRIGHT(map):MAPLEFT(map);
      mask <<= 1;
    }
  return (*found = (map!=empty_map && elem==MAPLETX(map)))?MAPLETY(map):NIL;
}

PUBLIC ExprPtr MapGet(elem,map,found)
IN CONST ExprPtr elem;
IN REGISTER CONST MapPtr map;
OUT Boolean *found;
{
#ifdef DEBUG
  if (!ISMAP(map))
	CRASH();
#endif
  if (map==empty_map)
    {
      *found = false;
      return NIL;
    }
  return MapGet1(elem,map,MASK0,found);
}

PUBLIC ExprPtr MapGetSure(elem,map)
IN CONST ExprPtr elem;
IN CONST ExprPtr map;
{
  Boolean found;

  elem = MapGet(elem,map,&found);
  if (!found)
    berror(message(464));
  return elem;
}

PUBLIC ExprPtr MapGetMaplet1(elem,map,mask,found)
IN REGISTER CONST ExprPtr elem;
IN REGISTER CONST MapPtr map;
IN REGISTER CODE mask;
OUT Boolean *found;
{
  while (/*map!=empty_map && map->kind!=(CHAR)E_MAPLET*/
	   map->kind==(CHAR)E_MAP)
    {
      if (((CODE)elem) & mask)
	map = MAPRIGHT(map);
      else
	map = MAPLEFT(map);
      mask <<= 1;
      if (map->kind!=(CHAR)E_MAP) /* unroll loop. */
	break;
      if (((CODE)elem) & mask)
	map = MAPRIGHT(map);
      else
	map = MAPLEFT(map);
      mask <<= 1;
      if (map->kind!=(CHAR)E_MAP) /* unroll loop. */
	break;
      if (((CODE)elem) & mask)
	map = MAPRIGHT(map);
      else
	map = MAPLEFT(map);
      mask <<= 1;
      if (map->kind!=(CHAR)E_MAP) /* unroll loop. */
	break;
      if (((CODE)elem) & mask)
	map = MAPRIGHT(map);
      else
	map = MAPLEFT(map);
      mask <<= 1;
      if (map->kind!=(CHAR)E_MAP) /* unroll loop. */
	break;
      if (((CODE)elem) & mask)
	map = MAPRIGHT(map);
      else
	map = MAPLEFT(map);
      mask <<= 1;
      if (map->kind!=(CHAR)E_MAP) /* unroll loop. */
	break;
      if (((CODE)elem) & mask)
	map = MAPRIGHT(map);
      else
	map = MAPLEFT(map);
      mask <<= 1;
      if (map->kind!=(CHAR)E_MAP) /* unroll loop. */
	break;
      if (((CODE)elem) & mask)
	map = MAPRIGHT(map);
      else
	map = MAPLEFT(map);
      mask <<= 1;
      if (map->kind!=(CHAR)E_MAP) /* unroll loop. */
	break;
      if (((CODE)elem) & mask)
	map = MAPRIGHT(map);
      else
	map = MAPLEFT(map);
      mask <<= 1;
    }
  if (map==empty_map)
    {
      *found = false;
      return empty_map;
    }
  return (*found = (elem==MAPLETX(map)))?map:empty_map;
}

PUBLIC MapPtr MapPutTrie(x,y,map0,mask)
IN REGISTER CONST ExprPtr x;
IN CONST ExprPtr y;
IN CONST MapPtr map0;
IN REGISTER CODE mask;
{
  REGISTER MapPtr *oldmapp;
  REGISTER MapPtr map;
  MapPtr oldmaps[MAP_CHAIN_LENGTH];

  map = map0;
  for (oldmapp=oldmaps;; mask<<=1)
    {
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
      mask <<= 1; /* unroll loop. */
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
      mask <<= 1; /* unroll loop. */
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
      mask <<= 1; /* unroll loop. */
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
    }
  if (map->kind==(CHAR)E_MAPLET)
    {
      REGISTER ExprPtr xx;

      xx = MAPLETX(map);
      if (xx==x && MAPLETY(map)==y)
	return map0;
      if (xx!=x)
	{
	  for (;; mask<<=1)
	    {
	      if ((((CODE)x) & mask)==(((CODE)xx) & mask))
		*oldmapp++ = empty_map;
	      else
		{
		  *oldmapp++ = map;
		  mask <<= 1;
		  break;
		}
	      mask <<= 1; /* unroll loop. */
	      if ((((CODE)x) & mask)==(((CODE)xx) & mask))
		*oldmapp++ = empty_map;
	      else
		{
		  *oldmapp++ = map;
		  mask <<= 1;
		  break;
		}
	      mask <<= 1; /* unroll loop. */
	      if ((((CODE)x) & mask)==(((CODE)xx) & mask))
		*oldmapp++ = empty_map;
	      else
		{
		  *oldmapp++ = map;
		  mask <<= 1;
		  break;
		}
	      mask <<= 1; /* unroll loop. */
	      if ((((CODE)x) & mask)==(((CODE)xx) & mask))
		*oldmapp++ = empty_map;
	      else
		{
		  *oldmapp++ = map;
		  mask <<= 1;
		  break;
		}
	    }
	}
    }
  BEGINQUICKKEEP();
  map = MapletKons(x,y);
  QUICKKEEP(map);
  for (; oldmapp>oldmaps;)
    {
      mask >>= 1;
      y = *--oldmapp;
      map = (((CODE)x) & mask)?MapKons(y,map):MapKons(map,y);
      QUICKKEEP(map);
    }
  ENDQUICKKEEP();
  return map;
}

PUBLIC MapPtr MapPut(x,y,map)
IN CONST ExprPtr x,y;
IN CONST MapPtr map;
{
  ENTER;

#ifdef DEBUG
  if (!ISMAP(map))
	CRASH();
#endif
  if (map==empty_map)
    map = MapSingleton(x,y);
  else
    map = MapPutTrie(x,y,map,MASK0);
  KEEPONLY(map);
  return map;
}

PUBLIC MapPtr MapPutMapletTrie(maplet,map0,mask)
IN CONST ExprPtr maplet;
IN CONST MapPtr map0;
IN REGISTER CODE mask;
/* does not keep its result! */
{
  REGISTER ExprPtr x;
  REGISTER MapPtr *oldmapp;
  REGISTER MapPtr map,y;
  MapPtr oldmaps[MAP_CHAIN_LENGTH];

  x = MAPLETX(maplet);
  map = map0;
  for (oldmapp=oldmaps;; mask<<=1)
    {
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
      mask <<= 1; /* unroll loop. */
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
      mask <<= 1; /* unroll loop. */
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
      mask <<= 1; /* unroll loop. */
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
    }
  if (map==maplet)
    map = map0;
  else
    {
      if (map->kind==(CHAR)E_MAPLET)
	{
	  REGISTER ExprPtr xx;

	  xx = MAPLETX(map);
	  if (xx!=x)
	    {
	      for (;; mask<<=1)
		{
		  if ((((CODE)x) & mask)==(((CODE)xx) & mask))
		    *oldmapp++ = empty_map;
		  else
		    {
		      *oldmapp++ = map;
		      mask <<= 1;
		      break;
		    }
		  mask <<= 1; /* unroll loop. */
		  if ((((CODE)x) & mask)==(((CODE)xx) & mask))
		    *oldmapp++ = empty_map;
		  else
		    {
		      *oldmapp++ = map;
		      mask <<= 1;
		      break;
		    }
		  mask <<= 1; /* unroll loop. */
		  if ((((CODE)x) & mask)==(((CODE)xx) & mask))
		    *oldmapp++ = empty_map;
		  else
		    {
		      *oldmapp++ = map;
		      mask <<= 1;
		      break;
		    }
		  mask <<= 1; /* unroll loop. */
		  if ((((CODE)x) & mask)==(((CODE)xx) & mask))
		    *oldmapp++ = empty_map;
		  else
		    {
		      *oldmapp++ = map;
		      mask <<= 1;
		      break;
		    }
		}
	    }
	}
      BEGINQUICKKEEP();
      map = maplet;
      for (; oldmapp>oldmaps;)
	{
	  mask >>= 1;
	  y = *--oldmapp;
	  if (((CODE)x) & mask)
	    map = MapKons(y,map);
	  else
	    map = MapKons(map,y);
	  QUICKKEEP(map);
	}
      ENDQUICKKEEP();
    }
  return map;
}

PUBLIC MapPtr MapPutBehindTrie(x,y,map0,mask)
IN REGISTER CONST ExprPtr x;
IN CONST ExprPtr y;
IN CONST MapPtr map0;
IN REGISTER CODE mask;
{
  REGISTER MapPtr *oldmapp;
  REGISTER MapPtr map;
  MapPtr oldmaps[MAP_CHAIN_LENGTH];

  map = map0;
  for (oldmapp=oldmaps;; mask<<=1)
    {
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
      mask <<= 1; /* unroll loop. */
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
      mask <<= 1; /* unroll loop. */
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
      mask <<= 1; /* unroll loop. */
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
    }
  if (/*map!=empty_map && */ map->kind==(CHAR)E_MAPLET)
    {
      REGISTER ExprPtr xx;

      xx = MAPLETX(map);
      if (xx==x)
	return map0;
      for (;; mask<<=1)
	{
	  if ((((CODE)x) & mask)==(((CODE)xx) & mask))
	    *oldmapp++ = empty_map;
	  else
	    {
	      *oldmapp++ = map;
	      mask <<= 1;
	      break;
	    }
	  mask <<= 1; /* unroll loop. */
	  if ((((CODE)x) & mask)==(((CODE)xx) & mask))
	    *oldmapp++ = empty_map;
	  else
	    {
	      *oldmapp++ = map;
	      mask <<= 1;
	      break;
	    }
	  mask <<= 1; /* unroll loop. */
	  if ((((CODE)x) & mask)==(((CODE)xx) & mask))
	    *oldmapp++ = empty_map;
	  else
	    {
	      *oldmapp++ = map;
	      mask <<= 1;
	      break;
	    }
	  mask <<= 1; /* unroll loop. */
	  if ((((CODE)x) & mask)==(((CODE)xx) & mask))
	    *oldmapp++ = empty_map;
	  else
	    {
	      *oldmapp++ = map;
	      mask <<= 1;
	      break;
	    }
	}
    }
  BEGINQUICKKEEP();
  map = MapletKons(x,y);
  QUICKKEEP(map);
  for (; oldmapp>oldmaps;)
    {
      mask >>= 1;
      y = *--oldmapp;
      if (((CODE)x) & mask)
	map = MapKons(y,map);
      else
	map = MapKons(map,y);
      QUICKKEEP(map);
    }
  ENDQUICKKEEP();
  return map;
}

PUBLIC MapPtr MapPutBehind(x,y,map)
IN CONST ExprPtr x,y;
IN CONST MapPtr map;
{
  ENTER;

#ifdef DEBUG
  if (!ISMAP(map))
	CRASH();
#endif
  if (map==empty_map)
    map = MapSingleton(x,y);
  else
    map = MapPutBehindTrie(x,y,map,MASK0);
  KEEPONLY(map);
  return map;
}

PUBLIC MapPtr MapPutMapletBehindTrie(maplet,map0,mask)
IN CONST ExprPtr maplet;
IN CONST MapPtr map0;
IN REGISTER CODE mask;
/* does not keep its result! */
{
  REGISTER ExprPtr x;
  REGISTER MapPtr *oldmapp;
  REGISTER MapPtr map,y;
  MapPtr oldmaps[MAP_CHAIN_LENGTH];

  x = MAPLETX(maplet);
  map = map0;
  for (oldmapp=oldmaps;; mask<<=1)
    {
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
      mask <<= 1; /* unroll loop. */
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
      mask <<= 1; /* unroll loop. */
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
      mask <<= 1; /* unroll loop. */
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
    }
  if (/*map!=empty_map && */ map->kind==(CHAR)E_MAPLET)
    {
      REGISTER ExprPtr xx;

      xx = MAPLETX(map);
      if (xx==x)
	return map0;
      for (;; mask<<=1)
	{
	  if ((((CODE)x) & mask)==(((CODE)xx) & mask))
	    *oldmapp++ = empty_map;
	  else
	    {
	      *oldmapp++ = map;
	      mask <<= 1;
	      break;
	    }
	  mask <<= 1; /* unroll loop. */
	  if ((((CODE)x) & mask)==(((CODE)xx) & mask))
	    *oldmapp++ = empty_map;
	  else
	    {
	      *oldmapp++ = map;
	      mask <<= 1;
	      break;
	    }
	  mask <<= 1; /* unroll loop. */
	  if ((((CODE)x) & mask)==(((CODE)xx) & mask))
	    *oldmapp++ = empty_map;
	  else
	    {
	      *oldmapp++ = map;
	      mask <<= 1;
	      break;
	    }
	  mask <<= 1; /* unroll loop. */
	  if ((((CODE)x) & mask)==(((CODE)xx) & mask))
	    *oldmapp++ = empty_map;
	  else
	    {
	      *oldmapp++ = map;
	      mask <<= 1;
	      break;
	    }
	}
    }
  BEGINQUICKKEEP();
  map = maplet;
  for (; oldmapp>oldmaps;)
    {
      mask >>= 1;
      y = *--oldmapp;
      if (((CODE)x) & mask)
	map = MapKons(y,map);
      else
	map = MapKons(map,y);
      QUICKKEEP(map);
    }
  ENDQUICKKEEP();
  return map;
}

PUBLIC MapPtr MapRemoveTrie(x,map0,mask)
IN REGISTER CONST ExprPtr x;
IN CONST MapPtr map0;
IN REGISTER CODE mask;
{
  REGISTER MapPtr *oldmapp;
  REGISTER MapPtr map,y;
  MapPtr oldmaps[MAP_CHAIN_LENGTH];

  map = map0;
  for (oldmapp=oldmaps;; mask<<=1)
    {
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
      mask <<= 1; /* unroll loop. */
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
      mask <<= 1; /* unroll loop. */
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
      mask <<= 1; /* unroll loop. */
      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
	   map->kind!=(CHAR)E_MAP)
	break;
      if (((CODE)x) & mask)
	{
	  *oldmapp++ = MAPLEFT(map);
	  map = MAPRIGHT(map);
	}
      else
	{
	  *oldmapp++ = MAPRIGHT(map);
	  map = MAPLEFT(map);
	}
    }
  if (map==empty_map || MAPLETX(map)!=x)
    map = map0;
  else
    {
      BEGINQUICKKEEP();
      map = empty_map;
      for (; oldmapp>oldmaps;)
	{
	  mask >>= 1;
	  y = *--oldmapp;
	  if (y==empty_map)
	    {
	      if (/*map==empty_map || map->kind==(CHAR)E_MAPLET*/
		   map->kind!=(CHAR)E_MAP)
		continue;
	    }
	  else if (map==empty_map)
	    {
	      if (/*y->kind==(CHAR)E_MAPLET*/
		   y->kind!=(CHAR)E_MAP)
		{
		  map = y;
		  continue;
		}
	    }
	  if (((CODE)x) & mask)
	    map = MapKons(y,map);
	  else
	    map = MapKons(map,y);
	  QUICKKEEP(map);
	}
      ENDQUICKKEEP();
    }
  return map;
}

PUBLIC MapPtr MapRemove(x,map)
IN CONST ExprPtr x;
IN CONST MapPtr map;
{
#ifdef DEBUG
  if (!ISMAP(map))
	CRASH();
#endif
  if (map!=empty_map)
    {
      ENTER;

      map = MapRemoveTrie(x,map,MASK0);
      KEEPONLY(map);
    }
  return map;
}

PUBLIC Boolean MapSweep(map,func,data)
IN REGISTER CONST MapPtr map;
IN mapSweepFunc func;
INOUT VOIDPTR data;
{
  ExprPtr stack[MAP_CHAIN_LENGTH];
  REGISTER ExprPtr *sp;
  REGISTER ExprPtr aux;

  sp = stack;
again:
  switch (map->kind)
    {
    case E_EMPTY:
ret:
      if (sp==stack)
	return false;
      map = *--sp;
      goto again;
    case E_MAPLET:
      if ((*func)(MAPLETX(map),MAPLETY(map),(CHAR *)data))
	return true;
      goto ret;
    case E_MAP:
      aux = MAPRIGHT(map);
      if (aux==empty_map)
	{
	  map = MAPLEFT(map);
	  goto again;
	}
      map = MAPLEFT(map);
      if (map==empty_map)
	{
	  map = aux;
	  goto again;
	}
      *sp++ = aux;
      goto again;
    default:
      kinderror(message(113),map->kind);
      return false;
    }
#ifdef THE_RECURSIVE_DEFINITION
  REGISTER ExprPtr aux;

again:
  switch (map->kind)
    {
    case E_EMPTY:
      return false;
    case E_MAPLET:
      return (*func)(MAPLETX(map),MAPLETY(map),(CHAR *)data);
    case E_MAP:
      aux = MAPLEFT(map);
      if (aux==empty_map)
	{
	  map = MAPRIGHT(map);
	  goto again;
	}
      map = MAPRIGHT(map);
      if (map==empty_map)
	{
	  map = aux;
	  goto again;
	}
      if (MapSweep(aux,func,data))
	return true;
      goto again;
    default:
      kinderror(message(113),map->kind);
      return false;
    }
#endif
}

PUBLIC ExprPtr MapSome(map,func,data) /* func is a predicate */
IN REGISTER CONST MapPtr map; /* MapSome returns the first maplet that satisfies func, or NIL */
IN mapSweepFunc func;
INOUT VOIDPTR data;
{
  REGISTER ExprPtr aux;

again:
  switch (map->kind)
    {
    case E_MAPLET:
      if ((*func)(MAPLETX(map),MAPLETY(map),(CHAR *)data))
	return map;
      /*FALLTHROUGH*/
    case E_EMPTY:
      return NIL;
    case E_MAP:
      aux = MAPLEFT(map);
      if (aux==empty_map)
	{
	  map = MAPRIGHT(map);
	  goto again;
	}
      map = MAPRIGHT(map);
      if (map==empty_map)
	{
	  map = aux;
	  goto again;
	}
      if ((aux = MapSome(aux,func,data))!=NIL)
	return aux;
      goto again;
    default:
      kinderror(message(649),map->kind);
      return NIL;
    }
}

PUBLIC ExprPtr mapStack[2*MAP_CHAIN_LENGTH];
PUBLIC ExprPtr *mapStackTop=mapStack;

struct mapcollect {
  ExprPtr map;
  Boolean side;
};

PUBLIC MapPtr MapCollect0(map,func,data)
IN REGISTER CONST MapPtr map;
IN mapCollectFunc func;
INOUT VOIDPTR data;
{
  struct mapcollect stack[MAP_CHAIN_LENGTH];
  REGISTER struct mapcollect *sp;
  REGISTER ExprPtr aux;
  struct _x { COUNTER n; gcStackView *next; VOLATILE ExprPtr p[2*MAP_CHAIN_LENGTH]; } _x;
  REGISTER VOLATILE ExprPtr *pp;
  ENTER;

  if (map==empty_map)
    return empty_map;
  _x.n = 0;
  _x.next = theStackFrames;
  theStackFrames = (gcStackView *)&_x;
  sp = stack;
  pp = _x.p;
again:
  if (map->kind==(CHAR)E_MAPLET)
    {
    /*case E_MAPLET:*/
      map = (*func)(MAPLETX(map),MAPLETY(map),(CHAR *)data);
      QUIT;
ret:
      if (sp==stack)
	{
	  theStackFrames = _x.next;
	  return map;
	}
      *pp++ = map;
      _x.n++;
      if (sp[-1].side)
	{
	  sp[-1].side = false;
	  map = sp[-1].map;
	  goto again;
	}
      sp--;
      aux = pp[-2];
      if (aux!=empty_map)
	map = (map==empty_map)?aux:(mapStackTop = mapStack, MapMerge1(aux,map,MASK0));
      pp -= 2;
      _x.n -= 2;
      goto ret;
    }
  else
    {
    /*case E_MAP:*/
      aux = MAPRIGHT(map);
      if (aux==empty_map)
	{
	  map = MAPLEFT(map);
	  goto again;
	}
      map = MAPLEFT(map);
      if (map==empty_map)
	{
	  map = aux;
	  goto again;
	}
      sp->map = aux;
      sp->side = true;
      sp++;
      goto again;
    }
    /*default:
      kinderror(message(114),map->kind);
      break;*/
#ifdef THE_RECURSIVE_DEFINITION
  REGISTER ExprPtr aux;

again:
  switch (map->kind)
    {
    case E_EMPTY:
      return NewMap();
    case E_MAPLET:
      return (*func)(MAPLETX(map),MAPLETY(map),(CHAR *)data);
    case E_MAP:
      aux = MAPLEFT(map);
      if (aux==empty_map)
	{
	  map = MAPRIGHT(map);
	  goto again;
	}
      map = MAPRIGHT(map);
      if (map==empty_map)
	{
	  map = aux;
	  goto again;
	}
      return MapMerge(MapCollect(aux,func,data),
		      MapCollect(map,func,data));
    default:
      kinderror(message(114),map->kind);
      break;
    }
#endif
}

PUBLIC MapPtr MapCollect(map,func,data)
IN REGISTER CONST MapPtr map;
IN mapCollectFunc func;
INOUT VOIDPTR data;
{
  map = MapCollect0(map,func,data);
  return KEEP(map);
}

PUBLIC ExprPtr MapFold(map,func,data,neutral,op)
IN REGISTER CONST MapPtr map;
IN mapFoldFunc func;
INOUT VOIDPTR data;
IN CONST ExprPtr neutral;
IN mapFoldOp op;
{
  struct mapcollect stack[MAP_CHAIN_LENGTH];
  REGISTER struct mapcollect *sp;
  REGISTER ExprPtr aux;
  struct _x { COUNTER n; gcStackView *next; VOLATILE ExprPtr p[2*MAP_CHAIN_LENGTH]; } _x;
  REGISTER VOLATILE ExprPtr *pp;
  ENTER;

  if (map==empty_map)
    return neutral;
  _x.n = 0;
  _x.next = theStackFrames;
  theStackFrames = (gcStackView *)&_x;
  sp = stack;
  pp = _x.p;
again:
  if (map->kind==(CHAR)E_MAPLET)
    {
    /*case E_MAPLET:*/
      map = (*func)(MAPLETX(map),MAPLETY(map),(CHAR *)data);
ret:
      if (sp==stack)
	{
	  theStackFrames = _x.next;
	  return map;
	}
      QUIT;
      *pp++ = map;
      _x.n++;
      if (sp[-1].side)
	{
	  sp[-1].side = false;
	  map = sp[-1].map;
	  goto again;
	}
      sp--;
      aux = pp[-2];
      map = (*op)(aux,map);
      pp -= 2;
      _x.n -= 2;
      goto ret;
    }
  else
    {
    /*case E_MAP:*/
      aux = MAPRIGHT(map);
      if (aux==empty_map)
	{
	  map = MAPLEFT(map);
	  goto again;
	}
      map = MAPLEFT(map);
      if (map==empty_map)
	{
	  map = aux;
	  goto again;
	}
      sp->map = aux;
      sp->side = true;
      sp++;
      goto again;
    }
    /*default:
      kinderror(message(115),map->kind);
      break;*/
#ifdef THE_RECURSIVE_DEFINITION
  REGISTER ExprPtr aux;

again:
  switch (map->kind)
    {
    case E_EMPTY:
      return neutral;
    case E_MAPLET:
      return (*func)(MAPLETX(map),MAPLETY(map),(CHAR *)data);
    case E_MAP:
      aux = MAPLEFT(map);
      if (aux==empty_map)
	{
	  map = MAPRIGHT(map);
	  goto again;
	}
      map = MAPRIGHT(map);
      if (map==empty_map)
	{
	  map = aux;
	  goto again;
	}
      return (*op)(MapFold(aux,func,data,neutral,op),
		   MapFold(map,func,data,neutral,op));
    default:
      kinderror(message(115),map->kind);
      break;
    }
#endif
}

PUBLIC INT MapCard(map)
IN REGISTER CONST MapPtr map;
{
  ExprPtr stack[MAP_CHAIN_LENGTH];
  REGISTER ExprPtr *sp;
  REGISTER INT n;

  sp = stack;
  n = 0;
again:
  switch (map->kind)
    {
    case E_MAPLET:
      n++;
      /*fallthrough*/
    case E_EMPTY:
      if (sp==stack)
	return n;
      map = *--sp;
      goto again;
    case E_MAP:
      *sp++ = MAPRIGHT(map);
      map = MAPLEFT(map);
      goto again;
    default:
      kinderror(message(116),map->kind);
      return 0;
    }
}

PUBLIC INT MapCardDec(map,n)
IN REGISTER CONST MapPtr map;
IN REGISTER INT n;
{
  ExprPtr stack[MAP_CHAIN_LENGTH];
  REGISTER ExprPtr *sp;

  sp = stack;
again:
  switch (map->kind)
    {
    case E_MAPLET:
      n--;
      /*fallthrough*/
    case E_EMPTY:
      if (sp==stack || n<0)
	return n;
      map = *--sp;
      goto again;
    case E_MAP:
      *sp++ = MAPRIGHT(map);
      map = MAPLEFT(map);
      goto again;
    default:
      kinderror(message(116),map->kind);
      return 0;
    }
}

struct mapmerge {
  ExprPtr map1,map2;
  CODE mask;
  Boolean side;
};

PUBLIC MapPtr MapMerge1(map1,map2,mask)
IN CONST REGISTER MapPtr map1,map2;
IN CONST REGISTER CODE mask;
{
  struct mapmerge stack[MAP_CHAIN_LENGTH];
  REGISTER struct mapmerge *sp;
  /* we simulate the stack by hand to be faster */

  sp = stack;
again:
  if (map1==map2)
    goto ret;
  switch (map1->kind)
    {
    case E_EMPTY:
      map1 = map2;
      goto ret;
    case E_MAPLET:
      map1 = MapPutMapletBehindTrie(map1,map2,mask);
      goto ret;
    case E_MAP:
      switch (map2->kind)
	{
	case E_MAPLET:
	  map1 = MapPutMapletTrie(map2,map1,mask);
	  /*fallthrough*/
	case E_EMPTY:
ret:
	  if (sp==stack)
	    return map1;
	  *mapStackTop++ = map1;
	  if (sp[-1].side)
	    {
	      sp[-1].side = false;
	      map1 = sp[-1].map1;
	      map2 = sp[-1].map2;
	      mask = sp[-1].mask;
	      goto again;
	    }
	  sp--;
	  map2 = mapStackTop[-2];
	  map1 = MapKons(map2,map1);
	  mapStackTop -= 2;
	  goto ret;
	case E_MAP:
	  mask <<= 1;
	  sp->map1 = MAPRIGHT(map1);
	  map1 = MAPLEFT(map1);
	  sp->map2 = MAPRIGHT(map2);
	  map2 = MAPLEFT(map2);
	  sp->mask = mask;
	  sp->side = true;
	  sp++;
	  goto again;
	default:
	  map1 = map2;
	  /* fallthrough */
	}
    default:
      mapStackTop = mapStack;
      kinderror(message(117),map1->kind);
      return NIL;
    }
#ifdef THE_RECURSIVE_VERSION
  REGISTER MapPtr left,right;

  if (map1==map2)
    return map1;
  switch (map1->kind)
    {
    case E_EMPTY:
      return map2;
    case E_MAPLET:
      return MapPutMapletBehindTrie(map1,map2,mask);
    case E_MAP:
      switch (map2->kind)
	{
	case E_EMPTY:
	  return map1;
	case E_MAPLET:
	  return MapPutMapletTrie(map2,map1,mask);
	case E_MAP:
	  mask <<= 1;
	  left = MapMerge1(MAPLEFT(map1),MAPLEFT(map2),mask);
	  *mapStackTop++ = left;
	  right = MapMerge1(MAPRIGHT(map1),MAPRIGHT(map2),mask);
	  *mapStackTop++ = right;
	  right = MapKons(left,right);
	  mapStackTop -= 2;
	  return right;
	default:
	  map1 = map2;
	  /* fallthrough */
	}
    default:
      mapStackTop = mapStack;
      kinderror(message(117),map1->kind);
      return NIL;
    }
#endif
}

PUBLIC ExprPtr MapMerge(map1,map2)
IN CONST MapPtr map1,map2;
{
  REGISTER MapPtr map;

  if (map1==empty_map)
    return map2;
  if (map2==empty_map)
    return map1;
  mapStackTop = mapStack;
  map = MapMerge1(map1,map2,MASK0);
  return KEEP(map);
}

PUBLIC MapPtr MapDomRestrTo1(map1,map2,mask)
IN CONST REGISTER MapPtr map1,map2;
IN CONST REGISTER CODE mask;
{
  struct mapmerge stack[MAP_CHAIN_LENGTH];
  REGISTER struct mapmerge *sp;
  Boolean found;
  /* we simulate the stack by hand to be faster */

  sp = stack;
again:
  if (map1==map2)
    goto ret;
  switch (map1->kind)
    {
    case E_EMPTY:
ret:
      if (sp==stack)
	return map1;
      *mapStackTop++ = map1;
      if (sp[-1].side)
	{
	  sp[-1].side = false;
	  map1 = sp[-1].map1;
	  map2 = sp[-1].map2;
	  mask = sp[-1].mask;
	  goto again;
	}
      sp--;
      map2 = mapStackTop[-2];
      if (map2==empty_map)
	{
	  if (/*map1==empty_map || map1->kind==(CHAR)E_MAPLET*/
	      map1->kind!=(CHAR)E_MAP)
	    {
	      mapStackTop -= 2;
	      goto ret;
	    }
	}
      else if (map1==empty_map)
	{
	  if (/*map2->kind==(CHAR)E_MAPLET*/
	      map2->kind!=(CHAR)E_MAP)
	    {
	      map1 = map2;
	      mapStackTop -= 2;
	      goto ret;
	    }
	}
      map1 = MapKons(map2,map1);
      mapStackTop -= 2;
      goto ret;
    case E_MAPLET:
      (void) MapGetMaplet1(MAPLETX(map1),map2,mask,&found);
      if (!found)
	map1 = empty_map;
      goto ret;
    case E_MAP:
      switch (map2->kind)
	{
	case E_MAPLET:
	  map1 = MapGetMaplet1(MAPLETX(map2),map1,mask,&found);
	  goto ret;
	case E_EMPTY:
	  map1 = empty_map;
	  goto ret;
	case E_MAP:
	  mask <<= 1;
	  sp->map1 = MAPRIGHT(map1);
	  map1 = MAPLEFT(map1);
	  sp->map2 = MAPRIGHT(map2);
	  map2 = MAPLEFT(map2);
	  sp->mask = mask;
	  sp->side = true;
	  sp++;
	  goto again;
	default:
	  map1 = map2;
	  /*fallthrough*/
	}
    default:
      mapStackTop = mapStack;
      kinderror(message(118),map1->kind);
      return NIL;
    }
#ifdef THE_RECURSIVE_VERSION
  REGISTER MapPtr left,right;
  Boolean found;

  if (map1==map2)
    return map1;
  switch (map1->kind)
    {
    case E_EMPTY:
      return empty_map;
    case E_MAPLET:
      (void) MapGetMaplet1(MAPLETX(map1),map2,mask,&found);
      return found?map1:empty_map;
    case E_MAP:
      switch (map2->kind)
	{
	case E_EMPTY:
	  return empty_map;
	case E_MAPLET:
	  return MapGetMaplet1(MAPLETX(map2),map1,mask,&found);
	case E_MAP:
	  mask <<= 1;
	  left = MapDomRestrTo1(MAPLEFT(map1),MAPLEFT(map2),mask);
	  *mapStackTop++ = left;
	  right = MapDomRestrTo1(MAPRIGHT(map1),MAPRIGHT(map2),mask);
	  if (left==empty_map)
	    {
	      if (/*right==empty_map || right->kind==(CHAR)E_MAPLET*/
		   right->kind!=(CHAR)E_MAP)
		{
		  --mapStackTop;
		  return right;
		}
	    }
	  else if (right==empty_map)
	    {
	      if (/*left->kind==(CHAR)E_MAPLET*/
		   left->kind!=(CHAR)E_MAP)
		{
		  --mapStackTop;
		  return left;
		}
	    }
	  *mapStackTop++ = right;
	  right = MapKons(left,right);
	  mapStackTop -= 2;
	  return right;
	default:
	  map1 = map2;
	  /* fallthrough */
	}
      default:
      mapStackTop = mapStack;
      kinderror(message(118),map1->kind);
      return NIL;
    }
#endif
}

PUBLIC ExprPtr MapDomRestrTo(map1,map2)
IN CONST MapPtr map1,map2;
{
  REGISTER MapPtr map;

  if (map1==empty_map || map2==empty_map)
    return empty_map;
  mapStackTop = mapStack;
  map = MapDomRestrTo1(map1,map2,MASK0);
  return KEEP(map);
}

PUBLIC MapPtr MapDomRestrBy1(map1,map2,mask)
IN CONST REGISTER MapPtr map1,map2;
IN CONST REGISTER CODE mask;
{
  struct mapmerge stack[MAP_CHAIN_LENGTH];
  REGISTER struct mapmerge *sp;
  Boolean found;
  /* we simulate the stack by hand to be faster */

  sp = stack;
again:
  if (map1==map2)
    {
      map1 = empty_map;
      goto ret;
    }
  switch (map1->kind)
    {
    case E_EMPTY:
ret:
      if (sp==stack)
	return map1;
      *mapStackTop++ = map1;
      if (sp[-1].side)
	{
	  sp[-1].side = false;
	  map1 = sp[-1].map1;
	  map2 = sp[-1].map2;
	  mask = sp[-1].mask;
	  goto again;
	}
      sp--;
      map2 = mapStackTop[-2];
      if (map2==empty_map)
	{
	  if (/*map1==empty_map || map1->kind==(CHAR)E_MAPLET*/
	      map1->kind!=(CHAR)E_MAP)
	    {
	      mapStackTop -= 2;
	      goto ret;
	    }
	}
      else if (map1==empty_map)
	{
	  if (/*map2->kind==(CHAR)E_MAPLET*/
	      map2->kind!=(CHAR)E_MAP)
	    {
	      map1 = map2;
	      mapStackTop -= 2;
	      goto ret;
	    }
	}
      map1 = MapKons(map2,map1);
      mapStackTop -= 2;
      goto ret;
    case E_MAPLET:
      (void) MapGetMaplet1(MAPLETX(map1),map2,mask,&found);
      if (found)
	map1 = empty_map;
      goto ret;
    case E_MAP:
      switch (map2->kind)
	{
	case E_MAPLET:
	  map1 = MapRemoveTrie(MAPLETX(map2),map1,mask);
	  goto ret;
	case E_EMPTY:
	  goto ret;
	case E_MAP:
	  mask <<= 1;
	  sp->map1 = MAPRIGHT(map1);
	  map1 = MAPLEFT(map1);
	  sp->map2 = MAPRIGHT(map2);
	  map2 = MAPLEFT(map2);
	  sp->mask = mask;
	  sp->side = true;
	  sp++;
	  goto again;
	default:
	  map1 = map2;
	  /*fallthrough*/
	}
    default:
      mapStackTop = mapStack;
      kinderror(message(119),map1->kind);
      return NIL;
    }
#ifdef THE_RECURSIVE_VERSION
  REGISTER MapPtr left,right;
  Boolean found;

  if (map1==map2)
    return empty_map;
  switch (map1->kind)
    {
    case E_EMPTY:
      return empty_map;
    case E_MAPLET:
      (void) MapGetMaplet1(MAPLETX(map1),map2,mask,&found);
      return found?empty_map:map1;
    case E_MAP:
      switch (map2->kind)
	{
	case E_EMPTY:
	  return map1;
	case E_MAPLET:
	  return MapRemoveTrie(MAPLETX(map2),map1,mask);
	case E_MAP:
	  mask <<= 1;
	  left = MapDomRestrBy1(MAPLEFT(map1),MAPLEFT(map2),mask);
	  *mapStackTop++ = left;
	  right = MapDomRestrBy1(MAPRIGHT(map1),MAPRIGHT(map2),mask);
	  if (left==empty_map)
	    {
	      if (/*right==empty_map || right->kind==(CHAR)E_MAPLET*/
		   right->kind!=(CHAR)E_MAP)
		{
		  --mapStackTop;
		  return right;
		}
	    }
	  else if (right==empty_map)
	    {
	      if (/*left->kind==(CHAR)E_MAPLET*/
		   left->kind!=(CHAR)E_MAP)
		{
		  --mapStackTop;
		  return left;
		}
	    }
	  *mapStackTop++ = right;
	  right = MapKons(left,right);
	  mapStackTop -= 2;
	  return right;
	default:
	  map1 = map2;
	  /* fallthrough */
	}
    default:
      mapStackTop = mapStack;
      kinderror(message(119),map1->kind);
      return NIL;
    }
#endif
}

PUBLIC ExprPtr MapDomRestrBy(map1,map2)
IN CONST MapPtr map1,map2;
{
  REGISTER MapPtr map;

  if (map1==empty_map || map2==empty_map)
    return map1;
  mapStackTop = mapStack;
  map = MapDomRestrBy1(map1,map2,MASK0);
  return KEEP(map);
}

PRIVATE MapPtr MapInter1(map1,map2,mask)
IN CONST REGISTER MapPtr map1,map2;
IN CONST REGISTER CODE mask;
{
  REGISTER MapPtr left,right;
  Boolean found;

  if (map1==map2)
    return map1;
  switch (map1->kind)
    {
    case E_EMPTY:
      return empty_map;
    case E_MAPLET:
      left = MapGetMaplet1(MAPLETX(map1),map2,mask,&found);
      return (found && left==map1)?map1:empty_map;
    case E_MAP:
      switch (map2->kind)
	{
	case E_EMPTY:
	  return empty_map;
	case E_MAPLET:
	  left = MapGetMaplet1(MAPLETX(map2),map1,mask,&found);
	  return (found && left==map2)?map2:empty_map;
	case E_MAP:
	  mask <<= 1;
	  left = MapInter1(MAPLEFT(map1),MAPLEFT(map2),mask);
	  *mapStackTop++ = left;
	  right = MapInter1(MAPRIGHT(map1),MAPRIGHT(map2),mask);
	  if (left==empty_map)
	    {
	      if (/*right==empty_map || right->kind==(CHAR)E_MAPLET*/
		   right->kind!=(CHAR)E_MAP)
		{
		  --mapStackTop;
		  return right;
		}
	    }
	  else if (right==empty_map)
	    {
	      if (/*left->kind==(CHAR)E_MAPLET*/
		   left->kind!=(CHAR)E_MAP)
		{
		  --mapStackTop;
		  return left;
		}
	    }
	  *mapStackTop++ = right;
	  right = MapKons(left,right);
	  mapStackTop -= 2;
	  return right;
	default:
	  map1 = map2;
	  /* fallthrough */
	}
    default:
      mapStackTop = mapStack;
      kinderror(message(118),map1->kind);
      return NIL;
    }
}

PUBLIC ExprPtr MapInter(map1,map2)
IN CONST MapPtr map1,map2;
{
  REGISTER MapPtr map;

  if (map1==empty_map || map2==empty_map)
    return empty_map;
  mapStackTop = mapStack;
  map = MapInter1(map1,map2,MASK0);
  return KEEP(map);
}

PUBLIC MapPtr MapRngRestrTo1(map1,map2)
IN CONST REGISTER MapPtr map1,map2;
{
  REGISTER MapPtr left,right;
  Boolean found;

  switch (map1->kind)
    {
    case E_EMPTY:
      return empty_map;
    case E_MAPLET:
      (void) MapGet(MAPLETY(map1),map2,&found);
      return found?map1:empty_map;
    case E_MAP:
      left = MapRngRestrTo1(MAPLEFT(map1),map2);
      *mapStackTop++ = left;
      right = MapRngRestrTo1(MAPRIGHT(map1),map2);
      if (left==empty_map)
	{
	  if (/*right==empty_map || right->kind==(CHAR)E_MAPLET*/
	       right->kind!=(CHAR)E_MAP)
	    {
	      --mapStackTop;
	      return right;
	    }
	}
      else if (right==empty_map)
	{
	  if (/*left->kind==(CHAR)E_MAPLET*/
	       left->kind!=(CHAR)E_MAP)
	    {
	      --mapStackTop;
	      return left;
	    }
	}
      *mapStackTop++ = right;
      right = MapKons(left,right);
      mapStackTop -= 2;
      return right;
    default:
      mapStackTop = mapStack;
      kinderror(message(120),map1->kind);
      return NIL;
    }
}

PUBLIC ExprPtr MapRngRestrTo(map1,map2)
IN CONST MapPtr map1,map2;
{
  REGISTER MapPtr map;

  mapStackTop = mapStack;
  map = MapRngRestrTo1(map1,map2);
  return KEEP(map);
}

PUBLIC MapPtr MapRngRestrBy1(map1,map2)
IN CONST REGISTER MapPtr map1,map2;
{
  REGISTER MapPtr left,right;
  Boolean found;

  switch (map1->kind)
    {
    case E_EMPTY:
      return empty_map;
    case E_MAPLET:
      (void) MapGet(MAPLETY(map1),map2,&found);
      return found?empty_map:map1;
    case E_MAP:
      left = MapRngRestrBy1(MAPLEFT(map1),map2);
      *mapStackTop++ = left;
      right = MapRngRestrBy1(MAPRIGHT(map1),map2);
      if (left==empty_map)
	{
	  if (/*right==empty_map || right->kind==(CHAR)E_MAPLET*/
	       right->kind!=(CHAR)E_MAP)
	    {
	      --mapStackTop;
	      return right;
	    }
	}
      else if (right==empty_map)
	{
	  if (/*left->kind==(CHAR)E_MAPLET*/
	       left->kind!=(CHAR)E_MAP)
	    {
	      --mapStackTop;
	      return left;
	    }
	}
      *mapStackTop++ = right;
      right = MapKons(left,right);
      mapStackTop -= 2;
      return right;
    default:
      mapStackTop = mapStack;
      kinderror(message(121),map1->kind);
      return NIL;
    }
}

PUBLIC ExprPtr MapRngRestrBy(map1,map2)
IN CONST MapPtr map1,map2;
{
  REGISTER MapPtr map;

  mapStackTop = mapStack;
  map = MapRngRestrBy1(map1,map2);
  return KEEP(map);
}

MAPCOLLECTFUNC(do_mapinverse)
{
  return MapletKons(y,x);
}

PUBLIC MapPtr MapInverse0(map)
IN CONST MapPtr map;
/* Note that MapInverse0() does not KEEP its result. */
{
  if (map==empty_map)
    return map;
  return MapCollect0(map,do_mapinverse,(char *)0);
}

PUBLIC MapPtr MapDelta1(map1,map2,mask)
IN CONST REGISTER MapPtr map1,map2;
IN CONST REGISTER CODE mask;
{
  REGISTER MapPtr left,right;
  Boolean found;

  if (map1==map2)
    return empty_map;
  switch (map1->kind)
    {
    case E_EMPTY:
      return map2;
    case E_MAPLET:
      (void) MapGetMaplet1(MAPLETX(map1),map2,mask,&found);
      return found?MapRemoveTrie(MAPLETX(map1),map2,mask):
	MapPutMapletTrie(map1,map2,mask);
    case E_MAP:
      switch (map2->kind)
	{
	case E_EMPTY:
	  return map1;
	case E_MAPLET:
	  (void) MapGetMaplet1(MAPLETX(map2),map1,mask,&found);
	  return found?MapRemoveTrie(MAPLETX(map2),map1,mask):
	    MapPutMapletTrie(map2,map1,mask);
	case E_MAP:
	  mask <<= 1;
	  left = MapDelta1(MAPLEFT(map1),MAPLEFT(map2),mask);
	  *mapStackTop++ = left;
	  right = MapDelta1(MAPRIGHT(map1),MAPRIGHT(map2),mask);
	  if (left==empty_map)
	    {
	      if (/*right==empty_map || right->kind==(CHAR)E_MAPLET*/
		   right->kind!=(CHAR)E_MAP)
		{
		  --mapStackTop;
		  return right;
		}
	    }
	  else if (right==empty_map)
	    {
	      if (/*left->kind==(CHAR)E_MAPLET*/
		   left->kind!=(CHAR)E_MAP)
		{
		  --mapStackTop;
		  return left;
		}
	    }
	  *mapStackTop++ = right;
	  right = MapKons(left,right);
	  mapStackTop -= 2;
	  return right;
	default:
	  map1 = map2;
	  /* fallthrough */
	}
    default:
      mapStackTop = mapStack;
      kinderror(message(122),map1->kind);
      return NIL;
    }
}

PUBLIC ExprPtr MapDelta(map1,map2)
IN CONST MapPtr map1,map2;
{
  REGISTER MapPtr map;

  if (map1==empty_map)
    return map2;
  if (map2==empty_map)
    return map1;
  mapStackTop = mapStack;
  map = MapDelta1(map1,map2,MASK0);
  return KEEP(map);
}

PUBLIC MapPtr MapDom1(map)
IN CONST MapPtr map;
{
  MapPtr left,right;

  switch (map->kind)
    {
    case E_EMPTY:
      return empty_map;
    case E_MAPLET:
      if (MAPLETY(map)!=UNIT)
	map = MapletKons(MAPLETX(map),UNIT);
      return map;
    case E_MAP:
      left = MapDom1(MAPLEFT(map));
      *mapStackTop++ = left;
      right = MapDom1(MAPRIGHT(map));
      *mapStackTop++ = right;
      left = MapKons(left,right);
      mapStackTop -= 2;
      return left;
    default:
      mapStackTop = mapStack;
      kinderror(message(123),map->kind);
      return NIL;
    }
}

PUBLIC MapPtr Dom(map)
IN CONST MapPtr map;
{
  if (map==empty_map)
    return map;
  mapStackTop = mapStack;
  map = MapDom1(map);
  return KEEP(map);
}

MAPCOLLECTFUNC(do_maprng)
{
  return MapletKons(y,UNIT);
}

PUBLIC MapPtr Rng0(map)
IN CONST MapPtr map;
/* Note that Rng() does not KEEP its result. */
{
  if (map==empty_map)
    return map;
  return MapCollect0(map,do_maprng,(char *)0);
}

PUBLIC Boolean MapMemberp(x,map)
IN CONST ExprPtr x;
IN CONST MapPtr map;
{
  Boolean found;

  (void) MapGetMaplet1(x,map,MASK0,&found);
  return found;
}

MAPSWEEPFUNC(do_mapincluded)
{
  MapPtr map;
  Boolean found;
  ExprPtr z;

  map = MAPDATA(MapPtr);
  z = MapGet1(x,map,MASK0,&found);
  return !(found && y==z);
}

PUBLIC Boolean MapIncludedp(map1,map2)
IN CONST MapPtr map1,map2;
{
  return !MapSweep(map1,do_mapincluded,map2);
}

MAPSWEEPFUNC(do_setincluded)
{
  MapPtr map;
  Boolean found;

  map = MAPDATA(MapPtr);
  (void) MapGet1(x,map,MASK0,&found);
  return !found;
}

PUBLIC Boolean SetIncludedp(map1,map2)
IN CONST MapPtr map1,map2;
{
  return !MapSweep(map1,do_setincluded,map2);
}

PRIVATE Boolean MapInterp1(map1,map2,mask)
IN CONST MapPtr map1,map2;
IN CODE mask;
{
  Boolean found;

  if (map1==map2)
    return map1!=empty_map;
  switch (map1->kind)
    {
    case E_EMPTY:
      return false;
    case E_MAPLET:
      (void) MapGetMaplet1(MAPLETX(map1),map2,mask,&found);
      return found;
    case E_MAP:
      switch (map2->kind)
	{
	case E_EMPTY:
	  return false;
	case E_MAPLET:
	  (void) MapGetMaplet1(MAPLETX(map2),map1,mask,&found);
	  return found;
	case E_MAP:
	  mask <<= 1;
	  return MapInterp1(MAPLEFT(map1),MAPLEFT(map2),mask) ||
	    MapInterp1(MAPRIGHT(map1),MAPRIGHT(map2),mask);
	  default:
	  map1 = map2;
	  /* fallthrough */
	}
    default:
      kinderror(message(124),map1->kind);
      return false;
    }
}

PUBLIC Boolean MapInterp(map1,map2)
IN CONST MapPtr map1,map2;
{
#ifdef DEBUG
  if (!ISMAP(map1) || !ISMAP(map2))
    CRASH();
#endif
  return map1!=empty_map && map2!=empty_map &&
    MapInterp1(map1,map2,MASK0);
}

PUBLIC ExprPtr ChooseInDom(map)
IN REGISTER CONST MapPtr map;
{
#ifdef DEBUG
  if (!ISMAP(map))
    CRASH();
#endif
  if (map==empty_map)
    return NIL;
  while (map->kind!=(CHAR)E_MAPLET)
    {
      if (MAPLEFT(map)!=empty_map)
	map = MAPLEFT(map);
      else
	map = MAPRIGHT(map);
    }
  return MAPLETX(map);
}

PUBLIC ExprPtr ChooseInRng(map)
IN REGISTER CONST MapPtr map;
{
#ifdef DEBUG
  if (!ISMAP(map))
    CRASH();
#endif
  if (map==empty_map)
    return NIL;
  while (map->kind!=(CHAR)E_MAPLET)
    {
      if (MAPLEFT(map)!=empty_map)
	map = MAPLEFT(map);
      else
	map = MAPRIGHT(map);
    }
  return MAPLETY(map);
}

PRIVATE Boolean setp(x,y,data)
IN CONST ExprPtr x,y;
IN CONST CHAR *data;
{
  return y!=UNIT;
}

PUBLIC Boolean Setp(map)
IN CONST MapPtr map;
{
  return !MapSweep(map,setp,(CHAR *)0L);
}

MAPCOLLECTFUNC(compose)
{
  struct compose *data;
  Boolean found;

  data = MAPDATA(struct compose *);
  return MapletKons(x,MapGet(y,data->map1,&found));
  /* found should be true */
}

PUBLIC MapPtr MapCompose(map1,map2)
IN CONST MapPtr map1,map2;
{
  struct compose data;
  MapPtr res;
  ENTER;

  data.map1 = map1;
  res = MapCollect0(MapRngRestrTo(map2,map1),compose,(CHAR *)&data);
  KEEPONLY(res);
  return res;
}

PUBLIC Boolean sameDomains(map1,map2)
IN CONST MapPtr map1,map2;
{
again:
  if (map1->kind!=map2->kind)
    return false;
  switch (map1->kind)
    {
      case E_EMPTY:
	return true;
      case E_MAPLET:
	return MAPLETX(map1)==MAPLETX(map2);
      case E_MAP:
	if (!sameDomains(CAR(map1),CAR(map2)))
	  return false;
	map1 = CDR(map1);
	map2 = CDR(map2);
	goto again;
    default:
      kinderror(message(919), map1->kind);
      return false;
    }
}

PUBLIC void MapSplit(map,map1,map2)
IN CONST MapPtr map;
OUT MapPtr *map1,*map2;
{
  MapPtr m1,m2;

#ifdef DEBUG
  if (!ISMAP(map))
    CRASH();
#endif
  switch (map->kind)
    {
    case E_EMPTY:
      *map1 = *map2 = map;
      break;
    case E_MAPLET:
      *map1 = NewMap();
      *map2 = map;
      break;
    case E_MAP:
      m1 = MAPLEFT(map);
      m2 = MAPRIGHT(map);
      if (m1==empty_map)
	{
	  MapSplit(m2,&m1,&m2);
	  if (m1->kind!=(CHAR)E_MAPLET)
	    {
	      *map1 = MapKons(empty_map,m1);
	      (void) KEEP(*map1);
	    }
	  else
	    *map1 = m1;
	  if (m2->kind!=(CHAR)E_MAPLET)
	    {
	      *map2 = MapKons(empty_map,m2);
	      (void) KEEP(*map2);
	    }
	  else
	    *map2 = m2;
	}
      else
	{
	  if (m2==empty_map)
	    {
	      MapSplit(m1,&m1,&m2);
	      if (m1->kind!=(CHAR)E_MAPLET)
		{
		  *map1 = MapKons(m1,empty_map);
		  (void) KEEP(*map1);
		}
	      else
		*map1 = m1;
	      if (m2->kind!=(CHAR)E_MAPLET)
		{
		  *map2 = MapKons(m2,empty_map);
		  (void) KEEP(*map2);
		}
	      else
		*map2 = m2;
	      break;
	    }
	  else
	    {
	      switch (m1->kind)
		{
		case E_MAPLET:
		  *map1 = m1;
		  break;
		case E_MAP:
		  *map1 = MapKons(m1,empty_map);
		  (void) KEEP(*map1);
		  break;
		}
	      switch (m2->kind)
		{
		case E_MAPLET:
		  *map2 = m2;
		  break;
		case E_MAP:
		  *map2 = MapKons(empty_map,m2);
		  (void) KEEP(*map2);
		  break;
		}
	    }
	}
      break;
    }
}

PUBLIC void setMapPut(x,y,loc)
IN CONST ExprPtr x,y;
INOUT REGISTER MapPtr *loc;
{
  REGISTER ExprPtr aux,map;

  map = *loc;
  aux = MapAdd(x,y,map);
  SETQ(*loc,aux);
}

PUBLIC void setcarMapPut(x,y,pair)
IN CONST ExprPtr x,y;
INOUT REGISTER ExprPtr pair;
{
  REGISTER ExprPtr aux,map;

  map = CAR(pair);
  aux = MapAdd(x,y,map);
  SETCAR(pair,aux);
}

PUBLIC void setcdrMapPut(x,y,pair)
IN CONST ExprPtr x,y;
INOUT REGISTER ExprPtr pair;
{
  REGISTER ExprPtr aux,map;

  map = CDR(pair);
  aux = MapAdd(x,y,map);
  SETCDR(pair,aux);
}

PUBLIC void setrefMapPut(x,y,ref)
IN CONST ExprPtr x,y;
INOUT REGISTER ExprPtr ref;
{
  REGISTER ExprPtr aux,map;

  map = REFVAL(ref);
  aux = MapAdd(x,y,map);
  SETREFVAL(ref,aux);
}

PUBLIC void setMapRemove(x,loc)
IN CONST ExprPtr x;
INOUT REGISTER MapPtr *loc;
{
  REGISTER ExprPtr aux,map;

  map = *loc;
  aux = MapKill(x,map);
  SETQ(*loc,aux);
}

PUBLIC void setcarMapRemove(x,pair)
IN CONST ExprPtr x;
INOUT REGISTER ExprPtr pair;
{
  REGISTER ExprPtr aux,map;

  map = CAR(pair);
  aux = MapKill(x,map);
  SETCAR(pair,aux);
}

PUBLIC void setcdrMapRemove(x,pair)
IN CONST ExprPtr x;
INOUT REGISTER ExprPtr pair;
{
  REGISTER ExprPtr aux,map;

  map = CDR(pair);
  aux = MapKill(x,map);
  SETCDR(pair,aux);
}

PUBLIC void setMapMerge(loc,map)
INOUT REGISTER MapPtr *loc;
IN CONST MapPtr map;
{
  REGISTER ExprPtr aux,m;

  m = *loc;
  aux = MapOverwrite(m,map);
  SETQ(*loc,aux);
}

PUBLIC void setcarMapMerge(pair,map)
INOUT REGISTER ExprPtr pair;
IN CONST MapPtr map;
{
  REGISTER ExprPtr aux,m;

  m = CAR(pair);
  aux = MapOverwrite(m,map);
  SETCAR(pair,aux);
}

PUBLIC void setcdrMapMerge(pair,map)
INOUT REGISTER ExprPtr pair;
IN CONST MapPtr map;
{
  REGISTER ExprPtr aux,m;

  m = CDR(pair);
  aux = MapOverwrite(m,map);
  SETCDR(pair,aux);
}

PUBLIC void setMapDomRestrTo(loc,map)
INOUT REGISTER MapPtr *loc;
IN CONST MapPtr map;
{
  REGISTER ExprPtr aux,m;

  m = *loc;
  aux = MapDRT(m,map);
  SETQ(*loc,aux);
}

PUBLIC void setMapDomRestrBy(loc,map)
INOUT REGISTER MapPtr *loc;
IN CONST MapPtr map;
{
  REGISTER ExprPtr aux,m;

  m = *loc;
  aux = MapDRB(m,map);
  SETQ(*loc,aux);
}

PUBLIC void setNRemove1(x,locp,mask,lastloc)
IN CONST REGISTER ExprPtr x;
INOUT REGISTER ExprPtr *locp;
IN CONST REGISTER CODE mask;
IN CONST ExprPtr lastloc;
{
  REGISTER ExprPtr xx;

  /*if (*locp==NIL)
    return;*/
  switch ((*locp)->kind)
    {
      case E_NIL:
	break;
      case E_MAPLET:
	xx = MAPLETX(*locp);
	if (xx==x)
	  {
	    SETQ(*locp,NIL);
	    if (lastloc!=NOEXPR)
	      HITBARRIER(lastloc);
	  }
	break;
      case E_CONS:
	setNRemove1(x,(((CODE)x) & mask)?(&CDR(*locp)):(&CAR(*locp)),mask << 1,*locp);
	if (CAR(*locp)==NIL && CDR(*locp)==NIL)
	  {
	    SETQ(*locp,NIL);
	    if (lastloc!=NOEXPR)
	      HITBARRIER(lastloc);
	  }
	break;
      default:
	kinderror(message(627),(*locp)->kind);
	break;
    }
}

PUBLIC void setNPut1(x,y,locp,mask,lastloc)
IN CONST REGISTER ExprPtr x;
IN CONST ExprPtr y;
INOUT REGISTER ExprPtr *locp;
IN CONST REGISTER CODE mask;
IN CONST REGISTER ExprPtr lastloc;
{
  REGISTER ExprPtr maplet,xx,aux;
  ENTER;

again:
  /*if (*locp==NIL)
  {
    MAKEMAPLET(x,y,maplet);
    SETQ(*locp,maplet);
  } else*/
  switch ((*locp)->kind)
  {
    case E_NIL:
      MAKEMAPLET(x,y,maplet);
      SETQ(*locp,maplet);
      if (lastloc!=NOEXPR)
	HITBARRIER(lastloc);
      break;
    case E_MAPLET:
      xx = MAPLETX(*locp);
      if (xx==x)
      {
        SETMAPLETY(*locp,y);
        break;
      }
      if (((CODE)xx) & mask)
        {
          MAKECONS(NIL,*locp,aux);
        }
      else
        {
          MAKECONS(*locp,NIL,aux);
        }
      SETQ(*locp,aux);
      if (lastloc!=NOEXPR)
	HITBARRIER(lastloc);
      /* fallthrough */
    case E_CONS:
      lastloc = *locp;
      if (((CODE)x) & mask)
        locp = &CDR(*locp);
      else
        locp = &CAR(*locp);
      mask <<= 1;
      goto again;
    default:
      kinderror(message(457),(*locp)->kind);
      break;
  }
  QUIT;
}

PUBLIC void setNPutBehind1(x,y,locp,mask,lastloc)
IN CONST REGISTER ExprPtr x;
IN CONST ExprPtr y;
INOUT REGISTER ExprPtr *locp;
IN CONST REGISTER CODE mask;
IN CONST REGISTER ExprPtr lastloc;
{
  REGISTER ExprPtr maplet,xx,aux;
  ENTER;

again:
  /*if (*locp==NIL)
  {
    MAKEMAPLET(x,y,maplet);
    SETQ(*locp,maplet);
  } else*/
  switch ((*locp)->kind)
  {
    case E_NIL:
      MAKEMAPLET(x,y,maplet);
      SETQ(*locp,maplet);
      if (lastloc!=NOEXPR)
	HITBARRIER(lastloc);
      break;
    case E_MAPLET:
      xx = MAPLETX(*locp);
      if (xx==x)
      {
        break;
      }
      if (((CODE)xx) & mask)
        {
          MAKECONS(NIL,*locp,aux);
        }
      else
        {
          MAKECONS(*locp,NIL,aux);
        }
      SETQ(*locp,aux);
      if (lastloc!=NOEXPR)
	HITBARRIER(lastloc);
      /* fallthrough */
    case E_CONS:
      lastloc = *locp;
      if (((CODE)x) & mask)
        locp = &CDR(*locp);
      else
        locp = &CAR(*locp);
      mask <<= 1;
      goto again;
    default:
      kinderror(message(944),(*locp)->kind);
      break;
  }
  QUIT;
}

PUBLIC ExprPtr NGet1(x,loc,foundp,mask)
IN CONST REGISTER ExprPtr x;
IN CONST REGISTER ExprPtr loc;
OUT Boolean *foundp;
IN REGISTER CODE mask;
{
  /*again:*/
  /*if (loc==NIL)
    {
      *foundp = false;
      return NIL;
    }*/
  while (loc->kind==(CHAR)E_CONS)
    {
      loc = ((((CODE)x) & mask)?CDR(loc):CAR(loc));
      mask <<= 1;
      if (loc->kind!=(CHAR)E_CONS) /* unroll loop. */
	break;
      loc = ((((CODE)x) & mask)?CDR(loc):CAR(loc));
      mask <<= 1;
      if (loc->kind!=(CHAR)E_CONS) /* unroll loop. */
	break;
      loc = ((((CODE)x) & mask)?CDR(loc):CAR(loc));
      mask <<= 1;
      if (loc->kind!=(CHAR)E_CONS) /* unroll loop. */
	break;
      loc = ((((CODE)x) & mask)?CDR(loc):CAR(loc));
      mask <<= 1;
      if (loc->kind!=(CHAR)E_CONS) /* unroll loop. */
	break;
      loc = ((((CODE)x) & mask)?CDR(loc):CAR(loc));
      mask <<= 1;
      if (loc->kind!=(CHAR)E_CONS) /* unroll loop. */
	break;
      loc = ((((CODE)x) & mask)?CDR(loc):CAR(loc));
      mask <<= 1;
      if (loc->kind!=(CHAR)E_CONS) /* unroll loop. */
	break;
      loc = ((((CODE)x) & mask)?CDR(loc):CAR(loc));
      mask <<= 1;
      if (loc->kind!=(CHAR)E_CONS) /* unroll loop. */
	break;
      loc = ((((CODE)x) & mask)?CDR(loc):CAR(loc));
      mask <<= 1;
    }
  if (loc->kind==(CHAR)E_NIL)
    {
      *foundp = false;
      return NIL;
    }
  *foundp = (MAPLETX(loc)==x);
  return MAPLETY(loc);
  /*
  switch (loc->kind)
  {
    case E_CONS:
      loc = ((((CODE)x) & mask)?CDR(loc):CAR(loc));
      mask <<= 1;
      goto again;
    case E_NIL:
      *foundp = false;
      return NIL;
    case E_MAPLET:
      *foundp = (MAPLETX(loc)==x);
      return MAPLETY(loc);
    default:
      kinderror(message(458),loc->kind);
      return NIL;
  }
  */
}

PUBLIC ExprPtr NGetMaplet1(x,loc,foundp,mask)
IN CONST REGISTER ExprPtr x;
IN CONST REGISTER ExprPtr loc;
OUT Boolean *foundp;
IN CODE REGISTER mask;
{
again:
  /*if (loc==NIL)
    {
      *foundp = false;
      return NIL;
    }*/
  switch (loc->kind)
  {
    case E_NIL:
      *foundp = false;
      return NIL;
    case E_MAPLET:
      *foundp = (MAPLETX(loc)==x);
      return loc;
    case E_CONS:
      loc = ((((CODE)x) & mask)?CDR(loc):CAR(loc));
      mask <<= 1;
      goto again;
    default:
      kinderror(message(426),loc->kind);
      return NIL;
  }
}

PUBLIC Boolean NSweep(loc,fun,data)
IN CONST ExprPtr loc;
IN CONST nSweepFunc fun;
IN CONST char *data;
{
again:
  /*if (loc==NIL)
    return false;*/
  switch (loc->kind)
  {
    case E_NIL:
      return false;
    case E_MAPLET:
      return (*fun)(loc,data);
    case E_CONS:
      if (NSweep(CAR(loc),fun,data))
	return true;
      loc = CDR(loc);
      goto again;
    default:
      kinderror(message(459),loc->kind);
      return false;
  }
}

PUBLIC MapPtr NCollect(loc,fun,data)
IN CONST ExprPtr loc;
IN CONST nCollectFunc fun;
IN CONST char *data;
{
  /*if (loc==NIL)
    return false;*/
  switch (loc->kind)
  {
    case E_NIL:
      return NewMap();
    case E_MAPLET:
      return (*fun)(loc,data);
    case E_CONS:
      return MapMerge(NCollect(CAR(loc),fun,data),
		      NCollect(CDR(loc),fun,data));
    default:
      kinderror(message(625),loc->kind);
      return NIL;
  }
}
