/* Definitions for stop and copy page management.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

EXTERN struct Expression *writeBarrier;

#define NEWGEN ((struct Expression *)-1L)
#define OLDGEN ((struct Expression *)1L)
/* NEWGEN and OLDGEN are different from NOEXPR, since NOEXPR
   is already taken to be the end of the linked list of shared
   cons cells in hash tables. */

#define HITBARRIER(x) {if (ELINK(x)==OLDGEN) { ELINK(x)=writeBarrier;writeBarrier=(x); }}
#define SETREFVAL1(x,e) { REFVAL(x) = e; HITBARRIER(x);}
#define SETCAR1(x,e) { CAR(x) = e; HITBARRIER(x);}
#define SETCDR1(x,e) { CDR(x) = e; HITBARRIER(x);}
#define SETSHAPE1(x,e) { ASHAPE(x) = e; HITBARRIER(x);}
#define SETQENV1(x,e) { QENV(x) = e; HITBARRIER(x);}

EXTERN CHAR IncRefTable[],DecRefTable[];

#define	REFCOUNT(e) ((e)->mark & MARK_REF_COUNT)
#define	SETREFCOUNT(e,n) (e)->mark = ((e)->mark & ~MARK_REF_COUNT) | (n)

#define	INCREF(e) (e)->mark = IncRefTable[(int)(e)->mark]
#define	DECREFP(e) ((((e)->mark = DecRefTable[(int)(e)->mark]) & MARK_REF_COUNT)==0)

#define	SETQ(x,e)	{ x = e; }
#define	POPQ(x)		{ x = CDR(x); }
#define	PUSHQ(x,e,aux)	{ MAKECONS(e,x,aux); x = aux; }

#define SETREFVAL(x,e) SETREFVAL1(x,e)
#define SETCAR(x,e) SETCAR1(x,e)
#define SETCDR(x,e) SETCDR1(x,e)
#define SETSHAPE(x,e) SETSHAPE1(x,e)
#define SETQENV(x,e) SETQENV1(x,e)

#define SETMAPLETY SETCDR

#if STOP_AND_COPY_VECTORS
#define PAGESIZE MODALIGN(4096+ALIGNMENT-1)

#define MAXALLOC_IN_PAGE (PAGESIZE/4)

typedef struct Page {
  struct Page *next;
  char *end,*top;
#if PAGE_GAP!=0
  char filler[PAGE_GAP];
#endif
  char contents[PAGESIZE];
} Page,*PagePtr;

EXTERN PagePtr twinPageList; /* list of couple of pages */
EXTERN PagePtr curPage;
EXTERN int spaceIndex;
EXTERN INT nPages,neededPages,allocedPages,minAllocedPages;

#define pageStart(page) (page)->contents
#define	pageEnd(page) (page)->end
#define	pageTop(page) (page)->top
#define	pageNext(page) (page)->next

IMPORT PagePtr NewPage P((char *(*new)()));
IMPORT void DelPage P((void));

IMPORT char *nextscAlloc P((unsigned long rsize));
#define scRealSize(size) MODALIGN(size+(ALIGNMENT-1))
     /* size need not be >=sizeof(char *), to allow gc to put
	a forwarding pointer in a block, because we don't use
	forwarding pointers (objects in pages are referenced
	exactly once); moreover, the size is
	not recorded in the block, because we use we know it
	from elsewhere. */
#define scAlloc1(p,xwzRealSize) {\
	if (xwzRealSize<=MAXALLOC_IN_PAGE && pageTop(curPage)+xwzRealSize<=pageEnd(curPage)) \
		{ p = pageTop(curPage); pageTop(curPage) += xwzRealSize; } \
	else p = nextscAlloc(xwzRealSize); }
#define scAlloc(p,size) {\
	unsigned long xwzRealSize = scRealSize(size); \
	scAlloc1(p,xwzRealSize); }
IMPORT char *scCopy P((char *p,unsigned long size));
#define scMARK(p,size) *(char **)&(p) = scCopy((char *)p,size)
#define scFREE(p,size) {if (scRealSize(size)>MAXALLOC_IN_PAGE) {MARKFREE(p);}}

#else
#define scAlloc(p,size) VECALLOC(p,size)
#define scMARK(p,size)
#define scFREE(p,size) VECFREE(p,size)
#endif
