/*
 * regcomp and regexec -- regsub and regerror are elsewhere
 *
 *	Copyright (c) 1986 by University of Toronto.
 *	Written by Henry Spencer.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. The author is not responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *** THIS IS AN ALTERED VERSION.  It was altered by Jean Goubault-Larrecq,
 *** goubault@lsv.ens-cachan.fr, several times since May 18, 1995,
 *** to interface it to HimML.
 *** THIS IS AN ALTERED VERSION.  It was altered by John Gilmore,
 *** hoptoad!gnu, on 27 Dec 1986, to add \n as an alternative to |
 *** to assist in implementing egrep.
 *** THIS IS AN ALTERED VERSION.  It was altered by John Gilmore,
 *** hoptoad!gnu, on 27 Dec 1986, to add \< and \> for word-matching
 *** as in BSD grep and ex.
 *** THIS IS AN ALTERED VERSION.  It was altered by John Gilmore,
 *** hoptoad!gnu, on 28 Dec 1986, to optimize characters quoted with \.
 *** THIS IS AN ALTERED VERSION.  It was altered by James A. Woods,
 *** ames!jaw, on 19 June 1987, to quash a regcomp() redundancy.
 *** THIS IS AN ALTERED VERSION, altered by Jean Goubault-Larrecq,
 *** on 18 May 1996, to adapt it to the HimML runtime (regsub.c is
 *** incorporated in this file, and regerror.c has been disposed of
 *** as well).
 *
 * Beware that some of this code is subtly aware of the way operator
 * precedence is structured in regular expressions.  Serious changes in
 * regular-expression syntax might require a total rethink.
 */
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/errno.h>
#else
#include <errno.h>
#endif
#include "eval.h"
#include "expr.h"
#include "mem.h"
#include "timing.h"
#include "debug.h"
#include "transfer.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "regmagic.h"
#include "regexp.h"
#ifdef HAS_PROTOTYPES
#include "sregexp.h"
#endif
/*
 * The "internal use only" fields in regexp.h are present to pass info from
 * compile to execute that permits the execute phase to run lots faster on
 * simple cases.  They are:
 *
 * regstart	char that must begin a match; '\0' if none obvious
 * reganch	is the match anchored (at beginning-of-line only)?
 * regmust	string (pointer into program) that match must include, or NULL
 * regmlen	length of regmust string
 *
 * Regstart and reganch permit very fast decisions on suitable starting points
 * for a match, cutting down the work a lot.  Regmust permits fast rejection
 * of lines that cannot possibly match.  The regmust tests are costly enough
 * that regcomp() supplies a regmust only if the r.e. contains something
 * potentially expensive (at present, the only such thing detected is * or +
 * at the start of the r.e., which can involve a lot of backup).  Regmlen is
 * supplied because the test in regexec() needs it and regcomp() is computing
 * it anyway.
 */


EXTERN ExprPtr reException,stringnthException,voidString;
EXTERN ExprPtr rematchsym,resubstsym,rematchfullsym,rematchpsym;
EXTERN ExprPtr none,optttm;

/*
 * Structure for regexp "program".  This is essentially a linear encoding
 * of a nondeterministic finite-state machine (aka syntax charts or
 * "railroad normal form" in parsing technology).  Each node is an opcode
 * plus a "next" pointer, possibly plus an operand.  "Next" pointers of
 * all nodes except BRANCH implement concatenation; a "next" pointer with
 * a BRANCH on both ends of it is connecting two alternatives.  (Here we
 * have one of the subtle syntax dependencies:  an individual BRANCH (as
 * opposed to a collection of them) is never concatenated with anything
 * because of operator precedence.)  The operand of some types of node is
 * a literal string; for others, it is a node leading into a sub-FSM.  In
 * particular, the operand of a BRANCH node is the first node of the branch.
 * (NB this is *not* a tree structure:  the tail of the branch connects
 * to the thing following the set of BRANCHes.)  The opcodes are:
 */

/* definition	number	opnd?	meaning */
#define	END	0	/* no	End of program. */
#define	BOL	1	/* no	Match "" at beginning of line. */
#define	EOL	2	/* no	Match "" at end of line. */
#define	ANY	3	/* no	Match any one character. */
#define	ANYOF	4	/* str	Match any character in this string. */
#define	ANYBUT	5	/* str	Match any character not in this string. */
#define	BRANCH	6	/* node	Match this alternative, or the next... */
#define	BACK	7	/* no	Match "", "next" ptr points backward. */
#define	EXACTLY	8	/* str	Match this string. */
#define	NOTHING	9	/* no	Match empty string. */
#define	STAR	10	/* node	Match this (simple) thing 0 or more times. */
#define	PLUS	11	/* node	Match this (simple) thing 1 or more times. */
#define	WORDA	12	/* no	Match "" at wordchar, where prev is nonword */
#define	WORDZ	13	/* no	Match "" at nonwordchar, where prev is word */
#define WORDL	14	/* no	Match "" at beginning or end of word */
#define WORDNL	15	/* no	Match "" neither at beginning nor end of word */
#define WORDIN	16	/* no	Match one word constituent character */
#define WORDOUT	17	/* no	Match one non-word constituent character */
#define	OPEN	20	/* no	Mark this point in input as start of #n. */
			/*	OPEN+1 is number 1, etc. */
#define	CLOSE	(OPEN+NSUBEXP)	/* no	Analogous to OPEN. */
#define PREV	(CLOSE+NSUBEXP)	/* no	Analogous to OPEN, match against previous data */

/*
 * Opcode notes:
 *
 * BRANCH	The set of branches constituting a single choice are hooked
 *		together with their "next" pointers, since precedence prevents
 *		anything being concatenated to any individual branch.  The
 *		"next" pointer of the last BRANCH in a choice points to the
 *		thing following the whole choice.  This is also where the
 *		final "next" pointer of each individual branch points; each
 *		branch starts with the operand node of a BRANCH node.
 *
 * BACK		Normal "next" pointers all implicitly point forward; BACK
 *		exists to make loop structures possible.
 *
 * STAR,PLUS	'?', and complex '*' and '+', are implemented as circular
 *		BRANCH structures using BACK.  Simple cases (one character
 *		per match) are implemented with STAR and PLUS for speed
 *		and to minimize recursive plunges.
 *
 * OPEN,CLOSE,PREV	...are numbered at compile time.
 */

/*
 * A node is one char of opcode followed by two chars of "next" pointer.
 * "Next" pointers are stored as two 8-bit pieces, high order first.  The
 * value is a positive offset from the opcode of the node containing it.
 * An operand, if any, simply follows the node.  (Note that much of the
 * code generation knows about this implicit relationship.)
 *
 * Using two bytes for the "next" pointer is vast overkill for most things,
 * but allows patterns to get big without disasters.
 */
#define	OP(p)	(*(p))
#define	NEXT(p)	(((*((p)+1)&0377)<<8) + (*((p)+2)&0377))
#define	OPERAND(p)	((p) + 3)

/*
 * See regmagic.h for one further detail of program structure.
 */


/*
 * Utility definitions.
 */
#ifndef CHARBITS
#define	UCHARAT(p)	((int)*(unsigned char *)(p))
#else
#define	UCHARAT(p)	((int)*(p)&CHARBITS)
#endif

/*#define	FAIL(m)	{ regerror(m); return(NULL); }*/
#define	ISMULT(c)	((c) == '*' || (c) == '+' || (c) == '?')

/*
 * Flags to be passed up and down.
 */
#define	HASWIDTH	01	/* Known never to match null string. */
#define	SIMPLE		02	/* Simple enough to be STAR/PLUS operand. */
#define	SPSTART		04	/* Starts with * or +. */
#define	WORST		0	/* Worst case. */

/* failure mode [jg] */
int regexperr;

/*
 * Global work variables for regcomp().
 */
static CHAR *regparse;		/* Input-scan pointer. */
static CHAR *regpend;		/* end of regparse */
static int regnpar;		/* () count. */
static CHAR regdummy;
static CHAR *regcode;		/* Code-emit pointer; &regdummy = don't. */
static long regsize;		/* Code size. */

/*
 * Forward declarations for regcomp()'s friends.
 */
#ifndef STATIC
#define	STATIC	static
#endif
STATIC CHAR *reg P((int paren,int *flagp));
STATIC CHAR *regbranch P((int *flagp));
STATIC CHAR *regpiece P((int *flagp));
STATIC CHAR *regatom P((int *flagp));
STATIC CHAR *regnode P((CHAR op));
STATIC CHAR *regnext P((CHAR *p));
STATIC void regc P((CHAR b));
STATIC void reginsert P((CHAR op,CHAR *opnd));
STATIC void regtail P((CHAR *p,CHAR *val));
STATIC void regoptail P((CHAR *p,CHAR *val));
#ifdef STRCSPN
STATIC int strcspn();
#endif

/*
 - regcomp - compile a regular expression into internal code
 *
 * We can't allocate space until we know how big the compiled form will be,
 * but we can't compile it (and thus know how big it is) until we've got a
 * place to put the code.  So we cheat:  we compile it twice, once with code
 * generation turned off and size counting turned on, and once "for real".
 * This also means that we don't allocate space until we are sure that the
 * thing really will compile successfully, and we never have to move the
 * code and thus invalidate pointers into it.  (Note that it has to be in
 * one piece because free() must be able to free it all.)
 *
 * Beware that the optimization-preparation code in here knows about some
 * of the structure of the compiled regexp.
 */
PUBLIC regexp *regcomp(str)
IN CONST ExprPtr str;
{
  register regexp *r;
  register CHAR *scan;
  register CHAR *longest;
  register int len;
  int flags;
#if STOP_AND_COPY_VECTORS
  CHAR *regtmp;
#endif

  regexperr = RE_NOERR;
  /* First pass: determine size, legality. */
#ifdef notdef
  if (size>=2 && exp[0] == '.' && exp[1] == '*') exp += 2;  /* aid grep */
#endif
#if STOP_AND_COPY_VECTORS
  regparse = regtmp = (CHAR *)alloctmp(SLEN(str));
  safememcpy(regparse,STEXT(str),SLEN(str));
#else
  regparse = (CHAR *)STEXT(str);
#endif
  regpend = regparse+SLEN(str);
  regnpar = 1;
  regsize = 0L;
  regcode = &regdummy;
  regc(MAGIC);
  if (reg(0, &flags) == NULL)
    return(NULL);

  /* Small enough for pointer-storage convention? */
  if (regsize >= 32767L)		/* Probably could be 65535L. */
    { regexperr = RE_TOOBIG; return NULL; } /* jg */
  /*FAIL("regexp too big");*/

  /* Allocate space. */
  r = (regexp *)MALLOC(sizeof(regexp) + (unsigned long)regsize);
  /*if (r == NULL)
		FAIL("out of space");
		cannot happen in HimML [jg] */
  r->regsize = regsize; /* jg */
  r->matched = voidString; /* jg */
  r->mstart = (char *)STEXT(r->matched); /* jg */

  /* Second pass: emit code. */
#if STOP_AND_COPY_VECTORS
  regparse = regtmp;
#else
  regparse = (CHAR *)STEXT(str);
#endif
  regpend = regparse+SLEN(str);
  regnpar = 1;
  regcode = (CHAR *)r->program;
  regc(MAGIC);
  if (reg(0, &flags) == NULL)
    return(NULL);

  /* Dig out information for optimizations. */
  r->regstart = '\0';	/* Worst-case defaults. */
  r->reganch = 0;
  r->regmust = NULL;
  r->regmlen = 0;
  scan = (CHAR *)r->program+1;			/* First BRANCH. */
  if (OP(regnext(scan)) == END) {		/* Only one top-level choice. */
    scan = OPERAND(scan);

    /* Starting-point info. */
    if (OP(scan) == EXACTLY)
      r->regstart = *OPERAND(scan);
    else if (OP(scan) == BOL)
      r->reganch++;

    /*
     * If there's something expensive in the r.e., find the
     * longest literal string that must appear and make it the
     * regmust.  Resolve ties in favor of later strings, since
     * the regstart check works with the beginning of the r.e.
     * and avoiding duplication strengthens checking.  Not a
     * strong reason, but sufficient in the absence of others.
     */
    if (flags&SPSTART) {
      longest = NULL;
      len = 0;
      for (; scan != NULL; scan = regnext(scan))
	if (OP(scan) == EXACTLY && strlen((char *)OPERAND(scan)) >= len) {
	  longest = OPERAND(scan);
	  len = strlen((char *)OPERAND(scan));
	}
      r->regmust = (char *)longest;
      r->regmlen = len;
    }
  }
  
  return(r);
}

DEF_PRIVATE_ACTION_1ARG(reg)
{
  struct reg1 *data;

  data = (struct reg1 *)ENV_UNDERLYING(env);
  data->res = reg(data->paren,&data->flag);
  return NIL;
}
ACTION_END

PRIVATE CHAR *regK(paren,flagp)
int paren;
int *flagp;
{
  struct reg1 *data = (struct reg1 *)alloctmp(sizeof(struct reg1));

  data->paren = paren;
  data->flag = *flagp;
  (void) CallCC(ACTION_NAME(reg),(ExprPtr)data,true);
  *flagp = data->flag;
  return data->res;
}

/*
 - reg - regular expression, i.e. main body or parenthesized thing
 *
 * Caller must absorb opening parenthesis.
 *
 * Combining parenthesis handling with the base level of regular expression
 * is a trifle forced, but the need to tie the tails of the branches to what
 * follows makes it hard to avoid.
 */
static CHAR *
reg(paren, flagp)
int paren;			/* Parenthesized? */
int *flagp;
{
	register CHAR *ret;
	register CHAR *br;
	register CHAR *ender;
	register int parno = 0;
	int flags;

	if (STKFULL(flags))
	  return regK(paren,flagp);
	*flagp = HASWIDTH;	/* Tentatively. */

	/* Make an OPEN node, if parenthesized. */
	if (paren) {
		if (regnpar >= NSUBEXP)
		  { regexperr = RE_TOOMANYPARENS; return NULL; } /* jg */
			/*FAIL("too many ()");*/
		parno = regnpar;
		regnpar++;
		ret = regnode(OPEN+parno);
	} else
		ret = NULL;

	/* Pick up the branches, linking them together. */
	br = regbranch(&flags);
	if (br == NULL)
		return(NULL);
	if (ret != NULL)
		regtail(ret, br);	/* OPEN -> first. */
	else
		ret = br;
	if (!(flags&HASWIDTH))
		*flagp &= ~HASWIDTH;
	*flagp |= flags&SPSTART;
	while (regparse<regpend &&
	       (*regparse == '|' || *regparse == '\n')) {
		regparse++;
		br = regbranch(&flags);
		if (br == NULL)
			return(NULL);
		regtail(ret, br);	/* BRANCH -> BRANCH. */
		if (!(flags&HASWIDTH))
			*flagp &= ~HASWIDTH;
		*flagp |= flags&SPSTART;
	}

	/* Make a closing node, and hook it on the end. */
	ender = regnode((paren) ? CLOSE+parno : END);	
	regtail(ret, ender);

	/* Hook the tails of the branches to the closing node. */
	for (br = ret; br != NULL; br = regnext(br))
		regoptail(br, ender);

	/* Check for proper termination. */
	if (paren && (regparse>=regpend || *regparse++ != ')')) {
	  regexperr = RE_UNMATCHEDPARENS; return NULL; /* jg */
		/*FAIL("unmatched ()");*/
	} else if (!paren && regparse<regpend /* *regparse != '\0' */) {
		if (*regparse == ')') {
		  regexperr = RE_UNMATCHEDPARENS; return NULL; /* jg */
			/*FAIL("unmatched ()");*/
		} else
		  { regexperr = RE_JUNKONEND; return NULL; } /* jg */
			/*FAIL("junk on end");*/	/* "Can't happen". */
		/* NOTREACHED */
	}

	return(ret);
}

/*
 - regbranch - one alternative of an | operator
 *
 * Implements the concatenation operator.
 */
static CHAR *
regbranch(flagp)
int *flagp;
{
	register CHAR *ret;
	register CHAR *chain;
	register CHAR *latest;
	int flags;

	*flagp = WORST;		/* Tentatively. */

	ret = regnode(BRANCH);
	chain = NULL;
	while (regparse<regpend && /* *regparse != '\0' && */
	       *regparse != ')' &&
	       *regparse != '\n' && *regparse != '|') {
		latest = regpiece(&flags);
		if (latest == NULL)
			return(NULL);
		*flagp |= flags&HASWIDTH;
		if (chain == NULL)	/* First piece. */
			*flagp |= flags&SPSTART;
		else
			regtail(chain, latest);
		chain = latest;
	}
	if (chain == NULL)	/* Loop ran zero times. */
		(void) regnode(NOTHING);

	return(ret);
}

/*
 - regpiece - something followed by possible [*+?]
 *
 * Note that the branching code sequences used for ? and the general cases
 * of * and + are somewhat optimized:  they use the same NOTHING node as
 * both the endmarker for their branch list and the body of the last branch.
 * It might seem that this node could be dispensed with entirely, but the
 * endmarker role is not redundant.
 */
static CHAR *
regpiece(flagp)
int *flagp;
{
	register CHAR *ret;
	register CHAR op;
	register CHAR *next;
	int flags;

	ret = regatom(&flags);
	if (ret == NULL)
		return(NULL);

	op = *regparse;
	if (!ISMULT(op)) {
		*flagp = flags;
		return(ret);
	}

	if (!(flags&HASWIDTH) && op != '?')
	  { regexperr = RE_STARPLUSOPEMPTY; return NULL; } /* jg */
		/*FAIL("*+ operand could be empty");*/
	*flagp = (op != '+') ? (WORST|SPSTART) : (WORST|HASWIDTH);

	if (op == '*' && (flags&SIMPLE))
		reginsert(STAR, ret);
	else if (op == '*') {
		/* Emit x* as (x&|), where & means "self". */
		reginsert(BRANCH, ret);			/* Either x */
		regoptail(ret, regnode(BACK));		/* and loop */
		regoptail(ret, ret);			/* back */
		regtail(ret, regnode(BRANCH));		/* or */
		regtail(ret, regnode(NOTHING));		/* null. */
	} else if (op == '+' && (flags&SIMPLE))
		reginsert(PLUS, ret);
	else if (op == '+') {
		/* Emit x+ as x(&|), where & means "self". */
		next = regnode(BRANCH);			/* Either */
		regtail(ret, next);
		regtail(regnode(BACK), ret);		/* loop back */
		regtail(next, regnode(BRANCH));		/* or */
		regtail(ret, regnode(NOTHING));		/* null. */
	} else if (op == '?') {
		/* Emit x? as (x|) */
		reginsert(BRANCH, ret);			/* Either x */
		regtail(ret, regnode(BRANCH));		/* or */
		next = regnode(NOTHING);		/* null. */
		regtail(ret, next);
		regoptail(ret, next);
	}
	regparse++;
	if (regparse<regpend && ISMULT(*regparse))
	  { regexperr = RE_NESTEDSTARQUESPLUS; return NULL; } /* jg */
		/*FAIL("nested *?+");*/

	return(ret);
}

/*
 - regatom - the lowest level
 *
 * Optimization:  gobbles an entire sequence of ordinary characters so that
 * it can turn them into a single node, which is smaller to store and
 * faster to run.  Backslashed characters are exceptions, each becoming a
 * separate node; the code is simpler that way and it's not worth fixing.
 */
static CHAR *
regatom(flagp)
int *flagp;
{
	register CHAR *ret;
	int flags;

	*flagp = WORST;		/* Tentatively. */

	if (regparse>=regpend)
	  goto urp; /* jg */
	switch (*regparse++) {
	/* FIXME: these chars only have meaning at beg/end of pat? */
	case '^':
		ret = regnode(BOL);
		break;
	case '$':
		ret = regnode(EOL);
		break;
	case '.':
		ret = regnode(ANY);
		*flagp |= HASWIDTH|SIMPLE;
		break;
	case '[': {
			register int class;
			register int classend;

			if (regparse<regpend &&
			    *regparse == '^') {	/* Complement of range. */
				ret = regnode(ANYBUT);
				regparse++;
			} else
				ret = regnode(ANYOF);
			if (regparse<regpend &&
			    (*regparse == ']' || *regparse == '-'))
				regc(*regparse++);
			while (regparse<regpend && /* *regparse != '\0' && */
			       *regparse != ']') {
				if (*regparse == '-') {
					regparse++;
					if (regparse>=regpend ||
					    *regparse == ']'
					    /* || *regparse == '\0' */)
						regc('-');
					else {
						class = UCHARAT(regparse-2)+1;
						classend = UCHARAT(regparse);
						if (class > classend+1)
						  { regexperr = RE_INVALBRAC;
						    return NULL; } /* jg */
							/*FAIL("invalid [] range");*/
						for (; class <= classend; class++)
							regc(class);
						regparse++;
					}
				} else
					regc(*regparse++);
			}
			regc('\0');
			if (regparse>=regpend || *regparse != ']')
			  { regexperr = RE_UNMATCHEDBRAC; return NULL; } /* jg */
				/*FAIL("unmatched []");*/
			regparse++;
			*flagp |= HASWIDTH|SIMPLE;
		}
		break;
	case '(':
		ret = reg(1, &flags);
		if (ret == NULL)
			return(NULL);
		*flagp |= flags&(HASWIDTH|SPSTART);
		break;
	case '\0':
	case '|':
	case '\n':
	case ')':
urp:
		regexperr = RE_INTERNALURP;
		return NULL; /* jg */
		/*FAIL("internal urp");*/	/* Supposed to be caught earlier. */
		/*break;*/
	case '?':
	case '+':
	case '*':
		regexperr = RE_QUESPLUSSTARAFTERNIL;
		return NULL; /* jg */
		/*FAIL("?+* follows nothing");
		break;*/
	case '\\':
		if (regparse>=regpend)
		  {
		    regexperr = RE_TRAILBACK;
		    return NULL; /* jg */
		    /*FAIL("trailing \\");
		      break;*/
		  }
		switch (*regparse++) {
		case '<':
			ret = regnode(WORDA);
			break;
		case '>':
			ret = regnode(WORDZ);
			break;
		case 'b':
			ret = regnode(WORDL);
			break;
		case 'B':
			ret = regnode(WORDNL);
			break;
		case 'w':
			ret = regnode(WORDIN);
			*flagp |= HASWIDTH|SIMPLE;
			break;
		case 'W':
			ret = regnode(WORDOUT);
			*flagp |= HASWIDTH|SIMPLE;
			break;
		case '1': case '2': case '3': case '4': case '5':
		case '6': case '7': case '8': case '9':
		  /*re_prev:*/
			ret = regnode(PREV + (regparse[-1]-'0'));
			break;
		default:
			/* Handle general quoted chars in exact-match routine */
			goto de_fault;
		}
		break;
	de_fault:
	default:
		/*
		 * Encode a string of characters to be matched exactly.
		 *
		 * This is a bit tricky due to quoted chars and due to
		 * '*', '+', and '?' taking the SINGLE char previous
		 * as their operand.
		 *
		 * On entry, the char at regparse[-1] is going to go
		 * into the string, no matter what it is.  (It could be
		 * following a \ if we are entered from the '\' case.)
		 * 
		 * Basic idea is to pick up a good char in  ch  and
		 * examine the next char.  If it's *+? then we twiddle.
		 * If it's \ then we frozzle.  If it's other magic char
		 * we push  ch  and terminate the string.  If none of the
		 * above, we push  ch  on the string and go around again.
		 *
		 *  regprev  is used to remember where "the current char"
		 * starts in the string, if due to a *+? we need to back
		 * up and put the current char in a separate, 1-char, string.
		 * When  regprev  is NULL,  ch  is the only char in the
		 * string; this is used in *+? handling, and in setting
		 * flags |= SIMPLE at the end.
		 */
		{
			CHAR *regprev;
			register CHAR ch;

			regparse--;			/* Look at cur char */
			ret = regnode(EXACTLY);
			for ( regprev = 0 ; ; ) {
			  if (regparse>=regpend)
			    { ch = '\0'; goto magic; } /* jg */
				ch = *regparse++;	/* Get current char */
				switch (*regparse) {	/* look at next one */

				default:
					regc(ch);	/* Add cur to string */
					break;

				case '.': case '[': case '(':
				case ')': case '|': case '\n':
				case '$': case '^':
				case '\0':
				/* FIXME, $ and ^ should not always be magic */
				magic:
					regc(ch);	/* dump cur char */
					goto done;	/* and we are done */

				case '?': case '+': case '*':
					if (!regprev) 	/* If just ch in str, */
						goto magic;	/* use it */
					/* End mult-char string one early */
					regparse = regprev; /* Back up parse */
					goto done;

				case '\\':
					regc(ch);	/* Cur char OK */
					if (regparse+1>=regpend)
					  goto done;
					switch (regparse[1]){ /* Look after \ */
					case '1': case '2': case '3': case '4':
					case '5': case '6': case '7': case '8':
					case '9':
					case '\0':
					case '<':
					case '>':
						goto done; /* Not quoted */
					default:
						/* Backup point is \, scan							 * point is after it. */
						regprev = regparse;
						regparse++; 
						continue;	/* NOT break; */
					}
				}
				regprev = regparse;	/* Set backup point */
			}
		done:
			regc('\0');
			*flagp |= HASWIDTH;
			if (!regprev)		/* One char? */
				*flagp |= SIMPLE;
		}
		break;
	}

	return(ret);
}

/*
 - regnode - emit a node
 */
#ifdef HAS_PROTOTYPES
static CHAR *regnode(CHAR op)
#else
static CHAR *			/* Location. */
regnode(op)
CHAR op;
#endif
{
	register CHAR *ret;
	register CHAR *ptr;

	ret = regcode;
	if (ret == &regdummy) {
		regsize += 3;
		return(ret);
	}

	ptr = ret;
	*ptr++ = op;
	*ptr++ = '\0';		/* Null "next" pointer. */
	*ptr++ = '\0';
	regcode = ptr;

	return(ret);
}

/*
 - regc - emit (if appropriate) a byte of code
 */
#ifdef HAS_PROTOTYPES
static void regc(CHAR b)
#else
static void
regc(b)
CHAR b;
#endif
{
	if (regcode != &regdummy)
		*regcode++ = b;
	else
		regsize++;
}

/*
 - reginsert - insert an operator in front of already-emitted operand
 *
 * Means relocating the operand.
 */
#ifdef HAS_PROTOTYPES
static void reginsert(CHAR op,CHAR *opnd)
#else
static void
reginsert(op, opnd)
CHAR op;
CHAR *opnd;
#endif
{
	register CHAR *src;
	register CHAR *dst;
	register CHAR *place;

	if (regcode == &regdummy) {
		regsize += 3;
		return;
	}

	src = regcode;
	regcode += 3;
	dst = regcode;
	while (src > opnd)
		*--dst = *--src;

	place = opnd;		/* Op node, where operand used to be. */
	*place++ = op;
	*place++ = '\0';
	*place = '\0';
}

/*
 - regtail - set the next-pointer at the end of a node chain
 */
static void
regtail(p, val)
CHAR *p;
CHAR *val;
{
	register CHAR *scan;
	register CHAR *temp;
	register int offset;

	if (p == &regdummy)
		return;

	/* Find last node. */
	scan = p;
	for (;;) {
		temp = regnext(scan);
		if (temp == NULL)
			break;
		scan = temp;
	}

	if (OP(scan) == BACK)
		offset = scan - val;
	else
		offset = val - scan;
	*(scan+1) = (offset>>8)&0377;
	*(scan+2) = offset&0377;
}

/*
 - regoptail - regtail on operand of first argument; nop if operandless
 */
static void
regoptail(p, val)
CHAR *p;
CHAR *val;
{
	/* "Operandless" and "op != BRANCH" are synonymous in practice. */
	if (p == NULL || p == &regdummy || OP(p) != BRANCH)
		return;
	regtail(OPERAND(p), val);
}

/*
 * regexec and friends
 */

/*
 * Global work variables for regexec().
 */
static ExprPtr regstring;	/* String to be matched [jg] */
static CHAR *reginput;		/* String-input pointer. */
static CHAR *reginputend;	/* end of reginput [jg] */
static CHAR *regbol;		/* Beginning of input, for ^ check. */
static CHAR **regstartp;	/* Pointer to startp array. */
static CHAR **regendp;		/* Ditto for endp. */

/*
 * Forwards.
 */
STATIC int regtry P((regexp *prog,CHAR *string,INT len));
STATIC int regmatch P((CHAR *prog));
STATIC int regrepeat P((CHAR *p));

#ifdef DEBUG
int regnarrate = 0;
void regdump P((regexp *r));
STATIC CHAR *regprop P((CHAR *op));
#endif

/*
 - regexec - match a regexp against a string
 */
CHAR *
regexec(prog, string, start, end)
register CONST regexp *prog;
register CONST ExprPtr string;
IN INT start,end;
{
  register CHAR *s;
  register CHAR *send;
  /*extern CHAR *strchr();*/

#ifdef OBSOLETE
  /* Be paranoid... */
  if (prog == NULL || string == NULL) {
    regerror("NULL parameter");
    return(NULL);
  }
#endif

  /* Check validity of program. */
  if (UCHARAT(prog->program) != MAGIC) {
    regexperr = RE_INTERNALURP; /* jg */
    /*regerror("corrupted program");*/
    return(NULL);
  }

  regexperr = RE_NOERR;
  prog->matched = string;
  prog->mstart = (char *)STEXT(string);
  send = (CHAR *)(prog->mstart+end);
  /* If there is a "must appear" string, look for it. */
  if (prog->regmust != NULL) {
    s = (CHAR *)(prog->mstart+start);
    while ((s = (CHAR *)memchr(s, prog->regmust[0], send-s)) != NULL) {
      if (send-s>=prog->regmlen &&
	  memcmp(s, prog->regmust, prog->regmlen) == 0)
	break;	/* Found it. */
      s++;
    }
    if (s == NULL)	/* Not present. */
      return(NULL);
  }
  
  /* Mark beginning of line for ^ . */
  s = regbol = (CHAR *)(prog->mstart+start);
  regstring = string;
  
  /* Simplest case:  anchored match need be tried only once. */
  if (prog->reganch)
    return(regtry(prog, s, end-start)?s:NULL);
  
  /* Messy cases:  unanchored match. */
  if (prog->regstart != '\0')
    /* We know what char it must start with. */
    while ((s = (CHAR *)memchr(s, prog->regstart, send-s))!= NULL) {
      if (regtry(prog, s, send-s))
	return(s);
      s++;
    }
  else
    /* We don't -- general case. */
    do {
      if (regtry(prog, s, send-s))
	return(s);
    } while (++s<send); /* instead of (*s++ != '\0'); */
  
  /* Failure. */
  return(NULL);
}

/*
 - regtry - try match at specific point
 */
static int			/* 0 failure, 1 success */
regtry(prog, string, len)
regexp *prog;
CHAR *string;
INT len;
{
	register int i;
	register CHAR **sp;
	register CHAR **ep;

	reginput = string;
	reginputend = string+len;
	regstartp = (CHAR **)prog->startp;
	regendp = (CHAR **)prog->endp;

	sp = regstartp;
	ep = regendp;
	for (i = NSUBEXP; i > 0; i--) {
		*sp++ = NULL;
		*ep++ = NULL;
	}
	if (regmatch((CHAR *)prog->program + 1)) {
		prog->startp[0] = (char *)string;
		prog->endp[0] = (char *)reginput;
		return(1);
	} else
		return(0);
}

DEF_PRIVATE_ACTION_1ARG(regmatch)
{
  struct regmatch1 *data;

  data = (struct regmatch1 *)ENV_UNDERLYING(env);
#if STOP_AND_COPY_VECTORS
  if (STEXT(regstring)!=data->stext) /* allocating a continuation may have
					triggered a GC, which may in turn have
					moved the string to be matched. */
    {
      INT offset = STEXT(regstring) - data->stext;

      reginput += offset;
      reginputend += offset;
      regbol += offset;
    }
#endif
  data->res = regmatch(data->prog);
  return NIL;
}
ACTION_END

PRIVATE int regmatchK(prog)
CHAR *prog;
{
  struct regmatch1 *data = (struct regmatch1 *)alloctmp(sizeof(struct regmatch1));

  data->prog = prog;
#if STOP_AND_COPY_VECTORS
  data->stext = STEXT(regstring);
#endif
  (void) CallCC(ACTION_NAME(regmatch),(ExprPtr)data,true);
#if STOP_AND_COPY_VECTORS
  if (STEXT(regstring)!=data->stext) /* throwing a continuation may have
					triggered a GC, which may in turn have
					moved the string to be matched. */
    {
      INT offset = STEXT(regstring) - data->stext;

      reginput += offset;
      reginputend += offset;
      regbol += offset;
    }
#endif
  return data->res;
}

/*
 - regmatch - main matching routine
 *
 * Conceptually the strategy is simple:  check to see whether the current
 * node matches, call self recursively to see whether the rest matches,
 * and then act accordingly.  In practice we make some effort to avoid
 * recursion, in particular by going through "ordinary" nodes (that don't
 * need to know whether the rest of the match failed) by a loop instead of
 * by recursion.
 */
static int			/* 0 failure, 1 success */
regmatch(prog)
CHAR *prog;
{
	register CHAR *scan;	/* Current node. */
	CHAR *next;		/* Next node. */
	/*extern char *strchr();*/

	if (STKFULL(next))
	  return regmatchK(prog);
	/*again:*/
	scan = prog;
#ifdef DEBUG
	if (scan != NULL && regnarrate)
		fprintf(stderr, "%s(\n", regprop(scan));
#endif
	while (scan != NULL) {
#ifdef DEBUG
		if (regnarrate)
			fprintf(stderr, "%s...\n", regprop(scan));
#endif
		next = regnext(scan);

		switch (OP(scan)) {
		case BOL:
			if (reginput != regbol)
				return(0);
			break;
		case EOL:
			if (reginput<reginputend /* *reginput != '\0' */)
				return(0);
			break;
		case WORDA:
			/* Must be looking at a letter, digit, or _ */
			if (reginput>=reginputend ||
			    ((!isalnum(*reginput)) && *reginput != '_'))
				return(0);
			/* Prev must be BOL or nonword */
			if (reginput > regbol &&
			    (isalnum(reginput[-1]) || reginput[-1] == '_'))
				return(0);
			break;
		case WORDZ:
wordz:
			/* Must be looking at non letter, digit, or _ */
			if (reginput<reginputend &&
			    (isalnum(*reginput) || *reginput == '_'))
				return(0);
			/* We don't care what the previous char was */
			break;
		case WORDL:
			/* = WORDA | WORDZ */
			if (reginput>=reginputend ||
			    ((!isalnum(*reginput)) && *reginput != '_'))
				goto wordz;
			if (reginput > regbol &&
			    (isalnum(reginput[-1]) || reginput[-1] == '_'))
				goto wordz;
			break;
		case WORDNL:
			/* = complement of WORDL */
			if (reginput>=reginputend ||
			    ((!isalnum(*reginput)) && *reginput != '_'))
				goto wordnz;
			if (reginput > regbol &&
			    (isalnum(reginput[-1]) || reginput[-1] == '_'))
				goto wordnz;
			return(0);
wordnz:
			if (reginput<reginputend &&
			    (isalnum(*reginput) || *reginput == '_'))
				break;
			return(0);
		case WORDIN:
			/* Must be looking at a letter, digit, or _ */
			if (reginput>=reginputend ||
			    ((!isalnum(*reginput)) && *reginput != '_'))
				return(0);
			break;
		case WORDOUT:
			/* Must not be looking at a letter, digit, or _ */
			if (reginput>=reginputend ||
			    ((!isalnum(*reginput)) && *reginput != '_'))
				break;
			return(0);
		case ANY:
			if (reginput>=reginputend /* *reginput == '\0' */)
				return(0);
			reginput++;
			break;
		case EXACTLY: {
				register int len;
				register CHAR *opnd;

				opnd = OPERAND(scan);
				if (reginput>=reginputend)
				  {
				    if (*opnd!='\0')
				      return 0;
				  }
				else
				  {
				    /* Inline the first character, for speed. */
				    if (*opnd != *reginput)
				      return(0);
				    len = strlen((char *)opnd);
				    if (/*len > 1 && */
					(len>reginputend-reginput ||
					 memcmp(opnd, reginput, len) != 0))
				      return(0);
				    reginput += len;
				  }
			}
			break;
		case ANYOF:
 			if (reginput>=reginputend /* *reginput == '\0' */
			    || strchr((char *)OPERAND(scan), *reginput) == NULL)
				return(0);
			reginput++;
			break;
		case ANYBUT:
 			if (reginput>=reginputend /* *reginput == '\0' */
			    || strchr((char *)OPERAND(scan), *reginput) != NULL)
				return(0);
			reginput++;
			break;
		case NOTHING:
			break;
		case BACK:
			break;
		case OPEN+1:	case OPEN+2:	case OPEN+3:	case OPEN+4:
		case OPEN+5:	case OPEN+6:	case OPEN+7:	case OPEN+8:
		case OPEN+9:
			{
				register int no;
				register CHAR *save;

				no = OP(scan) - OPEN;
				save = reginput;

				regstartp[no] = reginput; /* jg */
				regendp[no] = NULL; /* jg */
				if (regmatch(next)) {
					/*
					 * Don't set startp if some later
					 * invocation of the same parentheses
					 * already has.
					 */
					if (regstartp[no] == NULL)
						regstartp[no] = save;
					return(1);
				} else
				  {
				    regstartp[no] = NULL; /* jg */
					return(0);
				  }
			}
			break;
		case CLOSE+1:	case CLOSE+2:	case CLOSE+3:	case CLOSE+4:
		case CLOSE+5:	case CLOSE+6:	case CLOSE+7:	case CLOSE+8:
		case CLOSE+9:
			{
				register int no;
				register CHAR *save;

				no = OP(scan) - CLOSE;
				save = reginput;

				regendp[no] = reginput; /* jg */
				if (regmatch(next)) {
					/*
					 * Don't set endp if some later
					 * invocation of the same parentheses
					 * already has.
					 */
					if (regendp[no] == NULL)
						regendp[no] = save;
					return(1);
				} else
				  {
				    regendp[no] = NULL; /* jg */
					return(0);
				  }
			}
			break;
		case PREV+1:	case PREV+2:	case PREV+3:	case PREV+4:
		case PREV+5:	case PREV+6:	case PREV+7:	case PREV+8:
		case PREV+9:
			{
			  register int no;
			  register INT len;

			  no = OP(scan) - PREV;
			  if (regstartp[no]==NULL || regendp[no]==NULL)
			    break; /* match the null string if no expression recorded */
			  len = regendp[no] - regstartp[no];
			  if (reginput+len>reginputend)
			    return 0;
			  if (memcmp(reginput,regstartp[no],len)!=0)
			    return 0;
			  reginput += len;
			}
			break;
		case BRANCH: {
				register CHAR *save;

				if (OP(next) != BRANCH)		/* No choice. */
					next = OPERAND(scan);	/* Avoid recursion. */
				else {
					do {
						save = reginput;
						if (regmatch(OPERAND(scan)))
							return(1);
						reginput = save;
						scan = regnext(scan);
					} while (scan != NULL && OP(scan) == BRANCH);
					return(0);
					/* NOTREACHED */
				}
			}
			break;
		case STAR:
		case PLUS: {
				register CHAR nextch;
				register int no;
				register CHAR *save;
				register int min;

				/*
				 * Lookahead to avoid useless match attempts
				 * when we know what character comes next.
				 */
				nextch = '\0';
				if (OP(next) == EXACTLY)
					nextch = *OPERAND(next);
				min = (OP(scan) == STAR) ? 0 : 1;
				save = reginput;
				no = regrepeat(OPERAND(scan));
				while (no >= min) {
					/* If it could work, try it. */
					if (nextch == '\0' ||
					    (reginput<reginputend &&
					     *reginput == nextch))
						if (regmatch(next))
							return(1);
					/* Couldn't or didn't -- back up. */
					no--;
					reginput = save + no;
				}
				return(0);
			}
			break;
		case END:
			return(1);	/* Success! */
			break;
		default:
			regexperr = RE_INTERNALURP; /* jg */
			/*regerror("memory corruption");*/
			return(0);
			break;
		}

		scan = next;
	}

	/*
	 * We get here only if there's trouble -- normally "case END" is
	 * the terminating point.
	 */
	regexperr = RE_INTERNALURP; /* jg */
	/*regerror("corrupted pointers");*/
	return(0);
}

/*
 - regrepeat - repeatedly match something simple, report how many
 */
static int
regrepeat(p)
CHAR *p;
{
	register int count = 0;
	register CHAR *scan;
	register CHAR *opnd;

	scan = reginput;
	opnd = OPERAND(p);
	switch (OP(p)) {
	case ANY:
		count = reginputend-scan; /*strlen(scan);*/
		scan = reginputend; /* += count; */
		break;
	case EXACTLY:
		while (scan<reginputend && *opnd == *scan) {
			count++;
			scan++;
		}
		break;
	case ANYOF:
		while (scan<reginputend && /* *scan != '\0' && */
		       strchr((char *)opnd, *scan) != NULL) {
			count++;
			scan++;
		}
		break;
	case ANYBUT:
		while (scan<reginputend && /* *scan != '\0' && */
		       strchr((char *)opnd, *scan) == NULL) {
			count++;
			scan++;
		}
		break;
	default:		/* Oh dear.  Called inappropriately. */
		/*regerror("internal foulup");*/ /* jg */
		count = 0;	/* Best compromise. */
		break;
	}
	reginput = scan; /* useless? [jg] */

	return(count);
}

/*
 - regnext - dig the "next" pointer out of a node
 */
static CHAR *
regnext(p)
register CHAR *p;
{
	register int offset;

	if (p == &regdummy)
		return(NULL);

	offset = NEXT(p);
	if (offset == 0)
		return(NULL);

	if (OP(p) == BACK)
		return(p-offset);
	else
		return(p+offset);
}

#ifdef DEBUG

STATIC CHAR *regprop();

/*
 - regdump - dump a regexp onto stdout in vaguely comprehensible form
 */
void
regdump(r)
regexp *r;
{
	register CHAR *s;
	register CHAR op = EXACTLY;	/* Arbitrary non-END op. */
	register CHAR *next;

	s = r->program + 1;
	while (op != END) {	/* While that wasn't END last time... */
		op = OP(s);
		printf("%2d%s", s-r->program, regprop(s));	/* Where, what. */
		next = regnext(s);
		if (next == NULL)		/* Next ptr. */
			printf("(0)");
		else 
			printf("(%d)", (s-r->program)+(next-s));
		s += 3;
		if (op == ANYOF || op == ANYBUT || op == EXACTLY) {
			/* Literal string, where present. */
			while (*s != '\0') {
				putchar(*s);
				s++;
			}
			s++;
		}
		putchar('\n');
	}

	/* Header fields of interest. */
	if (r->regstart != '\0')
		printf("start `%c' ", r->regstart);
	if (r->reganch)
		printf("anchored ");
	if (r->regmust != NULL)
		printf("must have \"%s\"", r->regmust);
	printf("\n");
}

/*
 - regprop - printable representation of opcode
 */
static CHAR *
regprop(op)
CHAR *op;
{
	register CHAR *p;
	static CHAR buf[50];

	(void) strcpy(buf, ":");

	switch (OP(op)) {
	case BOL:
		p = "BOL";
		break;
	case EOL:
		p = "EOL";
		break;
	case ANY:
		p = "ANY";
		break;
	case ANYOF:
		p = "ANYOF";
		break;
	case ANYBUT:
		p = "ANYBUT";
		break;
	case BRANCH:
		p = "BRANCH";
		break;
	case EXACTLY:
		p = "EXACTLY";
		break;
	case NOTHING:
		p = "NOTHING";
		break;
	case BACK:
		p = "BACK";
		break;
	case END:
		p = "END";
		break;
	case OPEN+1:	case OPEN+2:	case OPEN+3:	case OPEN+4:
	case OPEN+5:	case OPEN+6:	case OPEN+7:	case OPEN+8:
	case OPEN+9:
		sprintf(buf+strlen(buf), "OPEN%d", OP(op)-OPEN);
		p = NULL;
		break;
	case CLOSE+1:	case CLOSE+2:	case CLOSE+3:	case CLOSE+4:
	case CLOSE+5:	case CLOSE+6:	case CLOSE+7:	case CLOSE+8:
	case CLOSE+9:
		sprintf(buf+strlen(buf), "CLOSE%d", OP(op)-CLOSE);
		p = NULL;
		break;
	case PREV+1:	case PREV+2:	case PREV+3:	case PREV+4:
	case PREV+5:	case PREV+6:	case PREV+7:	case PREV+8:
	case PREV+9:
		sprintf(buf+strlen(buf), "PREV%d", OP(op)-PREV);
		p = NULL;
		break;
	case STAR:
		p = "STAR";
		break;
	case PLUS:
		p = "PLUS";
		break;
	case WORDA:
		p = "WORDA";
		break;
	case WORDZ:
		p = "WORDZ";
		break;
	case WORDL:
		p = "WORDL";
		break;
	case WORDNL:
		p = "WORDNL";
		break;
	case WORDIN:
		p = "WORDIN";
		break;
	case WORDOUT:
		p = "WORDOUT";
		break;
	default:
		/*regerror("corrupted opcode");*/ /* jg */
		break;
	}
	if (p != NULL)
		(void) strcat(buf, p);
	return(buf);
}
#endif

/*
 * The following is provided for those people who do not have strcspn() in
 * their C libraries.  They should get off their butts and do something
 * about it; at least one public-domain implementation of those (highly
 * useful) string routines has been published on Usenet.
 */
#ifdef STRCSPN
/*
 * strcspn - find length of initial segment of s1 consisting entirely
 * of characters not from s2
 */

static int
strcspn(s1, s2)
CHAR *s1;
CHAR *s2;
{
	register CHAR *scan1;
	register CHAR *scan2;
	register int count;

	count = 0;
	for (scan1 = s1; *scan1 != '\0'; scan1++) {
		for (scan2 = s2; *scan2 != '\0';)	/* ++ moved down. */
			if (*scan1 == *scan2++)
				return(count);
		count++;
	}
	return(count);
}
#endif

#if STOP_AND_COPY_VECTORS
/* string may have moved because of a GC, we correct all pointers */
static void regmove(prog)
CONST IN regexp *prog;
{
  if (STEXT(prog->matched)!=(CHAR *)prog->mstart)
    {
      REGISTER INT off = STEXT(prog->matched) - (CHAR *)prog->mstart;
      REGISTER COUNTER i;
      REGISTER CHAR **p,**pp;

      for (i=0, p=prog->startp, pp=prog->endp; i<NSUBEXP; i++, p++, pp++)
	{
	  *p += off;
	  *pp += off;
	}
      prog->mstart = (CHAR *)STEXT(prog->matched);
    }
}
#endif

PRIVATE INT regsub1(prog,source,dest)
IN CONST regexp *prog;
IN CONST ExprPtr source;
OUT CHAR *dest;
{
  register CHAR *src,*srcend;
  register CHAR *dst;
  register CHAR c;
  register COUNTER no;
  register INT len,dlen;

  regexperr = RE_NOERR;
  len = 0L;
  if (UCHARAT(prog->program) != MAGIC) {
    regexperr = RE_INTERNALURP; /* jg */
    /*regerror("damaged regexp fed to regsub");*/
    return len;
  }

#if STOP_AND_COPY_VECTORS
  regmove(prog); /* jg */
#endif
  src = STEXT(source);
  srcend = src+SLEN(source);
  dst = dest;
  while (src<srcend) /* (c = *src++) != '\0') */ {
    c = *src++;
    if (c == '&')
      no = 0;
    else if (c == '\\' && src<srcend && '0' <= *src && *src <= '9')
      no = *src++ - '0';
    else
      no = -1;
    if (no < 0) {	/* Ordinary character. */
      if (c == '\\' && src<srcend && (*src == '\\' || *src == '&'))
	c = *src++;
      if (dest!=NULL)
	*dst++ = c;
      len++;
    } else if (prog->startp[no] != NULL && prog->endp[no] != NULL) {
      dlen = prog->endp[no] - prog->startp[no];
      if (dest!=NULL)
	{
	  safememcpy(dst,(CHAR *) prog->startp[no],dlen);
	  dst += dlen;
	}
      len += dlen;
    }
  }
  return len;
}

/*
 - regsub - perform substitutions after a regexp match
 */
PUBLIC ExprPtr regsub(prog, source)
CONST regexp *prog;
IN CONST ExprPtr source;
{
  INT len;
  ExprPtr str;

  len = regsub1(prog,source,NULL); /* compute length of substituted string */
  if (regexperr!=RE_NOERR)
    return NIL;
  {
    CHAR *text;

#if STOP_AND_COPY_VECTORS
    text = (CHAR *)alloctmp(len);
#else
    VECALLOC(text,len);
#endif
    (void) regsub1(prog,source,text); /* do the actual substitution */
#if STOP_AND_COPY_VECTORS
    str = Kstring1(len,text,true);
#else
    str = Kstring1(len,text,false);
#endif
  }
  return str; /* not kept! */
}


PRIVATE void regexpSweep(extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  regexp *prog;

  prog = (regexp *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      {
	INT len;
	COUNTER i;

	len = prog->regsize;
	outInt(len);
	if (errno) return;
	(*doext->action)(prog->matched);
	if (errno) return;
	for (i=0; i<NSUBEXP; i++)
	  {
	    CHAR *p;

	    p = (CHAR *)prog->startp[i];
	    if (p==NULL)
	      len = -1;
	    else len = p-(CHAR *)prog->mstart;
	    outInt(len);
	    if (errno) return;
	    p = (CHAR *)prog->endp[i];
	    if (p==NULL)
	      len = -1;
	    else len = p-(CHAR *)prog->mstart;
	    outInt(len);
	    if (errno) return;
	  }
	outChars((CHAR *)&prog->regstart,1L);
	if (errno) return;
	outChars((CHAR *)&prog->reganch,1L);
	if (errno) return;
	len = (prog->regmust==NULL)?(-1):(prog->regmust-prog->program);
	outInt(len);
	if (errno) return;
	outChars((CHAR *)prog->program,prog->regsize);
	if (errno) return;
      }
      break;
    case X_MARK:
    case X_MARKDATA:
      (*doext->action)(prog->matched);
      break;
    case X_FREE:
      MARKFREE(prog);
      break;
    }
}

PRIVATE ExprPtr regexpLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  INT len;
  COUNTER i;
  ExprPtr e;
  regexp *prog;

  len = inInt();
  if (errno)
    return NIL;
  prog = (regexp *)MALLOC(sizeof(regexp) + (unsigned long)regsize);
  prog->regsize = len;
  e = (*load)(sharedref);
  if (errno)
    {
err:
      MARKFREE(prog);
      return NIL;
    }
  prog->matched = e;
  prog->mstart = (char *)STEXT(e);
  for (i=0; i<NSUBEXP;i++)
    {
      len = inInt();
      if (errno) goto err;
      prog->startp[i] = (len==-1L)?NULL:(prog->mstart+len);
      len = inInt();
      if (errno) goto err;
      prog->endp[i] = (len==-1L)?NULL:(prog->mstart+len);
    }
  inChars((CHAR *)&prog->regstart,1L);
  if (errno) goto err;
  inChars((CHAR *)&prog->reganch,1L);
  if (errno) goto err;
  len = inInt();
  if (errno) goto err;
  prog->regmust = (len==-1L)?NULL:(prog->program+len);
  inChars((CHAR *)prog->program,prog->regsize);
  e = MakeExternal(desc,(char *)prog);
  (void) KEEP(e);
  return e;
}

PUBLIC extDesc regexpExtDesc = {
  (CHAR *)"regexp",
  { 'r','e','x','p' }, NULL,
  NULL, /* no hash function, regexps are not shared */
  NULL, /* no comparison function, regexps are not shared */
  NULL, /* no print function */
  regexpSweep,
  regexpLoad
};

PUBLIC ExprPtr raiseRE()
{
  ExprPtr err,exc;

  err = KINT((INT)regexperr);
  MAKECONS(reException,err,exc);
  return Raise(KEEP(exc));
}

DEF_PRIVATE_ACTION_3ARGS(q_regexec) /* matchsub : string * int * int -> (int * int) option */
{ /* env is the external with the regexp * inside */
  CHAR *res;
  ExprPtr val;
  INT start,end;
  ENTER;

  val = arg1;
  start = IVAL(arg2);
  if (start<0)
    return Raise(stringnthException);
  end = IVAL(arg3);
  if (end>SLEN(val))
    return Raise(stringnthException);
  if (end<start)
    end = start;

  res = regexec((regexp *)EXTDATA(ENV_UNDERLYING(env)), val, start, end);
  if (regexperr!=RE_NOERR)
    raiseRE();
  if (res!=NULL)
    {
      start = res-STEXT(val);
      end = reginput-STEXT(val);
      val = Ktuple2(KINT(start),KINT(end));
      (void) KEEP(val);
      val = Kdata(0,optttm,val); /* SOME val */
    }
  else
    val = none;
  QUIT;
  return val;
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(q_regsub) /* subst : string -> string */
{ /* env is the external with the regexp * inside */
  ExprPtr val;

  val = regsub((regexp *)EXTDATA(ENV_UNDERLYING(env)), arg);
  if (regexperr!=RE_NOERR)
    raiseRE();
  return val;
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(q_regmatchfull) /* matches : string -> bool */
{ /* env is the external with the regexp * inside */
  CHAR *res;
  ExprPtr val;
  INT start,end;
  ENTER;

  res = regexec((regexp *)EXTDATA(ENV_UNDERLYING(env)), arg, 0L, SLEN(arg));
  if (regexperr!=RE_NOERR)
    raiseRE();
  if (res!=NULL)
    {
      start = res-STEXT(arg);
      end = reginput-STEXT(arg);
      val = Ktuple2(KINT(start),KINT(end));
      (void) KEEP(val);
      val = Kdata(0,optttm,val); /* SOME val */
    }
  else
    val = none;
  QUIT;
  return val;
}
ACTION_END

DEF_PRIVATE_ACTION_1ARG(q_regmatchp) /* match : string -> (int * int) option */
{ /* env is the external with the regexp * inside */
  CHAR *res;

  res = regexec((regexp *)EXTDATA(ENV_UNDERLYING(env)), arg, 0L, SLEN(arg));
  if (regexperr!=RE_NOERR)
    raiseRE();
  return (res==NULL)?falseValue:trueValue;
}
ACTION_END

PRIVATE ExprPtr MkCompiled (kind, action, env)
     IN ExprKind kind;
     IN compiledAction action;
     IN CONST ExprPtr env;
{
  ExprPtr e;

  e = MakeCompiled (kind, action, env);
  return KEEP (e);
}

DEF_ACTION_1ARG(regcomp)
{
  regexp *prog;
  ExprPtr ext;
#define N 4
  field fields[N];
  ENTER;

  prog = regcomp(arg);
  if (prog==NULL)
    return raiseRE();
  ext = MakeExternal(&regexpExtDesc,(char *)prog);
  (void) KEEP(ext);
  fields[0].label = rematchsym;
  fields[0].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_regexec),ext);
  fields[1].label = resubstsym;
  fields[1].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_regsub),ext);
  fields[2].label = rematchfullsym;
  fields[2].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_regmatchfull),ext);
  fields[3].label = rematchpsym;
  fields[3].val = MkCompiled(E_COMPILED_CLOSURE,ACTION_NAME(q_regmatchp),ext);
  sortFieldArray(N,fields); /* we could sort it in advance */
  arg = MakeRecordGnrl(N,fields,fieldsToShape(N,fields));
  QUIT;
  return arg;
#undef N
}
ACTION_END

DEF_ACTION_1ARG(regerrormsg)
{
  INT i;

  i = IVAL(arg);
  if (i<RE_NOERR || i>=N_RE_MSG)
    i = RE_NOERR;
  return Kstr((CHAR *)message(548+i));
}
ACTION_END
