/* Stack and continuation handling functions.
   This is mostly black magic.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005,2011 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "eval.h"
#include "expr.h"
#include "mem.h"
#include "timing.h"
#include "debug.h"
#include "evlproto.h"
#include "stkproto.h"
#include "errproto.h"
#ifdef MPW
#include <memory.h>
#endif
#ifdef THINK_C
#include <unix.h>
#endif
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#if OS==AMIGA_OS
#undef SHORT
/* SHORT is defined in types.h as short */
#include <exec/types.h>
#include <exec/tasks.h>
#include <exec/memory.h>
#include <proto/exec.h>
#include <exec/execbase.h>
EXTERN struct ExecBase *SysBase;
PRIVATE char *MLthreads=NULL;
struct StackSwapStruct initStack,MLstacks;
IMPORT NORETURN __asm void StackSwapOld P((IN CONST register __a0
					   struct StackSwapStruct *));
#define STACKSWAP(ss) \
  { if (SysBase->SoftVer>=36) StackSwap(ss); \
    else {Disable();StackSwapOld(ss);Enable();} }
#endif

#ifdef OBSOLETE
#if PROFILING & PROFILING_PROPER_TIMES
EXTERN ExprPtr currentProf;
#endif
#endif

#undef GETSTATS

#ifdef GETSTATS
typedef struct stackStat {
  INT count;
  clock_t time;
  clock_t lastentered;
} stackStat;

stackStat *stackStats = NULL;
#endif

#define MAX_BLOCK_SIZE 1024

typedef struct block1 {
  char *sp;
  struct block1 *prev;
} block1;

typedef struct block {
  struct block *next; /* going up (from most recent to oldest) */
  struct block *prev; /* going down */
  block1 *b1;
  char *end;
#if BLOCK_GAP!=0
  char filler[BLOCK_GAP];
#endif
  struct bcontents {
    char data[MAX_BLOCK_SIZE];
  } b;
} block;

#ifdef HAS_PROTOTYPES
#include "sstack.h"
#endif

#ifdef sparc
#define SPARC_SYSTEM
#endif

#ifdef __sparc__
#define SPARC_SYSTEM
#endif

#undef DEBUG
#undef DEBUG2
#undef DEBUG3

#define ALLOCTMP_FREE

#define __SEG__ STACK

EXTERN INT stackAlloced;
INT alloctmpBytes = 0;

PUBLIC INT currentAge = 0L;
EXTERN ExprPtr futureException;

PRIVATE block *alloctmp_block = (block *)0L;

#define BLOCK1_SIZE (MODALIGN(sizeof(block1)+ALIGNMENT-1))

PUBLIC VOIDPTR alloctmp(n)
long n; /* note that this code should remain essentially the same as the following one,
	   for alloctmp_anchored()! */
{
  REGISTER block1 *p1 = NULL;
  REGISTER block *p;

  while ((p = alloctmp_block)!=(block *)0)
    {
      p1 = p->b1;
unstack:
      if (p1!=(block1 *)0)
	{
	  if (p1->sp
#ifdef ASCENDING_STACK
	      <=
#else
	      >=
#endif
	      (char *)&n)
	    {
	      p->b1 = p1;
	      break;
	    }
	  p1 = p1->prev;
	  goto unstack;
	}
      alloctmp_block = p->next;
#ifdef ALLOCTMP_FREE
      alloctmpBytes -= p->end - (char *)p;
      MARKFREE(p);
#endif
    }
  if (n==0)
    return (VOIDPTR)NULL;
  n = MODALIGN(n+ALIGNMENT-1);
  {
    REGISTER char *b;
    REGISTER block1 *p2;

    if (p==(block *)0 ||
	(b = ((char *)p1)+BLOCK1_SIZE,
	 p2 = (block1 *)(b+n),
	 ((char *)p2)+BLOCK1_SIZE>p->end))
      {
	if (n+BLOCK1_SIZE>MAX_BLOCK_SIZE)
	  {
	    INT size = VARSIZE(block,b.data,n+BLOCK1_SIZE);

	    p = (block *)MALLOC(size);
	    alloctmpBytes += size;
	    p->end = &p->b.data[n+BLOCK1_SIZE];
	  }
	else
	  {
	    p = (block *)MALLOC(sizeof(block));
	    alloctmpBytes += sizeof(block);
	    p->end = &p->b.data[MAX_BLOCK_SIZE];
	  }
	p->next = alloctmp_block;
	alloctmp_block = p;
	b = p->b.data;
	p2 = (block1 *)(b+n);
	p1 = (block1 *)0;
      }
    p2->sp = (char *)&n;
    p2->prev = p1;
    p->b1 = p2;
    return (VOIDPTR)b;
  }
}

PUBLIC VOIDPTR alloctmp_anchored(n,anchor)
REGISTER long n; /* note that this code should remain essentially the same as the previous one,
	   for alloctmp()! */
REGISTER char *anchor;
{
  REGISTER block1 *p1 = NULL;
  REGISTER block *p;

  while ((p = alloctmp_block)!=(block *)0)
    {
      p1 = p->b1;
unstack:
      if (p1!=(block1 *)0)
	{
	  if (p1->sp
#ifdef ASCENDING_STACK
	      <=
#else
	      >=
#endif
	      anchor)
	    {
	      p->b1 = p1;
	      break;
	    }
	  p1 = p1->prev;
	  goto unstack;
	}
      alloctmp_block = p->next;
#ifdef ALLOCTMP_FREE
      alloctmpBytes -= p->end - (char *)p;
      MARKFREE(p);
#endif
    }
  if (n==0)
    return (VOIDPTR)NULL;
  n = MODALIGN(n+ALIGNMENT-1);
  {
    REGISTER char *b;
    REGISTER block1 *p2;

    if (p==(block *)0 ||
	(b = ((char *)p1)+BLOCK1_SIZE,
	 p2 = (block1 *)(b+n),
	 ((char *)p2)+BLOCK1_SIZE>p->end))
      {
	if (n+BLOCK1_SIZE>MAX_BLOCK_SIZE)
	  {
	    INT size;

	    size = VARSIZE(block,b.data,n+BLOCK1_SIZE);
	    p = (block *)MALLOC(size);
	    alloctmpBytes += size;
	    p->end = &p->b.data[n+BLOCK1_SIZE];
	  }
	else
	  {
	    p = (block *)MALLOC(sizeof(block));
	    alloctmpBytes += sizeof(block);
	    p->end = &p->b.data[MAX_BLOCK_SIZE];
	  }
	p->next = alloctmp_block;
	alloctmp_block = p;
	b = p->b.data;
	p2 = (block1 *)(b+n);
	p1 = (block1 *)0;
      }
    p2->sp = anchor;
    p2->prev = p1;
    p->b1 = p2;
    return (VOIDPTR)b;
  }
}

PUBLIC void freetmp(b)
IN CONST VOIDPTR b;
{
  REGISTER block1 *p1,*p0;
  REGISTER block *p;

  p0 = (block1 *)(((char *)b)-BLOCK1_SIZE);
  while ((p = alloctmp_block)!=(block *)0)
    {
      p1 = p->b1;
unstack:
      if (p1!=(block1 *)0)
	{
	  if (p1==p0)
	    {
	      p->b1 = p1;
	      return;
	    }
	  p1 = p1->prev;
	  goto unstack;
	}
      alloctmp_block = p->next;
      alloctmpBytes -= p->end - (char *)p;
      if (b==(VOIDPTR)p->b.data)
	{
	  MARKFREE(p);
	  return;
	}
      MARKFREE(p);
    }
  abort();
}

PRIVATE void alloctmpFree(blocks)
IN REGISTER block *blocks;
{
  REGISTER block *p;

  while (blocks!=(block *)0L)
    {
      p = blocks->next;
      alloctmpBytes -= blocks->end - (char *)blocks;
      MARKFREE(blocks);
      blocks = p;
    }
}

#if OS==AMIGA_OS
PRIVATE void changeStackBounds(s)
IN CONST struct StackSwapStruct *s;
{
  struct Task *myself;

  myself = SysBase->ThisTask;
  myself->tc_SPLower = s->stk_Lower;
  myself->tc_SPUpper = (APTR)s->stk_Upper;
}
#endif

#ifdef SPARC_SYSTEM
#ifdef __GNUC__
#define FLUSH_REGS() __asm__("t 0x3")
#else
#define FLUSH_REGS() (void) sparcFlushRegs(32)
  /* this 32 number is pure guessing. */

PRIVATE int sparcFlushRegs(n)
int n;
{
  int a; /* used so that the compiler does not optimize the recursion away. */

  if (n==0)
    return 0;
  a = sparcFlushRegs(n-1);
  return a+1;
}
#endif
#else
#define FLUSH_REGS()
#endif

PUBLIC COUNTER nThreads=DEFAULT_THREADS,threadSize=STACK_SIZE;
PUBLIC stackMark *stackMarks=(stackMark *)0L;
PUBLIC stackMark *currentThread;
PRIVATE stackMark *firstAllocThread; /* first allocated (stack!=NOEXPR) thread
					after currentThread, or currentThread
					if none. */

PUBLIC char *lowMark = NULL; /* used to test for stack full conditions */
/* We might use currentThread->lowMark instead, but we prefer to install
   an hysteresis between CallCC and Throw so as to avoid floundering. */
#ifdef ASCENDING_STACK
#define SETLOWMARK(mark)  lowMark = mark->lowMark - HYSTERESIS
#else
#define SETLOWMARK(mark)  lowMark = mark->lowMark + HYSTERESIS
#endif

/* The following values obtained by running himml itself,
   executing 'irand 20000' 64 times. */
#define NH 32
PRIVATE COUNTER hysteresis[NH] =
  { 13958, 16053, 5182, 4622, 19305, 5984, 1794, 10067,
    15267, 17833, 12722, 10130, 10365, 9245, 6002, 11968,
    3589, 12095, 9996, 2898, 19512, 18490, 12004, 7178,
    19993, 7500, 5796, 6257, 13363, 18121, 7784, 8692
  };
/*{ 0, 600, 300, 450, 150 };*/
PRIVATE COUNTER nhyst = 0;
#define HYSTERESIS hysteresis[(++nhyst==NH)?(nhyst=0):nhyst]

EXTERN ExprPtr basisEnv;
EXTERN gcStackView *theStackFrames,*theHeapFrames;
EXTERN ExprPtr catchException,nonsenseException;

/*PRIVATE jmp_buf topStackBuf;*/
PRIVATE jmp_buf botStackBuf;

PRIVATE StackChunk *VOLATILE theChunk;
PRIVATE stackDesc theStackDesc;

/* the increment by which we progressively deepen the stack: */
#define STACK_INC 100

PRIVATE NORETURN void establishTopStack P((IN COUNTER n,
					   IN loweredStackAction fun,
					   IN char *data)) _NORETURN;

typedef struct defeatOptimizer {
  struct defeatOptimizer *next;
  char reserve[STACK_INC];
} defeatOptimizer;

PUBLIC defeatOptimizer *aGlobalVariableJustToDefeatTheOptimizerInStackC;
/* Some of the following functions declare a local array
   reserve[STACK_INC], just to eat up some stack space.
   At least two compilers (SAS/C 6.55 on Amiga and gcc >=2.7.0) are able to
   detect that we never do anything with this array, and
   does not allocate space on the stack for this array.  They
   in fact manage to detect that the recursion we use in
   establishBotStack() is in fact terminal.  This defeats
   the whole purpose of this function.
   Therefore, we store the local array reserve in this global
   variable, so that the optimizer cannot possibly know how
   to optimize it.
   It is however still necessary to disable all function inlining
   switches when compiling this file, hence the STACK_CFLAGS / CFLAGS
   dichotomy in OPTIONS (resp. the NOOPTNILINE option in smakefile,
   for SAS C).
*/

#define RESERVE_STACK(size) \
  struct _r { defeatOptimizer *next; char reserve[size];} _r; \
  _r.next = aGlobalVariableJustToDefeatTheOptimizerInStackC; \
  aGlobalVariableJustToDefeatTheOptimizerInStackC = (defeatOptimizer *)&_r

PUBLIC void lockThread(mark) /* signal a thread as unswappable.  The thread must not
			      be the current one, and must not be swapped either already. */
IN CONST stackMark *mark;
{
  stackMark *next,*prev;

  if (mark==currentThread)
    berror(message(667));
  next = mark->next;
  prev = mark->prev;
  next->prev = prev;
  prev->next = next;
}

PUBLIC void unlockThread(mark) /* signal a thread back as swappable.  The thread must be
				  declared as swapped. */
IN CONST stackMark *mark;
{
  if (mark==currentThread)
    {
      /* we use the old value of mark->next to reinsert mark in the circular list.
	 This fails pathetically if mark->next is also swapped out.
	 We should not have problems, because this is really done only to swap out
	 *one* thread at any given time (the toplevel thread) to ensure that we
	 can Throw() safely to it in case of memory overflow.
	 But it is unsafe to lock too many threads in any case. */
      stackMark *next,*prev;

      next = mark->next;
      prev = next->prev;
      mark->prev = prev;
      prev->next = mark;
      next->prev = mark;
    }
  else
    {
      /* insert mark just after current thread. */
      mark->next = currentThread->next;
      mark->prev = currentThread;
      currentThread->next->prev = mark;
      currentThread->next = mark;
    }
}

PRIVATE void flushThread(mark) /* copy a thread to the heap */
IN CONST stackMark *mark;
{
  ExprPtr e;
  REGISTER StackChunk *chunk;
  INT size;

  e = mark->stack;
  if (e!=NOEXPR)
    {
      chunk = SCHUNK(e);
      if (chunk->contents==(char *)0L || chunk->contents==(char *)1L)
	{
	  FLUSH_REGS();
#ifdef ASCENDING_STACK
	  size = chunk->top-mark->highMark;
	  chunk->contents = (char *)MALLOC(size);
	  safememcpy(chunk->contents,mark->highMark,size);
#else
	  size = mark->highMark-chunk->top;
	  chunk->contents = (char *)MALLOC(size);
	  safememcpy(chunk->contents,chunk->top,size);
#endif
	}
      mark->stack = NOEXPR;
    }
}

PRIVATE void useThread(mark) /* mark thread as most recently used,
			      and put it in currentThread. */
IN CONST stackMark *mark;
{
  if (mark!=currentThread)
    /* currentThread is always the most recently used already */
    {
      if (firstAllocThread==mark)
	firstAllocThread = mark->next;
      if (currentThread->next!=mark)
	{
	  stackMark *left,*right;

	  /* first, get mark out of the list */
	  left = mark->prev;
	  right = mark->next;
	  left->next = right;
	  right->prev = left;
	  /* then, reinsert mark into the list */
	  left = currentThread;
	  right = left->next;
	  left->next = mark;
	  mark->next = right;
	  right->prev = mark;
	  mark->prev = left;
	}
#ifdef GETSTATS
      {
	COUNTER n;
	struct tms t;

	n = currentThread-stackMarks;
	(void) times(&t);
	stackStats[n].time += t.tms_utime - stackStats[n].lastentered;
	n = mark-stackMarks;
	stackStats[n].count++;
	stackStats[n].lastentered = t.tms_utime;
      }
#endif
      currentThread = mark;
      {
	ExprPtr stack;

	stack = mark->stack;
	if (stack!=NOEXPR)
	  {
	    HITBARRIER(stack);
	    /* Because the stackDesc SJUMBUF(stack) may contain stackFrames
	       or heapFrames that vary each time we compute under this stack.
	       We have to inform the GC of any possible change in there. */
	  }
      }
    }
}

PRIVATE VOLATILE loweredStackAction lowFun;
PRIVATE char *VOLATILE lowData;

PRIVATE NORETURN void establishLowStack P((IN loweredStackAction fun,
					   IN char *data)) _NORETURN;

PRIVATE NORETURN void establishLowStack(fun,data)
IN loweredStackAction fun;
IN char *data;
{
#if OS==AMIGA_OS
  struct StackSwapStruct aux;

  lowFun = fun;
  lowData = data;
  aux = initStack;
  STACKSWAP(&aux);
  (*lowFun)(lowData);
#else
  RESERVE_STACK(SECURITY);
  (*fun)(data);
#endif
  ml_done(0);
  /*NOTREACHED*/
}

PRIVATE NORETURN void establishBotStack P((IN COUNTER n,
					   IN loweredStackAction fun,
					   IN char *data)) _NORETURN;

PRIVATE NORETURN void establishBotStack(n,fun,data)
IN COUNTER n;
IN loweredStackAction fun;
IN char *data;
{
  RESERVE_STACK(STACK_INC);

#ifdef ASCENDING_STACK
  if (_r.reserve+STACK_INC+SECURITY<=stackMarks[n].lowMark)
    {
      establishBotStack(n,fun,data); /* recurse until enough stack depth */
#ifndef HAS_NORETURN
      return;
#endif
    }
  stackMarks[n].lowMark = _r.reserve+STACK_INC-SAFETY_SIZE;
#else
  if (_r.reserve-SECURITY>=stackMarks[n].lowMark)
    {
      establishBotStack(n,fun,data); /* recurse until enough stack depth */
#ifndef HAS_NORETURN
      return;
#endif
    }
  stackMarks[n].lowMark = _r.reserve+SAFETY_SIZE;
#endif
  if (n!=0)
    {
      establishTopStack(n,fun,data); /* create other threads */
      /*NOTREACHED*/
    }
  else switch (SETJMP(botStackBuf))
    {
    case 0:
      establishLowStack(fun,data);
      /*NOTREACHED*/
      break;
    case 1: /* Throw inside the same stackMark */
      /* don't use any locals, and don't simply return ! */
      /* the reason is that this portion of code is called in
	 an artificially lowered stack (thus no stack is really
	 present yet).
	 */
      FLUSH_REGS();
#ifdef ASCENDING_STACK
      safememcpy(theChunk->thread->highMark,theChunk->contents,
		 theChunk->top-theChunk->thread->highMark);
#else
      safememcpy(theChunk->top,theChunk->contents,
		 theChunk->thread->highMark-theChunk->top);
#endif
      /* smash stack with contents of the chunk */
      LONGJUMP(theStackDesc,1);
      /*NOTREACHED*/
      break;
    case 2: /* change thread inside the same stackMark */
      flushThread(currentThread);
      /*useThread(currentThread); // useless */
      SETLOWMARK(currentThread);
      LONGJMP(currentThread->topBuf,1);
      /*NOTREACHED*/
      break;
    default:
      /*NOTREACHED*/
      abort();
      break;
    }
}

PRIVATE VOLATILE compiledAction theAction;
PRIVATE VEXPRPTR theEnv;
PRIVATE VEXPRPTR theStack;

PRIVATE NORETURN void TopOfThread P((IN compiledAction action,
				     IN CONST ExprPtr env,
				     IN CONST ExprPtr stack)) _NORETURN;

PRIVATE NORETURN void TopOfThread(action,env,stack)
IN compiledAction action;
IN CONST ExprPtr env;
IN CONST ExprPtr stack;
{
  Throw(stack,(*action)(ENV_CONS(env,1), stack));
  /*NOTREACHED*/
}

PRIVATE VOLATILE mlAction theFun;
PRIVATE char *VOLATILE theData;
PRIVATE stackDesc theSd;

PRIVATE NORETURN void doDoML P((IN mlAction action,
				IN CONST char *data,
				IN stackDesc *sd)) _NORETURN;

PRIVATE NORETURN void doDoML(action,data,sd)
IN mlAction action;
IN CONST char *data;
IN stackDesc *sd;
{
  stackDesc sdd;

  sdd = *sd;
  (*action)(data);
#if OS==AMIGA_OS
  Disable();
  changeStackBounds(&initStack);
#endif
  LONGJUMP(sdd,1);
  /*NOTREACHED*/
}

PRIVATE NORETURN void establishTopStack(n,fun,data)
IN COUNTER n;
IN loweredStackAction fun;
IN char *data;
{
  stackMark *mark;
  RESERVE_STACK(SECURITY);

  n--;
  mark = &stackMarks[n];
#if OS==AMIGA_OS
  mark->highMark = _r.reserve;
  mark->lowMark = ((char *)MLthreads)+n*(threadSize+SAFETY_SIZE);
#else
#ifdef ASCENDING_STACK
  mark->highMark = _r.reserve+SECURITY;
  mark->lowMark = _r.reserve+SECURITY+(threadSize+SAFETY_SIZE);
#else
  mark->highMark = _r.reserve;
  mark->lowMark = _r.reserve-(threadSize+SAFETY_SIZE);
#endif
#endif
  /* make circularly linked list of threads */
  mark->next = (n==0)?(&stackMarks[nThreads-1]):(mark-1);
  mark->prev = (n==nThreads-1)?stackMarks:(mark+1);
  mark->stack = NOEXPR;

  switch (SETJMP(mark->topBuf))
    {
    case 0:
      establishBotStack(n,fun,data);
      /*NOTREACHED*/
      break;
    case 1:
      /* don't use any locals, and don't simply return ! */
      /* the reason is that this portion of code is called in
	 a possibly lowered stack (thus no stack is really
	 present yet);
	 */
      TopOfThread(theAction,theEnv,theStack);
      /*NOTREACHED*/
      break;
    case 2: /* called by doML */
#if OS==AMIGA_OS
      Enable();
#endif
      doDoML(theFun,theData,&theSd);
      /*NOTREACHED*/
      break;
    default:
      /*NOTREACHED*/
      abort();
      break;
    }
  /*NOTREACHED*/
}

PRIVATE VEXPRPTR theThrownExpr;

PUBLIC NORETURN void Throw1(stack,e,sd)
IN CONST ExprPtr stack;
IN CONST ExprPtr e;
IN CONST stackDesc *sd;
{
  StackChunk *chunk;
  stackMark *mark,*here;
#define sdd *((sd==(stackDesc *)0L)?SJMPBUF(stack):sd)
  chunk = SCHUNK(stack);
  if (chunk==(StackChunk *)0L)
    {
      (void) Raise(nonsenseException);
      /* This stack is fake (it was attempted to be reloaded
	 from a file, but this cannot be done) */
#ifndef HAS_NORETURN
      return;
#endif
    }
  if (chunk->age>currentAge)
    {
      (void) Raise(futureException);
      /* We don't allow returns to the future, because the global
	 environment is physically modified, so this would allow
	 to travel along different time lines, but with the same
	 structure for the global environment. We could solve this
	 by regenerating a new global environment each time we
	 evaluate a toplevel declaration, but this would be far
	 too costly. */
#ifndef HAS_NORETURN
      return;
#endif
    }
  mark = chunk->thread;
  if (mark==(stackMark *)0L)
    {
      (void) Raise(catchException); /* Throw() is illegal */
#ifndef HAS_NORETURN
      return;
#endif
    }
  setHandlerStack(chunk->excStack);
  chunk->excStack = NULL;
  setProtectWrapper(chunk->protected,e);
#if DEBUGGING
#ifdef OBSOLETE
  setApplStk(chunk->applStack);
#endif
#endif
  { /* resynchronize the alloctmp_blocks pointers */
    ExprPtr e;

    e = currentThread->stack;
    /* we store alloctmp_block inside the current chunk. If there is none,
       then this is a definite throw out of the current stack, so we may
       free the blocks (case e==NOEXPR). */
    if (e==NOEXPR)
      alloctmpFree(alloctmp_block);
    else
      SCHUNK(e)->alloctmp_blocks = (char *)alloctmp_block;
    /* then, we get the new alloctmp_block from where it was saved,
       in chunk->alloctmp_blocks. */
    alloctmp_block = (block *)chunk->alloctmp_blocks;
    chunk->alloctmp_blocks = (char *)0L;
    /* we finally reset chunk->alloctmp_blocks, because in case the continuation
       was captured by catch and it was not swapped out to the heap, we just
       free the chunk (by stackChunkFree() below), and this also deallocates
       the alloctmp_blocks field; we don't want this, as this would deallocate
       the global alloctmp_block variable. */
  }
  here = currentThread;
  useThread(mark);
  SETLOWMARK(mark);
  theThrownExpr = e;

  if (mark->stack!=stack) /* if stack was swapped out to the heap */
    {
      flushThread(mark);
      if (here==mark) /* in this case, we must swap ourselves out,
			 then back in; to do this, we longjmp() to
			 botStackBuf, where we swap the thread in,
			 and do the final LONGJUMP(). */
	{
	  theChunk = chunk;
	  theStackDesc = sdd;
	  SJMPBUF(stack)->stack_frames = (gcStackView *)0L;
	  SJMPBUF(stack)->heap_frames = (gcStackView *)0L;
	  LONGJMP(botStackBuf,1);
	  /*NOTREACHED*/
	}
      /* otherwise, swap the saved stack in */
      FLUSH_REGS();
#ifdef ASCENDING_STACK
      safememcpy(mark->highMark,chunk->contents,
		 chunk->top-mark->highMark);
#else
      safememcpy(chunk->top,chunk->contents,
		 mark->highMark-chunk->top);
#endif
    }
  else if (chunk->contents==(char *)0L) /* if stack not flushed, and was
					   captured with callcc, not catch,
					   then flush it for possible reuse. */
    flushThread(mark);
  else
    stackChunkFree(chunk); /* otherwise, make subsequent Throws illegal
			      and free the chunk */
  theStackFrames = (sdd).stack_frames;
  SJMPBUF(stack)->stack_frames = (gcStackView *)0L;
  theHeapFrames = (sdd).heap_frames;
  SJMPBUF(stack)->heap_frames = (gcStackView *)0L;
  LONGJMP((sdd).buf,1);
  /*NOTREACHED*/
#undef sdd
}

PRIVATE NORETURN void withNewThread(action,oldstack,env)
IN compiledAction action;
IN CONST ExprPtr oldstack;
IN CONST ExprPtr env;
{
  stackMark *mark;

  SCHUNK(currentThread->stack)->top = (char *)&mark;
  mark = currentThread->next;
  theAction = action;
  theEnv = env;
  theStack = oldstack;
  if (mark==currentThread) /* then we must swep ourselves out and then back in */
    {
      LONGJMP(botStackBuf,2);
      /*NOTREACHED*/
    }
  else /* otherwise, flush the new thread and evaluate the action at the top
	  of the emptied thread. */
    {
      flushThread(mark);
      useThread(mark);
      SETLOWMARK(mark);
      LONGJMP(mark->topBuf,1);
      /*NOTREACHED*/
    }
}

#ifdef DEBUG2
#include "prtproto.h"
#endif

PUBLIC ExprPtr CallCC(action,env,catchp)
IN compiledAction action;
IN CONST ExprPtr env;
IN Boolean catchp; /* true if continuation can be destroyed after Throw() */
{
  REGISTER stackMark *mark;
  REGISTER ExprPtr stack;
  REGISTER StackChunk *chunk;
  stackDesc *new;
  excHandler *xh;
#ifdef OBSOLETE
#if PROFILING & PROFILING_PROPER_TIMES
  VOLATILE ExprPtr saveCurrentProf = currentProf;
#endif
#endif

#ifdef DEBUG3
  hfputs("CallCC!\n",hstderr);
  hfflush(hstderr);
#endif
  /* if we raise an exception that is defined only in a given
     stack, we must ensure that raising it first reinstalls
     the correct thread first. This is why we record it here. */
  xh = (excHandler *)alloctmp(sizeof(excHandler));
  xh->next = NULL;
  xh->tag = X_CONT;
  xh->data = NIL; /* will be patched by stack below. */
  /* This new excHandler xh must be allocated in the alloctmp_blocks
     of the parent thread, since Raise() will sometimes want
     to Throw1(_,_,jb) with a jb that is inside xh; as Throw1
     may deallocate the alloctmp_blocks of the child, this
     would destroy jb as well, causing subtle bugs. */
  mark = currentThread;
  VECALLOC(chunk,sizeof(StackChunk));
  /* build dummy stack chunk; we shall build the real one
     if we need to swap it out of the C stack. */
  chunk->thread = currentThread;
  chunk->contents = catchp?(char *)1:(char *)0;
  chunk->excStack = excStack;
  chunk->alloctmp_blocks = (char *)alloctmp_block;
  alloctmp_block = (block *)0L;
  excStack = NULL;
#if DEBUGGING
#ifdef OBSOLETE
  chunk->applStack = getApplStk();
#endif
#endif
  chunk->protected = getProtectWrapper();
  chunk->age = currentAge;
  /* now, build an E_STACK ExprPtr */
  CHECKHETEROFREE();
#if STOP_AND_COPY_VECTORS
  scAlloc(*(char **)&new,sizeof(stackDesc));
#else
  new = (stackDesc *)MALLOC(sizeof(stackDesc));
  /* cannot use scAlloc, since we need new to be aligned. */
#endif
  EALLOC_HETERO_NOGC(stack);
  stack->kind = (CHAR)E_STACK;
  stackAlloced++;
  new->stack_frames = theStackFrames;
  new->heap_frames = theHeapFrames; /* we do this here to keep GC happy in
				       case the PUSHQ below wanted to trigger one. */
  theStackFrames = (gcStackView *)0L;
  theHeapFrames = (gcStackView *)0L;
  SJMPBUF(stack) = new;
  SCHUNK(stack) = chunk;
  mark->stack = stack;
  (void) KEEP(stack);
  xh->data = stack; /* patch data field. */
  excStack = xh; /* and install new excStack */
  new = SJMPBUF(stack);
  if (SETJMP(new->buf)==0) /* no need to SETJUMP(), since stack_frames and heap_frames
			      were set up above. */
    { /* change thread before calling action,
	 and reestablish thread afterwards. */
      withNewThread(action,stack,env);
      /*NOTREACHED*/
      /* This function never returns, since all returns are done
	 through Throw(). This is why TopOfThread() does an extra
	 Throw() around the call to (*action)(ENV_CONS(env,1),stack). */
    }
#ifdef OBSOLETE
#if PROFILING & PROFILING_PROPER_TIMES
  syncProfProper(saveCurrentProf);
#endif
#endif
  return theThrownExpr;
}

PRIVATE NORETURN void establishStack P((IN loweredStackAction fun, IN char *data)) _NORETURN;

PRIVATE void establishStack(fun,data)
IN loweredStackAction fun;
IN char *data;
{
  lowFun = fun;
  lowData = data;
#if OS==MAC
  /* a Mac has a descending stack and a limit to the stack,
   but we can change it to suit our needs, at least as long
   as MaxApplZone() was not called. */
  SetApplLimit(GetApplLimit()-nThreads*(threadSize+SAFETY_SIZE));
#else
#if OS==AMIGA_OS /* an Amiga has a descending stack and a limit, too */
  { /* but we can allocate the ML stacks in a separate memory block */
    unsigned long size;

    size = nThreads*(threadSize+SAFETY_SIZE);
    MLthreads = AllocMem(size,MEMF_CLEAR|MEMF_PUBLIC);
    if (MLthreads==NULL)
      berror(message(479));
    MLstacks.stk_Lower = MLthreads;
    MLstacks.stk_Pointer = MLthreads+size;
    MLstacks.stk_Upper = (ULONG)MLstacks.stk_Pointer;
    initStack = MLstacks;
  }
  STACKSWAP(&initStack); /* also puts initial stack into initStack */
#endif
#endif
  establishTopStack(nThreads,lowFun,lowData);
}

PUBLIC void doML(fun,data) /* computes fun(data) in the ML stack */
IN mlAction fun;
INOUT char *data;
{
  theStackFrames = (gcStackView *)0L;
  theHeapFrames = (gcStackView *)0L;
  theFun = fun;
  theData = data;
  if (SETJUMP(theSd)==0)
    {
#if OS==AMIGA_OS
      Disable();
      changeStackBounds(&MLstacks);
#endif
      LONGJMP(currentThread->topBuf,2); /* current thread should be empty */
    }
#if OS==AMIGA_OS
  Enable();
#endif
}

PUBLIC void stackChunkFree(chunk)
INOUT StackChunk *chunk;
{
  alloctmpFree((block *)chunk->alloctmp_blocks);
  chunk->alloctmp_blocks = (char *)0L;
  {
    REGISTER stackMark *mark;
    REGISTER ExprPtr e;

    if (chunk->contents!=(char *)0L && chunk->contents!=(char *)1L)
      {
	MARKFREE(chunk->contents);
	chunk->contents = (char *)0L;
      }
    mark = chunk->thread;
    if (mark!=(stackMark *)0L)
      {
	e = mark->stack;
	if (e!=NOEXPR && SCHUNK(e)==chunk)
	  {
	    mark->stack = NOEXPR;
	    /* now, put the freed thread just after the last free thread after
	       currentThread, to say it was the least recently used, except
	       possibly for already free threads. This means we put mark just before
	       firstAllocThread, and advance the latter. */
	    if (firstAllocThread!=mark && firstAllocThread->prev!=mark)
	      {
		stackMark *left,*right;

		/* first, get mark out of the list */
		left = mark->prev;
		right = mark->next;
		left->next = right;
		right->prev = left;
		/* then, reinsert mark into the list */
		right = firstAllocThread;
		left = right->prev;
		left->next = mark;
		mark->next = right;
		right->prev = mark;
		mark->prev = left;
	      }
	    firstAllocThread = mark->next;
	  }
	chunk->thread = (stackMark *)0L;
      }
  }
}

#undef __SEG__
#define __SEG__ INIT

PUBLIC void InitStack(fun,data)
IN loweredStackAction fun;
IN char *data;
{
  stackMarks = (stackMark *)NEWPTR(nThreads*sizeof(stackMark));
  if (stackMarks==(stackMark *)0L)
    berror(message(480));
#ifdef GETSTATS
  stackStats = (stackStat *)NEWPTR(nThreads*sizeof(stackStat));
  {
    COUNTER i;

    for (i=0; i<nThreads; i++)
      {
	stackStats[i].count = 0;
	stackStats[i].time = 0;
      }
  }
#endif
  currentThread = &stackMarks[nThreads-1]; /* topmost one */
  lowMark = currentThread->lowMark;
#ifdef GETSTATS
    {
      struct tms t;

      stackStats[nThreads-1].count = 1;
      (void) times(&t);
      stackStats[nThreads-1].lastentered = t.tms_utime;
    }
#endif
  firstAllocThread = currentThread;
  establishStack(fun,data);
}

PUBLIC void EndStack(n)
int n;
{
  if (stackMarks!=(stackMark *)0L)
    {
      DISPOSPTR(stackMarks,nThreads*sizeof(stackMark));
      stackMarks = (stackMark *)0L;
    }
#if OS==AMIGA_OS
  {
    PRIVATE VOLATILE int theN;

    theN = n;
    if (MLthreads!=NULL)
      {
	STACKSWAP(&initStack);
	FreeMem(MLthreads,nThreads*(threadSize+SAFETY_SIZE));
      }
    exit(theN);
  }
#else
  exit(n);
#endif
}
