/* Initialization of types.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <math.h>
#include "eval.h"
#include "type.h"
#include "expr.h"
#include "typproto.h"
#ifdef HAS_PROTOTYPES
#include "stypinit.h"
#endif

PUBLIC ExprPtr basisAssum,basisState;

PUBLIC ExprPtr stringType,numType,intType,boolType,exnType,powerType,dynType,timeType,
  iarrayType;
PUBLIC ExprPtr tdgAlt;
PUBLIC ExprPtr tdgOrHashTable,tdgEquivHashTable;
	/* memoizing tables for tdgOr and tdgEquiv */

#if IMPERATIVE_TYPES==EFFECT_TYVAR
EXTERN ExprPtr theFx; /* contains the computed effect of the expression */
PUBLIC MapPtr autoDecorations; /* map from tynames to effects or regions automatically added at instantiation time */
PUBLIC MapPtr autoPatFx; /* map from 'init' to 'read', that is from
	effects of functions to effects of the same symbols as pattern
	constructors */
#endif

PUBLIC ExprPtr unitset;

#define __SEG__ INIT

/* as during initialization, there cannot be any garbage collection,
   KEEP() is useless */
#undef KEEP
#define KEEP(x) x
#define KONS Kons

PRIVATE ExprPtr KSYMBOL(n)
IN COUNTER n;
{
  ExprPtr sym;

  sym = Ksymbol(message(n));
  return KEEP(sym);
}

PUBLIC void InitTypes()
{
  ExprPtr alphaarg,alphaeqarg,_alphaarg,alphahdr,_alphahdr,alphaeqhdr;
  ExprPtr ref_alpha,array_alpha,promisealpha,contalpha,listalpha,setalphaeq,
    alphaeqmapbeta,alphaeqmapbetaeq,
    _alphaeq_beta_hdr, _alphaeq_beta_arg;
  ExprPtr refcons,listcons,boolcons;
  ExprPtr aux,aux2;
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  ExprPtr aux3,refrhoalpha;
#endif
  ENTER;

  unitset = Singleton(UNIT);
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  theFx = NewMap();
#endif
  moduleName = NIL;
  stringType = KtypeAppl(stringSym,NIL);
  numType = KnumType(NewMap());
  intType = KtypeAppl(intSym,NIL);
  boolType = KtypeAppl(boolSym,NIL);
  exnType = KtypeAppl(exnSym,NIL);
  powerType = funType(KONS(numType,KONS(numType,NIL)),numType);
  dynType = KtypeAppl(dynSym,NIL);
  iarrayType = KtypeAppl(iarraySym,NIL);
  tdgAlt = Kons(falseValue,trueValue);

  alphaarg = KONS(alpha,NIL);
  _alphaarg = KONS(_alpha,NIL);
  alphaeqarg = KONS(alphaeq,NIL);
  alphahdr = KONS(alphaarg,MapSingleton(alpha,falseValue));
  _alphahdr = KONS(_alphaarg,MapSingleton(_alpha,falseValue));
  alphaeqhdr = KONS(alphaeqarg,MapSingleton(alphaeq,trueValue));
  ref_alpha = KtypeAppl(refSym,_alphaarg);
  (void) KEEP(ref_alpha);
  _alphaeq_beta_arg = KONS(_alphaeq, KONS (_beta, NIL));
  _alphaeq_beta_hdr = KONS(_alphaeq_beta_arg, MapPair(_alphaeq,trueValue,_beta,falseValue));
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  aux = KtypeFx(refSym,genRgn);
  (void) KEEP(aux);
  refrhoalpha = KtypeAppl(aux,alphaarg);
  (void) KEEP(refrhoalpha);
#endif
  array_alpha = KtypeAppl(arraySym,_alphaarg);
  (void) KEEP(array_alpha);
  promisealpha = KtypeAppl(promiseSym,alphaarg);
  (void) KEEP(promisealpha);
  contalpha = KtypeAppl(contSym,alphaarg);
  (void) KEEP(contalpha);
  listalpha = listType(alpha);
  setalphaeq = setType(alphaeq);
  alphaeqmapbeta = mapType(alphaeq,beta);
  alphaeqmapbetaeq = mapType(alphaeq,betaeq);
  tdgOrHashTable = MakeArray(4*3*TDG_OR_HASH_SIZE,NIL);
  tdgEquivHashTable = MakeArray(2*3*TDG_EQUIV_HASH_SIZE,NIL);

  aux = NewTypeAssum();
  SETQ(basisAssum,aux);
  aux = NewTypeState();
  SETQ(basisState,aux);
  aux2 = KONS(NIL,NewMap());
  aux = MapMerge(MapMerge(MapPair(stringsym,KONS(aux2,stringType),
					  KSYMBOL(240),KONS(aux2,numType)),
				  MapPair(unitsym,KONS(aux2,NIL),
					  listsym,KONS(alphahdr,listalpha))),
			 MapMerge(MapPair(refsym,KONS(_alphahdr,ref_alpha),
					  arraysym,KONS(_alphahdr,array_alpha)),
				  MapPair(promisesym,KONS(alphahdr,promisealpha),
					  contsym,KONS(alphahdr,contalpha))));
  aux = MapMerge(MapMerge(MapPair(boolsym,KONS(aux2,boolType),
				  dynsym,KONS(aux2,dynType)),
			  MapPair(exnsym,KONS(aux2,exnType),
				  KSYMBOL(241),KONS(alphaeqhdr,mapType(alphaeq,NIL)))),
		 MapMerge(MapPair(intsym,KONS(aux2,intType),
				  iarraysym,KONS(aux2,iarrayType)),
			  aux));
  aux = MapMerge(MapPair(outstreamsym,
			 KONS(alphahdr,MapPair(writesym,funType(stringType,NIL),
					       REST_TAG,alpha)),
			 instreamsym,
			 KONS(alphahdr,MapPair(readsym,funType(intType,stringType),
					       REST_TAG,alpha))),
		 aux);
  aux = MapMerge(MapSingleton(tablesym,
			      KONS(_alphaeq_beta_hdr,KtypeAppl(tableSym,_alphaeq_beta_arg))),
		 aux);
  { /* bool = true | false */
    ExprPtr ttm;

    ttm = CAR(trueValue);
    setMapPut(boolSym,KONS(KONS(NIL,NewMap()),
			   MapPair(AVAL(CAR(ttm))->expr[0],boolType,
				   AVAL(CAR(ttm))->expr[1],boolType)),
	      &S_TYCONS(basisState));
  }
  {
    ExprPtr unit,quanname,quan; /* dimension time(s) */

    unit = KSYMBOL(628);
    quan = KSYMBOL(629);
    quanname = KSYMBOL(630);
    timeType = KtypeAppl(quanname,NIL);
	timeType = KnumType(MapSingleton(KEEP(timeType),one));
	aux = MapPut(quan,KONS(KONS(NIL,NewMap()),timeType),aux);
	/* {quan => ((),{%quan() => 1.0})}, with tyvarseq=() */
	setMapPut(unit,KONS(KONS(one,NewMap()),timeType),&A_UNITS(basisAssum));
	setMapPut(quanname,unit,&S_TYCONS(basisState));
  }
  SETQ(A_TYFCNS(basisAssum),aux);
  /* ref => {ref => '_a -> '_a ref} */
  /* in the effect case,
  	ref => {ref => 'a ->[fx] 'a ref[rho] with fx >= 'a init[rho]} */
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  aux2 = KtypeFx(fxInitSym,genRgn);
  (void) KEEP(aux2);
  aux2 = KtypeAppl(aux2,alphaarg);
  (void) KEEP(aux2);
  aux2 = MakePair(E_ARRAY_2,effectSym,Singleton(aux2));
  (void) KEEP(aux2);
  refcons = MapSingleton(refsym,funTypeFx(alpha,refrhoalpha,aux2));
#else
  refcons = MapSingleton(refsym,funType(_alpha,ref_alpha));
#endif
  /* list => {nil => 'a list, :: => 'a * 'a list -> 'a list} */
  listcons = MapPair(nilSym,listalpha,consSym,
			    funType(KONS(alpha,KONS(listalpha,NIL)),listalpha));
  /* bool => {true => bool, false => bool} */
  boolcons = MapPair(trueSym,boolType,falseSym,boolType);
  aux = MapMerge(refcons,MapMerge(listcons,boolcons));
  SETQ(A_CONS(basisAssum),aux);
  aux = MapMerge(MapMerge(MapMerge(MapPair(bindSym,exnType, /* exception Bind */
					   matchSym,exnType), /* exception Match */
				   MapPair(emptySym,exnType, /* exception Empty */
					   mapgetSym,exnType)), /* exception MapGet */
			  MapMerge(MapPair(parsweepSym,exnType, /* exception ParSweep */
					   nomemSym,exnType), /* exception NoMem */
				   MapPair(stackSym,exnType, /* exception Stack */
					   incompSym,exnType))), /* exception Incomparable */
		 MapMerge(MapMerge(MapPair(noninvSym,exnType, /* exception NonOneToOne */
					   nthSym,exnType), /* exception Nth */
				   MapPair(rangeSym,exnType, /* exception Range */
					   asciiSym,exnType)), /* exception Ascii */
			  MapMerge(MapPair(stringnthSym,exnType, /* exception StringNth */
					   nonsenseSym,exnType), /* exception NonSense */
				   MapPair(subscriptSym,exnType, /* exception Subscript */
					   arithSym,exnType)))); /* exception Arith */
  aux = MapMerge(MapPair(ufindSym,exnType, /* exception UFind */
			 catchSym,exnType), /* exception Catch */
		 MapPut(futureSym,exnType, /* exception ReturnToTheFuture */
			aux));
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  aux2 = readFx(genRgn);
  aux2 = MakePair(E_ARRAY_2,effectSym,aux2);
  aux2 = funTypeFx(refrhoalpha,alpha,KEEP(aux2));
  aux3 = writeFx(genRgn);
  aux3 = MakePair(E_ARRAY_2,effectSym,aux3);
  aux3 = funTypeFx(KONS(KtypeAppl(refSym,alphaarg),alphaarg),NIL,KEEP(aux3));
#endif
  aux = MapMerge(MapMerge(MapPair(ioSym,funType(intType,exnType), /* exception IO of int */
				  reSym,funType(intType,exnType)), /* exception RE of int */
  			  aux),
  		 MapPair(derefSym,
#if IMPERATIVE_TYPES==EFFECT_TYVAR
			aux2,
#else
			funType(KtypeAppl(refSym,alphaarg),alpha),
#endif
  			/* ! : 'a ref -> 'a */
  			/* in the effect case, ! : 'a ref[rho] ->[fx] 'a
  				with fx >= read[rho] */
  			 assignSym,
#if IMPERATIVE_TYPES==EFFECT_TYVAR
			aux3
#else
  			funType(KONS(KtypeAppl(refSym,alphaarg),
  					   alphaarg),NIL)
#endif
  			));
  			/* := : 'a ref * 'a -> unit */
  			/* in the effect case, := : 'a ref[rho] * 'a ->[fx] unit,
  				with fx >= write[rho] */
  aux =
    MapMerge(MapMerge(A_CONS(basisAssum),aux),
	     MapMerge(MapMerge(MapMerge(MapPair(appendSym,funType(listType(listalpha),listalpha),
						/* append : 'a list list -> 'a list */
						dunionSym,funType(mapType(setalphaeq,beta),setalphaeq)),
					/* union : (''a set -m> 'b) -> ''a set */
					MapPair(overwriteSym,funType(listType(alphaeqmapbeta),
								     alphaeqmapbeta),
						/* overwrite : (''a -m> 'b) list -> ''a -m> 'b */
						underwriteSym,funType(listType(alphaeqmapbeta),
								      alphaeqmapbeta))),
			       /* underwrite : (''a -m> 'b) list -> ''a -m> 'b */
			       MapMerge(MapPair(interSym,funType(mapType(alphaeqmapbetaeq,gama),alphaeqmapbetaeq),
						/* inter : ((''a -m> ''b) -m> 'c) -> (''a -m> ''b) */
						identitySym,funType(alpha,alpha)),
					/* identity : 'a -> 'a */
					MapPair(itSym,NIL,
						/* it : (), originally */
						mergeSym,funType(KONS(alphaeqmapbeta,
								      KONS(alphaeqmapbeta,NIL)),
								 alphaeqmapbeta)))),
		      /* ++ : (''a -m> 'b) * (''a -m> 'b) -> (''a -m> 'b) */
		      MapPair(forceSym,
#if IMPERATIVE_TYPES==EFFECT_TYVAR
			      funTypeFx(promiseTypeFx(alpha,pureGenFx),
					alpha,
					pureGenFx),
#else
			      funType(promisealpha,alpha),
#endif
			      /* force : 'a promise -> 'a */
			      /* in the effect case,
				 force : 'a promise[fx] ->[fx] 'a */
			      mapappSym,funType(alphaeqmapbeta,
						funType(alphaeq,beta))
			      )));
  /* ? : (''a -m> 'b) -> ''a -> 'b */
  aux2 = funType(KONS(alphaeq,KONS(alphaeq,NIL)),boolType);
  aux = MapMerge(aux,
		 MapPut(Ksymbol(message(275)),funType(boolType,boolType),
				 MapPair(Ksymbol(message(276)),aux2,
					 Ksymbol(message(277)),aux2)));
  SETQ(A_IDS(basisAssum),aux);
  
  aux = MapMerge(MapPair(promiseSym,falseValue,contSym,falseValue),
		     MapPut(dynSym,falseValue,
			    MapPair(funSym,falseValue,exnSym,falseValue)));
  SETQ(S_NONEQS(basisState),aux);
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  aux = MakePair(E_ARRAY_2,effectSym,NewMap());
  (void) KEEP(aux);
  autoDecorations = MapPair(funSym,aux,promiseSym,aux);
  aux = MakePair(E_ARRAY_2,rgnSym,NewMap());
  (void) KEEP(aux);
  autoDecorations = MapMerge(MapMerge(MapPair(exnSym,aux,contSym,aux),
	  			      MapPair(refSym,aux,arraySym,aux)),
  			     autoDecorations);
  autoPatFx = MapSingleton(fxInitSym,fxReadSym);
#endif
  QUIT;
}

PRIVATE MapPtr getEqConds(tyvarseq)
IN CONST ExprPtr tyvarseq;
{
  MapPtr eqconds;
  ENTER;

  eqconds = NewMap();
  for (; tyvarseq!=NIL; tyvarseq=CDR(tyvarseq))
    eqconds = MapPut(CAR(tyvarseq),falseValue,eqconds);
  KEEPONLY(eqconds);
  return eqconds;
}

PUBLIC void addDataType(name,tyvarseq,type,cons)
	/* adds a datatype that respects equality */
IN CONST ExprPtr name;
IN CONST ExprPtr tyvarseq;
IN CONST ExprPtr type;
IN CONST ExprPtr cons;
{
  ExprPtr tyhdr;
  ENTER;

  setMapMerge(&A_IDS(basisAssum),cons);
  setMapMerge(&A_CONS(basisAssum),cons);
  tyhdr = KONS(tyvarseq,getEqConds(tyvarseq));
  setMapPut(name,KONS(tyhdr,type),&A_TYFCNS(basisAssum));
  setMapPut(fxSym(APPLFUN(type)),KONS(tyhdr,cons),&S_TYCONS(basisState));
  QUIT;
}

PUBLIC void addIdType(name,type)
IN CONST ExprPtr name;
IN CONST ExprPtr type;
{
  setMapPut(name,type,&A_IDS(basisAssum));
}
