/* Union-Find structures (unused).
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "eval.h"
#include "expr.h"
#include "ufind.h"

struct mergeconginfodata {
  infoMergeFunc merge2;
  char *data2;
};

struct killsig {
  ExprPtr pending;
  congPtr cong;
};

#ifdef HAS_PROTOTYPES
#include "sufind.h"
#endif

#define __SEG__ UNIONFIND

/*
(* find : (''a,'b) union_find_tree -> ''a -> ''a *)
(* given a tree, find computes the color of an object *)
fun find (|[link,...]| : (''a,'b) union_find_tree) =
  let fun f x = if x inset link
		    then f (?link(x))
		else x
  in
      f
  end
*/

EXTERN ExprPtr ufindException;

PUBLIC colorPtr uFind1(utree,x,foundp)
IN CONST uFindPtr utree;
IN CONST ExprPtr x;
OUT Boolean *foundp;
{
  REGISTER ExprPtr y,yy;

  y = NGetMaplet(x,utree,foundp);
  if (!(*foundp))
    return NIL;
  while (yy = MAPLETY(y), yy->kind==(CHAR)E_MAPLET)
    y = yy;
  return y;
}

PUBLIC void uMerge(utree,x,y,doinfo,infodata)
IN CONST uFindPtr utree;
IN CONST ExprPtr x,y;
IN CONST infoMergeFunc doinfo;
IN char *infodata;
{
  ExprPtr newi,infox,infoy;
  long nx,ny;
  ENTER;

  if (x==y)
    return;
  infox = MAPLETY(x);
  nx = COUNTEDN(infox);
  infoy = MAPLETY(y);
  ny = COUNTEDN(infoy);
  if (nx<ny)
    {
      newi = (*doinfo)(x,y,COUNTEDVAL(infox),COUNTEDVAL(infoy),infodata);
      SETMAPLETY(x,infoy);
      SETCOUNTEDN(infoy,nx+ny);
      SETREFVAL(infoy,newi);
    }
  else
    {
      newi = (*doinfo)(y,x,COUNTEDVAL(infoy),COUNTEDVAL(infox),infodata);
      SETMAPLETY(y,infox);
      SETCOUNTEDN(infox,nx+ny);
      SETREFVAL(infox,newi);
    }
  QUIT;
}

/*
(* colored : (''a,'b,'c) union_find_tree -> ''a -> bool *)
(* returns true if the object has been colored *)
fun colored (|[info,link,...]| : (''a,'b,'c) union_find_tree) x =
    x inset link orelse (* if it has a link, it is colored *)
    x inset info (* otherwise, it is itself a color, which is colored iff it is in info *)
*/

PUBLIC Boolean uColored(utree,x)
IN CONST uFindPtr utree;
IN CONST ExprPtr x;
{
  Boolean found;

  (void) NGetMaplet(x,utree,&found);
  return found;
}

/*
(* new_color adds a new object as an independent color;
 the function is incorrect if x is already colored *)
fun new_color (tree as |[link,info,userdata]|) (extrainfo,douser,x) =
    |[link = link,
      info = info ++ {x => (1,extrainfo)},
      userdata = douser (x,userdata)]|
*/

PUBLIC void uNewColor(utree,x,info)
INOUT uFindPtr utree;
IN CONST ExprPtr x,info;
{
  ExprPtr newi;
  ENTER;

  newi = MakeCounted(E_COUNTED,info,(INT)1);
  (void) KEEP(newi);
  setNPut(x,newi,&utree,NOEXPR);
  KEEPONLY(utree);
}

/*
fun concat_heap (H0,h) = h
  | concat_heap (h,H0) = h
  | concat_heap hpair = H2 hpair
*/

PUBLIC ExprPtr heapConcat(h1,h2)
IN CONST REGISTER ExprPtr h1,h2;
{
  ExprPtr e;

  if (h1==H0)
    return h2;
  if (h2==H0)
    return h1;
  MAKETUPLE2(h1,h2,e);
  return KEEP(e);
}

/*
fun app_heap f =
    let fun app H0 = ()
	  | app (H1 x) = f x
	  | app (H2 (h,h')) = (app h;
			       app h')
    in
	app
    end
*/

PUBLIC Boolean heapMap(h,f,data)
IN CONST REGISTER ExprPtr h;
IN heapSweepFunc f;
INOUT CHAR *data;
{
again:
  if (h==H0)
    return false;
  if (h->kind==(CHAR)H1)
    return (*f)(REFVAL(h),data);
  if (heapMap(CAR(h),f,data))
    return true;
  h = CDR(h);
  goto again;
}

PUBLIC congPtr newCong(newinfo,data)
IN newCongInfoFunc newinfo;
IN char *data;
{
  ExprPtr u,e,info,ll;
  ENTER;

  u = newUFind();
  info = (*newinfo)(H0,data);
  e = MakeCounted(E_COUNTED,H0,(INT)0);
  (void) KEEP(e);
  MAKETUPLE2(e,info,ll);
  (void) KEEP(ll);
  uNewColor(u,H0,ll);
  MAKETUPLE2(u,NIL,e);
  KEEPONLY(e);
  return e;
}

INFOMERGEFUNC(mergeCongInfo)
{
  struct mergeconginfodata *data;
  ExprPtr otherinfo,nl1,nl2,ll;
  ENTER;

  data = INFOMERGEDATA(struct mergeconginfodata *);
  otherinfo = (data->merge2)(x1,x2,CDR(info1),CDR(info2),data->data2);
  SETCDR(info1,otherinfo);
  nl1 = CAR(info1);
  nl2 = CAR(info2);
  SETCOUNTEDN(nl1,COUNTEDN(nl1)+COUNTEDN(nl2));
  ll = heapConcat(COUNTEDVAL(nl1),COUNTEDVAL(nl2));
  SETREFVAL(nl1,ll);
  QUIT;
  return info1;
}

PRIVATE colorPtr congNewColor(cong,h,newinfo,data)
INOUT congPtr cong;
IN CONST ExprPtr h;
IN newCongInfoFunc newinfo;
IN char *data;
{
  ExprPtr info,aux,nl;
  ENTER;

  info = (*newinfo)(h,data);
  nl = MakeCounted(E_COUNTED,H0,(INT)0); /* no parents */
  (void) KEEP(nl);
  MAKETUPLE2(nl,info,aux);
  (void) KEEP(aux);
  uNewColor(CONGUTREE(cong),h,aux);
  QUIT;
  return h;
}

PRIVATE void addOneParent(c,parent)
IN CONST colorPtr c;
IN CONST ExprPtr parent;
{
  ExprPtr nl,aux,h;
  ENTER;

  nl = CAR(uinfoData(colorInfo(c)));
  SETCOUNTEDN(nl,COUNTEDN(nl)+1);
  aux = MakeTuple1(parent);
  (void) KEEP(aux);
  h = heapConcat(aux,COUNTEDVAL(nl));
  SETREFVAL(nl,h);
  QUIT;
}  

PUBLIC colorPtr congColorize(cong,h,newinfo,newinfodata,mergeinfo,mergeinfodata)
INOUT congPtr cong;
IN CONST ExprPtr h;
IN newCongInfoFunc newinfo;
IN char *newinfodata;
IN infoMergeFunc mergeinfo;
IN char *mergeinfodata;
{
  {
    ExprPtr color;
    Boolean found;

    color = uFind1(CONGUTREE(cong),h,&found);
    if (found)
      return color;
  }
  /* H0 is always in the utree, so we only have to test for H1 and H2 objects */
  if (h->kind==H1) /* then add new color */
    return congNewColor(cong,h,newinfo,newinfodata);
  else /* then h==H2 (x,y) */
    {
      ExprPtr c1,c2,sig,color,newcolor;
      Boolean found;
      ENTER;

      c1 = congColorize(cong,CAR(h),newinfo,newinfodata,mergeinfo,mergeinfodata);
      c2 = congColorize(cong,CDR(h),newinfo,newinfodata,mergeinfo,mergeinfodata);
      sig = Ktuple2(c1,c2);
      (void) KEEP(sig);
      color = NGet(sig,CONGSIGS(cong),&found);
      newcolor = congNewColor(cong,h,newinfo,newinfodata);
      if (found)
	{
	  struct mergeconginfodata data;

	  data.merge2 = mergeinfo;
	  data.data2 = mergeinfodata;
	  uMerge(CONGUTREE(cong),color,newcolor,mergeCongInfo,(char *)&data);
	}
      else
	{
	  addOneParent(c1,h);
	  if (c1!=c2)
	   addOneParent(c2,h);
	 color = newcolor;
	}
      KEEPONLY(color);
      return color;
    }
}

MAPSWEEPFUNC(do_congeqinitcombine)
{
  ExprPtr *combinep;
  ExprPtr aux;

  combinep = MAPDATA(ExprPtr *);
  MAKECONS(x,*combinep,aux);
  *combinep = KEEP(aux);
  return false;
}

PRIVATE ExprPtr congEqInitCombine(eqns)
IN CONST MapPtr eqns;
{
  ExprPtr combine;
  ENTER;

  combine = NIL;
  (void) MapSweep(eqns,do_congeqinitcombine,(char *)&combine);
  KEEPONLY(combine);
  return combine;
}

HEAPSWEEPFUNC(do_killsig)
{
  struct killsig *data;
  ExprPtr c1,c2,sig,aux;
  Boolean found;
  ENTER;

  data = HEAPSWEEPDATA(struct killsig *);
  c1 = uFind1(CONGUTREE(data->cong),CAR(x),&found); /* found should be true */
  c2 = uFind1(CONGUTREE(data->cong),CDR(x),&found); /* found should be true */
  sig =  Ktuple2(c1,c2);
  (void) KEEP(sig);
  setNRemove(sig,&CONGSIGS(data->cong),data->cong);
  MAKECONS(x,data->pending,aux);
  data->pending = aux;
  KEEPONLY(data->pending);
  return false;
}

PUBLIC void congEqualize(cong,eqns,newinfo,newinfodata,mergeinfo,mergeinfodata)
INOUT congPtr cong;
IN CONST MapPtr eqns;
IN newCongInfoFunc newinfo;
IN char *newinfodata;
IN infoMergeFunc mergeinfo;
IN char *mergeinfodata;
{
  ExprPtr combine;
  struct killsig killsigdata;
  ENTER;

  killsigdata.pending = NIL;
  killsigdata.cong = cong;
  combine = congEqInitCombine(eqns);
  while (combine!=NIL)
    {
      while (combine!=NIL)
	{
	  ExprPtr cv,cw,eqn;

	  eqn = CAR(combine);
	  combine = CDR(combine);
	  cv = congColorize(cong,CAR(eqn),newinfo,newinfodata,mergeinfo,mergeinfodata);
	  cw = congColorize(cong,CDR(eqn),newinfo,newinfodata,mergeinfo,mergeinfodata);
	  if (cv!=cw)
	    {
	      long nv,nw;

	      nv = COUNTEDN(CAR(uinfoData(colorInfo(cv))));
	      nw = COUNTEDN(CAR(uinfoData(colorInfo(cw))));
	      if (nv<nw)
		{
		  ExprPtr aux;

		  aux = cv; cv = cw; cw = aux;
		}
	      (void) heapMap(COUNTEDVAL(CAR(uinfoData(colorInfo(cw)))),
			     do_killsig,(CHAR *)&killsigdata);
	      {
		struct mergeconginfodata data;

		data.merge2 = mergeinfo;
		data.data2 = mergeinfodata;
		uMerge(CONGUTREE(cong),cv,cw,mergeCongInfo,(char *)&data);
	      }
	    }
	}
      KEEPONLY(killsigdata.pending);
      while (killsigdata.pending!=NIL)
	{
	  ExprPtr x,c1,c2,sig,c;
	  Boolean found;

	  x = CAR(killsigdata.pending);
	  killsigdata.pending = CDR(killsigdata.pending);
	  c1 = uFind1(CONGUTREE(cong),CAR(x),&found); /* found should be true */
	  c2 = uFind1(CONGUTREE(cong),CDR(x),&found); /* found should be true */
	  sig =  Ktuple2(c1,c2);
	  (void) KEEP(sig);
	  c = NGet(sig,CONGSIGS(cong),&found);
	  if (found)
	    {
	      ExprPtr aux,aux2;

	      MAKETUPLE2(x,c,aux);
	      (void) KEEP(aux);
	      MAKECONS(aux,combine,aux2);
	      (void) KEEP(aux2);
	      combine = aux2;
	    }
	  else
	    {
	      c = uFind1(CONGUTREE(cong),x,&found); /* found should be true */
	      setNPut(sig,c,&CONGSIGS(cong),cong);
	    }
	}
      KEEPONLY(combine);
    }
  QUIT;
}
